import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {ImageObject} from "./ImageObject";
import {ModeInteraction} from "../../public-api";
import { debounce } from 'debounce';

/**
 * Base container model
 */
export class BaseContainerModel {
}

/**
 * Base container image model
 */
export class BaseContainerImageModel extends BaseContainerModel {
    image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage;
}

/**
 * Base container
 * Override default behaviour of pixi container
 */
export abstract class BaseContainer<TData extends BaseContainerModel | BaseContainerImageModel> extends PIXI.Container {
    /**
     * Members
     */
    protected  _data: TData;
    protected  _modeInteraction: ModeInteraction<ImageObject>;

    /**
     * Methods
     */
    protected _debounceStartDrawingMethod: EventListener;
    protected _debounceDrawMethod: EventListener;

    /**
     * Constructor
     * @param data
     */
    protected constructor(data: TData) {
        // Parent constructor
        super();

        // Members
        this.data = data;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction = new MedsurfDraw.ModeInteraction(this);
        this.modeInteraction.on("init", this.init, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceStartDrawingMethod = debounce(this.startDrawing.bind(this), 500).bind(this);
        this._debounceDrawMethod = debounce(this.draw.bind(this), 20).bind(this);
        // TODO test: this._debounceDrawMethod = debounce(this.draw.bind(this), 200).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", (parent: MedsurfDraw.ImageObject) => {
            // Events
            if (this instanceof MedsurfDraw.Image || this instanceof MedsurfDraw.NavigatorElement
                || this instanceof MedsurfDraw.VirtualPointerElement) {
                // -- Mode
                parent.on("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
                parent.on("setMode", this.modeInteraction.setMode, this.modeInteraction);
            } else if (this instanceof MedsurfDraw.MenuElement || this instanceof MedsurfDraw.SelftestPoint/* || this instanceof MedsurfDraw.ScaleElement*/) {
                // -- Zoom
                this.image.on('imageZoom', this.onImageZoom, this);
                // -- Mode
                this.image.modeInteraction.on("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
                this.image.modeInteraction.on("setMode", this.modeInteraction.setMode, this.modeInteraction);
            } else {
                // -- Zoom
                this.image.on('imageZoom', this.onImageZoom, this);
                // -- Mode
                parent.modeInteraction.on("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
                parent.modeInteraction.on("setMode", this.modeInteraction.setMode, this.modeInteraction);
            }
            // -- Debounce
            this.on("debounceStart", this._debounceStartDrawingMethod);
            this.on("debounceDraw", this._debounceDrawMethod);
        });
        // -- Removed
        this.on("removed", (parent: MedsurfDraw.ImageObject) => {
            // Events
            if (this instanceof MedsurfDraw.Image || this instanceof MedsurfDraw.NavigatorElement || this instanceof MedsurfDraw.VirtualPointerElement) {
                // -- Mode
                parent.off("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
                parent.off("setMode", this.modeInteraction.setMode, this.modeInteraction);
            } else if (this instanceof MedsurfDraw.MenuElement || this instanceof MedsurfDraw.SelftestPoint/* || this instanceof MedsurfDraw.ScaleElement*/) {
                // -- Zoom
                this.image.off('imageZoom', this.onImageZoom, this);
                // -- Mode
                this.image.modeInteraction.off("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
                this.image.modeInteraction.off("setMode", this.modeInteraction.setMode, this.modeInteraction);
            } else {
                // -- Zoom
                this.image.off('imageZoom', this.onImageZoom, this);
                // -- Mode
                parent.modeInteraction.off("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
                parent.modeInteraction.off("setMode", this.modeInteraction.setMode, this.modeInteraction);
            }
            // -- Debounce
            this.off("debounceStart", this._debounceStartDrawingMethod);
            this.off("debounceDraw", this._debounceDrawMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     * @param parent
     */
    public abstract init(parent: MedsurfDraw.ImageObject): void;

    /**
     * Start Draw
     */
    public startDrawing(): void {
        // Events
        this.on("debounceDraw", this._debounceDrawMethod);
    }

    /**
     * No Draw
     * @param selfheal_timeout
     */
    public stopDrawing(selfheal_timeout: number): void {
        // Events
        this.off("debounceDraw", this._debounceDrawMethod);

        // Try again after selfheal_timeout seconds
        setTimeout(() => {
            this.emit("debounceStart");
        }, selfheal_timeout);
    }

    /**
     * Draw
     */
    public abstract draw(): void;

    /**
     * Destroy
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Mode interaction
        if (this.modeInteraction) {
            this.modeInteraction.removeAllListeners();
            // TODO remove delete this.modeInteraction;
        }
        // -- Element interaction
        this.removeAllListeners();
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        const position = this.position;
        const bounds = this.getLocalBounds();
        return new PIXI.Rectangle(position.x + bounds.x, position.y + bounds.y, bounds.width, bounds.height);
    }

    /**
     * Get element rectangle
     */
    public getElementRectangle(): PIXI.Rectangle {
        return this.getRectangle();
    }

    /**
     * Get rotation
     */
    public getRotation(): number {
        return this.rotation;
    }

    /**
     * Show item
     */
    public showItem(): void {
        // Setup
        this.visible = true;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Hide item
     */
    public hideItem(): void {
        // Setup
        this.visible = false;
    }

    /**
     * Toggle item
     */
    public toggleItem(override?: boolean): void {
        this.visible = override || !this.visible;
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get data
     */
    public get data(): TData {
        return this._data;
    }

    /**
     * Set data
     * @param value
     */
    public set data(value: TData) {
        this._data = value;
    }

    /**
     * Get image
     */
    public get image(): MedsurfDraw.Image | MedsurfDraw.DeepZoomImage {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseContainerImageModel).image;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get image canvas
     */
    public get imageCanvas(): HTMLCanvasElement {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseContainerImageModel).image.canvas;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get image scale
     */
    public get imageScale(): PIXI.IPoint {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseContainerImageModel).image.imageScale;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get image dimensions
     */
    public get imageDimensions(): MedsurfDraw.Dimensions {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseContainerImageModel).image.imageDimensions;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get mode interaction
     */
    public get modeInteraction(): ModeInteraction<ImageObject> {
        return this._modeInteraction;
    }

    /**
     * Set mode interaction
     * @param value
     */
    public set modeInteraction(value: ModeInteraction<ImageObject>) {
        this._modeInteraction = value;
    }
    //</editor-fold>
}