import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {ImageObject} from "./ImageObject";
import {ModeInteraction} from "../../public-api";
import { debounce } from 'debounce';

/**
 * Base graphics model
 */
export class BaseGraphicsModel {
}

/**
 * Base graphics image model
 */
export class BaseGraphicsImageModel extends BaseGraphicsModel {
    image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage;
}

/**
 * Base graphic
 * Override default behaviour of pixi graphics
 */
export class BaseGraphics<TData extends BaseGraphicsModel | BaseGraphicsImageModel> extends PIXI.Graphics {
    /**
     * Members
     */
    protected _data: TData;
    protected _modeInteraction: ModeInteraction<ImageObject>;

    /**
     * Methods
     */
    protected _debounceDrawMethod: EventListener;

    /**
     * Constructor
     * @param data
     */
    public constructor(data: TData) {
        // Parent constructor
        super();

        // Members
        this.data = data;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction = new MedsurfDraw.ModeInteraction(this);
        this.modeInteraction.on("init", this.init, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceDrawMethod = debounce(this.draw.bind(this), 20).bind(this);
        //</editor-fold>

        // Events
        // -- Added
        this.on("added", (parent: MedsurfDraw.ImageObject) => {
            // -- Debounce
            this.on("debounceDraw", this._debounceDrawMethod);
            // -- Mode
            parent.modeInteraction.on("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
            parent.modeInteraction.on("setMode", this.modeInteraction.setMode, this.modeInteraction);
        });
        // -- Removed
        this.on("removed", (parent: MedsurfDraw.ImageObject) => {
            // -- Debounce
            this.off("debounceDraw", this._debounceDrawMethod);
            // -- Mode
            parent.modeInteraction.off("setDefaultMode", this.modeInteraction.setDefaultMode, this.modeInteraction);
            parent.modeInteraction.off("setMode", this.modeInteraction.setMode, this.modeInteraction);
        });
    }

    /**
     * Init
     * @param parent
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        return
    }

    /**
     * Draw
     */
    public draw(): void {
        return;
    }

    /**
     * Destroy
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Mode interaction
        if (this.modeInteraction) {
            this.modeInteraction.removeAllListeners();
            // TODO remove delete this.modeInteraction;
        }
        // -- Element interaction
        this.removeAllListeners();
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        const position = this.position;
        const bounds = this.getLocalBounds();
        return new PIXI.Rectangle(position.x + bounds.x, position.y + bounds.y, bounds.width, bounds.height);
    }

    /**
     * Get element rectangle
     */
    public getElementRectangle(): PIXI.Rectangle {
        return this.getRectangle();
    }

    /**
     * Get rotation
     */
    public getRotation(): number {
        return this.rotation;
    }

    /**
     * Show item
     */
    public showItem(): void {
        // Setup
        this.visible = true;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Hide item
     */
    public hideItem(): void {
        // Setup
        this.visible = false;
    }

    /**
     * Toggle item
     */
    public toggleItem(override?: boolean): void {
        this.visible = override || !this.visible;
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get data
     */
    public get data(): TData {
        return this._data;
    }

    /**
     * Set data
     * @param value
     */
    public set data(value: TData) {
        this._data = value;
    }

    /**
     * Get image
     */
    public get image(): MedsurfDraw.Image | MedsurfDraw.DeepZoomImage {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseGraphicsImageModel).image;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get image canvas
     */
    public get imageCanvas(): HTMLCanvasElement {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseGraphicsImageModel).image.canvas;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get image scale
     */
    public get imageScale(): PIXI.IPoint {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseGraphicsImageModel).image.imageScale;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get image dimensions
     */
    public get imageDimensions(): MedsurfDraw.Dimensions {
        if (this.data.hasOwnProperty("image")) {
            return (this.data as BaseGraphicsImageModel).image.imageDimensions;
        }
        debugger;
        throw 'Not implemented';
    }

    /**
     * Get mode interaction
     */
    public get modeInteraction(): ModeInteraction<ImageObject> {
        return this._modeInteraction;
    }

    /**
     * Set mode interaction
     * @param value
     */
    public set modeInteraction(value: ModeInteraction<ImageObject>) {
        this._modeInteraction = value;
    }
    //</editor-fold>
}