import * as PIXI from "pixi.js-legacy";
import {ImageObject} from "../elements/ImageObject";
import { debounce } from 'debounce';

/**
 * Base generator model
 */
export class BaseGeneratorModel<TTarget extends ImageObject> {
    target: TTarget;
}

/**
 * Base generator
 */
export abstract class BaseGenerator<TTarget extends ImageObject, TData extends BaseGeneratorModel<TTarget>> extends PIXI.utils.EventEmitter {
    /**
     * Members
     */
    private _data: TData;

    /**
     * Methods
     */
    protected _debounceStartMethod: any;
    protected _debounceEndMethod: any;

    /**
     * Constructor
     * @param data
     */
    protected constructor(data: TData) {
        // Parent constructor
        super();

        // Members
        this.data = data;

        //<editor-fold desc="Methods">
        this._debounceStartMethod = debounce(this.start.bind(this), 50).bind(this);
        this._debounceEndMethod = debounce(this.end.bind(this), 200).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Image
        this.target.parent.once("setDefaultMode", this.destroy, this);
        // -- Debounce
        this.on("debounceStart", this._debounceStartMethod);
        this.on("debounceEnd", this._debounceEndMethod);
        //</editor-fold>
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        // Emit start generator
        this.emit("startGenerator", event);
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.parent.off("setDefaultMode", this.destroy, this);
        // -- Debounce
        this.off("debounceStart", this._debounceStartMethod);
        this.off("debounceEnd", this._debounceEndMethod);
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public abstract destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void;

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Events
        // -- Generator
        this.off('endGenerator', this.endGenerator, this);
        this.off('abortGenerator', this.abortGenerator, this);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Events
        // -- Generator
        this.off('endGenerator', this.endGenerator, this);
        this.off('abortGenerator', this.abortGenerator, this);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get data
     */
    public get data(): TData {
        return this._data;
    }

    /**
     * Set data
     * @param value
     */
    public set data(value: TData) {
        this._data = value;
    }

    /**
     * Get target
     */
    public get target(): TTarget {
        return this.data.target;
    }

    /**
     * Set target
     * @param value
     */
    public set target(value: TTarget) {
        this.data.target = value;
    }
    //</editor-fold>
}