import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import { BaseGraphics, BaseGraphicsModel } from '../../bases/elements/BaseGraphics';
import { debounce } from "debounce";

/**
 * Arc model
 */
export class ArcModel extends BaseGraphicsModel {
    center: PIXI.IPoint;
    radius: number;
    options: Models.ElementOptions;
    lineWidth: number;
}

/**
 * Arc
 */
export class Arc extends BaseGraphics<ArcModel> {
    /**
     * Methods
     */
    public _debounceUpdateHitAreaMethod: EventListener;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ArcModel) {
        // Parent constructor
        super(model);

        //<editor-fold desc="Methods">
        this._debounceUpdateHitAreaMethod = debounce(this.updateHitArea.bind(this), 500).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debouncedUpdateHitArea", this._debounceUpdateHitAreaMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debouncedUpdateHitArea", this._debounceUpdateHitAreaMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        this.clear();

        // Draw line
        this.lineStyle(this.lineWidth, this.options.lineColor, this.options.lineAlpha, 0.6);
        this.arc(this.center.x, this.center.y, this.radius, 0, Math.PI / 2);

        // Ensure geometry points are set
        this.emit("debouncedUpdateHitArea");

    }

    //<editor-fold desc="Public functions">
    /**
     * Update hit area
     */
    public updateHitArea(): void {
        if (!this._destroyed) {
            // this.hitArea = this.currentPath;
            // @ts-ignore
            this.hitArea = new PIXI.Polygon(this.geometry.points);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get center
     */
    public get center(): PIXI.IPoint {
        return this.data.center;
    }

    /**
     * Set center
     * @param value
     */
    public set center(value: PIXI.IPoint) {
        this.data.center = value;
    }

    /**
     * Get radius
     */
    public get radius(): number {
        return this.data.radius;
    }

    /**
     * Set radius
     * @param value
     */
    public set radius(value: number) {
        this.data.radius = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number) {
        this.data.lineWidth = value;
    }
    //</editor-fold>
}