import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import { BaseText, BaseTextImageModel } from "../../bases/elements/BaseText";
import { Design } from "../../config/design";

/**
 * Editable Text model
 */
export class EditableTextModel extends BaseTextImageModel {
    text: string;
    style: PIXI.TextStyle;
}

/**
 * Editable text
 */
export class EditableText extends BaseText<EditableTextModel> {
    /**
     * Members
     */
    private _startTextIndex: number;
    private _endTextIndex: number;
    private _textSelection: boolean;

    /**
     * Interactions
     */
    private _writeInteraction: MedsurfDraw.WriteInteraction<EditableText>;

    /**
     * Methods
     */
    protected _copyMethod: any;
    protected _cutMethod: any;
    protected _pasteMethod: any;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: EditableTextModel) {
        // Parent constructor
        super(model);

        // Members
        this.startTextIndex = this.data.text.length;
        this.endTextIndex = this.data.text.length;
        this.textSelection = false;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        // -- Write interaction
        this.writeInteraction = new MedsurfDraw.WriteInteraction(this);
        this.writeInteraction.on("onText", this.onText, this);
        this.writeInteraction.on("endText", this.endText, this);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Methods
        this._copyMethod = this.writeInteraction.onCopy.bind(this.writeInteraction);
        this._cutMethod = this.writeInteraction.onCut.bind(this.writeInteraction);
        this._pasteMethod = this.writeInteraction.onPaste.bind(this.writeInteraction);

        // Sort items
        this.sortChildren();

        // On image zoom
        this.onImageZoom(this.imageScale.x, this.imageScale.y);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Mode interaction
        if (this.modeInteraction) {
            this.modeInteraction.off('select_item', this._modeSelectItem, this);
            this.modeInteraction.off("remove-select_item", this._removeModeSelectItem, this);
        }
        // -- Write interaction
        if (this.writeInteraction) {
            this.writeInteraction.off("onText", this.onText, this);
            this.writeInteraction.off("endText", this.endText, this);
            // TODO remove delete this.writeInteraction;
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Mode">
    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        // Setup
        this.interactive = true;

        // Events
        // -- Keyboards
        MedsurfDraw.Keyboard.events.on("down", "onTextPrimitiveKeyDown", this.writeInteraction.onKeyDown.bind(this.writeInteraction));
        MedsurfDraw.Keyboard.events.on("pressed", "onTextPrimitiveKeyPress", this.writeInteraction.onKeyPressed.bind(this.writeInteraction));
        MedsurfDraw.Keyboard.events.on("released", "onTextPrimitiveKeyUp", this.writeInteraction.onKeyReleased.bind(this.writeInteraction));
        // -- Methods
        document.addEventListener("copy", this._copyMethod);
        document.addEventListener("cut", this._cutMethod);
        document.addEventListener("paste", this._pasteMethod);

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Events
        // -- Keyboards
        MedsurfDraw.Keyboard.events.remove("down", "onTextPrimitiveKeyDown");
        MedsurfDraw.Keyboard.events.remove("pressed", "onTextPrimitiveKeyPress");
        MedsurfDraw.Keyboard.events.remove("released", "onTextPrimitiveKeyUp");
        // -- Methods
        document.removeEventListener("copy", this._copyMethod);
        document.removeEventListener("cut", this._cutMethod);
        document.removeEventListener("paste", this._pasteMethod);

        if (mode !== 'select_child' && !mode.startsWith('select_item') && mode !== 'delete') {
            // Set select layer
            if (this.parent instanceof MedsurfDraw.TextPrimitive || this.parent instanceof MedsurfDraw.LegendRow) {
                parent.emit("debounceControlSelectItem");
            }
        }
    }
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this.parent.parent)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set default mode for parent
        if (this.parent instanceof MedsurfDraw.TextPrimitive) {
            this.parent.modeInteraction.setModeItem('select_child');
        }
        if (this.parent instanceof MedsurfDraw.LegendRow) {
            this.parent.modeInteraction.setModeItem('select_child')
        }

        // Set children mode
        this.modeInteraction.setModeChildren('select_parent');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }
    //</editor-fold>

    //<editor-fold desc="Public function">
    /**
     * Select end of text
     */
    public selectEndText(): void {
        this.writeInteraction.selectEndText();
    }

    /**
     * Select all text
     */
    public selectAllText(): void {
        this.writeInteraction.selectAllText();
    }
    //</editor-fold>

    //<editor-fold desc="Text">
    /**
     * On text
     * @param text
     */
    public onText(text: string): void {
        this.emit("onText", text);
    }

    /**
     * End text
     */
    public endText(): void {
        this.emit("endText");
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Update resolution
        let resolution = Math.round(scaleX);
        if (resolution < Design.text.minResolution) {
            resolution = Design.text.minResolution;
        }
        if (resolution > Design.text.maxResolution) {
            resolution = Design.text.maxResolution;
        }
        this.resolution = resolution;
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        try {
            if (this.parent instanceof MedsurfDraw.TextPrimitive || this.parent instanceof MedsurfDraw.LegendRow) {
                return this.parent.getRectangle();
            }
            return super.getRectangle();
        } catch (e) {
            // TODO what to do with this ???
            // console.warn(e);
            return new PIXI.Rectangle(this.position.x, this.position.y, 1, 1);
        }
    }

    /**
     * Get rotation
     */
    public getRotation(): number {
        if (this.parent instanceof MedsurfDraw.TextPrimitive || this.parent instanceof MedsurfDraw.LegendRow) {
            return this.parent.rotation;
        }
        return this.rotation;
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">


    /**
     * Get start text index
     */
    public get startTextIndex(): number {
        return this._startTextIndex;
    }

    /**
     * Set start text index
     * @param value
     */
    public set startTextIndex(value: number) {
        this._startTextIndex = value;
    }

    /**
     * Get end text index
     */
    public get endTextIndex(): number {
        return this._endTextIndex;
    }

    /**
     * Set end text index
     * @param value
     */
    public set endTextIndex(value: number) {
        this._endTextIndex = value;
    }

    /**
     * Get text selection
     */
    public get textSelection(): boolean {
        return this._textSelection;
    }

    /**
     * Set text selection
     * @param value
     */
    public set textSelection(value: boolean) {
        this._textSelection = value;
    }

    /**
     * Get write interaction
     */
    public get writeInteraction(): MedsurfDraw.WriteInteraction<EditableText> {
        return this._writeInteraction;
    }

    /**
     * Set write interaction
     * @param value
     */
    public set writeInteraction(value: MedsurfDraw.WriteInteraction<EditableText>) {
        this._writeInteraction = value;
    }
    //</editor-fold>
}