import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import { BaseGraphics, BaseGraphicsModel } from '../../bases/elements/BaseGraphics';
import { Design } from "../../config/design";
import { debounce } from "debounce";

/**
 * Control points model
 */
export class ControlPointsModel {
    c1: PIXI.IPoint;
    c2: PIXI.IPoint;
    l1: number;
    l2: number;
}

/**
 * Straight line points model
 */
export class StraightLinePointsModel {
    start: PIXI.IPoint;
    end: PIXI.IPoint;
}

/**
 * Straight line model
 */
export class StraightLineModel extends BaseGraphicsModel {
    points: StraightLinePointsModel;
    options: Models.ElementOptions;
    lineWidth: number;
}

/**
 * Straight line
 */
export class StraightLine extends BaseGraphics<StraightLineModel> {
    /**
     * Methods
     */
    public _debounceUpdateHitAreaMethod: EventListener;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: StraightLineModel) {
        // Parent constructor
        super(model);

        //<editor-fold desc="Methods">
        this._debounceUpdateHitAreaMethod = debounce(this.updateHitArea.bind(this), 50).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debouncedUpdateHitArea", this._debounceUpdateHitAreaMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debouncedUpdateHitArea", this._debounceUpdateHitAreaMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        this.clear();

        // Draw line
        if (this.options.hasLine) {
            this.lineStyle(this.lineWidth, this.options.lineColor, this.options.lineAlpha);
        } else {
            this.lineStyle(this.lineWidth, this.options.lineColor, 0.00000001);
        }
        this.moveTo(this.points.start.x, this.points.start.y);
        this.lineTo(this.points.end.x, this.points.end.y);
        // TODO test this: this._lineElement.drawPolygon([new PIXI.Point(points.start.x, points.start.y), new PIXI.Point(points.end.x, points.end.y)]);
        // TODO remove this.endFill();

        // Hit area update
        /* TODO find better solution for this */
        this.lineStyle(this.lineWidth + Design.line.hitboxLineWidthAddition, this.options.lineColor, 0.00000001);
        this.moveTo(this.points.start.x, this.points.start.y);
        this.lineTo(this.points.end.x, this.points.end.y);
        // TODO test this: this._lineElement.drawPolygon([new PIXI.Point(points.start.x, points.start.y), new PIXI.Point(points.end.x, points.end.y)]);
        // TODO remove this.endFill();
        /**/

        // Ensure geometry points are set
        this.emit("debouncedUpdateHitArea");
    }

    //<editor-fold desc="Public functions">
    /**
     * Update hit area
     */
    public updateHitArea(): void {
        if (!this._destroyed && this.geometry.graphicsData.length > 0) {
            // @ts-ignore
            // this.hitArea = new PIXI.Polygon(this.geometry.graphicsData[0].shape.points);
            // @ts-ignore
            this.hitArea = this.lineToPolygon(this.lineWidth * Design.line.hitboxLineWidthAddition, this.geometry.graphicsData[0].shape.points)
        }
    }

    /**
     * Line zo polygon
     *
     * @param distance
     * @param points
     */
    public lineToPolygon(distance: number, points: number[]): PIXI.Polygon {
        const numPoints = points.length / 2;
        const output = new Array(points.length * 2);
        for (let i = 0; i < numPoints; i++) {
            const j = i * 2;

            // Position of current point
            const x = points[j];
            const y = points[j + 1];

            // Start
            const x0 = points[j - 2] !== undefined ? points[j - 2] : x;
            const y0 = points[j - 1] !== undefined ? points[j - 1] : y;

            // End
            const x1 = points[j + 2] !== undefined ? points[j + 2] : x;
            const y1 = points[j + 3] !== undefined ? points[j + 3] : y;

            // Get the angle of the line
            const a = Math.atan2(-x1 + x0, y1 - y0);
            const deltaX = distance * Math.cos(a);
            const deltaY = distance * Math.sin(a);

            // Add the x, y at the beginning
            output[j] = x + deltaX;
            output[j + 1] = y + deltaY;

            // Add the reflected x, y at the end
            output[(output.length - 1) - j - 1] = x - deltaX;
            output[(output.length - 1) - j] = y - deltaY;
        }
        // close the shape
        output.push(output[0], output[1]);

        return new PIXI.Polygon(output);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get points
     */
    public get points(): StraightLinePointsModel {
        return this.data.points;
    }

    /**
     * Set points
     * @param value
     */
    public set points(value: StraightLinePointsModel) {
        this.data.points = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number) {
        this.data.lineWidth = value;
    }
    //</editor-fold>
}