import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Button element model
 */
export class ButtonElementModel extends BaseContainerImageModel {
    text: string;
    description: string;
    descriptionFontSize: number;
    rectangle?: PIXI.Rectangle;
    options?: Models.ElementOptions;
    fontSize?: number;
    fontColor?: number;
}

/**
 * Button element
 */
export class ButtonElement extends BaseContainer<ButtonElementModel> {
    /**
     * Interaction
     */
    private _selectInteraction: MedsurfDraw.SelectInteraction<ButtonElement>;

    /**
     * Elements
     */
    protected _rectangleElement: MedsurfDraw.Rectangle;
    protected _iconElement: MedsurfDraw.Text;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ButtonElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.button.zIndex;

        // Position
        this.position.set(this.rectangle.x, this.rectangle.y);

        //<editor-fold desc="Interactions">
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onRelease", this.onRelease, this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.on("mousedown", this.selectInteraction.onSelect, this.selectInteraction);
            this.on("mouseup", this.selectInteraction.onRelease, this.selectInteraction);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.off("mousedown", this.selectInteraction.onSelect, this.selectInteraction);
            this.off("mouseup", this.selectInteraction.onRelease, this.selectInteraction);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        /*
        let scale = this.imageScale.x * Design.subMenu.scaleAspect;
        let buttonWidth = this.rectangle.width;
        let buttonHeight = this.rectangle.height;
        buttonWidth = Math.floor(buttonWidth / scale);
        buttonHeight = Math.floor(buttonHeight / scale);
        */
        let fontSize = this.fontSize;
        if (fontSize < 1) {
            fontSize = 1;
        }
        //<editor-fold desc="Elements">
        // -- Rectangle element
        this._rectangleElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, this.rectangle.width, this.rectangle.height),
            options: {hasLine: true, lineColor: Design.button.lineColor, lineAlpha: Design.button.lineAlpha, hasFill: true, fillColor: Design.button.fillColor, fillAlpha: Design.button.fillAlpha},
            lineWidth: Design.button.lineWidth
        });
        this.addChild(this._rectangleElement);
        // -- Icon element
        this._iconElement = new MedsurfDraw.Text({
            text: this.text,
            style: new PIXI.TextStyle({ align: "left", fill: 0x000000, fontSize: fontSize, fontFamily: "Font Awesome 5 Pro", fontStyle: "normal", fontWeight: "500"})
        });
        // TODO remove this._iconElement.resolution = 4;
        this._iconElement.anchor.set(0.5, 0.5);
        this._rectangleElement.addChild(this._iconElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        /*
        let scale = this.imageScale.x * Design.subMenu.scaleAspect;
        let buttonWidth = this.rectangle.width;
        let buttonHeight = this.rectangle.height;
        buttonWidth = Math.floor(buttonWidth / scale);
        buttonHeight = Math.floor(buttonHeight / scale);
        */
        let fontSize = this.fontSize;
        if (fontSize < 1) {
            fontSize = 1;
        }
        //<editor-fold desc="Elements">
        // -- Rectangle element
        this._rectangleElement.rectangle = new PIXI.Rectangle(0, 0, this.rectangle.width, this.rectangle.height);
        this._rectangleElement.draw();
        // -- Icon element
        this._iconElement.style.fontSize = fontSize;
        this._iconElement.position.set(this._rectangleElement.rectangle.width / 2, this._rectangleElement.rectangle.height / 2);
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Icon element
        if (this._iconElement) {
            this._iconElement.destroy(options);
        }
        // -- Rectangle element
        if (this._rectangleElement) {
            this._rectangleElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Set icon angle
     * @param angle
     */
    public setIconAngle(angle: number): void {
        this._iconElement.angle = angle;
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On release
     * @param event
     * @private
     */
    public onRelease(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit button
        this.emit("button", event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text
     */
    public get text(): string {
        return this.data.text;
    }

    /**
     * Set text
     * @param value
     */
    public set text(value: string) {
        this.data.text = value;

        // Elements
        // -- Icon element
        this._iconElement.text = this.text;
    }

    /**
     * Get description
     */
    public get description(): string {
        return this.data.description;
    }

    /**
     * Set description
     * @param value
     */
    public set description(value: string) {
        this.data.description = value;
    }

    /**
     * Get description font size
     */
    public get descriptionFontSize(): number {
        return this.data.descriptionFontSize;
    }

    /**
     * Set description font size
     * @param value
     */
    public set descriptionFontSize(value: number) {
        this.data.descriptionFontSize = value;
    }

    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle || new PIXI.Rectangle(0, 0, 20, 20);
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;

        // Elements
        // -- Rectangle element
        this._rectangleElement.rectangle = this.rectangle;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options || {hasLine: true, lineColor: Design.menuButton.lineColor, lineAlpha: Design.menuButton.lineAlpha, hasFill: true, fillColor: Design.menuButton.fillColor, fillAlpha: Design.menuButton.fillAlpha};
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;

        // Elements
        // -- Rectangle element
        this._rectangleElement.options = this.options;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get fontSize
     */
    public get fontSize(): number {
        return this.data.fontSize || 20;
    }

    /**
     * Set fontSize
     * @param value
     */
    public set fontSize(value: number) {
        this.data.fontSize = value;

        // Elements
        // -- Icon element
        this._iconElement.style.fontSize = this.fontSize;
    }

    /**
     * Get fontColor
     */
    public get fontColor(): number {
        return this.data.fontColor || 0x000000;
    }

    /**
     * Set fontColor
     * @param value
     */
    public set fontColor(value: number) {
        this.data.fontColor = value;

        // Elements
        // -- Icon element
        this._iconElement.style.fill = this.fontColor;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<ButtonElement> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<ButtonElement>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}