import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {BaseTilingSprite, BaseTilingSpriteImageModel} from "../../bases/elements/BaseTilingSprite";

/**
 * Color button element model
 */
export class ColorButtonElementModel extends BaseContainerImageModel {
    description: string;
    descriptionFontSize: number;
    loader: PIXI.Loader;
    hasFill: boolean;
    color: number | undefined;
    colorAlpha: number | undefined;
    rectangle?: PIXI.Rectangle;
}

/**
 * Color button element
 */
export class ColorButtonElement extends BaseContainer<ColorButtonElementModel> {
    /**
     * Interaction
     */
    private _selectInteraction: MedsurfDraw.SelectInteraction<ColorButtonElement>;

    /**
     * Elements
     */
    protected _transparentElement: BaseTilingSprite<BaseTilingSpriteImageModel>;
    protected _colorElement: MedsurfDraw.Rectangle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ColorButtonElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.colorButton.zIndex;

        //<editor-fold desc="Interactions">
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onSelect", this.onSelect, this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", (parent: PIXI.Container) => {
            // Events
            this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.on("mousedown", this.selectInteraction.onSelect, this.selectInteraction);
            this.on("mouseup", this.selectInteraction.onRelease, this.selectInteraction);
        });
        // -- Removed
        this.on("removed", (parent: PIXI.Container) => {
            // Events
            this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.off("mousedown", this.selectInteraction.onSelect, this.selectInteraction);
            this.off("mouseup", this.selectInteraction.onRelease, this.selectInteraction);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        // Position
        this.position.set(this.rectangle.x, this.rectangle.y);

        // Texture
        const texture = this.loader.resources['transparent'].texture;
        if (!texture || !texture.baseTexture) {
            throw 'No texture created';
        }
        texture.baseTexture.wrapMode = PIXI.WRAP_MODES.REPEAT;

        //<editor-fold desc="Elements">
        // -- Transparent element
        this._transparentElement = new BaseTilingSprite({image: this.image, texture});
        this._transparentElement.width = this.rectangle.width;
        this._transparentElement.height = this.rectangle.height;
        this.addChild(this._transparentElement);
        // -- Color element
        this._colorElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, this.rectangle.width, this.rectangle.height),
            options: {hasLine: true, lineColor: Design.colorButton.lineColor, lineAlpha: Design.colorButton.lineAlpha, hasFill: true, fillColor: this.color, fillAlpha: this.colorAlpha},
            lineWidth: Design.colorButton.lineWidth
        });
        this.addChild(this._colorElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Color element
        this._colorElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Transparent element
        if (this._transparentElement) {
            this._transparentElement.destroy(options);
        }
        // -- Color element
        if (this._colorElement) {
            this._colorElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /*
    /**
     * Set color
     * @param color
     * /
    public setColor(color: number): void {
        this._colorElement.options.fillColor = color;
        this._colorElement.draw();
    }

    /**
     * Set alpha
     * @param alpha
     * /
    public setAlpha(alpha: number): void {
        this._colorElement.options.fillAlpha = alpha;
        this._colorElement.draw();
    }

    /**
     * Set button size
     * @param width
     * @param height
     * /
    public setButtonSize(width?: number, height?: number): void {
        if (width) {
            this._colorElement.rectangle.width = width;
        }
        if (height) {
            this._colorElement.rectangle.height = height;
        }
        this.emit("debounceDraw");
    }
    */
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On select
     * @param event
     * @private
     */
    public onSelect(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.emit("button", event, this._colorElement.options.fillColor, this._colorElement.options.fillAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get description
     */
    public get description(): string {
        return this.data.description;
    }

    /**
     * Set description
     * @param value
     */
    public set description(value: string) {
        this.data.description = value;
    }

    /**
     * Get description font size
     */
    public get descriptionFontSize(): number {
        return this.data.descriptionFontSize;
    }

    /**
     * Set description font size
     * @param value
     */
    public set descriptionFontSize(value: number) {
        this.data.descriptionFontSize = value;
    }

    /**
     * Get loader
     */
    public get loader(): PIXI.Loader {
        return this.data.loader;
    }

    /**
     * Set loader
     * @param value
     */
    public set loader(value: PIXI.Loader) {
        this.data.loader = value;
    }

    /**
     * Get has fill
     */
    public get hasFill(): boolean {
        return this.data.hasFill;
    }

    /**
     * Set has fill
     * @param value
     */
    public set hasFill(value: boolean) {
        this.data.hasFill = value;

        // Elements
        // -- Color element
        this._colorElement.options.hasFill = this.hasFill;
    }

    /**
     * Get color
     */
    public get color(): number | undefined {
        return this.data.color;
    }

    /**
     * Set color
     * @param value
     */
    public set color(value: number | undefined) {
        this.data.color = value;

        // Elements
        // -- Color element
        this._colorElement.options.fillColor = this.color;
    }

    /**
     * Get color alpha
     */
    public get colorAlpha(): number | undefined {
        return this.data.colorAlpha;
    }

    /**
     * Set color alpha
     * @param value
     */
    public set colorAlpha(value: number | undefined) {
        this.data.colorAlpha = value;

        // Elements
        // -- Color element
        this._colorElement.options.fillAlpha = this.colorAlpha;
    }

    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle || new PIXI.Rectangle(0, 0, 20, 20);
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<ColorButtonElement> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<ColorButtonElement>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}