import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseGraphics, BaseGraphicsImageModel} from "../../bases/elements/BaseGraphics";

/**
 * Indexed Round button mask element model
 */
export class IndexedRoundButtonMaskElementModel extends BaseGraphicsImageModel {
    textRectangle: PIXI.Rectangle;
    indexRectangle: PIXI.Rectangle;
}

/**
 * Indexed round button mask element
 */
export class IndexedRoundButtonMaskElement extends BaseGraphics<IndexedRoundButtonMaskElementModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: IndexedRoundButtonMaskElementModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        this.cacheAsBitmap = false;

        this.clear();

        /*
        Polygon style
        1                   2
        o-------------------o
        |                   |
        |       4           |
        |       o-----------o 3
        |       |
        |       |
        |       |
        o-------o
        6       5
         */

        // Draw text rectangle
        this.beginFill(0xFFFFFF, 1);
        this.drawPolygon([
            new PIXI.Point(this.textRectangle.x, this.textRectangle.y), // 1
            new PIXI.Point(this.textRectangle.x + this.textRectangle.width, this.textRectangle.y), // 2
            new PIXI.Point(this.textRectangle.x + this.textRectangle.width, this.indexRectangle.y - Design.indexedRoundButtonMask.pad), // 3
            new PIXI.Point(this.indexRectangle.x - Design.indexedRoundButtonMask.pad, this.indexRectangle.y - Design.indexedRoundButtonMask.pad), // 4
            new PIXI.Point(this.indexRectangle.x - Design.indexedRoundButtonMask.pad, this.textRectangle.y + this.textRectangle.height), // 5
            new PIXI.Point(this.textRectangle.x, this.textRectangle.y + this.textRectangle.height), // 6
        ]);
        this.endFill();

        this.cacheAsBitmap = true;
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text rectangle
     */
    public get textRectangle(): PIXI.Rectangle {
        return this.data.textRectangle;
    }

    /**
     * Set circle
     * @param value
     */
    public set textRectangle(value: PIXI.Rectangle) {
        this.data.textRectangle = value;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get index rectangle
     */
    public get indexRectangle(): PIXI.Rectangle {
        return this.data.indexRectangle;
    }

    /**
     * Set index rectangle
     * @param value
     */
    public set indexRectangle(value: PIXI.Rectangle) {
        this.data.indexRectangle = value;

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>
}