import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Round button element model
 */
export class RoundButtonElementModel extends BaseContainerImageModel {
    text: string;
    description: string;
    descriptionFontSize: number;
    circle?: PIXI.Circle;
    options?: Models.ElementOptions;
    fontSize?: number;
    fontColor?: number;
}

/**
 * Round button element
 */
export class RoundButtonElement extends BaseContainer<RoundButtonElementModel> {
    /**
     * Interaction
     */
    private _selectInteraction: MedsurfDraw.SelectInteraction<RoundButtonElement>;

    /**
     * Elements
     */
    // TODO protected _blurElement: MedsurfDraw.Circle;
    protected _circleElement: MedsurfDraw.Circle;
    protected _iconElement: MedsurfDraw.Text;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RoundButtonElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.roundButton.zIndex;

        // Position
        this.position.set(this.circle.x, this.circle.y);

        //<editor-fold desc="Interactions">
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onRelease", this.onRelease, this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Select
            this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
            this.on("mousedown", this.selectInteraction.onSelect, this.selectInteraction);
            this.on("mouseup", this.selectInteraction.onRelease, this.selectInteraction);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Select
            this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
            this.off("mousedown", this.selectInteraction.onSelect, this.selectInteraction);
            this.off("mouseup", this.selectInteraction.onRelease, this.selectInteraction);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        /*
        let scale = this.imageScale.x * Design.roundButton.scaleAspect;
        let buttonRadius = this.circle.radius;
        buttonRadius = Math.floor(buttonRadius / scale);
        */
        let lineWidth = Design.roundButton.lineWidth;
        if (lineWidth < Design.roundButton.minLineWidth) {
            lineWidth = Design.roundButton.minLineWidth;
        }
        let fontSize = this.fontSize
        if (fontSize < Design.roundButton.minFontSize) {
            fontSize = Design.roundButton.minFontSize;
        }
        //<editor-fold desc="Elements">
        // -- Blur element
        /* TODO
        // @ts-ignore
        // const filter = new PIXI.picture.MaskFilter(new PIXI.filters.BlurFilter());
        // const filter = new PIXI.filters.BlurFilter(8);
        // filter.safeFlipY = true;
        this._blurElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, buttonRadius),
            options: this.options,
            lineWidth: lineWidth
        });
        // this._blurElement.filters = [filter];
        this.addChild(this._blurElement);
        */
        // -- Circle element
        this._circleElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, this.circle.radius),
            options: this.options,
            lineWidth: lineWidth
        });
        this.addChild(this._circleElement);
        // -- Icon element
        this._iconElement = new MedsurfDraw.Text({
            text: this.text,
            style: new PIXI.TextStyle({ align: "left", fill: this.fontColor, fontSize: fontSize, fontFamily: "Font Awesome 5 Pro", fontStyle: "normal", fontWeight: "500"})
        });
        // TODO remove this._iconElement.resolution = 4;
        this._iconElement.anchor.set(0.5, 0.5);
        this._circleElement.addChild(this._iconElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        /*
        let scale = this.imageScale.x * Design.roundButton.scaleAspect;
        let buttonRadius = this.circle.radius;
        buttonRadius = Math.floor(buttonRadius / scale);
        */
        let lineWidth = Design.roundButton.lineWidth;
        if (lineWidth < Design.roundButton.minLineWidth) {
            lineWidth = Design.roundButton.minLineWidth;
        }
        let fontSize = this.fontSize;
        if (fontSize < Design.roundButton.minFontSize) {
            fontSize = Design.roundButton.minFontSize;
        }
        //<editor-fold desc="Elements">
        // -- Blur element
        /* TODO
        this._blurElement.circle = new PIXI.Circle(0, 0, buttonRadius);
        this._blurElement.lineWidth = lineWidth;
        this._blurElement.draw();
        */
        // -- Circle element
        this._circleElement.circle = new PIXI.Circle(0, 0, this.circle.radius);
        this._circleElement.lineWidth = lineWidth;
        this._circleElement.draw();
        // -- Icon element
        this._iconElement.style.fontSize = fontSize;
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Icon element
        if (this._iconElement) {
            this._iconElement.destroy(options);
        }
        // -- Circle element
        if (this._circleElement) {
            this._circleElement.destroy(options);
        }
        // -- Blur element
        /* TODO
        if (this._blurElement) {
            this._blurElement.destroy(options);
        }
        */
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Set icon angle
     * @param angle
     */
    public setIconAngle(angle: number): void {
        this._iconElement.angle = angle;
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     * @param event
     * @private
     */
    public onHover(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Element
        // -- Circle element
        this._circleElement.options.lineAlpha = 1;
        this._circleElement.options.fillAlpha = 1;

        // Draw
        this.emit("debounceDraw");

        // Emit hover
        this.emit("onHover", event);
    }

    /**
     * On out
     * @param event
     * @private
     */
    public onOut(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Element
        // -- Circle element
        this._circleElement.options = this.options;

        // Draw
        this.emit("debounceDraw");

        // Emit hover
        this.emit("onOut", event);
    }

    /**
     * On release
     * @param event
     * @private
     */
    public onRelease(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit button
        this.emit("button", event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text
     */
    public get text(): string {
        return this.data.text;
    }

    /**
     * Set text
     * @param value
     */
    public set text(value: string) {
        this.data.text = value;

        // Elements
        // -- Icon element
        this._iconElement.text = this.text;
    }

    /**
     * Get description
     */
    public get description(): string {
        return this.data.description;
    }

    /**
     * Set description
     * @param value
     */
    public set description(value: string) {
        this.data.description = value;
    }

    /**
     * Get description font size
     */
    public get descriptionFontSize(): number {
        return this.data.descriptionFontSize;
    }

    /**
     * Set description font size
     * @param value
     */
    public set descriptionFontSize(value: number) {
        this.data.descriptionFontSize = value;
    }

    /**
     * Get circle
     */
    public get circle(): PIXI.Circle {
        return this.data.circle || new PIXI.Circle(0, 0, 20);
    }

    /**
     * Set circle
     * @param value
     */
    public set circle(value: PIXI.Circle) {
        this.data.circle = value;

        // Elements
        // -- Blur element
        /* TODO
        this._blurElement.circle = new PIXI.Circle(this.circle.x, this.circle.y, this.circle.radius);
        */
        // -- Circle element
        this._circleElement.circle = this.circle;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options || {hasLine: true, lineColor: Design.roundButton.lineColor, lineAlpha: Design.roundButton.lineAlpha, hasFill: true, fillColor: Design.roundButton.fillColor, fillAlpha: Design.roundButton.fillAlpha};
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;

        // Elements
        // -- Blur element
        /* TODO
        this._blurElement.options = this.options;
        */
        // -- Circle element
        this._circleElement.options = this.options;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get fontSize
     */
    public get fontSize(): number {
        return this.data.fontSize || Design.roundButton.fontSize;
    }

    /**
     * Set fontSize
     * @param value
     */
    public set fontSize(value: number) {
        this.data.fontSize = value;

        // Elements
        // -- Icon element
        this._iconElement.style.fontSize = this.fontSize;
    }

    /**
     * Get fontColor
     */
    public get fontColor(): number {
        return this.data.fontColor || Design.roundButton.fontColor;
    }

    /**
     * Set fontColor
     * @param value
     */
    public set fontColor(value: number) {
        this.data.fontColor = value;

        // Elements
        // -- Icon element
        this._iconElement.style.fill = this.fontColor;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<RoundButtonElement> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<RoundButtonElement>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}