import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {BaseTilingSprite, BaseTilingSpriteImageModel} from "../../bases/elements/BaseTilingSprite";

export class AlphaSliderElementModel extends BaseContainerImageModel {
    color: number | undefined;
    colorAlpha: number | undefined;
    squareWidth: number = 20;
    squareHeight: number = 20;
}

/**
 * Alpha slider element
 */
export class AlphaSliderElement extends BaseContainer<AlphaSliderElementModel> {
    /**
     * Members
     */
    protected _canvas: HTMLCanvasElement;
    protected _canvasContext: CanvasRenderingContext2D;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<AlphaSliderElement>;

    /**
     * Elements
     */
    protected _transparentElement: BaseTilingSprite<BaseTilingSpriteImageModel>;
    protected _backgroundElement: PIXI.Sprite;
    protected _borderElement: MedsurfDraw.Rectangle;
    protected _sliderElement: MedsurfDraw.Rectangle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: AlphaSliderElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.alphaSliderElement.zIndex;

        //<editor-fold desc="Interactions">
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.onStartMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        this.on("pointerout", this.moveInteraction.endMove, this.moveInteraction);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        // Create canvas
        this._canvas = document.createElement("canvas");
        this._canvas.width = Design.alphaSliderElement.qualityWidth;
        this._canvas.height = Design.alphaSliderElement.qualityHeight;

        const ctx = this._canvas.getContext('2d');
        if (!ctx) {
            throw 'No canvas context created';
        }
        this._canvasContext = ctx;

        // Texture
        const texture = this.loader.resources['transparent'].texture;
        if (!texture || !texture.baseTexture) {
            throw 'No texture created';
        }
        texture.baseTexture.wrapMode = PIXI.WRAP_MODES.REPEAT;

        //<editor-fold desc="Elements">
        // -- Transparent element
        this._transparentElement = new BaseTilingSprite({image: this.image, texture});
        this._transparentElement.width = this.squareWidth;
        this._transparentElement.height = this.squareHeight;
        this.addChild(this._transparentElement);
        // -- Background element
        this._backgroundElement = new PIXI.Sprite();
        this._backgroundElement.width = this.squareWidth;
        this._backgroundElement.height = this.squareHeight;
        this.addChild(this._backgroundElement);
        // -- Border element
        this._borderElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, this.squareWidth, this.squareHeight),
            options: {hasLine: true, lineColor: Design.colorSliderElement.borderLineColor, lineAlpha: Design.colorSliderElement.borderLineAlpha, hasFill: false},
            lineWidth: Design.colorSliderElement.borderLineWidth
        });
        this.addChild(this._borderElement);
        // -- Slider element
        this._sliderElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, Design.alphaSliderElement.sliderWidth, this.squareHeight),
            options: {hasLine: false, hasFill: true, fillColor: Design.colorSliderElement.fillColor, fillAlpha: Design.colorSliderElement.fillAlpha},
            lineWidth: undefined
        });
        this.addChild(this._sliderElement);
        //</editor-fold>

        // Setup background
        this.createSliderBackground(this.color);
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Border element
        this._borderElement.draw();
        // -- Slider element
        this._sliderElement.draw();
        //</editor-fold>

        // Set position
        this._sliderElement.position.x = (this.squareWidth - Design.alphaSliderElement.sliderWidth) * (this.colorAlpha || 0);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
            this.off("pointerout", this.moveInteraction.endMove, this.moveInteraction);
            this.moveInteraction.off("startMove", this.onStartMove, this);
            this.moveInteraction.off("onMove", this.onMove, this);
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Transparent element
        if (this._transparentElement) {
            this._transparentElement.destroy(options);
        }
        // -- Background element
        if (this._backgroundElement) {
            this._backgroundElement.destroy(options);
        }
        // -- Border element
        if (this._borderElement) {
            this._borderElement.destroy(options)
        }
        // -- Slider element
        if (this._sliderElement) {
            this._sliderElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Create slider background
     * @param color
     */
    public createSliderBackground(color: number | undefined): void {
        if (color === undefined) {
            color = 0xff0000;
        }

        let r = ((color >> 16) & 0x0ff);
        let g = ((color >> 8) & 0x0ff);
        let b = (color & 0x0ff);

        this._canvasContext.clearRect(0, 0, Design.alphaSliderElement.qualityWidth, Design.alphaSliderElement.qualityHeight);
        const grad = this._canvasContext.createLinearGradient(0, 0, Design.alphaSliderElement.qualityWidth, 0);
        grad.addColorStop(0, 'rgba(' + r + ', ' + g + ', ' + b + ', 0)');
        grad.addColorStop(1, 'rgba(' + r + ', ' + g + ', ' + b + ', 1)');

        this._canvasContext.fillStyle = grad;
        this._canvasContext.fillRect(0, 0, Design.alphaSliderElement.qualityWidth, Design.alphaSliderElement.qualityHeight);

        this._backgroundElement.texture = PIXI.Texture.from(this._canvas);
        this._backgroundElement.texture.update();
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Emit alpha
     * @private
     */
    protected _emitAlpha(): void {
        // Get alpha
        this.colorAlpha = this._sliderElement.position.x / (this.squareWidth - Design.alphaSliderElement.sliderWidth);

        // Emit
        this.emit("alpha", this.colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On start move
     * @param event
     */
    public onStartMove(event: PIXI.InteractionEvent): void {
        const globalPosition = this.getGlobalPosition();

        this._sliderElement.position.x = Math.floor(event.data.global.x - globalPosition.x) / this.imageScale.x;
        if (this._sliderElement.position.x < 0) {
            this._sliderElement.position.x = 0;
        }
        if (this._sliderElement.position.x >= (this.squareWidth - Design.alphaSliderElement.sliderWidth)) {
            this._sliderElement.position.x = Design.alphaSliderElement.sliderWidth - 1;
        }

        this._emitAlpha();
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        this._sliderElement.position.x += dX;
        if (this._sliderElement.position.x < 0) {
            this._sliderElement.position.x = 0;
        }
        if (this._sliderElement.position.x >= (this.squareWidth - Design.alphaSliderElement.sliderWidth)) {
            this._sliderElement.position.x = (this.squareWidth - Design.alphaSliderElement.sliderWidth) - 1;
        }

        this._emitAlpha();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get loader
     */
    public get loader(): PIXI.Loader {
        return this.image.loader;
    }

    /**
     * Get color
     */
    public get color(): number | undefined {
        return this.data.color;
    }

    /**
     * Set color
     * @param value
     */
    public set color(value: number | undefined) {
        // Set value
        this.data.color = value;

        // Update background
        this.createSliderBackground(this.color);
    }

    /**
     * Get color alpha
     */
    public get colorAlpha(): number | undefined {
        return this.data.colorAlpha;
    }

    /**
     * Set color alpha
     * @param value
     */
    public set colorAlpha(value: number | undefined) {
        this.data.colorAlpha = value;
    }

    /**
     * Get squareWidth
     */
    public get squareWidth(): number {
        return this.data.squareWidth;
    }

    /**
     * Set squareWidth
     * @param value
     */
    public set squareWidth(value: number) {
        this.data.squareWidth = value;
    }

    /**
     * Get squareHeight
     */
    public get squareHeight(): number {
        return this.data.squareHeight;
    }

    /**
     * Set squareHeight
     * @param value
     */
    public set squareHeight(value: number) {
        this.data.squareHeight = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<AlphaSliderElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<AlphaSliderElement>) {
        this._moveInteraction = value;
    }
    //</editor-fold>
}