import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Color element model
 */
export class ColorElementModel extends BaseContainerImageModel {
    rectangle: PIXI.Rectangle = new PIXI.Rectangle(0, 0, 20, 20);
    color: number | undefined;
    colorAlpha: number | undefined;
    showAlpha: boolean;
}

/**
 * Color element
 */
export class ColorElement extends BaseContainer<ColorElementModel> {
    /**
     * Elements
     */
    protected _colorSquareElement: MedsurfDraw.ColorSquareElement;
    protected _alphaSliderElement: MedsurfDraw.AlphaSliderElement;
    protected _colorClearButtonElement: MedsurfDraw.ButtonElement;
    protected _colorSliderElement: MedsurfDraw.ColorSliderElement;
    protected _colorPaletteElement: MedsurfDraw.ColorPaletteElement;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ColorElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.colorElement.zIndex;

        // Position
        this.position.set(this.data.rectangle.x + this.data.rectangle.width, this.data.rectangle.y);
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        //<editor-fold desc="Elements">
        // -- Color square element
        this._colorSquareElement = new MedsurfDraw.ColorSquareElement({
            image: this.image,
            color: 0x000000,
            squareWidth: Design.colorElement.buttonWidth * Design.colorSquareElement.squareWidthAspect,
            squareHeight: Design.colorElement.buttonWidth * Design.colorSquareElement.squareWidthAspect
        });
        this._colorSquareElement.on("color", this.onSquareColor, this);
        this.addChild(this._colorSquareElement);
        // -- Alpha slider element
        this._alphaSliderElement = new MedsurfDraw.AlphaSliderElement({
            image: this.image,
            color: 0x000000,
            colorAlpha: 0,
            squareWidth: Design.colorElement.buttonWidth * Design.alphaSliderElement.squareWidthAspect,
            squareHeight: Design.colorElement.buttonWidth,
        });
        this._alphaSliderElement.on("alpha", this.onSliderAlpha, this);
        this.addChild(this._alphaSliderElement);
        // -- Color clear button element
        this._colorClearButtonElement = new MedsurfDraw.ButtonElement({
            image: this.image,
            text: '\uf5c7',
            description: "Farbe löschen",
            descriptionFontSize: 6,
            rectangle: new PIXI.Rectangle(0, 0, Design.colorElement.buttonWidth, Design.colorElement.buttonWidth),
            fontSize: Design.colorElement.fontSize
        });
        this._colorClearButtonElement.on("button", this.onClearButton, this);
        this.addChild(this._colorClearButtonElement);
        // -- Color slider element
        this._colorSliderElement = new MedsurfDraw.ColorSliderElement({
            image: this.image,
            color: 0x000000,
            squareWidth: Design.colorElement.buttonWidth * Design.colorSliderElement.squareWidthAspect,
            squareHeight: Design.colorElement.buttonWidth,
        });
        this._colorSliderElement.on("color", this.onSliderColor, this);
        this.addChild(this._colorSliderElement);
        // -- Color palette element
        this._colorPaletteElement = new MedsurfDraw.ColorPaletteElement({
            image: this.image,
            description: "Farbe wählen",
            descriptionFontSize: 6,
            localStorageKey: 'medsurf'
        });
        this._colorPaletteElement.on("color", this.onPaletteColor, this);
        this.addChild(this._colorPaletteElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Color square element
        this._colorSquareElement.position.set(this.rectangle.x + this.rectangle.width, this.rectangle.y - this.rectangle.height / 2);
        this._colorSquareElement.draw();
        // -- Alpha slider element
        this._alphaSliderElement.position.set(this._colorSquareElement.x, this._colorSquareElement.y + this._colorSquareElement.height + 5);
        this._alphaSliderElement.draw();
        // -- Color clear button element
        this._colorClearButtonElement.position.set(this._colorSquareElement.x + this._colorSquareElement.width + 5, this._colorSquareElement.y + this._colorSquareElement.height + 5);
        this._colorClearButtonElement.draw();
        // -- Color slider element
        this._colorSliderElement.position.set(this._colorSquareElement.x + this._colorSquareElement.width + 5, this._colorSquareElement.y);
        this._colorSliderElement.draw();
        // -- Color palette element
        this._colorPaletteElement.position.set(this._colorSliderElement.x + this._colorSliderElement.height + 5, this._colorSliderElement.y);
        this._colorPaletteElement.draw();
        //</editor-fold>

        // Pivot
        this.pivot.set(0, this.height / 4);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Color square element
        if (this._colorSquareElement) {
            this._colorSquareElement.off("color", this.onSquareColor, this);
            this._colorSquareElement.destroy(options);
        }
        // -- Alpha slider element
        if (this._alphaSliderElement) {
            this._alphaSliderElement.off("alpha", this.onSliderAlpha, this);
            this._alphaSliderElement.destroy(options);
        }
        // -- Color clear button element
        if (this._colorClearButtonElement) {
            this._colorClearButtonElement.off("button", this.onClearButton, this);
            this._colorClearButtonElement.destroy(options);
        }
        // -- Color slider element
        if (this._colorSliderElement) {
            this._colorSliderElement.off("color", this.onSliderColor, this);
            this._colorSliderElement.destroy(options);
        }
        // -- Color palette element
        if (this._colorPaletteElement) {
            this._colorPaletteElement.off("color", this.onPaletteColor, this);
            this._colorPaletteElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Color to hsl
     * @param color
     */
    public static colorToHSL(color: number | undefined) {
        if (color === undefined) {
            color = 0xFF0000;
        }

        // TODO remove console.log(((color >> 16) & 0x0ff), ((color >> 8) & 0x0ff), ((color) & 0x0ff));

        let r = ((color >> 16) & 0x0ff) / 255;
        let g = ((color >> 8) & 0x0ff) / 255;
        let b = ((color) & 0x0ff) / 255;

        const max = Math.max(r, g, b);
        const min = Math.min(r, g, b);

        let h: number;
        let s: number;
        let l: number = (max + min) / 2;
        let v = max;

        if (max == min) {
            h = s = 0; // achromatic
        } else {
            const d = max - min;
            s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
            switch(max){
                case r:
                    h = (g - b) / d + (g < b ? 6 : 0);
                    break;
                case g:
                    h = (b - r) / d + 2;
                    break;
                case b:
                    h = (r - g) / d + 4;
                    break;
            }

            // @ts-ignore
            h /= 6;
        }

        // @ts-ignore
        return {h: Math.round(360 * h), s: s, l: l, v: v};
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Save color
     * @param color
     * @param colorAlpha
     */
    public saveColor(color: number | undefined, colorAlpha: number | undefined) {
        this._colorPaletteElement.saveColor(color, colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Color">
    /**
     * On square color
     * @param color
     */
    public onSquareColor(color: number): void {
        // Update model but not with setter
        this.data.color = color;

        // Elements
        // -- Alpha slider element
        this._alphaSliderElement.createSliderBackground(this.color);
        this._alphaSliderElement.draw();

        // Emit
        this.emit("color", true, this.color, this.colorAlpha);
    }

    /**
     * On slider alpha
     * @param colorAlpha
     */
    public onSliderAlpha(colorAlpha: number): void {
        if (this.showAlpha) {
            // Update model but not with setter
            this.data.colorAlpha = colorAlpha;

            // Emit
            this.emit("color", true, this.color, this.colorAlpha);
        }
    }

    /**
     * On clear button
     * @param event
     */
    public onClearButton(event: PIXI.InteractionEvent): void {
        // Update model
        this.color = undefined;
        this.colorAlpha = undefined;

        // Elements
        // -- Color square element
        this._colorSquareElement.draw();
        // -- Alpha slider element
        this._alphaSliderElement.draw();
        // -- Color slider element
        this._colorSliderElement.draw();

        // Emit
        this.emit("color", false, this.color, this.colorAlpha);
    }

    /**
     * On slider color
     * @param color
     */
    public onSliderColor(color: number): void {
        // Update model
        this.color = color;

        // Elements
        // -- Color square element
        this._colorSquareElement.draw();
        // -- Alpha slider element
        this._alphaSliderElement.draw();

        // Emit
        this.emit("color", true, this.color, this.colorAlpha);
    }

    /**
     * on palette color
     * @param color
     * @param colorAlpha
     */
    public onPaletteColor(color: number, colorAlpha: number): void {
        // Update model but not with setter
        this.color = color;
        this.colorAlpha = colorAlpha;

        // Elements
        // -- Color square element
        this._colorSquareElement.draw();
        // -- Alpha slider element
        this._alphaSliderElement.draw();
        // -- Color slider element
        this._colorSliderElement.draw();

        // Emit
        this.emit("color", true, this.color, this.colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get color
     */
    public get color(): number | undefined {
        return this.data.color;
    }

    /**
     * Set color
     * @param value
     */
    public set color(value: number | undefined) {
        this.data.color = value;

        // Elements
        // -- Color square element
        this._colorSquareElement.color = this.color;
        // -- Alpha slider element
        this._alphaSliderElement.color = this.color;
        // -- Color slider element
        this._colorSliderElement.color = this.color;
    }

    /**
     * Get alpha color
     */
    public get colorAlpha(): number | undefined {
        return this.data.colorAlpha;
    }

    /**
     * Set alpha color
     * @param value
     */
    public set colorAlpha(value: number | undefined) {
        this.data.colorAlpha = value;

        // Elements
        // -- Alpha slider element
        this._alphaSliderElement.colorAlpha = this.colorAlpha;
    }

    /**
     * Get show alpha
     */
    public get showAlpha(): boolean {
        return this.data.showAlpha;
    }

    /**
     * Set show alpha
     * @param value
     */
    public set showAlpha(value: boolean) {
        this.data.showAlpha = value;

        // Elements
        // -- Alpha slider element
        this._alphaSliderElement.visible = this.showAlpha;
    }
    //</editor-fold>
}