import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

export class ColorPaletteElementModel extends BaseContainerImageModel {
    description: string;
    descriptionFontSize: number;
    localStorageKey: string;
}

/**
 * Color palette element
 */
export class ColorPaletteElement extends BaseContainer<ColorPaletteElementModel> {
    /**
     * Members
     */
    protected _selectionPalette: string[];

    /**
     * Elements
     */
    protected _presetButtons: MedsurfDraw.ColorButtonElement[];
    protected _storedButtons: MedsurfDraw.ColorButtonElement[];

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ColorPaletteElementModel) {
        // Parent constructor
        super(model);

        // Members
        this._selectionPalette = [];
        if (this.localStorageKey && window.localStorage) {
            try {
                this._selectionPalette = window.localStorage[this.localStorageKey].split(";");
            }
            catch (e) { }
        }

        // Setup
        this.zIndex = Design.colorPaletteElement.zIndex;
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        let dY = 0;

        //<editor-fold desc="Elements">
        // -- Preset buttons
        this._presetButtons = [];
        Design.colorPaletteElement.presetColors.forEach((color: number) => {
            const colorButtonElement = new MedsurfDraw.ColorButtonElement({
                image: this.image,
                description: this.description,
                descriptionFontSize: this.descriptionFontSize,
                loader: this.loader,
                hasFill: true,
                color: color,
                colorAlpha: 1,
                rectangle: new PIXI.Rectangle(0, 0, Design.colorPaletteElement.buttonWidth, Design.colorPaletteElement.buttonWidth),
            });
            colorButtonElement.on("button", this.onColor, this);
            this.addChild(colorButtonElement);

            const length = this._presetButtons.length;
            if (length !== 0 && length % 5 === 0) {
                dY++;
            }
            else if (length > 0) {
                const position = this._presetButtons[length - 1].position;
                colorButtonElement.position.x += position.x + Design.colorPaletteElement.buttonWidth + Design.colorPaletteElement.buttonCap;
            }
            colorButtonElement.position.y = dY * Design.colorPaletteElement.buttonWidth + dY * Design.colorPaletteElement.buttonCap;

            this._presetButtons.push(colorButtonElement);
        });
        // -- Stored buttons
        this._storedButtons = [];
        this._selectionPalette.forEach((selection: string) => {
            this._createButton(selection);
        });
        //</editor-fold>
    }

    /**
     * Draw
     */
    public draw(): void {
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Preset buttons
        this._presetButtons.forEach((button: MedsurfDraw.ColorButtonElement) => {
            button.off("button", this.onColor, this);
            button.destroy(options);
        });
        // -- Stored buttons
        this._storedButtons.forEach((button: MedsurfDraw.ColorButtonElement) => {
            button.off("button", this.onColor, this);
            button.destroy(options);
        });
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Save color
     * @param color
     * @param colorAlpha
     */
    public saveColor(color: number | undefined, colorAlpha: number | undefined) {
        if (color !== undefined && colorAlpha !== undefined) {
            const selection = ((color >> 16) & 0x0ff) + ',' + ((color >> 8) & 0x0ff) + ',' + ((color) & 0x0ff) + ',' + colorAlpha;
            if (this._selectionPalette.indexOf(selection) === -1 && !(Design.colorPaletteElement.presetColors.indexOf(color) > 0 && colorAlpha === 1)) {
                this._selectionPalette.push(selection);
                this._createButton(selection);
            }
        }
        if (this._selectionPalette.length > Design.colorPaletteElement.maxStoredValues) {
            this._storedButtons.forEach((button) => {
                this.removeChild(button);
                button.destroy();
            });
            this._storedButtons = [];

            this._selectionPalette.shift();
            this._selectionPalette.forEach((selection: string) => {
                this._createButton(selection);
            });
        }
        window.localStorage[this.localStorageKey] = this._selectionPalette.join(";");
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Create button
     * @param selection
     * @private
     */
    private _createButton(selection: string): void {
        const parts = selection.split(',');
        if (parts.length !== 4) {
            return;
        }

        const color = ((parseInt(parts[0]) & 0x0ff) << 16) | ((parseInt(parts[1]) & 0x0ff) << 8) | (parseInt(parts[2]) & 0x0ff);
        const colorAlpha = parseFloat(parts[3]);

        const colorButtonElement = new MedsurfDraw.ColorButtonElement({
            image: this.image,
            description: this.description,
            descriptionFontSize: this.descriptionFontSize,
            loader: this.loader,
            hasFill: true,
            color: color,
            colorAlpha: colorAlpha,
            rectangle: new PIXI.Rectangle(0, 0, Design.colorPaletteElement.buttonWidth, Design.colorPaletteElement.buttonWidth),
        });
        colorButtonElement.name = "selection";
        colorButtonElement.on("button", this.onColor, this);
        this.addChild(colorButtonElement);

        const length = this._storedButtons.length;
        const dY = Math.floor(length / 5) + 3;
        if (length > 0 && length % 5 !== 0) {
            const position = this._storedButtons[length - 1].position;
            colorButtonElement.position.x += position.x + Design.colorPaletteElement.buttonWidth + Design.colorPaletteElement.buttonCap;
        }
        colorButtonElement.position.y = dY * Design.colorPaletteElement.buttonWidth + dY * Design.colorPaletteElement.buttonCap;

        this._storedButtons.push(colorButtonElement);
    }
    //</editor-fold>

    //<editor-fold desc="Color">
    /**
     * On color
     * @param event
     * @param color
     * @param colorAlpha
     */
    public onColor(event: PIXI.InteractionEvent, color: number, colorAlpha: number): void {
        this.emit("color", color, colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get description
     */
    public get description(): string {
        return this.data.description;
    }

    /**
     * Set description
     * @param value
     */
    public set description(value: string) {
        this.data.description = value;
    }

    /**
     * Get description font size
     */
    public get descriptionFontSize(): number {
        return this.data.descriptionFontSize;
    }

    /**
     * Set description font size
     * @param value
     */
    public set descriptionFontSize(value: number) {
        this.data.descriptionFontSize = value;
    }

    /**
     * Get loader
     */
    public get loader(): PIXI.Loader {
        return this.image.loader;
    }

    /**
     * Get local storage key
     */
    public get localStorageKey(): string {
        return this.data.localStorageKey;
    }

    /**
     * Set local storage key
     * @param value
     */
    public set localStorageKey(value: string) {
        this.data.localStorageKey = value;
    }
    //</editor-fold>
}