import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Color slider element model
 */
export class ColorSliderElementModel extends BaseContainerImageModel {
    color: number | undefined;
    squareWidth: number = 20;
    squareHeight: number = 20;
}

/**
 * Color slider element
 */
export class ColorSliderElement extends BaseContainer<ColorSliderElementModel> {
    /**
     * Members
     */
    protected _canvas: HTMLCanvasElement;
    protected _canvasContext: CanvasRenderingContext2D;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<ColorSliderElement>;

    /**
     * Elements
     */
    protected _backgroundElement: PIXI.Sprite;
    protected _borderElement: MedsurfDraw.Rectangle;
    protected _sliderElement: MedsurfDraw.Rectangle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ColorSliderElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.colorSliderElement.zIndex;

        // Rotation
        this.rotation = Math.PI / 2;

        //<editor-fold desc="Interactions">
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.onStartMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        this.on("pointerout", this.moveInteraction.endMove, this.moveInteraction);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        // Create canvas
        this._canvas = document.createElement("canvas");
        this._canvas.width = Design.colorSliderElement.qualityWidth;
        this._canvas.height = Design.colorSliderElement.qualityHeight;

        const ctx = this._canvas.getContext('2d');
        if (!ctx) {
            throw 'No canvas context created';
        }
        this._canvasContext = ctx;

        //<editor-fold desc="Elements">
        // -- Background element
        this._backgroundElement = new PIXI.Sprite();
        this._backgroundElement.width = this.squareWidth;
        this._backgroundElement.height = this.squareHeight;
        this.addChild(this._backgroundElement);
        // -- Border element
        this._borderElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, this.squareWidth, this.squareHeight),
            options: {hasLine: true, lineColor: Design.colorSliderElement.borderLineColor, lineAlpha: Design.colorSliderElement.borderLineAlpha, hasFill: false},
            lineWidth: Design.colorSliderElement.borderLineWidth
        });
        this.addChild(this._borderElement);
        // -- Slider element
        this._sliderElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, Design.colorSliderElement.sliderWidth, this.squareHeight),
            options: {hasLine: false, hasFill: true, fillColor: Design.colorSliderElement.fillColor, fillAlpha: Design.colorSliderElement.fillAlpha},
            lineWidth: undefined
        });
        this.addChild(this._sliderElement);
        //</editor-fold>

        // Setup background
        this.createSliderBackground();
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Border element
        this._borderElement.draw();
        // -- Slider element
        this._sliderElement.draw();
        //</editor-fold>

        // Set position
        const hsl = MedsurfDraw.ColorElement.colorToHSL(this.color);
        this._sliderElement.position.x = this.squareWidth * (hsl.h / 360);

        // Pivot
        this.pivot.set(0, this.height);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
            this.off("pointerout", this.moveInteraction.endMove, this.moveInteraction);
            this.moveInteraction.off("startMove", this.onStartMove, this);
            this.moveInteraction.off("onMove", this.onMove, this);
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Background element
        if (this._backgroundElement) {
            this._backgroundElement.destroy(options);
        }
        // -- Border element
        if (this._borderElement) {
            this._borderElement.destroy(options)
        }
        // -- Slider element
        if (this._sliderElement) {
            this._sliderElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Create slider background
     */
    public createSliderBackground(): void {
        this._canvasContext.clearRect(0, 0, Design.colorSliderElement.qualityWidth, Design.colorSliderElement.qualityHeight);
        const grad = this._canvasContext.createLinearGradient(0, 0, Design.colorSliderElement.qualityWidth, 0);
        grad.addColorStop(0, '#ff0000');
        grad.addColorStop(0.17, '#ffff00');
        grad.addColorStop(0.33, '#00ff00');
        grad.addColorStop(0.5, '#00ffff');
        grad.addColorStop(0.67, '#0000ff');
        grad.addColorStop(0.83, '#ff00ff');
        grad.addColorStop(1, '#ff0000');

        this._canvasContext.fillStyle = grad;
        this._canvasContext.fillRect(0, 0, Design.colorSliderElement.qualityWidth, Design.colorSliderElement.qualityHeight);

        this._backgroundElement.texture = PIXI.Texture.from(this._canvas);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Emit color
     * @private
     */
    protected _emitColor(): void {
        // Get color
        const imageData = this._canvasContext.getImageData(
            this._sliderElement.position.x * (Design.colorSliderElement.qualityWidth / this.squareWidth),
            this._sliderElement.position.y * (Design.colorSliderElement.qualityHeight / this.squareHeight),
            1,
            1
        );
        this.color = ((imageData.data[0] & 0x0ff) << 16) | ((imageData.data[1] & 0x0ff) << 8) | (imageData.data[2] & 0x0ff);

        // Emit
        this.emit("color", this.color);
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On start move
     * @param event
     */
    public onStartMove(event: PIXI.InteractionEvent): void {
        // @ts-ignore TODO ignore
        const globalPosition = this.getGlobalPosition();

        this._sliderElement.position.x = Math.floor(event.data.global.y - globalPosition.y) / this.imageScale.x;
        // Elements
        // -- Slider element
        if (this._sliderElement.position.x < 0) {
            this._sliderElement.position.x = 0;
        }
        if (this._sliderElement.position.x >= this.squareWidth) {
            this._sliderElement.position.x = Design.colorSliderElement.sliderWidth - 1;
        }

        // Emit color
        this._emitColor();
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        const scale = this.imageScale;

        this._sliderElement.position.x += dY * scale.x / this.imageScale.x;
        if (this._sliderElement.position.x < 0) {
            this._sliderElement.position.x = 0;
        }
        if (this._sliderElement.position.x >= this.squareWidth) {
            this._sliderElement.position.x = this.squareWidth - 1;
        }

        // Emit color
        this._emitColor();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get color
     */
    public get color(): number | undefined {
        return this.data.color;
    }

    /**
     * Set color
     * @param value
     */
    public set color(value: number | undefined) {
        this.data.color = value;
    }

    /**
     * Get squareWidth
     */
    public get squareWidth(): number {
        return this.data.squareWidth;
    }

    /**
     * Set squareWidth
     * @param value
     */
    public set squareWidth(value: number) {
        this.data.squareWidth = value;
    }

    /**
     * Get squareHeight
     */
    public get squareHeight(): number {
        return this.data.squareHeight;
    }

    /**
     * Set squareHeight
     * @param value
     */
    public set squareHeight(value: number) {
        this.data.squareHeight = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<ColorSliderElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<ColorSliderElement>) {
        this._moveInteraction = value;
    }
    //</editor-fold>
}