import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

export class ColorSquareElementModel extends BaseContainerImageModel {
    color: number | undefined;
    squareWidth: number = 20;
    squareHeight: number = 20;
}

/**
 * Color square element
 */
export class ColorSquareElement extends BaseContainer<ColorSquareElementModel> {
    /**
     * Members
     */
    protected _canvas: HTMLCanvasElement;
    protected _canvasContext: CanvasRenderingContext2D;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<ColorSquareElement>;

    /**
     * Elements
     */
    protected _backgroundElement: PIXI.Sprite;
    protected _borderElement: MedsurfDraw.Rectangle;
    protected _pointElement: MedsurfDraw.Circle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ColorSquareElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.colorSquareElement.zIndex;

        //<editor-fold desc="Interactions">
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.onStartMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        this.on("pointerout", this.moveInteraction.endMove, this.moveInteraction);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        // Create canvas
        this._canvas = document.createElement("canvas");
        this._canvas.width = Design.colorSquareElement.qualityWidth;
        this._canvas.height = Design.colorSquareElement.qualityHeight;

        const ctx = this._canvas.getContext('2d');
        if (!ctx) {
            throw 'No canvas context created';
        }
        this._canvasContext = ctx;

        //<editor-fold desc="Elements">
        // -- Background element
        this._backgroundElement = new PIXI.Sprite();
        this._backgroundElement.width = this.squareWidth;
        this._backgroundElement.height = this.squareHeight;
        this.addChild(this._backgroundElement);
        // -- Border element
        this._borderElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, this.squareWidth, this.squareHeight),
            options: {hasLine: true, lineColor: Design.colorSquareElement.borderLineColor, lineAlpha: Design.colorSquareElement.borderLineAlpha, hasFill: false},
            lineWidth: Design.colorSquareElement.borderLineWidth
        });
        this.addChild(this._borderElement);
        // -- Point element
        this._pointElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.colorSquareElement.circleRadius),
            options: {hasLine: true, lineColor: Design.colorSquareElement.lineColor, lineAlpha: Design.colorSquareElement.lineAlpha, hasFill: true, fillColor: Design.colorSquareElement.fillColor, fillAlpha: Design.colorSquareElement.fillAlpha},
            lineWidth: Design.colorSquareElement.lineWidth
        });
        this.addChild(this._pointElement);
        //</editor-fold>

        // Setup background
        this.createSliderBackground(MedsurfDraw.ColorElement.colorToHSL(this.color).h);
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Border element
        this._borderElement.draw();
        // -- Point element
        this._pointElement.draw();
        //</editor-fold>

        // Set position
        const hsl = MedsurfDraw.ColorElement.colorToHSL(this.color);
        this._pointElement.position.x = this.squareWidth * hsl.s;
        this._pointElement.position.y = this.squareHeight - (this.squareHeight * hsl.v);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
            this.off("pointerout", this.moveInteraction.endMove, this.moveInteraction);
            this.moveInteraction.off("startMove", this.onStartMove, this);
            this.moveInteraction.off("onMove", this.onMove, this);
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Background element
        if (this._backgroundElement) {
            this._backgroundElement.destroy(options);
        }
        // -- Border element
        if (this._borderElement) {
            this._borderElement.destroy(options)
        }
        // -- Point element
        if (this._pointElement) {
            this._pointElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Create slider background
     * @param hue
     */
    public createSliderBackground(hue: number): void {
        this._canvasContext.clearRect(0, 0, Design.colorSquareElement.qualityWidth, Design.colorSquareElement.qualityHeight);
        for (let row = 0; row < Design.colorSquareElement.qualityHeight; row++) {
            const grad = this._canvasContext.createLinearGradient(0, 0, Design.colorSquareElement.qualityWidth, 0);
            grad.addColorStop(0, 'hsl(' + hue + ', 0%, ' + ((Design.colorSquareElement.qualityHeight - row) / Design.colorSquareElement.qualityHeight * 100) + '%)');
            grad.addColorStop(1, 'hsl(' + hue + ', 100%, ' + ((Design.colorSquareElement.qualityHeight / 2 - row / 2) / Design.colorSquareElement.qualityHeight * 100) + '%)');
            this._canvasContext.fillStyle = grad;
            this._canvasContext.fillRect(0, row, Design.colorSquareElement.qualityWidth, 1);
        }
        this._backgroundElement.texture = PIXI.Texture.from(this._canvas);
        this._backgroundElement.texture.update();
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Emit color
     * @private
     */
    protected _emitColor(): void {
        // Get color
        const imageData = this._canvasContext.getImageData(
            this._pointElement.position.x * (Design.colorSquareElement.qualityWidth / this.squareWidth),
            this._pointElement.position.y * (Design.colorSquareElement.qualityHeight / this.squareHeight),
            1,
            1
        );
        this.color = ((imageData.data[0] & 0x0ff) << 16) | ((imageData.data[1] & 0x0ff) << 8) | (imageData.data[2] & 0x0ff);

        // Emit
        this.emit("color", this.color);
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On start move
     * @param event
     */
    public onStartMove(event: PIXI.InteractionEvent): void {
        // @ts-ignore TODO ignore
        const globalPosition = this.getGlobalPosition();

        this._pointElement.position.x = Math.floor(event.data.global.x - globalPosition.x) / this.imageScale.x;
        if (this._pointElement.position.x < 0) {
            this._pointElement.position.x = 0;
        }
        if (this._pointElement.position.x >= this.squareWidth) {
            this._pointElement.position.x = this.squareWidth - 1;
        }

        this._pointElement.position.y = Math.floor(event.data.global.y - globalPosition.y) / this.imageScale.x;
        if (this._pointElement.position.y < 0) {
            this._pointElement.position.y = 0;
        }
        if (this._pointElement.position.y >= this.squareHeight) {
            this._pointElement.position.y = this.squareHeight - 1;
        }

        this._emitColor();
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        const scale = this.imageScale;

        this._pointElement.position.x += dX; // * scale.x / this.getImageScale().x;
        if (this._pointElement.position.x < 0) {
            this._pointElement.position.x = 0;
        }
        if (this._pointElement.position.x >= this.squareWidth) {
            this._pointElement.position.x = this.squareWidth - 1;
        }

        this._pointElement.position.y += dY; // * scale.x / this.getImageScale().x;
        if (this._pointElement.position.y < 0) {
            this._pointElement.position.y = 0;
        }
        if (this._pointElement.position.y >= this.squareHeight) {
            this._pointElement.position.y = this.squareHeight - 1;
        }

        this._emitColor();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get color
     */
    public get color(): number | undefined {
        return this.data.color;
    }

    /**
     * Set color
     * @param value
     */
    public set color(value: number | undefined) {
        // Set value
        this.data.color = value;

        // Update background
        this.createSliderBackground(MedsurfDraw.ColorElement.colorToHSL(this.color).h);
    }

    /**
     * Get squareWidth
     */
    public get squareWidth(): number {
        return this.data.squareWidth;
    }

    /**
     * Set squareWidth
     * @param value
     */
    public set squareWidth(value: number) {
        this.data.squareWidth = value;
    }

    /**
     * Get squareHeight
     */
    public get squareHeight(): number {
        return this.data.squareHeight;
    }

    /**
     * Set squareHeight
     * @param value
     */
    public set squareHeight(value: number) {
        this.data.squareHeight = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<ColorSquareElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<ColorSquareElement>) {
        this._moveInteraction = value;
    }
    //</editor-fold>
}