import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {
    BaseElementContainer,
    BaseElementContainerModel,
    ColorMode,
    NumberMode
} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";

/**
 * Line types
 */
export enum LineTypes {
    NOT_INITIALIZED,
    LINEAR,
    QUADRATIC,
    BEZIER,
}

/**
 * Line direction
 */
export enum LineDirection {
    START_END,
    END_START,
}

/**
 * Graphic line
 */
export class GraphicLine {
    type: LineTypes;
    start: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null;
    end: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null;
    direction: LineDirection;
    interactWithLines: boolean;
    controlStart: PIXI.IPoint | null;
    controlEnd: PIXI.IPoint | null;
}

/**
 * Control points
 */
export class ControlPoints {
    c1: PIXI.IPoint;
    c2: PIXI.IPoint;
    l1: number;
    l2: number;
}

/**
 * Line
 */
export class Line extends BaseElementContainer<Models.Line, BaseElementContainerModel<Models.Line>> {
    /**
     * Members
     */
    private _type: LineTypes;
    private _direction: LineDirection;
    private _start: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null;
    private _end: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null;
    private _controlStart: PIXI.IPoint | null;
    private _controlEnd: PIXI.IPoint | null;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<Line>;
    private _selectInteraction: MedsurfDraw.SelectInteraction<Line>;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: any;
    protected _debounceControlUnselectItemMethod: any;

    /**
     * Elements
     */
    protected _lineElement: MedsurfDraw.LineObject | undefined;
    protected _lineCornerElement: MedsurfDraw.LineCorner;
    protected _lineControlPointElement: MedsurfDraw.LineControlPoint;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.Line) {
        // Parent constructor
        super({image, model});

        // Members
        this.start = null;
        this.end = null;

        // Setup
        this.zIndex = Design.line.zIndexDefault + this.model.zIndex;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("remove-init", this._removeModeInit, this);
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on('select_layer_group', this._modeSelectLayerGroup, this);
        this.modeInteraction.on('remove-select_layer_group', this._removeModeSelectLayerGroup, this);
        this.modeInteraction.on('select_grouping', this._modeSelectGrouping, this);
        this.modeInteraction.on('remove-select_grouping', this._removeModeSelectGrouping, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('draw_item', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item", this._removeModeDrawItem, this);
        this.modeInteraction.on('choose_select', this._modeChooseSelect, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeChooseSelect, this);
        this.modeInteraction.on('choose_unselect', this._modeChooseUnselect, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeChooseUnselect, this);
        this.modeInteraction.on('clone_item', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item", this._removeModeCloneItem, this);
        this.modeInteraction.on('clone_item_new', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item_new", this._removeModeCloneItem, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        this.modeInteraction.on("delete_multiple", this._modeDelete, this);
        this.modeInteraction.on('delete_draw', this._modeDelete, this);
        this.modeInteraction.on('delete_clone', this._modeDelete, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.startMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.moveInteraction.on("endMove", this.endMove, this);
        // TODO this.moveInteraction.on("onRelease", this.onRelease, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelectStart", this.onSelectStart, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        this._debounceControlUnselectItemMethod = debounce(this.controlUnselectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Move
            this.on("onPointUpdate", this.onPointUpdate, this);
            this.on("startMove", this.showControlPoints, this); // TODO not working
            this.on("endMove", this.hideControlPoints, this); // TODO not working
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.on("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Move
            this.off("onPointUpdate", this.onPointUpdate, this);
            this.off("startMove", this.showControlPoints, this); // TODO not working
            this.off("endMove", this.hideControlPoints, this); // TODO not working
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.off("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Get image
        const image = this.image;

        // Members
        this.start = image.getChildByName(this.model.start) as BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>;
        this.end = image.getChildByName(this.model.end) as BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>;

        // Events
        if (this.start instanceof MedsurfDraw.PositionPoint) {
            this.start.on("onPositionPointMove", this.onMove, this);
            this.start.moveInteraction.on("startMove", this.startMove, this); // TODO not working
            this.start.moveInteraction.on("endMove", this.endMove, this); // TODO not working
            // this.start.moveInteraction.on("startMove", this.showControlPoints, this); // TODO not working
            // this.start.moveInteraction.on("endMove", this.hideControlPoints, this); // TODO not working
        } else {
            this.start.on("onFillMove", this.onMove, this);
            this.start.on("draw", this.onPointUpdate, this);
        }
        this.start.on("removed", this.onButtonDelete, this);

        if (this.end instanceof MedsurfDraw.PositionPoint) {
            this.end.on("onPositionPointMove", this.onMove, this);
            this.end.moveInteraction.on("startMove", this.startMove, this); // TODO not working
            this.end.moveInteraction.on("endMove", this.endMove, this); // TODO not working
            // this.end.moveInteraction.on("startMove", this.showControlPoints, this); // TODO not working
            // this.end.moveInteraction.on("endMove", this.hideControlPoints, this); // TODO not working
        } else {
            this.end.on("onFillMove", this.onMove, this);
            this.end.on("draw", this.onPointUpdate, this);
        }
        this.end.on("removed", this.onButtonDelete, this);

        //<editor-fold desc="Elements">
        // -- Line corner element
        this._lineCornerElement = new MedsurfDraw.LineCorner({
            image: this.image,
            points: undefined,
            options: Object.assign({}, this.model.options),
            lineWidth: Math.round(this.model.strokeWidth * this.imageDimensions.width * 100) / 100,
        });
        this.addChild(this._lineCornerElement);
        // -- Line control point element
        this._lineControlPointElement = new MedsurfDraw.LineControlPoint({
            image: this.image,
            points: undefined,
            options: {
                hasLine: true,
                lineColor: Design.line.bezierLineColor,
                lineAlpha: Design.line.bezierLineAlpha,
                hasFill: true,
                fillColor: Design.line.bezierPointColor,
                fillAlpha: Design.line.bezierPointAlpha
            },
            lineWidth: Design.line.bezierLineWidth,
            radius: Design.line.bezierPointRadius
        });
        this._lineControlPointElement.hideItem();
        this.addChild(this._lineControlPointElement);
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>
    }

    public startDrawing() {
        super.startDrawing();

        //<editor-fold desc="Events">
        // Events
        this.moveInteraction.on("onMove", this.onMove, this);
        if (this.start) {
            if (this.start instanceof MedsurfDraw.PositionPoint) {
                this.start.on("onPositionPointMove", this.onMove, this);
                this.start.moveInteraction.on("startMove", this.startMove, this); // TODO not working
                this.start.moveInteraction.on("endMove", this.endMove, this); // TODO not working
                // this.start.moveInteraction.on("startMove", this.showControlPoints, this); // TODO not working
                // this.start.moveInteraction.on("endMove", this.hideControlPoints, this); // TODO not working
            } else {
                this.start.on("onFillMove", this.onMove, this);
                this.start.on("draw", this.onPointUpdate, this);
            }
            // this.start.on("removed", this.onButtonDelete, this);
        }
        if (this.end) {
            if (this.end instanceof MedsurfDraw.PositionPoint) {
                this.end.on("onPositionPointMove", this.onMove, this);
                this.end.moveInteraction.on("startMove", this.startMove, this); // TODO not working
                this.end.moveInteraction.on("endMove", this.endMove, this); // TODO not working
                // this.end.moveInteraction.on("startMove", this.showControlPoints, this); // TODO not working
                // this.end.moveInteraction.on("endMove", this.hideControlPoints, this); // TODO not working
            } else {
                this.end.on("onFillMove", this.onMove, this);
                this.end.on("draw", this.onPointUpdate, this);
            }
            // this.end.on("removed", this.onButtonDelete, this);
        }
        //</editor-fold>

        // Emit
        this.emit("debounceDraw");
    }

    public stopDrawing(selfheal_timeout: number) {
        super.stopDrawing(selfheal_timeout);

        //<editor-fold desc="Events">
        this.moveInteraction.off("onMove", this.onMove, this);
        if (this.start) {
            this.start.off("onPositionPointMove", this.onMove, this);
            if (this.start instanceof MedsurfDraw.PositionPoint) {
                // this.start.moveInteraction.off("startMove", this.startMove, this); // TODO not working
                // this.start.moveInteraction.off("startMove", this.showControlPoints, this); // TODO not working
                this.start.moveInteraction.off("onMove", this.onMove, this);
                // this.start.moveInteraction.off("endMove", this.endMove, this); // TODO not working
                // this.start.moveInteraction.off("endMove", this.hideControlPoints, this); // TODO not working
            }
            // this.start.off("removed", this.onButtonDelete, this);
            this.start.off("onFillMove", this.onMove, this);
            this.start.off("draw", this.onPointUpdate, this);
        }
        if (this.end) {
            this.end.off("onPositionPointMove", this.onMove, this);
            if (this.end instanceof MedsurfDraw.PositionPoint) {
                // this.end.moveInteraction.off("startMove", this.startMove, this); // TODO not working
                // this.end.moveInteraction.off("startMove", this.showControlPoints, this); // TODO not working
                this.end.moveInteraction.off("onMove", this.onMove, this);
                // this.end.moveInteraction.off("endMove", this.endMove, this); // TODO not working
                // this.end.moveInteraction.off("endMove", this.hideControlPoints, this); // TODO not working
            }
            // this.end.off("removed", this.onButtonDelete, this);
            this.end.off("onFillMove", this.onMove, this);
            this.end.off("draw", this.onPointUpdate, this);
        }
        /*
        this.lines.forEach((line) => {
            line.off("onPointUpdate", this.draw, this);
        });
        */
        //</editor-fold>
    }

    /**
     * Draw
     */
    public draw(): void {
        // Elements
        // -- Line element
        if (this._lineElement || this.type === LineTypes.NOT_INITIALIZED) {
            // Store type
            const oldType = this.type;

            // Get line model
            try {
                const points = this._getLinePointsModel();
                if (points) {
                    // -- Line corner element
                    this._lineCornerElement.points = points;
                    // -- Line control point element
                    this._lineControlPointElement.points = points;

                    // Check type change
                    if (this.type !== oldType) {
                        // Elements
                        // -- Line element
                        if (this._lineElement) {
                            this._lineElement.destroy();
                        }


                        // Get line object
                        const result = this._getLineObject(points);
                        if (result) {
                            this._lineElement = result;
                            this.addChild(this._lineElement);
                        }
                    } else {
                        // -- Line element
                        if (this._lineElement) {
                            this._lineElement.points = points;
                        }
                    }

                    // Elements
                    // -- Line element
                    if (this._lineElement) {
                        this._lineElement.draw();
                    }
                    // -- Line corner element
                    this._lineCornerElement.renderable = true;
                    // -- Line control point element
                    this._lineControlPointElement.renderable = true;
                } else {
                    // Elements
                    // -- Line element
                    if (this._lineElement) {
                        this._lineElement.destroy();
                        delete this._lineElement;
                        this.type = LineTypes.NOT_INITIALIZED;
                    }
                    // -- Line corner element
                    if (this._lineCornerElement) {
                        this._lineCornerElement.renderable = false;
                    }
                    // -- Line control point element
                    if (this._lineControlPointElement) {
                        this._lineControlPointElement.renderable = false;
                    }
                }
            }
            catch (e) {
                // Stop drawing with 2 sec selfheal
                this.stopDrawing(2000);
            }
        }
        // -- Line corner element
        if (this._lineCornerElement) {
            this._lineCornerElement.draw();
        }
        // -- Line control point element
        if (this._lineControlPointElement) {
            this._lineControlPointElement.draw();
        }
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Events">
        if (this.start) {
            this.start.off("onPositionPointMove", this.onMove, this);
            if (this.start instanceof MedsurfDraw.PositionPoint) {
                this.start.moveInteraction.off("startMove", this.startMove, this); // TODO not working
                // this.start.moveInteraction.off("startMove", this.showControlPoints, this); // TODO not working
                this.start.moveInteraction.off("onMove", this.onMove, this);
                this.start.moveInteraction.off("endMove", this.endMove, this); // TODO not working
                // this.start.moveInteraction.off("endMove", this.hideControlPoints, this); // TODO not working
            }
            this.start.off("onFillMove", this.onMove, this);
            this.start.off("removed", this.onButtonDelete, this);
            this.start.off("draw", this.onPointUpdate, this);
        }

        if (this.end) {
            this.end.off("onPositionPointMove", this.onMove, this);
            if (this.end instanceof MedsurfDraw.PositionPoint) {
                this.end.moveInteraction.off("startMove", this.startMove, this); // TODO not working
                // this.end.moveInteraction.off("startMove", this.showControlPoints, this); // TODO not working
                this.end.moveInteraction.off("onMove", this.onMove, this);
                this.end.moveInteraction.off("endMove", this.endMove, this); // TODO not working
                // this.end.moveInteraction.off("endMove", this.hideControlPoints, this); // TODO not working
            }
            this.end.off("onFillMove", this.onMove, this);
            this.end.off("removed", this.onButtonDelete, this);
            this.end.off("draw", this.onPointUpdate, this);
        }
        //</editor-fold>

        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Line element
        if (this._lineElement) {
            this._lineElement.destroy(options);
        }
        // -- Line corner element
        if (this._lineCornerElement) {
            this._lineCornerElement.destroy(options);
        }
        // -- Line control point element
        if (this._lineControlPointElement) {
            this._lineControlPointElement.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param startPositionPoint
     * @param endPositionPoint
     * @param bezierCourve
     * @param imageWidth
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, startPositionPoint: string,
                              endPositionPoint: string, bezierCourve: boolean, imageWidth: number): Line {
        // Element
        // -- Line
        return new Line(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LINE,
            isStatic: false,
            zIndex: 0,
            start: startPositionPoint,
            end: endPositionPoint,
            isBezier: bezierCourve,
            interactWithLines: true,
            strokeWidth: 4 / imageWidth,
            options: {
                hasLine: true,
                lineColor: 0xffffff,
                hasFill: false
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Public function">
    /**
     * Has position point
     * @param point
     */
    public hasPositionPoint(point: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection): boolean {
        return point === this.start || point === this.end;
    }
    //</editor-fold>

    //<editor-fold desc="Private function">
    /**
     * Get line result
     * @private
     */
    protected _getLineObject(points: MedsurfDraw.StraightLinePointsModel | MedsurfDraw.QuadraticCurvePointsModel | MedsurfDraw.BezierLinePointsModel): MedsurfDraw.LineObject | undefined {
        const model = {
            points: points,
            options: Object.assign({}, this.model.options),
            lineWidth: Math.round(this.model.strokeWidth * this.imageDimensions.width * 100) / 100,
        };
        switch (this.type) {
            case LineTypes.LINEAR:
                return new MedsurfDraw.StraightLine(model as MedsurfDraw.StraightLineModel);
            case LineTypes.QUADRATIC:
                return new MedsurfDraw.QuadraticCurve(model as MedsurfDraw.QuadraticCurveModel);
            case LineTypes.BEZIER:
                return new MedsurfDraw.BezierLine(model as MedsurfDraw.BezierLineModel);
        }
        return undefined;
    }

    /**
     * Get line points model
     * @private
     */
    protected _getLinePointsModel(): MedsurfDraw.StraightLinePointsModel | MedsurfDraw.QuadraticCurvePointsModel | MedsurfDraw.BezierLinePointsModel | undefined {
        if (!this.start || !this.end || !this.parent) {
            return;
        }
        try {
            if (isNaN(this.start.x) || isNaN(this.start.y) || isNaN(this.end.x) || isNaN(this.start.y)) {
                return;
            }
        } catch (e) {
            // Stop drawing with 20 sec selfheal
            this.stopDrawing(20000);

            // Prevent error when positionPoint was deleted
            if (!this.start.destroyed && !this.end.destroyed) {
                // Send an error to the nav
                this.image.sendError(e);
            }

            // TODO delete when no start or end ???
            // Set mode
            // this.modeInteraction.setMode("delete");

            return;
        }

        this.type = LineTypes.NOT_INITIALIZED;
        this.direction = LineDirection.START_END;
        this.controlStart = null;
        this.controlEnd = null;

        if (this.model.isBezier) {
            const image = this.image;

            let startLines: MedsurfDraw.Line[] = [];
            if (this.start instanceof MedsurfDraw.PositionPoint) {
                startLines = image.getLinesByPositionPoint(this.start).filter((line: Line) => line !== this && line.model.interactWithLines);
            }

            let endLines: MedsurfDraw.Line[] = [];
            if (this.end instanceof MedsurfDraw.PositionPoint) {
                endLines = image.getLinesByPositionPoint(this.end).filter((line: Line) => line !== this && line.model.interactWithLines);
            }

            if (startLines.length == 0 && endLines.length == 0) {
                // No matching control points
                const points = this._getStraightLinePointsModel(this.start, this.end);
                if (points) {
                    this.type = LineTypes.LINEAR;
                    return points;
                }
            }
            else if (startLines.length == 0 || endLines.length == 0) {
                // Quadratic curve
                const points = this._getQuadraticCurvePointsModel(this.start, this.end, startLines, endLines);
                if (points) {
                    this.type = LineTypes.QUADRATIC;
                    this.controlStart = points.controlPointsPrevious.c2;
                    return points;
                }
            }
            else {
                // Bezier curve
                const points = this._getBezierLinePointsModel(this.start as MedsurfDraw.PositionPoint, this.end as MedsurfDraw.PositionPoint, startLines, endLines);
                if (points) {
                    this.type = LineTypes.BEZIER;
                    this.controlStart = points.controlPointsPrevious.c2;
                    this.controlEnd = points.controlPointsNext.c1;
                    return points;
                }
            }

            // TODO here we need some emits/slots to get the update stuff !!!!
            /* TODO remove
            startLines.forEach((line) => {
                console.log(line);
            });
            */
        } else {
            // Straight line
            const points = this._getStraightLinePointsModel(this.start, this.end);
            if (points) {
                this.type = LineTypes.LINEAR;
                return points;
            }
        }
        return undefined;
    }

    /**
     * Get straight line
     * @param start
     * @param end
     * @private
     */
    protected _getStraightLinePointsModel(start: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>, end: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>): MedsurfDraw.StraightLinePointsModel | undefined {
        let startPoint: PIXI.IPoint | undefined = undefined;
        let endPoint: PIXI.IPoint | undefined = undefined;

        let startSet: boolean = false;
        if (start instanceof MedsurfDraw.PositionPoint) {
            try {
                startPoint = start.position;
                startSet = true;
            } catch (e) {
                return undefined;
            }
        }

        let endSet: boolean = false;
        if (end instanceof MedsurfDraw.PositionPoint) {
            try {
                endPoint = end.position;
                endSet = true;
            } catch (e) {
                return undefined;
            }
        }

        if (!startSet) {
            if (!endSet) {
                throw 'Not implemented';
            } else {
                if (endPoint === undefined) {
                    return undefined;
                }
                if ((start as MedsurfDraw.FillCollection).isPointInPolygon(endPoint)) {
                    return undefined;
                }
                startPoint = (start as MedsurfDraw.FillCollection).getNearestPointForPoint(endPoint);
            }
        } else if (!endSet) {
            if (startPoint === undefined) {
                return undefined;
            }
            if ((end as MedsurfDraw.FillCollection).isPointInPolygon(startPoint)) {
                return undefined;
            }
            endPoint = (end as MedsurfDraw.FillCollection).getNearestPointForPoint(startPoint);
        }

        // @ts-ignore
        return {start: startPoint, end: endPoint};
    }

    /**
     * Get quadratic curve
     * @param start
     * @param end
     * @param startLines
     * @param endLines
     * @private
     */
    protected _getQuadraticCurvePointsModel(start: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>, end: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>, startLines: MedsurfDraw.Line[], endLines: MedsurfDraw.Line[]): MedsurfDraw.QuadraticCurvePointsModel {
        // Quadratic curve
        let controlPoint: PIXI.IPoint;

        let controlPointsPrevious;
        if (startLines.length == 0) {
            let startPoint: PIXI.IPoint;
            if (start instanceof MedsurfDraw.FillCollection) {
                startPoint = start.getNearestPointForPoint(end.position);
            } else {
                startPoint = start.position;
            }

            this.direction = LineDirection.END_START;

            if (endLines.length > 1) {
                let controlX = 0;
                let controlY = 0;
                endLines.forEach((line: Line) => {
                    if (!line.start || !line.end) {
                        return;
                    }

                    // TODO remove const lineStartElement = this.getImage().getChildByName(line.graphicLine.start.id || "new") as PositionPoint;
                    // TODO remove const lineEndElement = this.getImage().getChildByName(line.graphicLine.end.id || "new") as PositionPoint;
                    if (line.start == start || line.start == end) {
                        controlX += line.end.x;
                        controlY += line.end.y;
                    } else {
                        controlX += line.start.x;
                        controlY += line.start.y;
                    }
                });
                controlPoint = new PIXI.Point(
                    controlX / endLines.length,
                    controlY / endLines.length
                );
            } else {
                if (endLines[0].start == start || endLines[0].start == end) {
                    if (!endLines[0].end) {
                        throw "error";
                    }

                    // TODO remove const endEndLinesElement = this.getImage().getChildByName(endLines[0].graphicLine.end.id || "new") as PositionPoint;
                    controlPoint = new PIXI.Point(
                        endLines[0].end.x,
                        endLines[0].end.y
                    );
                } else {
                    if (!endLines[0].start) {
                        throw "error";
                    }

                    // TODO remove const startEndLinesElement = this.getImage().getChildByName(endLines[0].graphicLine.start.id || "new") as PositionPoint;
                    controlPoint = new PIXI.Point(
                        endLines[0].start.x,
                        endLines[0].start.y
                    );
                }
            }
            controlPointsPrevious = Line._findControlPoints(controlPoint, end.position, startPoint);

            return {
                start: end.position,
                end: startPoint,
                controlPoint,
                controlPointsPrevious
            };
        }

        let endPoint: PIXI.IPoint;
        if (end instanceof MedsurfDraw.FillCollection) {
            endPoint = end.getNearestPointForPoint(start.position);
        } else {
            endPoint = end.position;
        }

        if (startLines.length > 1) {
            let controlX = 0;
            let controlY = 0;
            startLines.forEach((line: Line) => {
                if (!line.start || !line.end) {
                    return;
                }

                // TODO remove const lineStartElement = this.getImage().getChildByName(line.graphicLine.start.id || "new") as PositionPoint;
                // TODO remove const lineEndElement = this.getImage().getChildByName(line.graphicLine.end.id || "new") as PositionPoint;
                if (line.start == this.start || line.start == this.end) {
                    controlX += line.end.x;
                    controlY += line.end.y;
                } else {
                    controlX += line.start.x;
                    controlY += line.start.y;
                }
            });
            controlPoint = new PIXI.Point(
                controlX / startLines.length,
                controlY / startLines.length
            );
        } else {
            if (startLines[0].start == this.start || startLines[0].start == this.end) {
                if (!startLines[0].end) {
                    throw "error";
                }

                // TODO remove const endStartLinesElement = this.getImage().getChildByName(startLines[0].graphicLine.end.id || "new") as PositionPoint;
                controlPoint = new PIXI.Point(
                    startLines[0].end.x,
                    startLines[0].end.y
                );
            } else {
                if (!startLines[0].start) {
                    throw "error";
                }

                // TODO remove const startStartLinesElement = this.getImage().getChildByName(startLines[0].graphicLine.start.id || "new") as PositionPoint;
                controlPoint = new PIXI.Point(
                    startLines[0].start.x,
                    startLines[0].start.y
                );
            }
        }
        controlPointsPrevious = Line._findControlPoints(controlPoint, start.position, endPoint);

        return {
            start: start.position,
            end: endPoint,
            controlPoint,
            controlPointsPrevious
        };
    }

    /**
     * Get bezier line
     * @param start
     * @param end
     * @param startLines
     * @param endLines
     * @private
     */
    protected _getBezierLinePointsModel(start: MedsurfDraw.PositionPoint, end: MedsurfDraw.PositionPoint, startLines: MedsurfDraw.Line[], endLines: MedsurfDraw.Line[]): MedsurfDraw.BezierLinePointsModel {
        // Bezier curve
        let controlPointPrevious: PIXI.IPoint;

        if (startLines.length > 1) {
            let controlX = 0;
            let controlY = 0;
            startLines.forEach((line: Line) => {
                if (!line.start || !line.end) {
                    // TODO better error handling
                    return;
                }

                if (line.start == start || line.start == end) {
                    controlX += line.end.x;
                    controlY += line.end.y;
                } else {
                    controlX += line.start.x;
                    controlY += line.start.y;
                }
            });
            controlPointPrevious = new PIXI.Point(
                controlX / startLines.length,
                controlY / startLines.length
            );
        } else {
            if (startLines[0].start == start || startLines[0].start == end) {
                if (!startLines[0].end) {
                    throw "error";
                }

                // TODO remove const endStartLinesElement = this.getImage().getChildByName(startLines[0].graphicLine.end.id || "new") as PositionPoint;
                controlPointPrevious = new PIXI.Point(
                    startLines[0].end.x,
                    startLines[0].end.y
                );
            } else {
                if (!startLines[0].start) {
                    throw "error";
                }

                // TODO remove const startStartLinesElement = this.getImage().getChildByName(startLines[0].graphicLine.start.id || "new") as PositionPoint;
                controlPointPrevious = new PIXI.Point(
                    startLines[0].start.x,
                    startLines[0].start.y
                );
            }
        }

        const controlPointsPrevious = Line._findControlPoints(controlPointPrevious, start.position, end.position);

        let controlPointNext: PIXI.IPoint;
        if (endLines.length > 1) {
            let controlX = 0;
            let controlY = 0;
            endLines.forEach((line: Line) => {
                if (!line.start || !line.end) {
                    return;
                }

                if (line.start == start || line.start == end) {
                    controlX += line.end.x;
                    controlY += line.end.y;
                } else {
                    controlX += line.start.x;
                    controlY += line.start.y;
                }
            });
            controlPointNext = new PIXI.Point(
                controlX / endLines.length,
                controlY / endLines.length
            );
        } else {
            if (endLines[0].start == start || endLines[0].start == end) {
                if (!endLines[0].end) {
                    throw "error";
                }

                // TODO remove const endEndLinesElement = this.getImage().getChildByName(endLines[0].graphicLine.end.id || "new") as PositionPoint;
                controlPointNext = new PIXI.Point(
                    endLines[0].end.x,
                    endLines[0].end.y
                );
            } else {
                if (!endLines[0].start) {
                    throw "error";
                }

                // TODO remove const startEndLinesElement = this.getImage().getChildByName(endLines[0].graphicLine.start.id || "new") as PositionPoint;
                controlPointNext = new PIXI.Point(
                    endLines[0].start.x,
                    endLines[0].start.y
                );
            }
        }
        const controlPointsNext = Line._findControlPoints(start.position, end.position, controlPointNext);

        return {
            start: start.position,
            end: end.position,
            controlPointsPrevious,
            controlPointsNext
        }
    }

    /**
     * Find control points
     * @param p1
     * @param p2
     * @param p3
     * @private
     */
    protected static _findControlPoints(p1: PIXI.IPoint, p2: PIXI.IPoint, p3: PIXI.IPoint): MedsurfDraw.ControlPoints {
        const dx1 = p1.x - p2.x, dy1 = p1.y - p2.y;
        const dx2 = p2.x - p3.x, dy2 = p2.y - p3.y;

        const l1 = Math.sqrt(dx1 * dx1 + dy1 * dy1);
        const l2 = Math.sqrt(dx2 * dx2 + dy2 * dy2);

        const m1 = {x: (p1.x + p2.x) / 2.0, y: (p1.y + p2.y) / 2.0};
        const m2 = {x: (p2.x + p3.x) / 2.0, y: (p2.y + p3.y) / 2.0};

        const dxm = (m1.x - m2.x);
        const dym = (m1.y - m2.y);

        const k = l2 / (l1 + l2);

        const cm = {x: m2.x + dxm * k, y: m2.y + dym * k};

        const tx = p2.x - cm.x, ty = p2.y - cm.y;

        const c1 = new PIXI.Point(m1.x + tx, m1.y + ty);
        const c2 = new PIXI.Point(m2.x + tx, m2.y + ty);

        return {c1: c1, c2: c2, l1: Math.floor(l1), l2: Math.floor(l2)};
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Remove mode init
     * @private
     */
    protected _removeModeInit(): void {
        // Elements
        try {
            const points = this._getLinePointsModel();
            if (!points) {
                return;
            }
            const result = this._getLineObject(points);
            if (result) {
                // -- Line element
                this._lineElement = result;
                this.addChild(this._lineElement);
                // -- Line corner element
                this._lineCornerElement.points = result.points;
                // -- Line control point element
                this._lineControlPointElement.points = result.points;
            }

            // Sort items
            this.sortChildren();

            // Draw
            this.emit("debounceDraw");
        }
        catch (e) {
            this.type = LineTypes.NOT_INITIALIZED;
            this.emit("debounceDraw");
        }
    }

    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        if (this.model.isStatic) {
            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        // Show item
        this.showItem();
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        if (this.model.isStatic) {
            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select layer
     * @private
     */
    protected _modeSelectLayerGroup(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select layer
     * @private
     */
    protected _removeModeSelectLayerGroup(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select grouping
     * @private
     */
    protected _modeSelectGrouping(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select grouping
     * @private
     */
    protected _removeModeSelectGrouping(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        /* TODO remove
        this.on("mousedown", this.onPointerDown, this);
        this.on("mouseup", this.onPointerUp, this);
        this.on("onRelease", this.onRelease, this);
        */
        if (this.start) {
            (this.start as MedsurfDraw.PositionPoint).moveInteraction.on("startMove", this.startMovePoint, this);
            (this.start as MedsurfDraw.PositionPoint).moveInteraction.on("endMove", this.endMovePoint, this);

            // Change position point mode
            this.start.modeInteraction.setModeItem("select_line");
        }
        if (this.end) {
            (this.end as MedsurfDraw.PositionPoint).moveInteraction.on("startMove", this.startMovePoint, this);
            (this.end as MedsurfDraw.PositionPoint).moveInteraction.on("endMove", this.endMovePoint, this);

            // Change position point mode
            this.end.modeInteraction.setModeItem("select_line");
        }

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        /* TODO removes
        this.off("mousedown", this.onPointerDown, this);
        this.off("mouseup", this.onPointerUp, this);
        this.off("onRelease", this.onRelease, this);
        */
        if (this.start) {
            (this.start as MedsurfDraw.PositionPoint).moveInteraction.on("startMove", this.startMovePoint, this);
            (this.start as MedsurfDraw.PositionPoint).moveInteraction.on("endMove", this.endMovePoint, this);
        }
        if (this.end) {
            (this.end as MedsurfDraw.PositionPoint).moveInteraction.on("startMove", this.startMovePoint, this);
            (this.end as MedsurfDraw.PositionPoint).moveInteraction.on("endMove", this.endMovePoint, this);
        }

        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // TODO remove console.log("line _removeModeSelectItem", mode);
        if (mode === 'author') {
            this.emit("debounceControlUnselectItem", this.image);
            // this.controlUnselectItem();
        }
        /* TODO remove
        if (!mode.startsWith('select_item') && !mode.startsWith('choose_item') && !mode.startsWith('choose_select')
            && !mode.startsWith('blocked_item')
            && ['default', 'marker', 'selftest'/*, 'author'* /, 'delete'].indexOf(mode) === -1) {
            // Unselect item
            this.emit("debounceControlUnselectItem");
        }
        */
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options.lineAlpha = Design.blocked.lineAlpha;
            this._lineElement.options.fillAlpha = Design.blocked.fillAlpha;
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        this.filters = [];
        */
        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options = Object.assign({}, this.model.options);
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode draw item
     * @private
     */
    protected _modeDrawItem(): void {
    }

    /**
     * Remove mode draw item
     * @private
     */
    protected _removeModeDrawItem(): void {
    }

    /**
     * Mode choose select
     * @private
     */
    protected _modeChooseSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options.lineColor = Design.selftest.lineColorSelect;
            this._lineElement.options.fillColor = Design.selftest.fillColorSelect;
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose select
     * @private
     */
    protected _removeModeChooseSelect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options = Object.assign({}, this.model.options);
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose unselect
     * @private
     */
    protected _modeChooseUnselect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options.lineColor = Design.selftest.lineColorUnselect;
            this._lineElement.options.fillColor = Design.selftest.fillColorUnselect;
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose unselect
     * @private
     */
    protected _removeModeChooseUnselect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options = Object.assign({}, this.model.options);
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode clone select
     * @private
     */
    protected _modeCloneItem(): void {
    }

    /**
     * Remove mode clone select
     * @private
     */
    protected _removeModeCloneItem(): void {
        this.moveInteraction.reset();
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        const image = this.image;

        if (image) {
            // Select layer group
            if (['delete_multiple', 'delete_draw', 'delete_clone'].indexOf(this.modeInteraction.lastMode) === -1) {
                this.controlUnselectItem(image);
            }

            // Delete from parent
            image.removeChild(this);

            // Check if position points has/contains in any other elements
            if (this.start) {
                this.start.emit("debounceDeleteItem", image);
            }
            if (this.end) {
                this.end.emit("debounceDeleteItem", image);
            }

            // Emits
            image.deleteImageObject(this.model);
        }

        this.destroy();

        // Hide menu
        image.controlHideMenu();

        // Update control elements
        image.controlUpdateElements();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    /**
     * Collapse all menus
     * @param event
     */
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = Design.line.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = Design.line.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            // Set mode
            this.modeInteraction.setMode("delete");
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.degroupLayerGroupImageObject(this.model);
        if (unselectElement) {
            this.image.unselectLayerGroup();
            this.image.selectImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        this.image.copyImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        this.image.clipboardImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Bezier">
    /**
     * On button bezier
     * @param override
     */
    public onButtonBezier(override?: boolean): void {
        this.model.isBezier = override || !this.model.isBezier;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Color">
    /**
     * On color
     * @param mode
     * @param has
     * @param color
     * @param colorAlpha
     */
    public onColor(mode: ColorMode, has: boolean, color: number | undefined, colorAlpha: number | undefined): void {
        if (mode === ColorMode.LINE) {
            // Update model
            this.model.options.hasLine = has;
            this.model.options.lineColor = color;
            this.model.options.lineAlpha = colorAlpha;

            // Elements
            // -- Line element
            if (this._lineElement) {
                this._lineElement.options = Object.assign({}, this.model.options);
            }

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updateImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Number">
    /**
     * On number
     * @param mode
     * @param number
     */
    public onNumber(mode: NumberMode, number: number): void {
        if (mode === NumberMode.WIDTH) {
            // Update model
            this.model.strokeWidth = number / this.imageDimensions.width;

            // Elements
            // -- Line element
            if (this._lineElement) {
                this._lineElement.lineWidth = Math.round(this.model.strokeWidth * this.imageDimensions.width * 100) / 100;
            }

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.emit("onBackgroundNumber");
            this.image.updateImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Interact with lines">
    /**
     * On button interact with lines
     * @param override
     */
    public onButtonInteractWithLines(override?: boolean): void {
        this.model.interactWithLines = override || !this.model.interactWithLines;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Static">
    /**
     * On button static
     * @param override
     */
    public onButtonStatic(override?: boolean): void {
        this.model.isStatic = override || !this.model.isStatic;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Middle point">
    /**
     * On button add middle point
     * @param image
     * @param layerGroup
     */
    public onButtonAddMiddlePoint(image: Models.Image, layerGroup?: Models.LayerGroup): void {
        if (!this.start || !this.end || !this._lineElement) {
            return;
        }

        // @ts-ignore
        const points = this._lineElement.geometry.graphicsData[0].shape.points as number[];

        // Split points to x and y components
        const valuesX = points.filter((v, i) => i % 2 === 0)
            .sort((a, b) => a - b);
        const valuesY = points.filter((v, i) => i % 2 === 1)
            .sort((a, b) => a - b);

        // Switch for straight line
        let posX: number;
        let posY: number;
        if (this.type !== LineTypes.LINEAR) {
            if (valuesX.length % 2 === 0) {
                const pointIndex = valuesX.length / 2;
                posX = (valuesX[pointIndex] + valuesX[pointIndex + 1]) / 2;
                posY = (valuesY[pointIndex] + valuesY[pointIndex + 1]) / 2;
            } else {
                const pointIndex = valuesX.length / 2;
                posX = valuesX[pointIndex];
                posY = valuesY[pointIndex];
            }
        } else {
            let startX = this.start.position.x;
            let startY = this.start.position.y;
            let endX = this.end.position.x;
            let endY = this.end.position.y;

            // Check for fill collections -> fill collections don't have correct position values
            //  -> we use the min/max line point values to adjust this problem
            const precision: number = 1;
            const index = valuesX.length - 1;
            if (this.start instanceof MedsurfDraw.FillCollection) {
                if (this.end instanceof MedsurfDraw.FillCollection) {
                    startX = valuesX[0];
                    startY = valuesY[0];
                } else {
                    if ((valuesX[0] - precision) <= endX && endX <= (valuesX[0] + precision)) {
                        startX = valuesX[index];
                        startY = valuesY[index];
                    } else {
                        startX = valuesX[0];
                        startY = valuesY[0];
                    }
                }
            }
            if (this.end instanceof MedsurfDraw.FillCollection) {
                if ((valuesX[0] - precision) <= startX && startX <= (valuesX[0] + precision)) {
                    endX = valuesX[index];
                    endY = valuesY[index];
                } else {
                    endX = valuesX[0];
                    endY = valuesY[0];
                }
            }

            // Calc middle point
            posX = (startX + endX) / 2;
            posY = (startY + endY) / 2;
        }

        // Get image
        const img = this.image;

        // Get image width
        const imageWidth = img.imageDimensions.width;

        // Create new position point
        const positionPoint = MedsurfDraw.PositionPoint.getInstance(this.image);
        positionPoint.model.position.x = posX / imageWidth;
        positionPoint.model.position.y = posY / imageWidth;
        img.addChild(positionPoint);
        positionPoint.modeInteraction.setDefaultMode("author");

        // Create new start line
        const startLine = MedsurfDraw.Line.getInstance(img, this.start.model.id as string, positionPoint.model.id as string, this.model.isBezier, imageWidth);
        startLine.model.isStatic = this.model.isStatic;
        startLine.model.interactWithLines = this.model.interactWithLines;
        startLine.model.strokeWidth = this.model.strokeWidth;
        startLine.model.options.lineColor = this.model.options.lineColor;
        startLine.model.options.lineAlpha = this.model.options.lineAlpha;
        img.addChild(startLine);
        startLine.modeInteraction.setDefaultMode("author");

        // Create new end line
        const endLine = MedsurfDraw.Line.getInstance(img, positionPoint.model.id as string, this.end.model.id as string, this.model.isBezier, imageWidth);
        endLine.model.isStatic = this.model.isStatic;
        endLine.model.interactWithLines = this.model.interactWithLines;
        endLine.model.strokeWidth = this.model.strokeWidth;
        endLine.model.options.lineColor = this.model.options.lineColor;
        endLine.model.options.lineAlpha = this.model.options.lineAlpha;
        img.addChild(endLine);
        endLine.modeInteraction.setDefaultMode("author");

        // Draw
        startLine.draw();
        endLine.draw();

        // Update model
        // -- Image
        if (!image.imageObjects) {
            image.imageObjects = [];
        }
        image.imageObjects.push(positionPoint.model);
        image.imageObjects.push(startLine.model);
        image.imageObjects.push(endLine.model);
        // -- Layer group
        if (layerGroup) {
            if (!layerGroup.layerImageObjects) {
                layerGroup.layerImageObjects = [];
            }
            // -- position point
            const pointLayerImageObject: Models.LayerImageObject = {
                dirty: true,
                id: uuidv4(),
                order: layerGroup.layerImageObjects.length,
                layerGroup: layerGroup,
                imageObject: positionPoint.name as string
            };
            layerGroup.layerImageObjects.push(pointLayerImageObject);
            // -- start line
            const startLayerImageObject: Models.LayerImageObject = {
                dirty: true,
                id: uuidv4(),
                order: layerGroup.layerImageObjects.length,
                layerGroup: layerGroup,
                imageObject: startLine.name as string
            };
            layerGroup.layerImageObjects.push(startLayerImageObject);
            // -- end line
            const endLayerImageObject: Models.LayerImageObject = {
                dirty: true,
                id: uuidv4(),
                order: layerGroup.layerImageObjects.length,
                layerGroup: layerGroup,
                imageObject: endLine.name as string
            };
            layerGroup.layerImageObjects.push(endLayerImageObject);
        }

        // Update fill collections
        this.image.getImageObjects()
            .filter((child: MedsurfDraw.ImageObject) => child instanceof MedsurfDraw.FillCollection)
            .forEach((child) => {
                let fillCollection = child as MedsurfDraw.FillCollection;
                if (fillCollection.removeLine(this)) {
                    fillCollection.addLine(startLine);
                    fillCollection.addLine(endLine);
                    fillCollection.draw();
                }
            });

        // Remove and add selftest items
        this.image.getImageObjects()
            .filter((child: MedsurfDraw.ImageObject) => child instanceof MedsurfDraw.PositionPoint
                || child instanceof MedsurfDraw.FillCollection || child instanceof MedsurfDraw.LegendCollection)
            .forEach((imageObject) => {
                let index: number | undefined = undefined;
                switch ((imageObject as MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection
                    | MedsurfDraw.LegendCollection).model.type) {
                    case Models.ImageObjectType.POSITIONPOINT:
                        const positionPoint = (imageObject as MedsurfDraw.PositionPoint);
                        if (!positionPoint.model.selftestItems) {
                            positionPoint.model.selftestItems = [];
                        }
                        index = positionPoint.model.selftestItems.findIndex((item: string) => item === this.name);
                        if (index > -1) {
                            // Remove from selftest items
                            positionPoint.model.selftestItems.splice(index, 1);
                            // Add new selftest items
                            positionPoint.model.selftestItems.push(startLine.name);
                            positionPoint.model.selftestItems.push(endLine.name);
                        }
                        break;
                    case Models.ImageObjectType.LEGENDCOLLECTION:
                        const legendCollection = (imageObject as MedsurfDraw.LegendCollection);
                        throw 'Not implemented';
                        // TODO legend selftest
                        break;
                    case Models.ImageObjectType.FILLCOLLECTION:
                        const fillCollection = (imageObject as MedsurfDraw.FillCollection);
                        if (!fillCollection.model.interactiveItems) {
                            fillCollection.model.interactiveItems = [];
                        }
                        index = fillCollection.model.interactiveItems.findIndex((item: string) => item === this.name);
                        if (index > -1) {
                            // Remove from interactive items
                            fillCollection.model.interactiveItems.splice(index, 1);
                            // Add new selftest items
                            fillCollection.model.interactiveItems.push(startLine.name);
                            fillCollection.model.interactiveItems.push(endLine.name);
                        }
                        break;
                    default:
                        throw 'Not implemented';
                }
            })

        // Emits
        this.image.deleteImageObject(this.model);

        // Cleanup current line
        this.onButtonDelete(new PIXI.InteractionEvent());

        // Select position point
        positionPoint.controlSelectItem();
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectImageObject(this.model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Control unselect item
     * @param image
     */
    public controlUnselectItem(image: MedsurfDraw.Image): void {
        // Set select layer
        if(!this.controlSelectLayerGroup(image)) {
            // Set select image
            image.externControlSelectImage();

            // Emits
            image.unselectImageObject();
        }
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('select_parent');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control select layer group
     * @param image
     */
    public controlSelectLayerGroup(image: MedsurfDraw.Image): boolean {
        // Get layer group
        const layerGroup = image.getLayerGroupByImageObject(this.model);
        if (layerGroup) {
            // Emits
            image.selectLayerGroup(layerGroup);

            // Select layer group
            return image.externControlSelectLayerGroup(layerGroup);
        }

        // Layer not selected
        return false;
    }

    /**
     * Control unselect layer group
     */
    public controlUnselectLayerGroup(): void {
        // Emits
        this.image.unselectLayerGroup();
    }

    /**
     * Control draw item
     * @param isNew
     */
    public controlDrawItem(isNew: boolean): void {
        // Set children mode
        this.modeInteraction.setModeChildren('draw_parent');

        // Set draw item
        this.modeInteraction.setModeItem(isNew ? 'draw_item_new' : 'draw_item');
    }

    /**
     * Control clone item
     */
    public controlCloneItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this && !imageObject.modeInteraction.lastMode.startsWith('clone'))
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('clone_parent');

        // Set draw item
        this.modeInteraction.setModeItem('clone_item_new');
    }
    //</editor-fold>

    //<editor-fold desc="Control Points">
    /**
     * Show control points
     */
    public showControlPoints(): void {
        if (this._lineControlPointElement) {
            this._lineControlPointElement.showItem()
        }
    }

    /**
     * Hide control points
     */
    public hideControlPoints(): void {
        if (this._lineControlPointElement) {
            this._lineControlPointElement.hideItem();
        }
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * Start move
     */
    public startMove(): void {
        // Emit startMove for connected lines
        const image = this.image;

        // Get start lines
        let startLines: Line[] = [];
        if (this.start instanceof MedsurfDraw.PositionPoint) {
            startLines = image.getLinesByPositionPoint(this.start);
        }

        // Get end lines
        let endLines: Line[] = [];
        if (this.end instanceof MedsurfDraw.PositionPoint) {
            endLines = image.getLinesByPositionPoint(this.end);
        }

        // Emit startMove
        startLines.concat(endLines.filter(e => startLines.indexOf(e) < 0)).forEach((line) => {
            line.emit( 'startMove');
        });
    }

    /**
     * On move
     */
    public onMove(event: PIXI.InteractionEvent): void {
        // No start or end
        if (!this.start || !this.end) {
            return;
        }

        // on point Update
        this.onPointUpdate();

        // Emit onPointUpdate for connected lines
        const image = this.image;
        // Get start lines
        let startLines: Line[] = [];
        if (this.start instanceof MedsurfDraw.PositionPoint) {
            startLines = image.getLinesByPositionPoint(this.start);
        }
        // Get end lines
        let endLines: Line[] = [];
        if (this.end instanceof MedsurfDraw.PositionPoint) {
            endLines = image.getLinesByPositionPoint(this.end);
        }
        // Emit onPointUpdate
        startLines.concat(endLines.filter(e => startLines.indexOf(e) < 0)).forEach((line) => {
            line.emit("onPointUpdate");
        });
        // Emit onRotateWithLine
        this.start.emit("onRotateWithLine", event);
        this.end.emit("onRotateWithLine", event);
    }

    /**
     * End move
     */
    public endMove(): void {
        // Emit endMove for connected lines
        const image = this.image;
        // Get start lines
        let startLines: Line[] = [];
        if (this.start instanceof MedsurfDraw.PositionPoint) {
            startLines = image.getLinesByPositionPoint(this.start);
        }
        // Get end lines
        let endLines: Line[] = [];
        if (this.end instanceof MedsurfDraw.PositionPoint) {
            endLines = image.getLinesByPositionPoint(this.end);
        }
        // Emit endMove
        startLines.concat(endLines.filter(e => startLines.indexOf(e) < 0)).forEach((line) => {
            line.emit('endMove');
        });
    }

    /**
     * Start move point
     */
    public startMovePoint(): void {
        this.emit("startMovePoint");
    }

    /**
     * End move point
     */
    public endMovePoint(): void {
        this.emit("endMovePoint");
    }

    /**
     * On release
     */
    /* TODO
    public onRelease(): void {
        this.modeInteraction.setMode(this.modeInteraction.defaultMode);
    }
    */

    pointStartMove(): void {

    }

    /**
     * On point update
     * TODO change type of lineElement, when Linetype changes
     * TODO remove lineControlPointElement when straight line + create it when not
     * TODO return linetype so we can compare ???
     */
    public onPointUpdate(): void {
        // Draw
        this.draw();
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Element
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options.lineColor = Design.line.lineColorHover;
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Elements
        // -- Line element
        if (this._lineElement) {
            this._lineElement.options = Object.assign({}, this.model.options);
        }

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_layer_group') {
            this.controlSelectItem();
        } else {
            if (!this.controlSelectLayerGroup(this.image) && !preventSelectItem) {
                this.controlSelectItem();
            }
        }
    }

    /**
     * On selftest
     * @private
     */
    public onSelftest(): void {
        /* TODO no selftest for line ???
        // Set default mode for image children
        this.getImage().getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.selftestItems && this.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.showItem();
                } else {
                    imageObject.hideItem();
                }
            });

        // Release event
        this.getImage().once('pointerup', this.onSelftestRelease, this);
        */
    }

    /* TODO remove
    public onRelease(): void {
        this.alpha = 1;
        this.modeInteraction.setMode(this.modeInteraction.defaultMode);
    }
    */
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<Line> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<Line>) {
        this._moveInteraction = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<Line> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<Line>) {
        this._selectInteraction = value;
    }

    /**
     * Get type
     */
    public get type(): LineTypes {
        return this._type;
    }

    /**
     * Set type
     * @param value
     */
    public set type(value: LineTypes) {
        this._type = value;
    }

    /**
     * Get direction
     */
    public get direction(): LineDirection {
        return this._direction;
    }

    /**
     * Set direction
     * @param value
     */
    public set direction(value: LineDirection) {
        this._direction = value;
    }

    /**
     * Get start
     */
    public get start(): BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null {
        return this._start;
    }

    /**
     * Set start
     * @param value
     */
    public set start(value: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null) {
        this._start = value;
    }

    /**
     * Get end
     */
    public get end(): BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null {
        return this._end;
    }

    /**
     * Set end
     * @param value
     */
    public set end(value: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | null) {
        this._end = value;
    }

    /**
     * Get control start
     */
    public get controlStart(): PIXI.IPoint | null {
        return this._controlStart;
    }

    /**
     * Set control start
     * @param value
     */
    public set controlStart(value: PIXI.IPoint | null) {
        this._controlStart = value;
    }

    /**
     * Get control end
     */
    public get controlEnd(): PIXI.IPoint | null {
        return this._controlEnd;
    }

    /**
     * Set control end
     * @param value
     */
    public set controlEnd(value: PIXI.IPoint | null) {
        this._controlEnd = value;
    }
    //</editor-fold>
}