import * as Models from "@ascii-dev-user/medsurf-lib/models";
import {BaseGraphics, BaseGraphicsImageModel} from "../../bases/elements/BaseGraphics";
import {StraightLinePointsModel, BezierLinePointsModel, QuadraticCurvePointsModel} from "../../public-api";

/**
 * Line control point model
 */
export class LineControlPointModel extends BaseGraphicsImageModel {
    points: StraightLinePointsModel | QuadraticCurvePointsModel | BezierLinePointsModel | undefined;
    options: Models.ElementOptions;
    lineWidth: number;
    radius: number;
}

/**
 * Control Point
 */
export class LineControlPoint extends BaseGraphics<LineControlPointModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: LineControlPointModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        this.cacheAsBitmap = false;

        this.clear();

        if (!(this.points instanceof QuadraticCurvePointsModel || this.points instanceof BezierLinePointsModel)) {
            return;
        }

        this.beginFill(this.options.fillColor, this.options.fillAlpha);
        this.drawCircle(this.points.controlPointsPrevious.c1.x, this.points.controlPointsPrevious.c1.y, this.radius);
        this.drawCircle(this.points.controlPointsPrevious.c2.x, this.points.controlPointsPrevious.c2.y, this.radius);
        this.endFill();

        this.lineStyle(this.lineWidth, this.options.lineColor, this.options.lineAlpha);
        this.moveTo(this.points.controlPointsPrevious.c1.x, this.points.controlPointsPrevious.c1.y);
        this.lineTo(this.points.start.x, this.points.start.y);
        this.moveTo(this.points.controlPointsPrevious.c2.x, this.points.controlPointsPrevious.c2.y);
        this.lineTo(this.points.start.x, this.points.start.y);
        this.endFill();

        if (this.points instanceof BezierLinePointsModel) {
            this.beginFill(this.options.fillColor, this.options.fillAlpha);
            this.drawCircle(this.points.controlPointsNext.c1.x, this.points.controlPointsNext.c1.y, this.radius);
            this.drawCircle(this.points.controlPointsNext.c2.x, this.points.controlPointsNext.c2.y, this.radius);
            this.endFill();

            this.lineStyle(this.lineWidth, this.options.lineColor, this.options.lineAlpha);
            this.moveTo(this.points.controlPointsNext.c1.x, this.points.controlPointsNext.c1.y);
            this.lineTo(this.points.end.x, this.points.end.y);
            this.moveTo(this.points.controlPointsNext.c2.x, this.points.controlPointsNext.c2.y);
            this.lineTo(this.points.end.x, this.points.end.y);
            this.endFill();
        }

        this.cacheAsBitmap = true;
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get points
     */
    public get points(): StraightLinePointsModel | QuadraticCurvePointsModel | BezierLinePointsModel | undefined {
        return this.data.points;
    }

    /**
     * Set points
     * @param value
     */
    public set points(value: StraightLinePointsModel | QuadraticCurvePointsModel | BezierLinePointsModel | undefined) {
        this.data.points = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number) {
        this.data.lineWidth = value;
    }

    /**
     * Get radius
     */
    public get radius(): number {
        return this.data.radius;
    }

    /**
     * Set radius
     * @param value
     */
    public set radius(value: number) {
        this.data.radius = value;
    }
    //</editor-fold>
}