import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {Design} from "../../config/design";

/**
 * Multi select item model
 */
export class MultiSelectItemModel extends BaseContainerImageModel {
}

/**
 * Multi select item element
 * TODO not implemented
 */
export class MultiSelectItemElement extends BaseContainer<MultiSelectItemModel> {
    /**
     * Members
     */
    private _displayObjects: MedsurfDraw.ImageObject[];

    /**
     * Elements
     */
    protected _selectElement: MedsurfDraw.Rectangle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: MultiSelectItemModel) {
        throw 'not implemented';
        // Parent constructor
        super(model);

        // Members
        this.displayObjects = [];

        // Setup
        this.zIndex = Design.selectElement.zIndex;

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Mode interaction
            this.image.modeInteraction.on("default", this._modeDefault, this);
            this.image.modeInteraction.on("remove-default", this._removeModeDefault, this);
            this.image.modeInteraction.on("author", this._modeAuthor, this);
            this.image.modeInteraction.on("remove-author", this._removeModeAuthor, this);
            this.image.modeInteraction.on("setDefaultMode", this.controlSetDefaultMode, this);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            this.image.modeInteraction.off("default", this._modeDefault, this);
            this.image.modeInteraction.off("remove-default", this._removeModeDefault, this);
            this.image.modeInteraction.off("author", this._modeAuthor, this);
            this.image.modeInteraction.off("remove-author", this._removeModeAuthor, this);
            this.image.modeInteraction.off("setDefaultMode", this.controlSetDefaultMode, this);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Setup
        this.hideItem();

        // Position
        this.position.set(0, 0);

        //<editor-fold desc="Elements">
        // -- Select element
        this._selectElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, 0, 0),
            options: {hasLine: true, lineColor: Design.selectElement.lineColor, lineAlpha: Design.selectElement.lineAlpha, hasFill: true, fillColor: Design.selectElement.fillColor, fillAlpha: Design.selectElement.fillAlpha},
            lineWidth: Design.selectElement.lineWidth
        });
        this.addChild(this._selectElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.draw();
    }

    /**
     * Draw
     */
    public draw(): void {
        // Hide item
        this.hideItem();

        // Get selected display objects
        const displayObjects = this._getSelectedDisplayObjects();

        // Get rectangle
        const rectangle = this._getSelectedItemsRectangle(displayObjects);

        // Elements
        if (rectangle) {
            // Set rectangle pad
            rectangle.pad(Design.selectElement.groupPad, Design.selectElement.groupPad);

            // -- Select element
            this._selectElement.rectangle = rectangle;

            // Show item
            this.showItem();

            // Set display objects
            this.displayObjects = displayObjects as MedsurfDraw.ImageObject[];
        } else {
            // Unset display objects
            this.displayObjects = [];
        }

        // -- Select element
        this._selectElement.draw();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.off("default", this._modeDefault, this);
        this.modeInteraction.off("remove-default", this._removeModeDefault, this);
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Select element
        if (this._selectElement) {
            this._selectElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Private functions">
    /**
     * Get selected image objects
     * @private
     */
    protected _getSelectedDisplayObjects(): PIXI.DisplayObject[] {
        // Get select layer elements
        return this.image.getImageObjects()
            .filter((displayObject) => displayObject.modeInteraction && displayObject.modeInteraction.lastMode === 'select_layer_group');
    }

    /**
     * Get selected items rectangle
     * @private
     */
    protected _getSelectedItemsRectangle(displayObjects: PIXI.DisplayObject[]): PIXI.Rectangle | undefined {
        // Calc rectangle
        if (displayObjects.length > 0) {
            // Get rectangle
            let topLeft: PIXI.Point;
            let topRight: PIXI.Point;
            let bottomLeft: PIXI.Point;

            // Get values
            displayObjects.forEach((displayObject) => {
                // Get rectangle
                const position = displayObject.position;
                // @ts-ignore
                const rect = displayObject.getRectangle();
                // const rect = new PIXI.Rectangle(position.x + bounds.x, position.y + bounds.y, bounds.width, bounds.height);

                if (!topLeft) {
                    topLeft = new PIXI.Point(rect.x, rect.y);
                } else {
                    if (rect.x < topLeft.x) {
                        topLeft.x = rect.x;
                    }
                    if (rect.y < topLeft.y) {
                        topLeft.y = rect.y;
                    }
                }

                if (!topRight) {
                    topRight = new PIXI.Point(rect.x + rect.width);
                } else {
                    if (rect.x + rect.width > topRight.x) {
                        topRight.x = rect.x + rect.width;
                    }
                }

                if (!bottomLeft) {
                    bottomLeft = new PIXI.Point(undefined, rect.y + rect.height);
                } else {
                    if (rect.y + rect.height > bottomLeft.y) {
                        bottomLeft.y = rect.y + rect.height;
                    }
                }
            });

            return new PIXI.Rectangle(
                // @ts-ignore
                topLeft.x,
                // @ts-ignore
                topLeft.y,
                // @ts-ignore
                topRight.x - topLeft.x,
                // @ts-ignore
                bottomLeft.y - topLeft.y
            );
        } else {
            return undefined;
        }
    }

    /**
     * Set events for display objects
     * @param displayObjects
     * @private
     */
    protected _setEventsForDisplayObjects(displayObjects: PIXI.DisplayObject[]): void {
        // Add events on elements
        displayObjects.forEach((displayObject) => {
            // @ts-ignore
            if (displayObject.moveInteraction) {
                // @ts-ignore
                displayObject.moveInteraction.on("endMove", this.controlSetDefaultMode, this);
                this._displayObjects.push(displayObject as MedsurfDraw.ImageObject);
            }
        });
    }

    /**
     * Unset events for display objects
     * @private
     */
    protected _unsetEventsForDisplayObjects(displayObjects: PIXI.DisplayObject[]): void {
        // Remove events from elements
        displayObjects.forEach((displayObject) => {
            // @ts-ignore
            if (displayObject.moveInteraction) {
                // @ts-ignore
                displayObject.moveInteraction.off("endMove", this.controlSetDefaultMode, this);
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        this.hideItem();
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
        this.hideItem();
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        this.hideItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        this.hideItem();
    }
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control set default mode
     */
    public controlSetDefaultMode(): void {
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get display objects
     */
    public get displayObjects(): MedsurfDraw.ImageObject[] {
        return this._displayObjects;
    }

    /**
     * Set display objects
     * @param value
     */
    public set displayObjects(value: MedsurfDraw.ImageObject[]) {
        // Remove display object events
        this._unsetEventsForDisplayObjects(this.displayObjects || []);

        // Set new value
        this._displayObjects = value;

        // Add display object events
        this._setEventsForDisplayObjects(this.displayObjects);
    }
    //</editor-fold>
}