import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {Design} from "../../config/design";
import { debounce } from 'debounce';

/**
 * Select group type
 */
export enum SelectGroupType {
    GROUPING,
    LAYERGROUP,
}

/**
 * Select group model
 */
export class SelectGroupModel extends BaseContainerImageModel {
}

/**
 * Select group element
 */
export class SelectGroupElement extends BaseContainer<SelectGroupModel> {
    /**
     * Members
     */
    protected _type: SelectGroupType;
    protected _layerGroup: Models.LayerGroup;
    protected _grouping: Models.Grouping;
    protected _displayObjects: MedsurfDraw.ImageObject[];
    protected _imageObjectMenu: MedsurfDraw.MenuEntryModel[];
    protected _groupingMenu: MedsurfDraw.MenuEntryModel[];
    private _menuShown: boolean = false;

    /**
     * Interactions
     */
    protected _contextInteraction: MedsurfDraw.ContextInteraction<SelectGroupElement>;
    protected _moveInteraction: MedsurfDraw.MoveInteraction<SelectGroupElement>;

    /**
     * Methods
     */
    protected _debounceControlSelectDefaultModeMethod: EventListener;

    /**
     * Elements
     */
    protected _selectElement: MedsurfDraw.Rectangle;
    protected _imageObjectMenuElement: MedsurfDraw.MenuElement;
    protected _groupingMenuElement: MedsurfDraw.MenuElement;

    /**
     * Menu elements
     */
    // -- Image object menu element
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Grouping menu element
    // -- Interactive grouping
    protected _interactiveGroupingButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Poi grouping
    protected _poiGroupingButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Text grouping
    protected _textGroupingButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Legend grouping
    protected _legendGroupingButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: SelectGroupModel) {
        // Parent constructor
        super(model);

        // Members
        this.displayObjects = [];
        this.imageObjectMenu = [];
        this.groupingMenu = []

        // Setup
        this.zIndex = Design.selectElement.zIndex;

        //<editor-fold desc="Interactions">
        // -- Context interaction
        this.contextInteraction = new MedsurfDraw.ContextInteraction(this);
        this.contextInteraction.on("endRight", this.endRight, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.startMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.moveInteraction.on("endMove", this.endMove, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectDefaultModeMethod = debounce(this.controlSetDefaultMode.bind(this), 50).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectDefaultMode", this._debounceControlSelectDefaultModeMethod);
            // -- Mode interaction
            this.image.modeInteraction.on("default", this._modeDefault, this);
            this.image.modeInteraction.on("remove-default", this._removeModeDefault, this);
            this.image.modeInteraction.on("author", this._modeAuthor, this);
            this.image.modeInteraction.on("remove-author", this._removeModeAuthor, this);
            this.image.modeInteraction.on("setDefaultMode", this._debounceControlSelectDefaultModeMethod);
            // -- Move
            this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
            this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectDefaultMode", this._debounceControlSelectDefaultModeMethod);
            // -- Mode interaction
            this.image.modeInteraction.off("default", this._modeDefault, this);
            this.image.modeInteraction.off("remove-default", this._removeModeDefault, this);
            this.image.modeInteraction.off("author", this._modeAuthor, this);
            this.image.modeInteraction.off("remove-author", this._removeModeAuthor, this);
            this.image.modeInteraction.off("setDefaultMode", this._debounceControlSelectDefaultModeMethod);
            // -- Move
            this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
            this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Setup
        this.hideItem();

        // Position
        this.position.set(0, 0);

        //<editor-fold desc="Elements">
        // -- Select element
        this._selectElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, 0, 0),
            options: {hasLine: true, lineColor: Design.selectElement.lineColor, lineAlpha: Design.selectElement.lineAlpha, hasFill: true, fillColor: Design.selectElement.fillColor, fillAlpha: Design.selectElement.fillAlpha},
            lineWidth: Design.selectElement.lineWidth
        });
        this._selectElement.zIndex = Design.selectElement.zIndex + 1;
        this.addChild(this._selectElement);
        // -- Image object menu element
        this._imageObjectMenuElement = new MedsurfDraw.MenuElement({
            image: this.image,
            menuElements: undefined
        });
        this.image.parent.addChild(this._imageObjectMenuElement);
        // -- Grouping menu element
        this._groupingMenuElement = new MedsurfDraw.MenuElement({
            image: this.image,
            menuElements: undefined
        });
        this.image.parent.addChild(this._groupingMenuElement);
        //</editor-fold>

        //<editor-fold desc="Image object menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Gruppe\nauflösen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        this._imageObjectMenuElement.menuElements = this.imageObjectMenu;
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>
        //<editor-fold desc="Grouping menu elements">
        //<editor-fold desc="Interactive Grouping">
        // -- Interactive grouping button element
        this._interactiveGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf277',
            description: 'Interaktive\nFläche\n(Gruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._interactiveGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._interactiveGroupingButtonElement.on("button", this.onButtonInteractiveGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._interactiveGroupingButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Poi Grouping">
        // -- Poi grouping button element
        this._poiGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf649',
            description: 'POI\n(Gruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._poiGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._poiGroupingButtonElement.on("button", this.onButtonPoiGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._poiGroupingButtonElement,
            order: 5,
        });
        //</editor-fold>
        //<editor-fold desc="Text Grouping">
        // -- Text grouping button element
        this._textGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf53f',
            description: 'Farbgruppe',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._textGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._textGroupingButtonElement.on("button", this.onButtonTextGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._textGroupingButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Legend Grouping">
        // -- Legend grouping button element
        this._legendGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0cb',
            description: 'Legende\n(Gruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._legendGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._legendGroupingButtonElement.on("button", this.onButtonLegendGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._legendGroupingButtonElement,
            order: 7,
        });
        //</editor-fold>
        this._groupingMenuElement.menuElements = this.groupingMenu;
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Set elements
        this._setElements();
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Select element
        this._selectElement.draw();
        // -- Image object menu element
        this._imageObjectMenuElement.draw();
        // -- Grouping menu element
        this._groupingMenuElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Context interaction
        if (this.contextInteraction) {
            this.contextInteraction.removeAllListeners();
            // TODO remove delete this.contextInteraction;
        }
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Select element
        if (this._selectElement) {
            this._selectElement.destroy(options);
        }
        // -- Image object menu element
        if (this._imageObjectMenuElement) {
            this._imageObjectMenuElement.destroy(options);
        }
        // -- Grouping menu element
        if (this._groupingMenuElement) {
            this._groupingMenuElement.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Set grouping
     * @param model
     */
    public setGrouping(model: Models.Grouping): boolean {
        // Get image objects
        const imageObjects = this.image.getDisplayObjectsByGrouping(model);
        if (imageObjects) {
            // Set default mode for image children
            this.image.getImageObjects()
                .filter((imageObject) => imageObjects.indexOf(imageObject) === -1)
                .forEach((imageObject) => {
                    imageObject.modeInteraction.setMode('blocked_item');
                });

            // Set select layer
            imageObjects.forEach((imageObject) => {
                // @ts-ignore
                imageObject.modeInteraction.setMode('select_grouping');
            });

            // Setup grouping
            this._type = SelectGroupType.GROUPING;
            this._grouping = model;

            // Grouping selected
            return true;
        }
        // Grouping not selected
        return false;
    }

    /**
     * Set layer group
     * @param model
     */
    public setLayerGroup(model: Models.LayerGroup): boolean {
        // Get image objects
        const imageObjects = this.image.getLayerGroupDisplayObjectsByLayerGroup(model);
        if (imageObjects) {
            // Set default mode for image children
            this.image.getImageObjects()
                .filter((imageObject) => imageObjects.indexOf(imageObject) === -1)
                .forEach((imageObject) => {
                    imageObject.modeInteraction.setMode('blocked_item');
                });

            // Set select layer
            imageObjects.forEach((imageObject) => {
                // @ts-ignore
                imageObject.modeInteraction.setMode('select_layer_group');
            });

            // Setup layer group
            this._type = SelectGroupType.LAYERGROUP;
            this._layerGroup = model;

            // Layer group selected
            return true;
        }
        // Layer group not selected
        return false;
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Set elements
     * @protected
     */
    protected _setElements(): void {
        // Hide item
        this.hideItem();

        // Prevent errors
        if (!this.parent) {
            return;
        }

        // Get selected display objects
        const displayObjects = this._getSelectedDisplayObjects();

        // Get rectangle
        const rectangle = this._getSelectedItemsRectangle(displayObjects);

        //<editor-fold desc="Elements">
        if (rectangle) {
            // Set rectangle pad
            rectangle.pad(Design.selectElement.groupPad, Design.selectElement.groupPad);

            // -- Select element
            this._selectElement.rectangle = rectangle;

            // Show item
            this.showItem();

            // Setup
            this.interactive = true;

            // Sort children
            this.image.sortChildren();

            // Set display objects
            this.displayObjects = displayObjects as MedsurfDraw.ImageObject[];
        } else {
            // Emits
            if (this.interactive) { // Check is ever shown
                if (this._type === SelectGroupType.GROUPING) {
                    this.image.emit("debouncedUnselectGrouping");
                } else {
                    this.image.emit("debouncedUnselectLayerGroup");
                }
            }

            // Setup
            this.interactive = false;

            // Unset display objects
            this.displayObjects = [];
        }
        //</editor-fold>

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get selected image objects
     * @private
     */
    protected _getSelectedDisplayObjects(): PIXI.DisplayObject[] {
        // Get select layer elements
        return this.image.getImageObjects()
            .filter((displayObject) => displayObject.modeInteraction && (displayObject.modeInteraction.lastMode === 'select_layer_group' || displayObject.modeInteraction.lastMode === 'select_grouping'));
    }

    /**
     * Get selected items rectangle
     * @private
     */
    protected _getSelectedItemsRectangle(displayObjects: PIXI.DisplayObject[]): PIXI.Rectangle | undefined {
        // Calc rectangle
        if (displayObjects.length > 0) {
            // Get rectangle
            let topLeft: PIXI.Point;
            let topRight: PIXI.Point;
            let bottomLeft: PIXI.Point;

            // Get values
            displayObjects.forEach((displayObject) => {
                // Get rectangle
                const position = displayObject.position;
                // @ts-ignore
                const rect = displayObject.getRectangle();
                // const rect = new PIXI.Rectangle(position.x + bounds.x, position.y + bounds.y, bounds.width, bounds.height);

                if (!topLeft) {
                    topLeft = new PIXI.Point(rect.x, rect.y);
                } else {
                    if (rect.x < topLeft.x) {
                        topLeft.x = rect.x;
                    }
                    if (rect.y < topLeft.y) {
                        topLeft.y = rect.y;
                    }
                }

                if (!topRight) {
                    topRight = new PIXI.Point(rect.x + rect.width);
                } else {
                    if (rect.x + rect.width > topRight.x) {
                        topRight.x = rect.x + rect.width;
                    }
                }

                if (!bottomLeft) {
                    bottomLeft = new PIXI.Point(undefined, rect.y + rect.height);
                } else {
                    if (rect.y + rect.height > bottomLeft.y) {
                        bottomLeft.y = rect.y + rect.height;
                    }
                }
            });

            return new PIXI.Rectangle(
                // @ts-ignore
                topLeft.x,
                // @ts-ignore
                topLeft.y,
                // @ts-ignore
                topRight.x - topLeft.x,
                // @ts-ignore
                bottomLeft.y - topLeft.y
            );
        } else {
            return undefined;
        }
    }

    /**
     * Set events for display objects
     * @param displayObjects
     * @private
     */
    protected _setEventsForDisplayObjects(displayObjects: PIXI.DisplayObject[]): void {
        // Add events on elements
        displayObjects.forEach((displayObject) => {
            // @ts-ignore
            if (displayObject.moveInteraction) {
                // @ts-ignore
                displayObject.moveInteraction.on("endMove", this.controlSetDefaultMode, this);
                this._displayObjects.push(displayObject as MedsurfDraw.ImageObject);
            }
        });
    }

    /**
     * Unset events for display objects
     * @private
     */
    protected _unsetEventsForDisplayObjects(displayObjects: PIXI.DisplayObject[]): void {
        // Remove events from elements
        displayObjects.forEach((displayObject) => {
            // @ts-ignore
            if (displayObject.moveInteraction) {
                // @ts-ignore
                displayObject.moveInteraction.off("endMove", this.controlSetDefaultMode, this);
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        this.hideItem();
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
        this.hideItem();
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        this.hideItem();

        //<editor-fold desc="Events">
        // -- Context
        this.image.parent.on("rightup", this.contextInteraction.endRight, this.contextInteraction);
        //</editor-fold>
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        this.hideItem();

        //<editor-fold desc="Events">
        // -- Context
        this.image.off("rightup", this.contextInteraction.endRight, this.contextInteraction);
        //</editor-fold>
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    /**
     * Display menu
     */
    public displayMenu(): void {
        if (!this.visible) {
            return;
        }

        const point = this.image.getMousePosition();

        // Setup
        this._menuShown = true;

        // Elements
        if (MedsurfDraw.Keyboard.isKeyDown("KeyG")) {
            // -- Grouping menu element
            this._groupingMenuElement.setMousePosition(point);
            this._groupingMenuElement.showItem();
        } else {
            // -- Image object menu element
            this._imageObjectMenuElement.setMousePosition(point);
            this._imageObjectMenuElement.showItem();
        }

        // Hide menu
        this.image.moveInteraction.once("startMove", () => {
            // Elements
            // -- Menu element
            this._imageObjectMenuElement.hideItem();
            // -- Grouping menu element
            this._groupingMenuElement.hideItem();
        });
        this.image.moveInteraction.once("endMove", () => {
            // Setup
            this._menuShown = false;
        });
        /* TODO remove
        // Image -> remove unselect behaviour and reset it after one click on the image
        const image = this.getImage();
        image.moveInteraction.off("endMove", image.endMove, image);
        image.moveInteraction.once("endMove", () => {
            // Elements
            // -- Menu element
            this._imageObjectMenuElement.hideItem();
            // -- Grouping menu element
            this._groupingMenuElement.hideItem();

            // Image
            image.moveInteraction.on("endMove", image.endMove, image);
        });
        */
    }
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editImageObjectGrouping();
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        this.displayObjects.forEach((io) => {
            (io as any).onButtonMoveToFront();
        });
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        this.displayObjects.forEach((io) => {
            (io as any).onButtonMoveToBack();
        });
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Delete
        this.displayObjects.forEach((io) => {
            (io as any).modeInteraction.setMode("delete_multiple");
        });

        // Emits
        this.image.controlSetDefaultMode(new PIXI.InteractionEvent());
        if (this._type === SelectGroupType.GROUPING) {
            this.image.unselectGrouping();
        } else {
            this.image.unselectLayerGroup();
        }
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     */
    public onButtonDegroup(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        if (this._type === SelectGroupType.GROUPING) {
            this.image.degroupGroupImageLayerGroups();
        } else {
            this.displayObjects.forEach((io) => {
                (io as any).onButtonDegroup(false);
            });
            this.image.unselectLayerGroup();
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        switch (this._type) {
            case SelectGroupType.LAYERGROUP:
                this.image.copyLayerGroup(this._layerGroup);
                break;
            default:
                throw new Error("Not implemented");
        }
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Clipboard image object
        switch (this._type) {
            case SelectGroupType.LAYERGROUP:
                this.image.clipboardLayerGroup(this._layerGroup);
                break;
            default:
                throw new Error("Not implemented");
        }
    }
    //</editor-fold>
    //<editor-fold desc="Interactive Grouping">
    /**
     * On button interactive grouping
     */
    public onButtonInteractiveGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create interative grouping
        this.image.createInteractiveGrouping(this.layerGroup);
    }
    //</editor-fold>
    //<editor-fold desc="Poi Grouping">
    /**
     * On button poi grouping
     */
    public onButtonPoiGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create poi grouping
        this.image.createPoiGrouping(this.layerGroup);
    }
    //</editor-fold>
    //<editor-fold desc="Text Grouping">
    /**
     * On button text grouping
     */
    public onButtonTextGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create text grouping
        this.image.createTextGrouping(this.layerGroup);
    }
    //</editor-fold>
    //<editor-fold desc="Legend Grouping">
    /**
     * On button legend grouping
     */
    public onButtonLegendGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create legend grouping
        this.image.createLegendGrouping(this.layerGroup);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    public controlHideMenu(): void {
        // Elements
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        this._groupingMenuElement.hideItem();
    }

    /**
     * Control set default mode
     */
    public controlSetDefaultMode(): void {
        this._setElements();
    }
    //</editor-fold>

    //<editor-fold desc="Context">
    /**
     * End right
     * @param event
     */
    public endRight(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Show menu
        this.displayMenu();
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * Start move
     * @param event
     * @private
     */
    public startMove(event: PIXI.InteractionEvent): void {
        // Emit
        this.displayObjects.forEach((displayObject) => {
            if (displayObject instanceof MedsurfDraw.PositionPoint || displayObject instanceof MedsurfDraw.LegendCollection) {
                displayObject.moveInteraction.startMove(event);
            }
        });
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     * @private
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number) {
        // Position
        this.position.x += dX;
        this.position.y += dY;

        // Emit
        this.displayObjects.forEach((displayObject) => {
            if (displayObject instanceof MedsurfDraw.PositionPoint || displayObject instanceof MedsurfDraw.LegendCollection) {
                displayObject.moveInteraction.onMove(event);
            }
        });
    }

    /**
     * End move
     * @param event
     * @private
     */
    public endMove(event: PIXI.InteractionEvent): void {
        // Position reset
        this.position.x = 0;
        this.position.y = 0;

        // Emit
        this.displayObjects.forEach((displayObject) => {
            if (displayObject instanceof MedsurfDraw.PositionPoint || displayObject instanceof MedsurfDraw.LegendCollection) {
                displayObject.moveInteraction.endMove(event);
            }
        });
    }

    /**
     * On release
     */
    public onRelease(): void {
        this.modeInteraction.setMode(this.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get grouping
     */
    public get grouping(): Models.Grouping {
        return this._grouping;
    }

    /**
     * Get layer group
     */
    public get layerGroup(): Models.LayerGroup {
        return this._layerGroup;
    }

    /**
     * Get display objects
     */
    public get displayObjects(): MedsurfDraw.ImageObject[] {
        return this._displayObjects;
    }

    /**
     * Set display objects
     * @param value
     */
    public set displayObjects(value: MedsurfDraw.ImageObject[]) {
        // Remove display object events
        this._unsetEventsForDisplayObjects(this.displayObjects || []);

        // Set new value
        this._displayObjects = value;

        // Add display object events
        this._setEventsForDisplayObjects(this.displayObjects);
    }

    /**
     * Get image object menu
     */
    public get imageObjectMenu(): MedsurfDraw.MenuEntryModel[] {
        return this._imageObjectMenu;
    }

    /**
     * Set image object menu
     * @param value
     */
    public set imageObjectMenu(value: MedsurfDraw.MenuEntryModel[]) {
        this._imageObjectMenu = value;
    }

    /**
     * Get grouping menu
     */
    public get groupingMenu(): MedsurfDraw.MenuEntryModel[] {
        return this._groupingMenu;
    }

    /**
     * Set grouping menu
     * @param value
     */
    public set groupingMenu(value: MedsurfDraw.MenuEntryModel[]) {
        this._groupingMenu = value;
    }

    /**
     * Get context interaction
     */
    public get contextInteraction(): MedsurfDraw.ContextInteraction<SelectGroupElement> {
        return this._contextInteraction;
    }

    /**
     * Set context interaction
     * @param value
     */
    public set contextInteraction(value: MedsurfDraw.ContextInteraction<SelectGroupElement>) {
        this._contextInteraction = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<SelectGroupElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<SelectGroupElement>) {
        this._moveInteraction = value;
    }

    /**
     * Get menu shown
     */
    public get menuShown(): boolean {
        return this._menuShown;
    }
    //</editor-fold>
}