import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {BaseElementContainer, BaseElementContainerModel} from "../../bases/elements/BaseElementContainer";
import {Design} from "../../config/design";
import {debounce} from "debounce";

/**
 * Select model
 */
export class SelectImageObjectElementModel extends BaseContainerImageModel {
}

/**
 * Select image object element
 */
export class SelectItemElement extends BaseContainer<SelectImageObjectElementModel> {
    /**
     * Members
     */
    protected _displayObject: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Text | MedsurfDraw.Image | undefined;
    protected _menuShown: boolean = false;

    /**
     * Methods
     */
    protected _debounceUpdateRectangleMethod: EventListener;
    protected _debounceOnMoveMethod: EventListener;

    /**
     * Elements
     */
    protected _scaleElement: MedsurfDraw.ScaleElement;
    protected _rotateElement: MedsurfDraw.RotateElement;
    protected _rotateScaleElement: MedsurfDraw.RotateScaleElement;
    protected _caretElement: MedsurfDraw.CaretElement;
    protected _selectionElement: MedsurfDraw.SelectionElement;
    protected _menuElement: MedsurfDraw.MenuElement;
    protected _colorElement: MedsurfDraw.ColorElement;
    protected _numberElement: MedsurfDraw.NumberElement;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: SelectImageObjectElementModel) {
        // Parent constructor
        super(model);

        // Members
        this.displayObject = undefined;

        // Setup
        this.zIndex = Design.selectImageObjectElement.zIndex;

        //<editor-fold desc="Methods">
        this._debounceUpdateRectangleMethod = debounce(this.updateRectangle.bind(this), 20).bind(this);
        this._debounceOnMoveMethod = debounce(this.controlSetDefaultMode.bind(this), 50).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Mode interaction
            this.image.modeInteraction.on("default", this._modeDefault, this);
            this.image.modeInteraction.on("remove-default", this._removeModeDefault, this);
            this.image.modeInteraction.on("author", this._modeAuthor, this);
            this.image.modeInteraction.on("remove-author", this._removeModeAuthor, this);
            this.image.modeInteraction.on("setDefaultMode", this.controlSetDefaultMode, this);
            // -- Debounce
            this.on("debounceUpdateRectangle", this._debounceUpdateRectangleMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Mode interaction
            this.image.modeInteraction.off("default", this._modeDefault, this);
            this.image.modeInteraction.off("remove-default", this._removeModeDefault, this);
            this.image.modeInteraction.off("author", this._modeAuthor, this);
            this.image.modeInteraction.off("remove-author", this._removeModeAuthor, this);
            this.image.modeInteraction.off("setDefaultMode", this.controlSetDefaultMode, this);
            // -- Debounce
            this.off("debounceUpdateRectangle", this._debounceUpdateRectangleMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Members
        this.displayObject = undefined;

        // Setup
        this.hideItem();

        // Position
        this.position.set(0, 0);

        //<editor-fold desc="Elements">
        // -- Scale element
        this._scaleElement = new MedsurfDraw.ScaleElement({
            image: this.image,
            type: MedsurfDraw.ScaleElementType.NOT_SCALEABLE,
            rectangle: new PIXI.Rectangle(0, 0, 0, 0)
        });
        this._scaleElement.on("startScale", this.startScale, this);
        this._scaleElement.on("onScale", this.onScale, this);
        this._scaleElement.on("endScale", this.endScale, this);
        this.addChild(this._scaleElement);
        // -- Rotate element
        this._rotateElement = new MedsurfDraw.RotateElement({
            image: this.image,
            rectangle: new PIXI.Rectangle(0, 0, 0, 0)
        });
        this._rotateElement.rotateInteraction.on("startRotate", this.startRotate, this);
        this._rotateElement.rotateInteraction.on("onRotate", this.onRotate, this);
        this._rotateElement.rotateInteraction.on("endRotate", this.endRotate, this);
        this.addChild(this._rotateElement);
        // -- Rotate scale element
        this._rotateScaleElement = new MedsurfDraw.RotateScaleElement({
            image: this.image,
            rotateElement: this._rotateElement
        });
        this._rotateScaleElement.hideItem();
        this.addChild(this._rotateScaleElement);
        // -- Caret element
        this._caretElement = new MedsurfDraw.CaretElement({
            image: this.image,
            textElement: undefined
        });
        this.addChild(this._caretElement);
        // -- Selection element
        this._selectionElement = new MedsurfDraw.SelectionElement({
            image: this.image,
            textElement: undefined
        });
        this.addChild(this._selectionElement);
        // -- Menu Element
        this._menuElement = new MedsurfDraw.MenuElement({
            image: this.image,
            menuElements: undefined
        });
        this.image.parent.addChild(this._menuElement);
        // -- Clear color button element
        this._colorElement = new MedsurfDraw.ColorElement({
            image: this.image,
            rectangle: new PIXI.Rectangle(0, 0, 0, 0),
            color: 0x000000,
            colorAlpha: 0,
            showAlpha: true,
        });
        this._colorElement.on("color", this.onColor, this);
        this.addChild(this._colorElement);
        // -- Number element
        this._numberElement = new MedsurfDraw.NumberElement({
            image: this.image,
            rectangle: new PIXI.Rectangle(0, 0, 0, 0),
            value: 1,
            minValue: 0,
            maxValue: 10,
            gab: 1,
        });
        this._numberElement.on('number', this.onNumber, this);
        this.addChild(this._numberElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Set elements
        this._setElements();
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Scale element
        this._scaleElement.draw();
        // -- Roate element
        this._rotateElement.draw();
        // -- Roate scale element
        this._rotateScaleElement.draw();
        // -- Caret element
        this._caretElement.draw();
        // -- Selection element
        this._selectionElement.draw();
        // -- Menu element
        this._menuElement.draw();
        // -- Clear color button element
        this._colorElement.draw();
        // -- Number element
        this._numberElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Scale element
        if (this._scaleElement) {
            this._scaleElement.off("startScale", this.startScale, this);
            this._scaleElement.off("onScale", this.onScale, this);
            this._scaleElement.off("endScale", this.endScale, this);
            this._scaleElement.destroy(options)
        }
        // -- Rotate element
        if (this._rotateElement) {
            this._rotateElement.rotateInteraction.off("startRotate", this.startRotate, this);
            this._rotateElement.rotateInteraction.off("onRotate", this.onRotate, this);
            this._rotateElement.rotateInteraction.off("endRotate", this.endRotate, this);
            this._rotateElement.destroy(options);
        }
        // -- Rotate scale element
        if (this._rotateScaleElement) {
            this._rotateScaleElement.destroy(options);
        }
        // -- Caret element
        if (this._caretElement) {
            this._caretElement.destroy(options);
        }
        // -- Selection element
        if (this._selectionElement) {
            this._selectionElement.destroy(options);
        }
        // -- Menu element
        if (this._menuElement) {
            this._menuElement.destroy(options);
        }
        // -- Clear color button element
        if (this._colorElement) {
            this._colorElement.off("color", this.onColor, this);
            this._colorElement.destroy(options);
        }
        // -- Number element
        if (this._numberElement) {
            this._numberElement.off('number', this.onNumber, this);
            this._numberElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Update rectangle
     */
    public updateRectangle(): {displayObject: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Image | undefined, rectangle: PIXI.Rectangle | undefined} {
        // Get selected display objects
        const displayObject = this._getSelectedDisplayObject();

        // Get rectangle
        let rectangle = undefined;
        if (displayObject) {
            rectangle = this._getSelectedItemsRectangle(displayObject);

            // Elements
            if (rectangle) {
                const pivotWidth = rectangle.width / 2;
                const pivotHeight = rectangle.height / 2;

                // Rotation
                const rotation = displayObject.getRotation();

                // Set rectangle pad
                rectangle.pad(Design.selectElement.groupPad, Design.selectElement.groupPad);

                // Setup
                this.position.set(rectangle.x + rectangle.width / 2, rectangle.y + rectangle.height / 2);

                // Elements
                // -- Scale element
                this._scaleElement.rectangle = new PIXI.Rectangle(0, 0, rectangle.width, rectangle.height);
                this._scaleElement.rotation = rotation;
                this._scaleElement.draw();
                // -- Rotate element
                this._rotateElement.rectangle = new PIXI.Rectangle(0, 0, rectangle.width, rectangle.height);
                this._rotateElement.rotation = rotation;
                this._rotateElement.draw();
                // -- Rotate scale element
                this._rotateScaleElement.draw();
                // -- Caret element
                this._caretElement.pivot.set(pivotWidth, pivotHeight);
                this._caretElement.rotation = rotation;
                this._caretElement.draw();
                // -- Selection element
                this._selectionElement.pivot.set(pivotWidth, pivotHeight);
                this._selectionElement.rotation = rotation;
                this._selectionElement.draw();
                // -- Menu element
                // this._menuElement.circle = new PIXI.Circle(0, 0, Math.sqrt(Math.pow(Math.abs(rectangle.width), 2) + Math.pow(Math.abs(rectangle.height), 2)) / 2 + Design.subMenu.buttonWidth + Design.selectImageObjectElement.pad);
                // TODO use this for constant radius this._menuElement.circle = new PIXI.Circle(0, 0, Design.menu.radius + Design.menu.buttonWidth + Design.selectImageObjectElement.pad);
                // this._menuElement.draw();
                // -- Clear color button element
                this._colorElement.rectangle = new PIXI.Rectangle(0, 0, rectangle.width, rectangle.height);
                this._colorElement.draw();
                // -- Number element
                this._numberElement.rectangle = new PIXI.Rectangle(0, 0, rectangle.width, rectangle.height);
                this._numberElement.draw();
            }
        }

        // Return
        return {displayObject, rectangle};
    }

    /**
     * Save color
     * @param color
     * @param colorAlpha
     */
    public saveColor(color: number | undefined, colorAlpha: number | undefined): void {
        this._colorElement.saveColor(color, colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Set elements
     * @protected
     */
    protected _setElements(): void {
        // Hide item
        this.hideItem();

        // Prevent errors
        if (!this.parent) {
            return;
        }

        // Get data
        const data = this.updateRectangle();

        //<editor-fold desc="Elements">
        if (data.rectangle) {
            // Set display objects
            this.displayObject = data.displayObject;

            // Show item
            this.showItem();
        } else {
            // Unset display objects
            this.displayObject = undefined;
        }
        //</editor-fold>

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get selected image objects
     * @private
     */
    protected _getSelectedDisplayObject(): BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Image | undefined {
        // Check if image is selected
        if (this.image.modeInteraction.lastMode.startsWith('select_item')) {
            return this.image;
        }

        // Get imageObjects
        let imageObjects = this.image.getImageObjects();

        // Get select item elements
        let displayObjects;
        for (let i = 0; i <= Design.selectImageObjectElement.displayObjectDepth; i++) {
            displayObjects = imageObjects.filter((displayObject) => displayObject.modeInteraction && displayObject.modeInteraction.lastMode.startsWith('select_item')) as BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>[];
            if (displayObjects.length > 0) {
                return displayObjects[0];
            }
            imageObjects = imageObjects
                .filter((displayObject) => displayObject.modeInteraction && displayObject.modeInteraction.lastMode === 'select_child')
                .map((displayObject) => displayObject.children)
                .reduce((cur, acc) => [...cur, ...acc], []) as BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>[];
        }

        // Get select item elements
        /* TODO remove Old approach
        let displayObjects = imageObjects
            .filter((displayObject) => displayObject.modeInteraction && displayObject.modeInteraction.lastMode === 'select_item');
        if (displayObjects.length === 0) {
            displayObjects = imageObjects
                .filter((displayObject) => displayObject.modeInteraction && displayObject.modeInteraction.lastMode === 'select_child')
                .map((displayObject) => displayObject.children)
                .reduce((cur, acc) => [...cur, ...acc], [])
                // @ts-ignore
                .filter((displayObject) => displayObject.modeInteraction && displayObject.modeInteraction.lastMode === 'select_item') as MedsurfDraw.ImageObject[];
        }
        */
        return undefined;
    }

    /**
     * Get selected items rectangle
     * @private
     */
    protected _getSelectedItemsRectangle(displayObject: PIXI.DisplayObject | undefined): PIXI.Rectangle | undefined {
        // Calc rectangle
        if (displayObject) {
            // Calc rectangle
            // @ts-ignore
            return displayObject.getElementRectangle();

            /* TODO WHY ???
            // Calc rectangle
            // @ts-ignore
            const rect = displayObject.getElementRectangle();

            // Get rectangle
            let topLeft: PIXI.Point = new PIXI.Point(rect.x, rect.y);
            let topRight: PIXI.Point = new PIXI.Point(rect.x + rect.width);
            let bottomLeft: PIXI.Point = new PIXI.Point(undefined, rect.y + rect.height);

            return new PIXI.Rectangle(
                // @ts-ignore
                topLeft.x,
                // @ts-ignore
                topLeft.y,
                // @ts-ignore
                topRight.x - topLeft.x,
                // @ts-ignore
                bottomLeft.y - topLeft.y
            );
            */
        } else {
            return undefined;
        }
    }

    /**
     * Set events for display objects
     * @param displayObject
     * @private
     */
    protected _setEventsForDisplayObjects(displayObject: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Text | MedsurfDraw.Image | undefined): void {
        if (displayObject instanceof MedsurfDraw.EditableText) {
            displayObject.writeInteraction.on("onCaret", this._caretElement.onCaret, this._caretElement);
            displayObject.writeInteraction.on("onSelection", this._selectionElement.onSelection, this._selectionElement);
        }
        if (displayObject instanceof MedsurfDraw.Line) {
            displayObject.on("startMovePoint", this.controlStartMove, this);
            displayObject.on("endMovePoint", this.controlSetDefaultMode, this);
        }
        // @ts-ignore
        if (displayObject && displayObject.moveInteraction) {
            // @ts-ignore
            displayObject.moveInteraction.on("startMove", this.controlStartMove, this);
            // @ts-ignore
            displayObject.moveInteraction.on("endMove", this.controlSetDefaultMode, this);
            this._displayObject = displayObject;
        }
    }

    /**
     * Unset events for display objects
     * @param displayObject
     * @private
     */
    protected _unsetEventsForDisplayObjects(displayObject: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Text | MedsurfDraw.Image | undefined): void {
        if (displayObject instanceof MedsurfDraw.EditableText) {
            displayObject.writeInteraction.off("onCaret", this._caretElement.onCaret, this._caretElement);
            displayObject.writeInteraction.off("onSelection", this._selectionElement.onSelection, this._selectionElement);
        }
        if (displayObject instanceof MedsurfDraw.Line) {
            displayObject.off("startMovePoint", this.controlStartMove, this);
            displayObject.off("endMovePoint", this.controlSetDefaultMode, this);
        }
        // @ts-ignore
        if (displayObject && displayObject.moveInteraction) {
            // @ts-ignore
            displayObject.moveInteraction.off("startMove", this.controlStartMove, this);
            // @ts-ignore
            displayObject.moveInteraction.off("endMove", this.controlSetDefaultMode, this);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        this.hideItem();
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
        this.hideItem();
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        this.hideItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        this.hideItem();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    /**
     * Display menu
     * @param event
     */
    public displayMenu(event: MouseEvent): void {
        if (!this.visible) {
            return;
        }

        event.stopPropagation();

        // Setup
        this._menuShown = true;

        // Elements
        // -- Menu element
        const point = this.image.getMousePosition();
        this._menuElement.setMousePosition(point);
        this._menuElement.showItem();

        // Hide menu
        /* TODO */
        this.image.moveInteraction.once("startMove", () => {
            // Elements
            // -- Menu element
            this._menuElement.hideItem();
        });
        /*
        this.image.moveInteraction.once("endMove", () => {
            // Setup
            this._menuElement.showItem();
        });
        */
        this.image.moveInteraction.once("endMoveLock", () => {
            // Setup
            this._menuShown = false;
        });
        /* TODO remove
        // Image -> remove unselect behaviour and reset it after one click on the image
        const image = this.image;
        image.moveInteraction.off("endMove", image.endMove, image);
        image.moveInteraction.once("endMove", () => {
            // Elements
            // -- Menu element
            this._menuElement.hideItem();

            // Image
            image.moveInteraction.on("endMove", image.endMove, image);
        });
        */
    }
    //</editor-fold>

    //<editor-fold desc="Control">
    public controlHideMenu(): void {
        // Elements
        // -- Menu element
        this._menuElement.hideItem();
    }

    /**
     * Control start move
     */
    public controlStartMove(): void {
        this.hideItem();
    }

    /**
     * Control set default mode
     */
    public controlSetDefaultMode(): void {
        this._setElements();
    }
    //</editor-fold>

    //<editor-fold desc="Scale">
    /**
     * Start scale
     * @param event
     * @param scaleSide
     */
    public startScale(event: PIXI.InteractionEvent, scaleSide: MedsurfDraw.ScaleSideTypes) {
        // @ts-ignore
        if (this.displayObject.scaleInteraction) {
            // @ts-ignore
            this.displayObject.scaleInteraction.startScale(event, scaleSide);
        }
    }

    /**
     * On scale
     * @param event
     */
    public onScale(event: PIXI.InteractionEvent): void {
        // @ts-ignore
        if (this.displayObject.scaleInteraction) {
            // @ts-ignore
            this.displayObject.scaleInteraction.onScale(event);
        }

        // Adjust rectangle
        this.updateRectangle();
    }

    /**
     * End scale
     * @param event
     */
    public endScale(event: PIXI.InteractionEvent): void {
        // @ts-ignore
        if (this.displayObject.scaleInteraction) {
            // @ts-ignore
            this.displayObject.scaleInteraction.endScale(event);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Rotate">
    /**
     * On start rotate
     */
    public startRotate(): void {
        (this.displayObject as MedsurfDraw.ArrowPrimitive | MedsurfDraw.EllipsePrimitive | MedsurfDraw.RectanglePrimitive | MedsurfDraw.TextPrimitive).startRotate();

        // Elements
        // -- Scale element
        this._scaleElement.hideItem();
        // -- Rotate scale element
        this._rotateScaleElement.showItem();
        // -- Menu element
        this._menuElement.hideItem();
    }

    /**
     * On scale
     * @param degree
     */
    public onRotate(degree: number): void {
        (this.displayObject as MedsurfDraw.ArrowPrimitive | MedsurfDraw.EllipsePrimitive | MedsurfDraw.RectanglePrimitive | MedsurfDraw.TextPrimitive).onRotate(degree);

        // Adjust rectangle
        this.updateRectangle();
    }

    /**
     * On end rotate
     */
    public endRotate(): void {
        (this.displayObject as MedsurfDraw.ArrowPrimitive | MedsurfDraw.EllipsePrimitive | MedsurfDraw.RectanglePrimitive | MedsurfDraw.TextPrimitive).endRotate();

        // Elements
        // -- Scale element
        this._scaleElement.showItem();
        // -- Rotate scale element
        this._rotateScaleElement.hideItem();
        // -- Menu element
        // TODO remove this this._menuElement.showItem();
    }
    //</editor-fold>

    //<editor-fold desc="Color">
    /**
     * On color
     * @param hasLine
     * @param color
     * @param colorAlpha
     */
    public onColor(hasLine: boolean, color: number, colorAlpha: number): void {
        // @ts-ignore TODO ignore
        this.displayObject.onColor(hasLine, color, colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Number">
    /**
     * On color
     * @param number
     */
    public onNumber(number: number): void {
        // @ts-ignore TODO ignore
        this.displayObject.onNumber(number);
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Adjust rectangle
        this.emit("debounceUpdateRectangle");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get display objects
     */
    public get displayObject(): BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Text | MedsurfDraw.Image | undefined {
        return this._displayObject;
    }

    /**
     * Set display objects
     * @param value
     */
    public set displayObject(value: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>> | MedsurfDraw.Text | MedsurfDraw.Image | undefined) {
        // Remove display object events
        this._unsetEventsForDisplayObjects(this.displayObject);

        // Set new value
        this._displayObject = value;

        // Events
        this.image.off('rightup', this.displayMenu, this);

        // Change type
        if (this.displayObject) {
            if (this.displayObject.modeInteraction.lastMode === 'select_item') {
                // -- Rotate element
                if (this.displayObject instanceof MedsurfDraw.PositionPoint
                    || this.displayObject instanceof MedsurfDraw.FillCollection
                    || this.displayObject instanceof MedsurfDraw.LegendCollection // TODO remove
                    || this.displayObject instanceof MedsurfDraw.LegendColumn
                    || this.displayObject instanceof MedsurfDraw.LegendRow
                    || this.displayObject instanceof MedsurfDraw.Line) {
                    // -- Scale element
                    this._scaleElement.type = MedsurfDraw.ScaleElementType.NOT_SCALEABLE;
                    this._scaleElement.showItem();
                    // -- Rotate element
                    this._rotateElement.hideItem();
                } else if (this.displayObject instanceof MedsurfDraw.Text) {
                    // -- Scale element
                    this._scaleElement.type = MedsurfDraw.ScaleElementType.TEXT;
                    this._scaleElement.showItem();
                    // -- Rotate element
                    this._rotateElement.hideItem();
                } else {
                    // -- Scale element
                    this._scaleElement.type = MedsurfDraw.ScaleElementType.SCALEABLE;
                    this._scaleElement.showItem();
                    // -- Rotate element
                    if (this.displayObject instanceof MedsurfDraw.LegendCollection) {
                        this._rotateElement.hideItem();
                    } else if (this._displayObject !== undefined) {
                        this._rotateElement.showItem();
                    }
                }
                // -- Caret element
                if (this.displayObject instanceof MedsurfDraw.EditableText) {
                    this._caretElement.textElement = this.displayObject;
                    this._caretElement.onCaret();
                    this._caretElement.showItem();
                } else {
                    this._caretElement.hideItem();
                }
                // -- Selection element
                if (this.displayObject instanceof MedsurfDraw.EditableText) {
                    this._selectionElement.textElement = this.displayObject;
                    this._selectionElement.showItem();
                } else {
                    this._selectionElement.hideItem();
                }
                // -- Menu element
                if (!(this.displayObject instanceof MedsurfDraw.Text)) {
                    this._menuElement.menuElements = this.displayObject.imageObjectMenu;
                    this.image.parent.on('rightup', this.displayMenu, this);
                } else {
                    this._menuElement.hideItem();
                }
                // -- Clear color button element
                this._colorElement.hideItem();
                // -- Number element
                this._numberElement.hideItem();
            } else if (this.displayObject.modeInteraction.lastMode.startsWith('select_item_color')) {
                // -- Scale element
                this._scaleElement.hideItem();
                // -- Rotate element
                this._rotateElement.hideItem();
                // -- Caret element
                this._caretElement.hideItem();
                // -- Selection element
                this._selectionElement.hideItem();
                // -- Menu element
                this._menuElement.hideItem();
                // -- Clear color button element
                if (this.displayObject instanceof MedsurfDraw.ArrowPrimitive
                    || this.displayObject instanceof MedsurfDraw.EllipsePrimitive
                    || this.displayObject instanceof MedsurfDraw.RectanglePrimitive) {
                    if (this.displayObject.modeInteraction.lastMode === 'select_item_color_line') {
                        this._colorElement.showAlpha = true;
                        this._colorElement.color = this.displayObject.model.options.lineColor;
                        this._colorElement.colorAlpha = this.displayObject.model.options.lineAlpha;
                    } else if (this.displayObject.modeInteraction.lastMode === 'select_item_color_fill') {
                        this._colorElement.showAlpha = true;
                        this._colorElement.color = this.displayObject.model.options.fillColor;
                        this._colorElement.colorAlpha = this.displayObject.model.options.fillAlpha;
                    }
                }
                if (this.displayObject instanceof MedsurfDraw.TextPrimitive) {
                    if (this.displayObject.modeInteraction.lastMode === 'select_item_color_font') {
                        this._colorElement.showAlpha = false;
                        this._colorElement.color = this.displayObject.model.style.fill;
                        this._colorElement.colorAlpha = undefined;
                    }
                }
                this._colorElement.showItem();
                // -- Number element
                this._numberElement.hideItem();
            } else if (this.displayObject.modeInteraction.lastMode.startsWith('select_item_number')) {
                // -- Scale element
                this._scaleElement.hideItem();
                // -- Rotate element
                this._rotateElement.hideItem();
                // -- Caret element
                this._caretElement.hideItem();
                // -- Selection element
                this._selectionElement.hideItem();
                // -- Menu element
                this._menuElement.hideItem();
                // -- Clear color button element
                this._colorElement.hideItem();
                // -- Number element
                if (this.displayObject instanceof MedsurfDraw.ArrowPrimitive
                    || this.displayObject instanceof MedsurfDraw.EllipsePrimitive
                    || this.displayObject instanceof MedsurfDraw.RectanglePrimitive
                    || this.displayObject instanceof MedsurfDraw.TextPrimitive) {
                    const numberData = this.displayObject.getNumberData();
                    this._numberElement.value = numberData.value;
                    this._numberElement.minValue = numberData.minValue;
                    this._numberElement.maxValue = numberData.maxValue;
                    this._numberElement.gab = numberData.gab;
                }
                this._numberElement.showItem();
            } else {
                // -- Scale element
                this._scaleElement.hideItem();
                // -- Rotate element
                this._rotateElement.hideItem();
                // -- Caret element
                this._caretElement.hideItem();
                // -- Selection element
                this._selectionElement.hideItem();
                // -- Menu element
                this._menuElement.hideItem();
                // -- Color element
                this._colorElement.hideItem();
                // -- Number element
                this._numberElement.hideItem();
            }
        } else {
            // -- Scale element
            if (this._scaleElement) {
                this._scaleElement.hideItem();
            }
            // -- Rotate element
            if (this._rotateElement) {
                this._rotateElement.hideItem();
            }
            // -- Caret element
            if (this._caretElement) {
                this._caretElement.hideItem();
            }
            // -- Selection element
            if (this._selectionElement) {
                this._selectionElement.hideItem();
            }
            // -- Menu element
            if (this._menuElement) {
                // Reset menu elements
                this._menuElement.menuElements = undefined;
                this._menuElement.hideItem();
            }
            // -- Clear color button element
            if (this._colorElement) {
                this._colorElement.hideItem();
            }
            // -- Number element
            if (this._numberElement) {
                this._numberElement.hideItem();
            }
        }

        // TODO check for menu -> hide instead

        // Add display object events
        this._setEventsForDisplayObjects(this.displayObject);
    }

    /**
     * Get menu shown
     */
    public get menuShown(): boolean {
        return this._menuShown;
    }
    //</editor-fold>
}