import * as PIXI from 'pixi.js-legacy';
import * as MedsurfDraw from '../../public-api';
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import { Design } from '../../config/design';
import {
    BaseElementContainer,
    BaseElementContainerModel,
    ColorMode,
    NumberMode
} from '../../bases/elements/BaseElementContainer';
import { BaseGraphics } from '../../bases/elements/BaseGraphics';
import { debounce } from 'debounce';

/**
 * Fill collection
 */
export class FillCollection extends BaseElementContainer<Models.FillCollection, BaseElementContainerModel<Models.FillCollection>> {
    /**
     * Members
     */
    private _lines: MedsurfDraw.Line[];
    private _options: Models.ElementOptions;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<FillCollection>;
    private _selectInteraction: MedsurfDraw.SelectInteraction<FillCollection>;
    private _chooseInteraction: MedsurfDraw.ChooseInteraction<FillCollection>;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: any;
    protected _debounceControlUnselectItemMethod: any;

    /**
     * Elements
     */
    private _fillElement: PIXI.Graphics;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.FillCollection) {
        // Parent constructor
        super({image, model});

        // Members
        this.lines = [];
        this.options = Object.assign({}, this.model.options);

        // Setup
        this.zIndex = Design.fillCollection.zIndexDefault + this.model.zIndex || 0;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("remove-init", this._removeModeInit, this);
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on('select_layer_group', this._modeSelectLayerGroup, this);
        this.modeInteraction.on('remove-select_layer_group', this._removeModeSelectLayerGroup, this);
        this.modeInteraction.on('select_grouping', this._modeSelectGrouping, this);
        this.modeInteraction.on('remove-select_grouping', this._removeModeSelectGrouping, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('draw_item', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item", this._removeModeDrawItem, this);
        this.modeInteraction.on('choose_item', this._modeChooseItem, this);
        this.modeInteraction.on("remove-choose_item", this._removeModeChooseItem, this);
        this.modeInteraction.on('choose_select', this._modeChooseSelect, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeChooseSelect, this);
        this.modeInteraction.on('choose_unselect', this._modeChooseUnselect, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeChooseUnselect, this);
        this.modeInteraction.on('clone_item', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item", this._removeModeCloneItem, this);
        this.modeInteraction.on('clone_item_new', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item_new", this._removeModeCloneItem, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        this.modeInteraction.on("delete_multiple", this._modeDelete, this);
        this.modeInteraction.on('delete_draw', this._modeDelete, this);
        this.modeInteraction.on('delete_clone', this._modeDelete, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("onMove", this.onMove, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelectStart", this.onSelectStart, this);
        this.selectInteraction.on("onInteractive", this.onInteractive, this);
        // -- Choose interaction
        this.chooseInteraction = new MedsurfDraw.ChooseInteraction<FillCollection>(this, MedsurfDraw.ChooseTypes.Primitives);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        this._debounceControlUnselectItemMethod = debounce(this.controlUnselectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.on("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.off("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Members
        this.lines = this.model.lines.map((id: string) => {
            return this.image.getChildByName(id) as MedsurfDraw.Line;
            // TODO: return this.getImage().children.find((c) => c.name === id) as MedsurfDraw.Line;
        });

        //<editor-fold desc="Elements">
        // -- Fill element
        this._fillElement = new BaseGraphics({image: this.image});
        this.addChild(this._fillElement);
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>

        //<editor-fold desc="Events">
        // TODO
        this.lines.forEach((line: MedsurfDraw.Line) => {
            // TODO remove this -> prevent some strange behaviour
            // line.on("draw", this.draw, this);
            line.on("onPointUpdate", this.draw.bind(this), this);
            line.on("onPointUpdate", this.emit.bind(this, "onFillMove"), this);
            // line.on("positionPointAdded", this._positionPointAdded, this);
            // line.on("removed", this._onRemoveButton, this);
        });
        //</editor-fold>
    }

    public startDrawing() {
        super.startDrawing();

        //<editor-fold desc="Events">
        this.lines.forEach((line: MedsurfDraw.Line) => {
            line.on("onPointUpdate", this.draw.bind(this), this);
            line.on("onPointUpdate", this.emit.bind(this, "onFillMove"), this);
        });
        //</editor-fold>

        // Emit
        this.emit("debounceDraw");
    }

    public stopDrawing(selfheal_timeout: number) {
        super.stopDrawing(selfheal_timeout);

        //<editor-fold desc="Events">
        this.lines.forEach((line) => {
            line.off("onPointUpdate", this.draw, this);
            line.off("onPointUpdate", this.emit.bind(this, "onFillMove"), this);
        });
        //</editor-fold>
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (this._fillElement && this._fillElement._geometry) {
            this._fillElement.clear();
        }

        if (!this.validate()) {
            return;
        }

        // Begin draw
        if (this.options.hasFill) {
            this._fillElement.beginFill(this.options.fillColor, this.options.fillAlpha);
        } else {
            this._fillElement.beginFill(Design.fillCollection.fillColorDefault, 0.00000001);
        }

        // Get first element and moveTo this
        let line: MedsurfDraw.Line = this.lines[0];
        if (!line.start || !line.end) {
            return;
        }

        let startPoint: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>;
        let endPoint: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>;
        if (line.direction === MedsurfDraw.LineDirection.START_END) {
            startPoint = line.start;
            endPoint  = line.end;
        } else {
            startPoint = line.end;
            endPoint  = line.start;
        }

        this._fillElement.moveTo(startPoint.x, startPoint.y);
        for (let i = 0; i < this.lines.length; i++) {
            switch (line.type) {
                case MedsurfDraw.LineTypes.LINEAR:
                    this._fillElement.lineTo(endPoint.x, endPoint.y);
                    break;
                case MedsurfDraw.LineTypes.QUADRATIC:
                    if (line.controlStart) {
                        this._fillElement.quadraticCurveTo(line.controlStart.x, line.controlStart.y, endPoint.x, endPoint.y);
                    }
                    break;
                case MedsurfDraw.LineTypes.BEZIER:
                    if (line.controlStart && line.controlEnd) {
                        const controlStart = (startPoint === line.start) ? line.controlStart : line.controlEnd;
                        const controlEnd = (startPoint === line.start) ? line.controlEnd : line.controlStart;
                        this._fillElement.bezierCurveTo(controlStart.x, controlStart.y, controlEnd.x, controlEnd.y, endPoint.x, endPoint.y);
                    }
                    break;
            }

            // Get next line
            const nextLine = this.lines.find((nextLine: MedsurfDraw.Line) => {
                return nextLine !== line && (nextLine.start === endPoint || nextLine.end === endPoint);
            });

            if (nextLine) {
                line = nextLine;
                if (nextLine.start === endPoint) {
                    startPoint = nextLine.start;
                    // @ts-ignore
                    endPoint = nextLine.end;
                } else {
                    // @ts-ignore
                    startPoint = nextLine.end;
                    // @ts-ignore
                    endPoint = nextLine.start;
                }
            } else {
                throw new Error('we messed up something');
            }
        }

        this._fillElement.endFill();

        // @ts-ignore
        // TODO remove this._fillElement.hitArea = new PIXI.Polygon(this._fillElement.geometry.points);

        // TODO find other solution for updating fill line
        // TODO hääää debounce(this.emit.bind(this, 'draw'), 20)();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        // -- Choose interaction
        if (this.chooseInteraction) {
            this.chooseInteraction.removeAllListeners();
            // TODO remove delete this.chooseInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Fill element
        if (this._fillElement) {
            this._fillElement.destroy(options);
        }
        // -- Lines
        if (this.lines) {
            this.lines.forEach((line) => {
                // TODO remove this -> prevent some strange behaviour
                // line.off("draw", this.draw, this);
                line.off("onPointUpdate", this.draw, this);
            });
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param ids
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, ids: string[]): FillCollection {
        // Element
        // -- Fill collection
        return new FillCollection(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.FILLCOLLECTION,
            isStatic: false,
            zIndex: 0,
            isInteractive: false,
            lines: ids,
            interactiveItems: [],
            options: {
                hasLine: true,
                lineColor: 0x000000,
                lineAlpha: 1,
                hasFill: true,
                fillColor: 0xffffff,
                fillAlpha: 0.3
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Validate
     */
    public validate(): boolean {
        if (this.lines.length === 0) {
            return false;
        }

        // Find positionPoints with an odd count of lines
        let positionPointReference: BaseElementContainer<Models.ImageObject, BaseElementContainerModel<Models.ImageObject>>[] = [];
        let positionPointReferenceCount: number[] = [];
        this.lines.find((line: MedsurfDraw.Line) => {
            if (!line.start || !line.end) {
                return false;
            }
            try {
                if (!line.start.x || !line.start.y || !line.end.x || !line.start.y) {
                    return false;
                }
            } catch (e) {
                // Stop drawing with 20 sec selfheal
                this.stopDrawing(20000);

                // TODO
                // Send an error to the nav
                this.image.sendError(e);

                // Set mode
                // this.modeInteraction.setMode("delete");

                return false;
            }

            const startIndex = positionPointReference.findIndex((ref) => ref === line.start);
            const endIndex = positionPointReference.findIndex((ref) => ref === line.end);

            if (startIndex > -1) {
                positionPointReferenceCount[startIndex]++;
            } else {
                positionPointReference.push(line.start);
                positionPointReferenceCount.push(1);
            }

            if (endIndex > -1) {
                positionPointReferenceCount[endIndex]++;
            } else {
                positionPointReference.push(line.end);
                positionPointReferenceCount.push(1);
            }
        });

        // Return when an positionPoint has an odd count of lines (At the moment we only cover simple shapes with 2 lines on one PositionPoint)
        // TODO use this when we want to cover something like an "8" shape: if (positionPointReferenceCount.some((count) => count % 2 !== 0)) {
        if (positionPointReferenceCount.some((count) => count !== 2)) {
            return false;
        }

        return true;
    }

    /**
     * Is point in polygon
     *
     * @see https://stackoverflow.com/questions/217578/how-can-i-determine-whether-a-2d-point-is-within-a-polygon
     * @param point
     */
    public isPointInPolygon(point: PIXI.IPoint): boolean {
        // Check for polygon
        // ts-ignore
        if (this._fillElement.geometry.graphicsData.length === 0) {
            return false;
        }

        /* TODO use this
        console.log("contains", this._fillElement.containsPoint(point));
        return this._fillElement.containsPoint(point);
        */

        // Get polygon
        // @ts-ignore
        const polygon = this._fillElement.geometry.graphicsData[0].shape.points.slice(0, -2);

        //
        let inside = false;
        let minX = polygon[0], maxX = polygon[0];
        let minY = polygon[1].y, maxY = polygon[1];
        for (let n = 1; n < polygon.length; n+=2) {
            let qx = polygon[n];
            let qy = polygon[n+1];
            minX = Math.min(qx, minX);
            maxX = Math.max(qx, maxX);
            minY = Math.min(qy, minY);
            maxY = Math.max(qy, maxY);
        }

        if (point.x < minX || point.x > maxX || point.y < minY || point.y > maxY) {
            return false;
        }

        let i, j;
        for (i = 0, j = polygon.length - 1; i < polygon.length; j = i, i += 2) {
            // @ts-ignore
            if(((polygon[i+1] > point.y) != (polygon[j+1] > point.y))
                && (point.x < (polygon[j]-polygon[i]) * (point.y-polygon[i+1])
                    / (polygon[j+1]-polygon[i+1]) + polygon[i]) )
                inside = !inside;
        }
        return inside;
    }

    /**
     * Get nearest point for point
     * @param point
     */
    public getNearestPointForPoint(point: PIXI.IPoint): PIXI.IPoint {
        let line: MedsurfDraw.Line;
        let nearestDelta: number | undefined = undefined;
        let nearestPoint: PIXI.Point;

        let footDelta: number;
        let foot: PIXI.Point;
        this.lines.forEach((l: MedsurfDraw.Line, i) => {
            // @ts-ignore
            let tmpStart = new PIXI.Point(l.start.position.x, l.start.position.y);
            // @ts-ignore
            let tmpEnd = new PIXI.Point(l.end.position.x, l.end.position.y);

            // Get nearest edge point
            let dStart = Math.abs(point.x - tmpStart.x) + Math.abs(point.y - tmpStart.y);
            let dEnd = Math.abs(point.x - tmpEnd.x) + Math.abs(point.y - tmpEnd.y);
            if (!nearestDelta) {
                nearestPoint = (dStart < dEnd) ? tmpStart : tmpEnd;
                nearestDelta = (dStart < dEnd) ? dStart : dEnd;
            } else {
                if (nearestDelta > ((dStart < dEnd) ? dStart : dEnd)) {
                    nearestPoint = (dStart < dEnd) ? tmpStart : tmpEnd;
                    nearestDelta = (dStart < dEnd) ? dStart : dEnd;
                }
            }

            // Check for point on line
            const tmpFoot = this._foot(tmpStart, tmpEnd, point);
            if (!this._isOnLine(tmpStart, tmpEnd, tmpFoot)) {
                return;
            }

            const tmpDelta = Math.abs(point.x - tmpFoot.x) + Math.abs(point.y - tmpFoot.y);

            if (line === undefined) {
                footDelta = tmpDelta;
                foot = tmpFoot;
                line = l;
            } else {
                if (tmpDelta < footDelta) {
                    footDelta = tmpDelta;
                    foot = tmpFoot;
                    line = l;
                }
            }
        });

        // @ts-ignore
        if (nearestPoint) {
            // @ts-ignore
            if (foot) {
                // @ts-ignore
                if (Math.round(footDelta * 100) / 100 <= Math.round(nearestDelta * 100) / 100) {
                    // @ts-ignore
                    if (line) {
                        if (line.model.isBezier && line.controlStart !== null) {
                            let rDelta: number | undefined = undefined;
                            let rX: number;
                            let rY: number;

                            // @ts-ignore
                            for (let i = 0; i < line._lineElement.geometry.graphicsData[0].shape.points.length; i = i + 2) {
                                // @ts-ignore
                                const tmpX = line._lineElement.geometry.graphicsData[0].shape.points[i];
                                // @ts-ignore
                                const tmpY = line._lineElement.geometry.graphicsData[0].shape.points[i + 1];

                                const tmpDelta = Math.abs(point.x - tmpX) + Math.abs(point.y - tmpY);

                                if (rDelta === undefined) {
                                    rDelta = tmpDelta;
                                    rX = tmpX;
                                    rY = tmpY;
                                } else {
                                    if (tmpDelta < rDelta) {
                                        rDelta = tmpDelta;
                                        rX = tmpX;
                                        rY = tmpY;
                                    }
                                }
                            }

                            // @ts-ignore
                            if (rX && rY) {
                                return new PIXI.Point(rX, rY);
                            }
                        }
                    }
                    return foot;
                }
            }
            return nearestPoint;
        }
        throw 'No point found';
    }

    /**
     * Get middle point for fills
     * @param fill
     */
    /* TODO remove not needed
    public getMiddlePointForFills(fill: FillCollection): PIXI.IPoint {
        // TODO use the graphics data here
        let myLine: MedsurfDraw.Line;
        let otherLine: MedsurfDraw.Line;

        let distance: number | undefined = undefined;
        let calcDistances: number[];
        this.lines.forEach((tmpMyline: MedsurfDraw.Line) => {
            fill.lines.forEach((tmpOtherLine: MedsurfDraw.Line) => {
                const tmpCalcDistances = [
                    // @ts-ignore
                    Math.abs(tmpMyline.start.x - tmpOtherLine.start.x) + Math.abs(tmpMyline.start.y - tmpOtherLine.start.y),
                    // @ts-ignore
                    Math.abs(tmpMyline.start.x - tmpOtherLine.start.x) + Math.abs(tmpMyline.end.y - tmpOtherLine.end.y),
                    // @ts-ignore
                    Math.abs(tmpMyline.end.x - tmpOtherLine.end.x) + Math.abs(tmpMyline.start.y - tmpOtherLine.start.y),
                    // @ts-ignore
                    Math.abs(tmpMyline.end.x - tmpOtherLine.end.x) + Math.abs(tmpMyline.end.y - tmpOtherLine.end.y)
                ];

                let minDistance = Math.min.apply(Math, tmpCalcDistances);
                if (distance === undefined) {
                    distance = minDistance;
                    calcDistances = tmpCalcDistances;

                    myLine = tmpMyline;
                    otherLine = tmpOtherLine;
                } else {
                    if (minDistance < distance) {
                        distance = minDistance;
                        calcDistances = tmpCalcDistances;

                        myLine = tmpMyline;
                        otherLine = tmpOtherLine;
                    }
                }
            });
        });

        // @ts-ignore
        return new PIXI.Point(
            // @ts-ignore
            (myLine.start.x + myLine.end.x + otherLine.start.x + otherLine.end.x) / 4,
            // @ts-ignore
            (myLine.start.y + myLine.end.y + otherLine.start.y + otherLine.end.y) / 4
        );
    }
    */

    /**
     * Add line
     * @param line
     */
    public addLine(line: MedsurfDraw.Line): void {
        // Add line
        this.lines.push(line);

        // Update model
        this.model.lines.push(line.model.id as string);

        // Add events
        line.on("onPointUpdate", this.draw.bind(this), this);
        line.on("onPointUpdate", this.emit.bind(this, "onFillMove"), this);
    }

    /**
     * Remove line
     * @param line
     */
    public removeLine(line: MedsurfDraw.Line): boolean {
        // Get index
        const lineIndex = this.lines.indexOf(line);

        if (lineIndex > -1) {
            // Remove line
            this.lines.splice(lineIndex, 1);

            // Update model
            this.model.lines.splice(lineIndex, 1);

            // Remove events
            line.off("onPointUpdate", this.draw, this);

            return true;
        }
        return false;
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Foot
     * @param start
     * @param end
     * @param point
     * @private
     */
    protected _foot(start: PIXI.IPoint, end: PIXI.IPoint, point: PIXI.IPoint): PIXI.Point {
        const line = {
            x: end.x - start.x,
            y: end.y - start.y
        };
        const k = ((point.x - start.x) * line.x + (point.y - start.y) * line.y) / (line.x * line.x + line.y * line.y);
        return new PIXI.Point(start.x + k * line.x, start.y + k * line.y);
    }

    /**
     * Is on line
     * @param start
     * @param end
     * @param foot
     * @private
     */
    protected _isOnLine(start: PIXI.IPoint, end: PIXI.IPoint, foot: PIXI.IPoint): boolean {
        const dxc = foot.x - start.x;
        const dyc = foot.y - start.y;

        const dxl = end.x - start.x;
        const dyl = end.y - start.y;

        const cross = dxc * dyl - dyc * dxl;

        if (Math.abs(cross) < 0.000001) {
            if (Math.abs(dxl) >= Math.abs(dyl)) {
                if ((dxl > 0) ? (start.x <= foot.x && foot.x <= end.x) : (end.x <= foot.x && foot.x <= start.x)) {
                    return true;
                }
            } else {
                if ((dyl > 0) ? (start.y <= foot.y && foot.y <= end.y) : (end.y <= foot.y && foot.y <= start.y)) {
                    return true;
                }
            }
        }
        return false;
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Remove mode init
     * @private
     */
    protected _removeModeInit(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        if (this.model.isInteractive) {
            // Setup
            this.options.fillAlpha = Design.fillCollection.fillAlphaInteractive;

            // Events
            this.interactive = true;
            this.on("pointerover", this.selectInteraction.onInteractive, this.selectInteraction);
            this.on("pointerout", this.selectInteraction.onRelease, this.selectInteraction);

            // Show item
            this.showItem();
        } else {
            if (this.model.isStatic) {
                // Show item
                this.showItem();
            } else {
                // Hide item
                this.hideItem();
            }
        }
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
        // Setup
        this.options = Object.assign({}, this.model.options);

        // Events
        this.interactive = false;
        this.off("pointerover", this.selectInteraction.onInteractive, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onRelease, this.selectInteraction);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        // Show item
        this.showItem();
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        if (this.model.isInteractive) {
            // Setup
            this.interactive = true;
            this.cursor = "select";
            this.options.fillAlpha = Design.fillCollection.fillAlphaInteractive;

            // Events
            this.selectInteraction.reset();
            this.on("pointerover", this.selectInteraction.onInteractive, this.selectInteraction);
            this.on("pointerout", this.selectInteraction.onRelease, this.selectInteraction);

            // Show item
            this.showItem();
        } else {
            if (this.model.isStatic) {
                // Show item
                this.showItem();
            } else {
                // Hide item
                this.hideItem();
            }
        }
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Setup
        this.options = Object.assign({}, this.model.options);

        // Events
        this.interactive = false;
        this.off("pointerover", this.selectInteraction.onInteractive, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onRelease, this.selectInteraction);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select layer
     * @private
     */
    protected _modeSelectLayerGroup(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select layer
     * @private
     */
    protected _removeModeSelectLayerGroup(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select grouping
     * @private
     */
    protected _modeSelectGrouping(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select grouping
     * @private
     */
    protected _removeModeSelectGrouping(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.on("startMove", this._startMove, this);
        // TODO this.on("endMove", this._endMove, this);
        // TODO this.on("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.on("down_ArrowUp", "onPositionPointMoveUp", this.moveInteraction.moveUp.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowRight", "onPositionPointMoveRight", this.moveInteraction.moveRight.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowDown", "onPositionPointMoveDown", this.moveInteraction.moveDown.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowLeft", "onPositionPointMoveLeft", this.moveInteraction.moveLeft.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowUp", "onPositionPointMoveUpEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowRight", "onPositionPointMoveRightEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowDown", "onPositionPointMoveDownEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowLeft", "onPositionPointMoveLeftEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.off("startMove", this._startMove, this);
        // TODO this.off("endMove", this._endMove, this);
        // TODO this.off("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.remove("down_ArrowUp", "onPositionPointMoveUp");
        MedsurfDraw.Keyboard.events.remove("down_ArrowRight", "onPositionPointMoveRight");
        MedsurfDraw.Keyboard.events.remove("down_ArrowDown", "onPositionPointMoveDown");
        MedsurfDraw.Keyboard.events.remove("down_ArrowLeft", "onPositionPointMoveLeft");
        MedsurfDraw.Keyboard.events.remove("released_ArrowUp", "onPositionPointMoveUpEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowRight", "onPositionPointMoveRightEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowDown", "onPositionPointMoveDownEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowLeft", "onPositionPointMoveLeftEnd");

        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // TODO remove console.log("fill _removeModeSelectItem", mode);
        if (mode === 'author') {
            this.emit("debounceControlUnselectItem", this.image);
            // this.controlUnselectItem();
        }
        /* TODO remove
        if (!mode.startsWith('select_child') && !mode.startsWith('select_item') && !mode.startsWith('choose_item')
            && !mode.startsWith('choose_select') && !mode.startsWith('blocked_item')
            && ['default', 'marker', 'selftest'/*, 'author'* /, 'delete'].indexOf(mode) === -1) {
            // Unselect item
            this.emit("debounceControlUnselectItem");
        }
        */
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
        this.options.lineAlpha = Design.blocked.lineAlpha;
        this.options.fillAlpha = Design.blocked.fillAlpha;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        this.filters = [];
        */
        this.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode draw item
     * @private
     */
    protected _modeDrawItem(): void {
    }

    /**
     * Remove mode draw item
     * @private
     */
    protected _removeModeDrawItem(): void {
    }

    /**
     * Mode choose item
     * @private
     */
    protected _modeChooseItem(): void {
        // Interactions
        // -- Choose interaction
        this.chooseInteraction.on("chooseElement", this.chooseElement, this);
    }

    /**
     * Remove mode choose item
     * @private
     */
    protected _removeModeChooseItem(): void {
        // Interactions
        // -- Choose interaction
        this.chooseInteraction.off("chooseElement", this.chooseElement, this);
    }

    /**
     * Mode choose select
     * @private
     */
    protected _modeChooseSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";
        this.options.lineColor = Design.selftest.lineColorSelect;
        this.options.fillColor = Design.selftest.fillColorSelect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose select
     * @private
     */
    protected _removeModeChooseSelect(): void {
        // Setup
        this.interactive = false;
        this.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose unselect
     * @private
     */
    protected _modeChooseUnselect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";
        this.options.lineColor = Design.selftest.lineColorUnselect;
        this.options.fillColor = Design.selftest.fillColorUnselect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose unselect
     * @private
     */
    protected _removeModeChooseUnselect(): void {
        // Setup
        this.interactive = false;
        this.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode clone item
     * @private
     */
    protected _modeCloneItem(): void {
        this.validate();
        this.emit("debounceDraw");
    }

    /**
     * Remove mode clone item
     * @private
     */
    protected _removeModeCloneItem(): void {
        this.moveInteraction.reset();
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        const image = this.image;

        if (image) {
            // Select layer group
            if (['delete_multiple', 'delete_draw', 'delete_clone'].indexOf(this.modeInteraction.lastMode) === -1) {
                this.controlUnselectItem(image);
            }

            // Delete from parent
            image.removeChild(this);

            // Emits
            image.deleteImageObject(this.model);
        }

        this.destroy();

        // Hide menu
        image.controlHideMenu();

        // Update control elements
        image.controlUpdateElements();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = Design.fillCollection.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = Design.fillCollection.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            // Set mode
            this.modeInteraction.setMode("delete");
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.degroupLayerGroupImageObject(this.model);
        if (unselectElement) {
            this.image.unselectLayerGroup();
            this.image.selectImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        this.image.copyImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        this.image.clipboardImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Interactive">
    /**
     * On button set interactive
     * @param override
     */
    // TODO
    public onButtonSetInteractive(override?: boolean): void {
        this.model.isInteractive = !this.model.isInteractive;

        /* TODO no color defined
        this._lineColor = this.model.isSelftest ? this._lineColorSelftest : this._lineColorSelected;
        this._fillColor = this.model.isSelftest ? this._fillColorSelftest : this._fillColorSelected;
        this.draw();
        */

        if (this.model.isInteractive) {
            /* TODO not needed
            this._selftestContainerElement.addImageObject(this._selftestChooseButton);
            this._selftestContainerElement.draw();
            */
        } else {
            const imageObjects = this.image.getImageObjects();
            const items = this.model.interactiveItems;
            this.model.interactiveItems = []; // Remove existing items
            if (items) {
                items.forEach((item) => {
                    const element = imageObjects.find((io) => io.name === item) as MedsurfDraw.DrawableObject;
                    if (element) {
                        element.modeInteraction.setMode('blocked_item');
                    }
                });
            }
            // TODO not needed this._selftestContainerElement.removeImageObject(this._selftestChooseButton);
        }

        // Draw
        this.emit("debounceDraw");

        // Update
        // TODO not needed this.emit("interactiveChanged");
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Bezier">
    public onButtonBezier(override?: boolean): void {
        this.lines.forEach((line) => line.onButtonBezier(override));

        // Draw
        this.emit("debounceDraw");

        // Control update elements
        this.image.controlUpdateElements();
    }
    //</editor-fold>
    //<editor-fold desc="Color">
    /**
     * On color
     * @param mode
     * @param has
     * @param color
     * @param colorAlpha
     */
    public onColor(mode: ColorMode, has: boolean, color: number | undefined, colorAlpha: number | undefined): void {
        if (mode === ColorMode.LINE) {
            this.lines.forEach((line) => {
                line.onColor(mode, has, color, colorAlpha);
            });
        } else if (mode === ColorMode.FILL) {
            // Update model
            this.model.options.hasFill = has;
            this.model.options.fillColor = color;
            this.model.options.fillAlpha = colorAlpha;

            // Element
            this.options = Object.assign({}, this.model.options);

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updateImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Number">
    /**
     * On number
     * @param mode
     * @param number
     */
    public onNumber(mode: NumberMode, number: number): void {
        if (mode === NumberMode.WIDTH) {
            this.lines.forEach((line) => {
                line.onNumber(mode, number);
            });
        }
    }
    //</editor-fold>
    //<editor-fold desc="Static">
    public onButtonStatic(override?: boolean): void {
        this.model.isStatic = override || !this.model.isStatic;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectImageObject(this.model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Control unselect item
     * @param image
     */
    public controlUnselectItem(image: MedsurfDraw.Image): void {
        // Set select layer
        if(!this.controlSelectLayerGroup(image)) {
            // Set select image
            image.externControlSelectImage();

            // Emits
            image.unselectImageObject();
        }
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.interactiveItems && this.model.interactiveItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else if (this.model.interactiveItems && imageObject instanceof MedsurfDraw.LegendCollection) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => { return {name: lr.model.id, element: lr}; });
                    for (let legendRow of legendRows) {
                        if (this.model.interactiveItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.modeInteraction.setMode('choose_select');
                        } else {
                            legendRow.element.modeInteraction.setMode('blocked_item');
                        }
                    }
                } else {
                    imageObject.modeInteraction.setMode('blocked_item');
                }
            });

        // Set children mode
        this.modeInteraction.setModeChildren('select_parent');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control select layer group
     * @param image
     */
    public controlSelectLayerGroup(image: MedsurfDraw.Image): boolean {
        // Get layer group
        const layerGroup = image.getLayerGroupByImageObject(this.model);
        if (layerGroup) {
            // Emits
            image.selectLayerGroup(layerGroup);

            // Select layer group
            return image.externControlSelectLayerGroup(layerGroup);
        }

        // Layer not selected
        return false;
    }

    /**
     * Control unselect layer group
     */
    public controlUnselectLayerGroup(): void {
        // Emits
        this.image.unselectLayerGroup();
    }

    /**
     * Control draw item
     */
    public controlDrawItem(): void {
        // Set children mode
        this.modeInteraction.setModeChildren('draw_parent');

        // Set draw item
        this.modeInteraction.setModeItem('draw_item');
    }

    /**
     * Control choose item
     */
    public controlChooseItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.interactiveItems && this.model.interactiveItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else if (this.model.interactiveItems && imageObject instanceof MedsurfDraw.LegendCollection) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => { return {name: lr.model.id, element: lr}; });
                    for (let legendRow of legendRows) {
                        if (this.model.interactiveItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.modeInteraction.setMode('choose_select');
                        } else {
                            legendRow.element.modeInteraction.setMode('choose_unselect');
                        }
                    }
                } else {
                    imageObject.modeInteraction.setMode('choose_unselect');
                }
            });

        // Set children mode
        this.modeInteraction.setModeChildren('choose_parent');

        // Set draw item
        this.modeInteraction.setModeItem('choose_item');
    }

    /**
     * Control clone item
     */
    public controlCloneItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this && !imageObject.modeInteraction.lastMode.startsWith('clone'))
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('clone_parent');

        // Set draw item
        this.modeInteraction.setModeItem('clone_item_new');
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        this.lines.forEach((line: MedsurfDraw.Line) => {
            if (line.direction === MedsurfDraw.LineDirection.START_END) {
                if (line.start instanceof MedsurfDraw.PositionPoint || line.start instanceof MedsurfDraw.FillCollection) {
                    line.start.moveInteraction.emit("onMove", event, dX, dY);
                }
            } else {
                if (line.end instanceof MedsurfDraw.PositionPoint || line.end instanceof MedsurfDraw.FillCollection) {
                    line.end.moveInteraction.emit("onMove", event, dX, dY);
                }
            }
        });

        // Emit on fill move
        this.emit("onFillMove");
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Setup
        this.options.lineColor = Design.primitive.lineColorHover;
        this.options.fillColor = Design.primitive.fillColorHover;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Setup
        this.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_layer_group') {
            this.controlSelectItem();
        } else {
            if (!this.controlSelectLayerGroup(this.image) && !preventSelectItem) {
                this.controlSelectItem();
            }
        }
    }

    /**
     * On interactive
     * @private
     */
    public onInteractive(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.interactiveItems) {
                    if (this.model.interactiveItems && this.model.interactiveItems.indexOf(imageObject.name as string) > -1) {
                        imageObject.showItem();
                    } else if (imageObject instanceof MedsurfDraw.LegendCollection) {
                        const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                            .reduce((cur, acc: any) => [...cur, ...acc], [])
                            .map((lr) => { return {name: lr.model.id, element: lr}; });
                        for (let legendRow of legendRows) {
                            if (this.model.interactiveItems.indexOf(legendRow.name) > -1) {
                                legendRow.element.showItem();
                            }
                        }
                    } else {
                        imageObject.hideItem();
                    }
                } else {
                    imageObject.hideItem();
                }
            });

        // Release event
        this.selectInteraction.once('onRelease', this.onInteractiveRelease, this);

        // Setup
        this.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On interactive release
     * @param event
     */
    public onInteractiveRelease(event: PIXI.InteractionEvent): void {
        // Reset select interaction
        this.selectInteraction.onRelease(event);

        // Set default mode for image children
        const defaultMode = this.image.modeInteraction.defaultMode;
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode(defaultMode);
            });

        // Setup
        this.options.fillAlpha = Design.fillCollection.fillAlphaInteractive;

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Choose">
    /**
     * Choose element
     * @param event
     * @param element
     */
    public chooseElement(event: PIXI.InteractionEvent, element: MedsurfDraw.DrawableObject): void {
        // Check model
        if (!this.model.interactiveItems) {
            this.model.interactiveItems = [];
        }

        const index = this.model.interactiveItems.findIndex((item: string) => item === element.name);
        if (index > -1) {
            // Remove from interactive items
            this.model.interactiveItems.splice(index, 1);

            // Set element mode unselect
            element.modeInteraction.setMode('choose_unselect');
        }  else {
            // Add to interactive items
            this.model.interactiveItems.push(element.name as string);

            // Set element mode select
            element.modeInteraction.setMode('choose_select');
        }
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<FillCollection> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<FillCollection>) {
        this._moveInteraction = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<FillCollection> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<FillCollection>) {
        this._selectInteraction = value;
    }

    /**
     * Get choose interaction
     */
    public get chooseInteraction(): MedsurfDraw.ChooseInteraction<FillCollection> {
        return this._chooseInteraction;
    }

    /**
     * Set choose interaction
     * @param value
     */
    public set chooseInteraction(value: MedsurfDraw.ChooseInteraction<FillCollection>) {
        this._chooseInteraction = value;
    }

    /**
     * Get lines
     */
    public get lines(): MedsurfDraw.Line[] {
        return this._lines;
    }

    /**
     * Set lines
     * @param value
     */
    public set lines(value: MedsurfDraw.Line[]) {
        this._lines = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this._options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this._options = value;
    }
    //</editor-fold>
}