import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import { Image } from "./Image";
import { Design } from "../../config/design";
import { debounce } from "debounce";

/**
 * Deep zoom image
 */
export class DeepZoomImage extends Image {
    /**
     * Members
     */
    protected _callbackId: any;// TODO PIXI.Loader.Loader.ICallbackID;
    protected _tilesUrl: string;
    protected _minLevel: number;
    protected _maxLevel: number;
    protected _imageSizes: PIXI.IPoint[];
    protected _gridSizes: PIXI.IPoint[];
    protected _levelScaleCache: number[];

    /**
     * Methods
     */
    protected _debounceLoadVisibleTilesMethod: EventListener;

    /**
     * Elements
     */
    protected _tileContainer: MedsurfDraw.Container;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: MedsurfDraw.ImageModel) {
        // Parent constructor
        super(model, MedsurfDraw.ImageTypes.DEEPZOOM);

        // Setup
        this._tilesUrl = model.url;

        // Setup tile size
        const tileSize = this.dimensions.tileSize || Design.image.tileSize;

        // Image and gris sizes
        // -- Image sizes
        this._imageSizes = [new PIXI.Point(this.dimensions.width, this.dimensions.height)];
        // -- Grid sizes
        this._gridSizes = [new PIXI.Point(this.dimensions.width / tileSize, this.dimensions.height / tileSize)];
        // ---- Iterate through image size
        const currentImageSize = new PIXI.Point(this.dimensions.width, this.dimensions.height);
        while (currentImageSize.x > tileSize || currentImageSize.y > tileSize) {
            currentImageSize.x = currentImageSize.x / 2;
            currentImageSize.y = currentImageSize.y / 2;

            // Setup
            this._imageSizes.push(new PIXI.Point(currentImageSize.x, currentImageSize.y));
            this._gridSizes.push(new PIXI.Point(currentImageSize.x / tileSize, currentImageSize.y / tileSize));
        }
        this._gridSizes.reverse();
        // -- Min level
        this._minLevel = 0;
        // -- Max level
        this._maxLevel = this._gridSizes.length - 1;
        // -- Level scale cache
        this._levelScaleCache = [];
        // ---- Iterate through levels
        for (let i = 0; i <= this._maxLevel; i++) {
            this._levelScaleCache[i] = 1 / Math.pow(2, this._maxLevel - i);
        }

        //<editor-fold desc="Interactions">
        // -- Move interaction
        this.moveInteraction.on("onMove", this.onTileMove, this);
        this.moveInteraction.on("endMove", this.onTileMove, this);
        // -- Zoom interaction
        this.zoomInteraction.on("onZoom", this.onTileZoom, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceLoadVisibleTilesMethod = debounce(this.loadVisibleTiles.bind(this), Design.deepZoomImage.loadDebounce).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Loader
        // TODO in index this.loader._queue.kill();
        this._callbackId = this.loader.onLoad.add(this.onLoad.bind(this));
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceLoadVisibleTiles", this._debounceLoadVisibleTilesMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceLoadVisibleTiles", this._debounceLoadVisibleTilesMethod);
        });
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.off("onMove", this.onTileMove, this);
            this.moveInteraction.off("endMove", this.onTileMove, this);
        }
        // -- Zoom interaction
        if (this.zoomInteraction) {
            this.zoomInteraction.off("onZoom", this.onTileZoom, this);
        }
        //</editor-fold>

        // Events
        // -- Loader
        this.loader.onLoad.detach(this._callbackId);

        //<editor-fold desc="Elements">
        // -- Tile container
        if (this._tileContainer) {
            this._tileContainer.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Get current level
     * @private
     */
    public getCurrentLevel() {
        // Get factor
        const factor = this.canvas.width * this.getNavigatorScale().x;

        // Get current zero ratio
        const currentZeroRatio = this._getPixelRatio(0).x * factor;

        // return highest level
        return Math.max(Math.min(Math.abs(this._maxLevel), Math.abs(Math.floor(Math.log(currentZeroRatio / Design.deepZoomImage.minPixelRatio) / Math.log(2)))), this._minLevel || 0);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Get pixel ratio
     * @param level
     * @private
     */
    protected _getPixelRatio(level: number): PIXI.Point {
        // Get factor
        const factor = this._levelScaleCache[level];

        // Get image size scaled
        const imageSizeScaled = new PIXI.Point(this.dimensions.width * factor, this.dimensions.height * factor);

        // Return pixel ratio
        return new PIXI.Point(1.0 / imageSizeScaled.x, 1.0 / imageSizeScaled.y);
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Remove mode init
     * @private
     */
    protected _modeRemoveInit(): void {
        //<editor-fold desc="Elements">
        // -- Tile container
        this._tileContainer = new MedsurfDraw.Container({image: this});
        this.addChild(this._tileContainer);
        //</editor-fold>

        // Load level 0 to initialLoadingLevel
        for (let level = 0; level <= Design.deepZoomImage.initialLoadingLevel; level++) {
            this.loadWholeLevel(level);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Tile">
    /**
     * Load whole level
     * @param level
     */
    public loadWholeLevel(level: number): void {
        // Get grid sizes
        const gridSize = this._gridSizes[level];

        // Load tiles
        for (let x = 0; x <= Math.ceil(gridSize.x); x++) {
            for (let y = 0; y <= Math.ceil(gridSize.y); y++) {
                this.loadTile(level, x, y);
            }
        }
    }

    /**
     * Load visible tiles
     */
    public loadVisibleTiles(): void {
        // Get level
        const level = this.getCurrentLevel();

        // Prevent load of low level images
        if (level <= Design.deepZoomImage.initialLoadingLevel) {
            return;
        }
        // Get tile size
        const tileSize = this.dimensions && this.dimensions.tileSize ? this.dimensions.tileSize : 256;

        // Get bounds
        const bounds = this.getBounds(true);

        // Calculate points
        // -- left x
        const scale = this.getNavigatorScale();
        let leftX = (((bounds.x < 0) ? bounds.x / scale.x * -1 : 0) || 0);
        if (leftX < 0) {
            leftX = 0;
        }
        // -- top y
        let topY = (((bounds.y < 0) ? bounds.y / scale.y * -1 : 0) || 0);
        if (topY < 0) {
            topY = 0;
        }
        // -- right x
        let rightX = bounds.width / scale.x;
        if ((bounds.x + bounds.width) > this.canvas.width) {
            rightX = (bounds.x - this.canvas.width) * -1 / scale.x;
        }
        // -- bottom y
        let bottomY = bounds.height / scale.y;
        if ((bounds.y + bounds.height) > this.canvas.height) {
            bottomY = (bounds.y - this.canvas.height) * -1 / scale.y;
        }

        // Get grid size
        const gridSize = this._gridSizes[level];

        // Calculate grid sizes
        const gridImgWidth = gridSize.x * tileSize;
        const gridImgHeight = gridSize.y * tileSize;

        // Calculate min and max values
        // -- min x
        let minX = Math.floor(leftX / this.actualWidth * gridImgWidth / tileSize) - Design.deepZoomImage.tileMargin; // Math.floor(leftX / this.canvas.width * this.actualWidth / displayWidth) - 1;
        if (minX < 0) {
            minX = 0;
        }
        // -- min y
        let minY = Math.floor(topY / this.actualHeight * gridImgHeight / tileSize) - Design.deepZoomImage.tileMargin; // Math.floor(topY / this.canvas.height * this.actualHeight / displayHeight) - 1;
        if (minY < 0) {
            minY = 0;
        }
        // -- max x
        let maxX = Math.ceil(rightX / this.actualWidth * gridImgWidth / tileSize) + Design.deepZoomImage.tileMargin; // Math.ceil(rightX / this.canvas.width * this.actualWidth / displayWidth) + 2;
        if (maxX > gridSize.x) {
            maxX = Math.ceil(gridSize.x);
        }
        // -- max y
        let maxY = Math.ceil(bottomY / this.actualHeight * gridImgHeight / tileSize) + Design.deepZoomImage.tileMargin; // Math.ceil(bottomY / this.canvas.height * this.actualHeight / displayHeight) + 2;
        if (maxY > gridSize.y) {
            maxY = Math.ceil(gridSize.y);
        }

        // Reset loader
        if (this.loader.loading) {
            this.loader.reset();
        }
        // TODO remove this -> slower than the loader reset
        // this.loader._queue.kill();

        // Load tiles
        // TODO algo not correct
        this.loadFromMidToBounds(level, minX, maxX, minY, maxY);
        /* TODO remove: old logic starts loading from top_left to bottom_right */
        for (let x = minX; x <= maxX; x++) {
            for (let y = minY; y <= maxY; y++) {
                this.loadTile(level, x, y);
            }
        }
        /**/

        // Start loading
        this.loader.load();
        this.loader.loading = false;
    }

    /**
     * Hide tiles
     */
    public hideTiles(): void {
        // Get level
        const level = this.getCurrentLevel();

        // Get all sprites from other level an hide them
        if (level > 3) {
            this._tileContainer.children
                .filter((sprite) => !((sprite as MedsurfDraw.Sprite).name as string).startsWith(this.id + '-' + level))
                .forEach((sprite) => {
                    (sprite as MedsurfDraw.Sprite).hideItem();
                });
        }
    }

    /**
     * Load from mid to bounds
     * @param level
     * @param minX
     * @param maxX
     * @param minY
     * @param maxY
     */
    public loadFromMidToBounds(level: number, minX: number, maxX: number, minY: number, maxY: number): void {
        // Calculate mid points
        // -- x
        const midX = (maxX + minX) / 2;
        const xDecimal = midX % 1 !== 0;
        const lowMidX = Math.floor(midX);
        const highMidX = Math.ceil(midX);
        // -- y
        const midY = (maxY + minY) / 2;
        const yDecimal = midY % 1 !== 0;
        const lowMidY = Math.floor(midY);
        const highMidY = Math.ceil(midY);

        // TODO this is wrong -> we need another solution...
        // -- Load from mid to bounds
        for (let x = 0; x + highMidX <= maxX; x++) {
            for (let y = 0; y + highMidY <= maxY; y++) {
                // load tiles above
                this.loadTile(level, highMidX + x, highMidY + y);
                if (xDecimal) {
                    this.loadTile(level, lowMidX - x, highMidY + y);
                    if (yDecimal) {
                        this.loadTile(level, lowMidY - x, lowMidY - y);
                    }
                }
                if (yDecimal) {
                    this.loadTile(level, highMidX + x, lowMidY - y);
                }
            }
        }
    }

    /**
     * Load tile
     * @param level
     * @param x
     * @param y
     */
    public loadTile(level: number, x: number, y: number): void {
        // Check if tile already loaded
        if (!this.loader.resources[this.id + '-' + level + '-' + x + '-' + y]) {
            // Load tile with loader
            const url = this.getTileUrl(level, x, y);
            this.loader.add(this.id + '-' + level + '-' + x + '-' + y, url, { crossOrigin: true });
        } else {
            // Draw tile
            this.drawTile(this.loader.resources[this.id + '-' + level + '-' + x + '-' + y]);
        }
    }

    /**
     * Draw tile
     * @param resource
     */
    public drawTile(resource: PIXI.LoaderResource): void {
        if (!resource.isComplete || !resource.texture) {
            return;
        }

        // Get data from name
        const name = /^[0-9]+-([0-9]+)-([0-9]+)-([0-9]+)/.exec(resource.name);
        if (!name) {
            return;
        }
        const level = parseInt(name[1]);
        const x = parseInt(name[2]);
        const y = parseInt(name[3]);

        const tileSize = this.dimensions && this.dimensions.tileSize ? this.dimensions.tileSize : 256;
        const gridSize = this._gridSizes[level];

        // Check if tile already exists
        let tile = this._tileContainer.getChildByName(name[0]) as MedsurfDraw.Sprite;
        // TODO: let tile = this._tileContainer.children.find((c) => c.name === name[0]) as MedsurfDraw.Sprite;
        if (!tile) {
            // Setup sprite with correct settings
            let sprite = new MedsurfDraw.Sprite({
                texture: resource.texture
            });
            sprite.zIndex = 13 + level; // starts at level 2 -> 15 && 7 -> 20
            sprite.name = name[0];

            // Setup width and height
            const width = this.dimensions.width / gridSize.x;
            const height = this.dimensions.height / gridSize.y;

            sprite.width = sprite.texture.width / tileSize * width;
            sprite.height = sprite.texture.height / tileSize * height;

            /*
            if (this._imageBounds.isEmpty()) {
            */
                sprite.position.x = x * width;
                sprite.position.y = y * height;
            /*
            } else {
                sprite.position.x = (x - this._imageBounds.minX || 0) * width;
                sprite.position.y = (y - this._imageBounds.minY || 0) * height;
            }
            */

            this._tileContainer.addChild(sprite);
        } else {
            tile.showItem();
        }

        // Hide lower level tiles
        // TODO not working debounce(this.hideTiles.bind(this), 1000)();

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get tile url
     * @param level
     * @param x
     * @param y
     * @private
     */
    public getTileUrl(level: number, x: number, y: number): string {
        const tileSize = this.dimensions && this.dimensions.tileSize ? this.dimensions.tileSize : 256;

        const result = Math.floor(this.calculateAbsoluteTileNumber(level, x, y) / tileSize);
        return this._tilesUrl + 'TileGroup' + result + '/' + level + '-' + x + '-' + y + '.jpg';
    }

    /**
     * Calculate absolute tile number
     * @param level
     * @param x
     * @param y
     * @private
     */
    public calculateAbsoluteTileNumber(level: number, x: number, y: number): number {
        let num: number = 0;
        let size: {x: number, y: number};

        //Sum up all tiles below the level we want the number of tiles
        for (let z = 0; z < level; z++) {
            size = this._gridSizes[z];
            num += Math.ceil(size.x) * Math.ceil(size.y);
        }

        // Add the tiles of the level
        size = this._gridSizes[level];
        num += Math.ceil(size.x) * y + x;
        return num;
    }
    //</editor-fold>

    //<editor-fold desc="Loader">
    /**
     * On load
     * @param loader
     * @param resource
     * @private
     */
    public onLoad(loader: PIXI.Loader, resource: PIXI.LoaderResource): void {
        // Check if tile is from current image
        if (resource.name.startsWith(this.id + '-')) {
            // Draw tile
            this.drawTile(resource);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On tile move
     */
    public onTileMove(): void {
        this.emit("debounceLoadVisibleTiles");
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On tile zoom
     */
    public onTileZoom(): void {
        this.emit("debounceLoadVisibleTiles");
    }
    //</editor-fold>

    //<editor-fold desc="Image">
    /**
     * Get min zoom factor
     */
    public get minZoomFactor(): number {
        return this.data.minZoomFactor || Design.deepZoomImage.minZoomFactor;
    }

    /**
     * Get max zoom factor
     */
    public get maxZoomFactor(): number {
        return this.data.maxZoomFactor || Design.deepZoomImage.maxZoomFactor;
    }
    //</editor-fold>
}