import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { BaseContainer, BaseContainerModel } from "../../bases/elements/BaseContainer";
import { Design } from "../../config/design";
import { debounce } from "debounce";

/**
 * Image types
 */
export enum ImageTypes {
    NORMAL,
    DEEPZOOM,
    VIDEO
}

/**
 * Dimensions
 */
export class Dimensions {
    width: number;
    height: number;
    tileSize?: number;
}

/**
 * Image Model
 */
export class ImageModel extends BaseContainerModel {
    renderer: PIXI.Renderer | PIXI.AbstractRenderer;
    canvas: HTMLCanvasElement;
    loader: PIXI.Loader;
    dimensions: Dimensions;
    imageId: string;
    id: string;
    url: string;
    minZoomFactor?: number;
    maxZoomFactor?: number;
}

/**
 * Control model
 */
export class ControlModel {
    name: string;
    displayObjects: PIXI.DisplayObject[];
}

/**
 * Image
 * Zoom And Pan: https://github.com/anvaka/ngraph/tree/master/examples/pixi.js/03%20-%20Zoom%20And%20Pan
 */
export class Image extends BaseContainer<ImageModel> {
    /**
     * Members
     */
    protected _scaleX: number;
    protected _scaleY: number;
    private _imageObjectMenu: MedsurfDraw.MenuEntryModel[];
    private _groupingMenu: MedsurfDraw.MenuEntryModel[];
    private _actualWidth: number;
    private _actualHeight: number;
    private _imageType: ImageTypes;
    private _layerGroups: Models.LayerGroup[];

    /**
     * Interactions
     */
    private _contextInteraction: MedsurfDraw.ContextInteraction<Image>;
    private _moveInteraction: MedsurfDraw.MoveInteraction<Image>;
    private _zoomInteraction: MedsurfDraw.ZoomInteraction<Image>;

    /**
     * Methods
     */
    // protected _debounceOnResizeMethod: EventListener;
    // protected _debounceImageZoomMethod: EventListener;
    protected _debounceUnselectLayerGroupMethod: any;
    protected _debounceUnselectGroupingMethod: any;
    protected _wheelMethod: any;

    /**
     * Elements
     */
    protected _selectGroupElement: MedsurfDraw.SelectGroupElement;
    protected _selectItemElement: MedsurfDraw.SelectItemElement;
    protected _imageObjectMenuElement: MedsurfDraw.MenuElement;
    protected _groupingMenuElement: MedsurfDraw.MenuElement;
    protected _spriteElement: MedsurfDraw.Sprite;

    /**
     * Menu elements
     */
    // -- Shared
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Image object menu element
    // -- Interactive Area / Polygon
    protected _interactiveAreaButtonElement: MedsurfDraw.IndexedRoundButtonElement;
    protected _polygonButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Text
    protected _legendButtonElement: MedsurfDraw.IndexedRoundButtonElement;
    protected _textButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Arrow
    protected _arrowTextGroupingButtonElement: MedsurfDraw.IndexedRoundButtonElement;
    protected _arrowButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Ellipse
    protected _ellipseInteractiveGroupingButtonElement: MedsurfDraw.IndexedRoundButtonElement;
    protected _ellipseButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Rectangle
    protected _rectangleInteractiveGroupingButtonElement: MedsurfDraw.IndexedRoundButtonElement;
    protected _rectangleButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Arrow Marker
    protected _arrowMarkerButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Ellipse Marker
    protected _ellipseMarkerButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Grouping menu element
    // -- Interactive grouping
    protected _interactiveGroupingButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Poi grouping
    protected _poiGroupingButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Text grouping
    protected _textGroupingButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Legend grouping
    protected _legendGroupingButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param model
     * @param type
     */
    public constructor(model: ImageModel, type: ImageTypes = ImageTypes.NORMAL) {
        // Parent constructor
        super(model);

        // Checks
        // -- Dimensions
        if (this.data.dimensions.width === undefined) {
            throw 'Dimension width not set correctly';
        }
        if (this.data.dimensions.height === undefined) {
            throw 'Dimension height not set correctly';
        }

        // Members
        this.name = this.imageId;
        this.imageObjectMenu = [];
        this.groupingMenu = [];
        this.layerGroups = [];

        // Setup
        // -- zIndex
        this.zIndex = Design.image.zIndex;
        // -- Cursor fix -> ensure default cursor
        this.cursor = "move";
        this.cursor = "default";

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("remove-init", this._modeRemoveInit, this);
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on("drawing", this._modeDrawing, this);
        this.modeInteraction.on("remove-drawing", this._removeModeDrawing, this);
        this.modeInteraction.on("choosing", this._modeChoosing, this);
        this.modeInteraction.on("remove-choosing", this._removeModeChoosing, this);
        this.modeInteraction.on("cloning", this._modeCloning, this);
        this.modeInteraction.on("remove-cloning", this._removeModeCloning, this);
        // -- Context interaction
        this.contextInteraction = new MedsurfDraw.ContextInteraction(this);
        this.contextInteraction.on("endRight", this.endRight, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this, true);
        this.moveInteraction.on("startMove", this.startMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.moveInteraction.on("endMove", this.endMove, this);
        this.moveInteraction.on("endMoveLock", this.endMoveLock, this);
        // -- Zoom interaction
        this.zoomInteraction = new MedsurfDraw.ZoomInteraction(this);
        this.zoomInteraction.on("onZoom", this.onZoom, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        // this._debounceOnResizeMethod = debounce(this.onResize.bind(this), 10).bind(this);
        // this._debounceImageZoomMethod = debounce(this..bind(this), 10).bind(this);
        this._debounceUnselectLayerGroupMethod = debounce(this.unselectLayerGroup.bind(this), 50).bind(this);
        this._debounceUnselectGroupingMethod = debounce(this.unselectGrouping.bind(this), 50).bind(this);
        this._wheelMethod = this.zoomInteraction.onZoom.bind(this.zoomInteraction);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", (parent: PIXI.Container) => {
            // Events
            // -- Resize
            parent.on("resize", this.onResize, this);
            // -- Move
            parent.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            parent.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            parent.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
            // -- Debounce
            this.on("debouncedUnselectLayerGroup", this._debounceUnselectLayerGroupMethod);
            this.on("debouncedUnselectGrouping", this._debounceUnselectGroupingMethod);
            // -- Wheel
            this.canvas.addEventListener("wheel", this._wheelMethod);
        });
        // -- Removed
        this.on("removed", (parent: PIXI.Container) => {
            // Events
            // -- Resize
            parent.off("resize", this.onResize, this);
            // -- Move
            parent.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            parent.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            parent.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
            // -- Debounce
            this.off("debouncedUnselectLayerGroup", this._debounceUnselectLayerGroupMethod);
            this.off("debouncedUnselectGrouping", this._debounceUnselectGroupingMethod);
            // -- Wheel
            this.canvas.removeEventListener("wheel", this._wheelMethod);
            // TODO remove delete this._wheelMethod;
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public async init(): Promise<void> {
        // On resize
        await this.onResize();

        //<editor-fold desc="Elements">
        // -- Select group element
        this._selectGroupElement = new MedsurfDraw.SelectGroupElement({
            image: this
        });
        this.addChild(this._selectGroupElement);
        // -- Select item element
        this._selectItemElement = new MedsurfDraw.SelectItemElement({
            image: this
        });
        this.addChild(this._selectItemElement);
        // -- Image object menu element
        this._imageObjectMenuElement = new MedsurfDraw.MenuElement({
            image: this,
            menuElements: undefined
        });
        this.parent.addChild(this._imageObjectMenuElement);
        // -- Grouping menu element
        this._groupingMenuElement = new MedsurfDraw.MenuElement({
            image: this,
            menuElements: undefined
        });
        this.parent.addChild(this._groupingMenuElement);
        //</editor-fold>

        //<editor-fold desc="Image object menu elements">
        //<editor-fold desc="Shared">
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf0ea',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        this.groupingMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>
        //<editor-fold desc="Interactive Area / Polygon">
        // -- Interactive area button element
        this._interactiveAreaButtonElement = new MedsurfDraw.IndexedRoundButtonElement({
            image: this,
            text: '\uf304',
            description: 'Zeichenstift\n(Interaktiv)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            index: '\uf277',
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._interactiveAreaButtonElement.on("button", this.onButtonInteractiveArea, this);
        // -- Polygon button element
        this._polygonButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this,
            text: '\uf304',
            description: 'Zeichenstift',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._interactiveAreaButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._polygonButtonElement.on("onHover", this._polygonButtonElement.showMenu, this._polygonButtonElement);
        this._polygonButtonElement.on("button", this.onButtonPolygon, this);
        this._polygonButtonElement.on("toggleMenu", this.onMenuPolygon, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._polygonButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Text">
        // -- Legend button
        this._legendButtonElement = new MedsurfDraw.IndexedRoundButtonElement({
            image: this,
            text: '\uf893',
            description: 'Text\n(Legende)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            index: '\uf0cb',
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._legendButtonElement.on("button", this.onButtonLegendGrouping, this);
        // -- Text button element
        this._textButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this,
            text: '\uf893',
            description: 'Text',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._legendButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._textButtonElement.on("onHover", this._textButtonElement.showMenu, this._textButtonElement);
        this._textButtonElement.on("button", this.onButtonText, this);
        this._textButtonElement.on("toggleMenu", this.onMenuText, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._textButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Arrow">
        // -- Arrow text grouping button element
        this._arrowTextGroupingButtonElement = new MedsurfDraw.IndexedRoundButtonElement({
            image: this,
            text: '\uf124',
            description: 'Pfeil\n(Farbgruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            index: '\uf53f',
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._arrowTextGroupingButtonElement.on("button", this.onButtonArrowTextGrouping, this);
        // -- Arrow button elemnt
        this._arrowButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this,
            text: '\uf124',
            description: 'Pfeil',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._arrowTextGroupingButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._arrowButtonElement.on("onHover", this._arrowButtonElement.showMenu, this._arrowButtonElement);
        this._arrowButtonElement.on("button", this.onButtonArrow, this);
        this._arrowButtonElement.on("toggleMenu", this.onMenuArrow, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._arrowButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Ellipse">
        // -- Ellipse interactive grouping button element
        this._ellipseInteractiveGroupingButtonElement = new MedsurfDraw.IndexedRoundButtonElement({
            image: this,
            text: '\uf5ed',
            description: 'Ellipse\n(Interaktiv)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            index: '\uf277',
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._ellipseInteractiveGroupingButtonElement.on("button", this.onButtonEllipseInteractiveGrouping, this);
        // -- Ellipse button elemnt
        this._ellipseButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this,
            text: '\uf5ed',
            description: 'Ellipse',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._ellipseInteractiveGroupingButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._ellipseButtonElement.on("onHover", this._ellipseButtonElement.showMenu, this._ellipseButtonElement);
        this._ellipseButtonElement.on("button", this.onButtonEllipse, this);
        this._ellipseButtonElement.on("toggleMenu", this.onMenuEllipse, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._ellipseButtonElement,
            order: 5,
        });
        //</editor-fold>
        //<editor-fold desc="Rectangle">
        // -- Rectangle interactive grouping button element
        this._rectangleInteractiveGroupingButtonElement = new MedsurfDraw.IndexedRoundButtonElement({
            image: this,
            text: '\uf5ef',
            description: 'Rechteck\n(Interaktiv)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            index: '\uf277',
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._rectangleInteractiveGroupingButtonElement.on("button", this.onButtonRectangleInteractiveGrouping, this);
        // -- Rectangle button elemnt
        this._rectangleButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this,
            text: '\uf5ef',
            description: 'Rechteck',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._rectangleInteractiveGroupingButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._rectangleButtonElement.on("onHover", this._rectangleButtonElement.showMenu, this._rectangleButtonElement);
        this._rectangleButtonElement.on("button", this.onButtonRectangle, this);
        this._rectangleButtonElement.on("toggleMenu", this.onMenuRectangle, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._rectangleButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Arrow Marker">
        // -- Arrow marker button element
        this._arrowMarkerButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf30a',
            description: 'Marker\n(Pfeil)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor,
        });
        this._arrowMarkerButtonElement.on("onHover", this.collapseAllMenus, this);
        this._arrowMarkerButtonElement.on("button", this.onButtonArrowMarker, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._arrowMarkerButtonElement,
            order: 3,
        });
        //</editor-fold>
        //<editor-fold desc="Ellipse Marker">
        // -- Ellipse marker button element
        this._ellipseMarkerButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf276',
            description: 'Marker\n(Punkt)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._ellipseMarkerButtonElement.on("onHover", this.collapseAllMenus, this);
        this._ellipseMarkerButtonElement.on("button", this.onButtonEllipseMarker, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._ellipseMarkerButtonElement,
            order: 2,
        });
        //</editor-fold>
        this._imageObjectMenuElement.menuElements = this.imageObjectMenu;
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>
        //<editor-fold desc="Grouping menu elements">
        //<editor-fold desc="Interactive Grouping">
        // -- Interactive grouping button element
        this._interactiveGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf277',
            description: 'Interaktive\nFläche\n(Gruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._interactiveGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._interactiveGroupingButtonElement.on("button", this.onButtonInteractiveGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._interactiveGroupingButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Poi Grouping">
        // -- Poi grouping button element
        this._poiGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf649',
            description: 'POI\n(Gruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._poiGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._poiGroupingButtonElement.on("button", this.onButtonPoiGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._poiGroupingButtonElement,
            order: 5,
        });
        //</editor-fold>
        //<editor-fold desc="Text Grouping">
        // -- Text grouping button element
        this._textGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf53f',
            description: 'Farbgruppe',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._textGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._textGroupingButtonElement.on("button", this.onButtonTextGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._textGroupingButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Legend Grouping">
        // -- Legend grouping button element
        this._legendGroupingButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this,
            text: '\uf0cb',
            description: 'Legende\n(Gruppe)',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._legendGroupingButtonElement.on("onHover", this.collapseAllMenus, this);
        this._legendGroupingButtonElement.on("button", this.onButtonLegendGrouping, this);
        // Add to menu
        this.groupingMenu.push({
            element: this._legendGroupingButtonElement,
            order: 7,
        });
        //</editor-fold>
        this._groupingMenuElement.menuElements = this.groupingMenu;
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Resolve
        return Promise.resolve();
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Image object menu element
        this._imageObjectMenuElement.draw();
        // -- Grouping menu element
        this._groupingMenuElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Mode interaction
        if (this.modeInteraction) {
            this.modeInteraction.off("remove-init", this._modeRemoveInit, this);
            this.modeInteraction.off("default", this._modeDefault, this);
            this.modeInteraction.off("remove-default", this._removeModeDefault, this);
            this.modeInteraction.off("marker", this._modeMarker, this);
            this.modeInteraction.off("remove-marker", this._removeModeMarker, this);
            this.modeInteraction.off("selftest", this._modeSelftest, this);
            this.modeInteraction.off("remove-selftest", this._removeModeSelftest, this);
            this.modeInteraction.off("author", this._modeAuthor, this);
            this.modeInteraction.off("remove-author", this._removeModeAuthor, this);
            this.modeInteraction.off("drawing", this._modeDrawing, this);
            this.modeInteraction.off("remove-drawing", this._removeModeDrawing, this);
            this.modeInteraction.off("choosing", this._modeChoosing, this);
            this.modeInteraction.off("remove-choosing", this._removeModeChoosing, this);
            this.modeInteraction.off("cloning", this._modeCloning, this);
            this.modeInteraction.off("remove-cloning", this._removeModeCloning, this);
        }
        // -- Context interaction
        if (this.contextInteraction) {
            this.contextInteraction.off("endRight", this.endRight, this);
            // TODO remove delete this.contextInteraction;
        }
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.off("onMove", this.onMove, this);
            // TODO remove delete this.moveInteraction;
        }
        // -- Zoom interaction
        if (this.zoomInteraction) {
            this.zoomInteraction.off("onZoom", this.onZoom, this);
            // TODO remove delete this.zoomInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Select group element
        if (this._selectGroupElement) {
            this._selectGroupElement.destroy(options);
        }
        // -- Select item element
        if (this._selectItemElement) {
            this._selectItemElement.destroy(options);
        }
        // -- Image object menu element
        if (this._imageObjectMenuElement) {
            this._imageObjectMenuElement.destroy(options);
        }
        // -- Grouping menu element
        if (this._groupingMenuElement) {
            this._groupingMenuElement.destroy(options);
        }
        // -- Sprite element
        if (this._spriteElement) {
            this._spriteElement.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Shared">
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>
        //<editor-fold desc="Interactive Area / Polygon">
        // -- Interactive area button element
        if (this._interactiveAreaButtonElement) {
            this._interactiveAreaButtonElement.destroy(options);
        }
        // -- Polygon button element
        if (this._polygonButtonElement) {
            this._polygonButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Text">
        // -- Legend button element
        if (this._legendButtonElement) {
            this._legendButtonElement.destroy(options);
        }
        // -- Text button element
        if (this._textButtonElement) {
            this._textButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Arrow">
        // -- Arrow text grouping button element
        if (this._arrowTextGroupingButtonElement) {
            this._arrowTextGroupingButtonElement.destroy(options);
        }
        // -- Arrow button element
        if (this._arrowButtonElement) {
            this._arrowButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Ellipse">
        // -- Ellipse interactive grouping button element
        if (this._ellipseInteractiveGroupingButtonElement) {
            this._ellipseInteractiveGroupingButtonElement.destroy(options);
        }
        // -- Ellipse button element
        if (this._ellipseButtonElement) {
            this._ellipseButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Rectangle">
        // -- Rectangle interactive grouping button element
        if (this._rectangleInteractiveGroupingButtonElement) {
            this._rectangleInteractiveGroupingButtonElement.destroy(options);
        }
        // -- Rectangle button element
        if (this._rectangleButtonElement) {
            this._rectangleButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Arrow Marker">
        // -- Arrow marker button element
        if (this._arrowMarkerButtonElement) {
            this._arrowMarkerButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Ellipse Marker">
        // -- Ellipse marker button element
        if (this._ellipseMarkerButtonElement) {
            this._ellipseMarkerButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Get image objects
     */
    public getImageObjects(): MedsurfDraw.ImageObject[] {
        return this.children
            .filter((displayObject) =>
                displayObject instanceof MedsurfDraw.PositionPoint
                || displayObject instanceof MedsurfDraw.Line
                || displayObject instanceof MedsurfDraw.FillCollection
                || displayObject instanceof MedsurfDraw.LegendCollection) as MedsurfDraw.ImageObject[];
    }

    /**
     * Get lines by position point
     * @param point
     */
    public getLinesByPositionPoint(point: MedsurfDraw.PositionPoint): MedsurfDraw.Line[] {
        return this.getImageObjects().filter((element) => element instanceof MedsurfDraw.Line && element.hasPositionPoint(point)) as MedsurfDraw.Line[];
    }

    /**
     * Reset generator
     */
    public resetGenerator(): void {
        // Check for generator
        /* TODO
        if (this.generator && (this.generator instanceof BaseDrawGenerator)) {
            /* TODO not needed
            const mouse = this.getMousePosition();
            const targetElements = (this.generator.drawInteraction as MedsurfDraw.DrawInteraction<MedsurfDraw.ImageObject>).element;
            if (Array.isArray(targetElements)) {
                targetElements.forEach((element) => {
                    element.position.set(mouse.x, mouse.y);
                });
            } else {
                targetElements.position.set(mouse.x, mouse.y);
            }
            * /

            const global = this.renderer.plugins.interaction.mouse.global;
            const e = new PIXI.InteractionEvent();
            e.data = new PIXI.InteractionData();
            e.data.global.x = global.x;
            e.data.global.y = global.y;

            (this.generator.drawInteraction as MedsurfDraw.DrawInteraction<MedsurfDraw.ImageObject>).resetDraw(e);
        }
        */
    }

    /**
     * Get texture
     */
    public getTexture(): PIXI.Texture | string {
        return this._spriteElement.texture;
    }

    /**
     * Get mouse position
     */
    public getMousePosition(): PIXI.IPoint {
        return this.renderer.plugins.interaction.mouse.getLocalPosition(this);
    }

    /**
     * TODO
     * @param control
     */
    /*
    public setupControl(control: ControlModel): void {
        this.control.push(control);
    }
    */

    /**
     * Get group image from grouping
     * @param grouping
     */
    public getGroupImageFromGrouping(grouping: Models.Grouping): Models.GroupImage | undefined {
        if (grouping.groupImages) {
            return grouping.groupImages.find((gi) => !gi.deleted && gi.image.id === this.imageId);
        }
        return undefined
    }

    /**
     * Get display objects by grouping
     * @param grouping
     */
    public getDisplayObjectsByGrouping(grouping: Models.Grouping): PIXI.DisplayObject[] | undefined {
        const groupImage = this.getGroupImageFromGrouping(grouping);
        if (groupImage && groupImage.layerGroupObjects) {
            const imageObjects: PIXI.DisplayObject[] = [];
            groupImage.layerGroupObjects.filter((lgo) => !lgo.deleted).forEach((lgo) => {
                const layerGroup = this.layerGroups.find((lg) => !lg.deleted && lg.id === lgo.layerGroup);
                if (layerGroup) {
                    this.getLayerGroupDisplayObjectsByLayerGroup(layerGroup)?.forEach((imageObject) => {
                        if (imageObjects.findIndex(io => io === imageObject) === -1) {
                            imageObjects.push(imageObject);
                        }
                    });
                }
            });
            return imageObjects;
        }
        return undefined;
    }

    /**
     * Get layer group by imgae object
     * @param imageObject
     */
    public getLayerGroupByImageObject(imageObject: Models.ImageObject): Models.LayerGroup | undefined {
        if (this.layerGroups) {
            for (const group of this.layerGroups) {
                // const group = g;
                const tmpImageObjects = group.layerImageObjects ? group.layerImageObjects.filter((lio) => !lio.deleted).map(io => io.imageObject) || [] : [];
                if (tmpImageObjects.filter(id => id === imageObject.id)[0]) {
                    return group;
                }
            }
        }
        return undefined;
    }

    /**
     * Get layer group display objects by image object
     * @param imageObject
     */
    public getLayerGroupDisplayObjectsByImageObject(imageObject: Models.ImageObject): PIXI.DisplayObject[] | undefined {
        // Get layer group
        const layerGroup = this.getLayerGroupByImageObject(imageObject);
        // Check layer group
        if (layerGroup) {
            return this.getLayerGroupDisplayObjectsByLayerGroup(layerGroup);
        }
        return undefined;
    }

    /**
     * Get layer group display objects by layer group
     * @param layerGroup
     */
    public getLayerGroupDisplayObjectsByLayerGroup(layerGroup: Models.LayerGroup): PIXI.DisplayObject[] | undefined {
        const layerImageObjects = layerGroup.layerImageObjects?.filter((lio) => !lio.deleted);

        // Remove sketch from layer group
        if (layerImageObjects) { // TODO layerGroup.id !== 'sketch' &&
            const layerGroupImageObjects = layerImageObjects.map((imageObject) => imageObject.imageObject);
            return this.getImageObjects().filter((imageObject) => layerGroupImageObjects.indexOf(imageObject.name as string) > -1)
        }
        return undefined;
    }

    /**
     * Get line for elements
     * @param start
     * @param end
     */
    public getLineForElements(start: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection, end: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection): MedsurfDraw.Line | undefined {
        return this.children.find((element) => element instanceof MedsurfDraw.Line && element.hasPositionPoint(start) && element.hasPositionPoint(end)) as MedsurfDraw.Line | undefined;
    }

    /**
     * Get navigator scale
     */
    public getNavigatorScale(): PIXI.IPoint {
        return new PIXI.Point(
            1 / this._scaleX * this.scale.x,
            1 / this._scaleY * this.scale.y,
        );
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Fit image to screen
     * @param width
     * @param height
     */
    protected async _fitToScreen(width: number, height: number): Promise<void> {
        if (!this) {
            return;
        }

        const height1 = this.dimensions.height * width / this.dimensions.width;
        const width2 = this.dimensions.width * height / this.dimensions.height;

        if (height1 > height) {
            this._scaleX = Math.floor(width2 / this.dimensions.width * 1000) / 1000;
            this._scaleY = Math.floor(height / this.dimensions.height * 1000) / 1000;
        } else {
            this._scaleX = Math.floor(width / this.dimensions.width * 1000) / 1000;
            this._scaleY = Math.floor(height1 / this.dimensions.height * 1000) / 1000
        }

        // Scale
        this.scale.set(
            this._scaleX,
            this._scaleY
        );

        // Position
        this.position.set(
            (this.canvas.width - this.dimensions.width * this._scaleX) / 2,
            (this.canvas.height - this.dimensions.height * this._scaleY) / 2
        );

        // Emit
        debounce(this.emit.bind(this, "imageZoom", this.scale.x, this.scale.y), 50)();

        // Promise
        return Promise.resolve();
    }

    /**
     * Set actual values
     * @private
     */
    protected async _setActualValues(): Promise<void> {
        const tmpWidth = this.dimensions.width / this.dimensions.height * this.canvas.height;
        const tmpHeight = this.dimensions.height / this.dimensions.width * this.canvas.width;

        if (tmpHeight > this.canvas.height) {
            this._actualWidth = tmpWidth;
            this._actualHeight = this.canvas.height;
        } else {
            this._actualWidth = this.canvas.width;
            this._actualHeight = tmpHeight;
        }

        return Promise.resolve();
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Remove mode init
     * @private
     */
    protected _modeRemoveInit(): void {
        //<editor-fold desc="Elements">
        // -- Sprite element
        this._spriteElement = new MedsurfDraw.Sprite({
            texture: PIXI.Texture.from(this.url, {resourceOptions: { crossorigin: true }}) || PIXI.Texture.EMPTY
        });
        this.addChild(this._spriteElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        // Setup
        // TODO remove this.interactive = true;
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
        // Setup
        // TODO remove this.interactive = false;
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        // Setup
        // TODO remove this.interactive = true;
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
        // Setup
        // TODO remove this.interactive = false;
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        // Setup
        // TODO remove this.interactive = true;
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Setup
        // TODO remove this.interactive = false;
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        // TODO remove this.interactive = true;

        //<editor-fold desc="Events">
        // -- Context
        this.parent.on("rightup", this.contextInteraction.endRight, this.contextInteraction);
        //</editor-fold>
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        // TODO remove this.interactive = false;

        //<editor-fold desc="Events">
        // -- Context
        this.parent.off("rightup", this.contextInteraction.endRight, this.contextInteraction);
        //</editor-fold>
    }

    /**
     * Mode drawing
     * @private
     */
    protected _modeDrawing(): void {
        // Setup
        // TODO remove this.interactive = true;
    }

    /**
     * Remove mode drawing
     * @private
     */
    protected _removeModeDrawing(): void {
        // Setup
        // TODO remove this.interactive = false;
    }

    /**
     * Mode choosing
     * @private
     */
    protected _modeChoosing(): void {
        // Setup
        // TODO remove this.interactive = true; // TODO not working on image why??? when uncommend, the drawing is not working
    }

    /**
     * Remove mode choosing
     * @private
     */
    protected _removeModeChoosing(): void {
        // Setup
        // TODO remove this.interactive = false; // TODO not working on image why???
    }

    /**
     * Mode cloning
     * @private
     */
    protected _modeCloning(): void {
        // Setup
        // TODO remove this.interactive = true;
    }

    /**
     * Remove mode cloning
     * @private
     */
    protected _removeModeCloning(): void {
        // Setup
        // TODO remove this.interactive = false;
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    /**
     * Display menu
     */
    public displayMenu(): void {
        if (this._selectGroupElement.visible || this._selectItemElement.visible) {
            return;
        }

        const point = this.image.getMousePosition();
        const scale = this.imageScale;
        // Elements
        if (MedsurfDraw.Keyboard.isKeyDown("KeyG")) {
            // -- Grouping menu element
            this._groupingMenuElement.setMousePosition(point);
            this._groupingMenuElement.showItem();
        } else {
            // -- Image Object menu element
            this._imageObjectMenuElement.setMousePosition(point);
            this._imageObjectMenuElement.showItem();
        }

        // Image -> remove unselect behaviour and reset it after one click on the image
        this.image.moveInteraction.off("endMoveLock", this.image.endMoveLock, this.image);
        this.image.moveInteraction.once("endMoveLock", this.image.endMoveMenu, this.image);
    }
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._polygonButtonElement.toggleMenu(event, false);
        this._textButtonElement.toggleMenu(event, false);
        this._arrowButtonElement.toggleMenu(event, false);
        this._ellipseButtonElement.toggleMenu(event, false);
        this._rectangleButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Shared">
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Clone from clipboard
        this.cloneFromClipboard();
    }
    //</editor-fold>
    //</editor-fold>
    //<editor-fold desc="Interactive Area / Polygon">
    /**
     * On button interactive area
     */
    public onButtonInteractiveArea(): void {
        //<editor-fold desc="Elements">
        // -- Image object menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create interactive area
        this.createInteractiveArea();
    }

    /**
     * On button polygon
     */
    public onButtonPolygon(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create fill
        this.createFill();
    }

    /**
     * On menu polygon
     * @param event
     * @param visible
     */
    public onMenuPolygon(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._textButtonElement.toggleMenu(event, false);
            this._arrowButtonElement.toggleMenu(event, false);
            this._ellipseButtonElement.toggleMenu(event, false);
            this._rectangleButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Text">
    /**
     * On button text
     */
    public onButtonText(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create text
        this.createText();
    }

    /**
     * On menu text
     * @param event
     * @param visible
     */
    public onMenuText(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._polygonButtonElement.toggleMenu(event, false);
            this._arrowButtonElement.toggleMenu(event, false);
            this._ellipseButtonElement.toggleMenu(event, false);
            this._rectangleButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Arrow">
    /**
     * On button arrow interactive grouping
     */
    public onButtonArrowInteractiveGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create arrow with interactive grouping
        this.createArrowWithInteractiveGrouping();
    }

    /**
     * On button arrow poi grouping
     */
    public onButtonArrowPoiGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create arrow with text grouping
        this.createArrowWithPoiGrouping();
    }

    /**
     * On button arrow text grouping
     */
    public onButtonArrowTextGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create arrow with text grouping
        this.createArrowWithTextGrouping();
    }

    /**
     * On button arrow
     */
    public onButtonArrow(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create arrow
        this.createArrow();
    }

    /**
     * On menu arrow
     * @param event
     * @param visible
     */
    public onMenuArrow(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._polygonButtonElement.toggleMenu(event, false);
            this._textButtonElement.toggleMenu(event, false);
            this._ellipseButtonElement.toggleMenu(event, false);
            this._rectangleButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Ellipse">
    /**
     * On button ellipse interactive grouping
     */
    public onButtonEllipseInteractiveGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create ellipse with interactive grouping
        this.createEllipseWithInteractiveGrouping();
    }

    /**
     * On button ellipse poi grouping
     */
    public onButtonEllipsePoiGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create ellipse with poi grouping
        this.createEllipseWithPoiGrouping();
    }

    /**
     * On button ellipse text grouping
     */
    public onButtonEllipseTextGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create ellipse with poi grouping
        this.createEllipseWithTextGrouping();
    }

    /**
     * On button ellipse
     */
    public onButtonEllipse(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create ellipse
        this.createEllipse();
    }

    /**
     * On menu ellipse
     * @param event
     * @param visible
     */
    public onMenuEllipse(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._polygonButtonElement.toggleMenu(event, false);
            this._textButtonElement.toggleMenu(event, false);
            this._arrowButtonElement.toggleMenu(event, false);
            this._rectangleButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Rectangle">
    /**
     * On button rectangle interactive grouping
     */
    public onButtonRectangleInteractiveGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create rectangle with interactice grouping
        this.createRectangleWithInteractiveGrouping();
    }

    /**
     * On button rectangle poi grouping
     */
    public onButtonRectanglePoiGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create rectangle with poi grouping
        this.createRectangleWithPoiGrouping();
    }

    /**
     * On button rectangle text grouping
     */
    public onButtonRectangleTextGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create rectangle with text grouping
        this.createRectangleWithTextGrouping();
    }

    /**
     * On button rectangle
     */
    public onButtonRectangle(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create rectangle
        this.createRectangle();
    }

    /**
     * On menu ellipse
     * @param event
     * @param visible
     */
    public onMenuRectangle(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._polygonButtonElement.toggleMenu(event, false);
            this._textButtonElement.toggleMenu(event, false);
            this._arrowButtonElement.toggleMenu(event, false);
            this._ellipseButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Arrow Marker">
    /**
     * On button arrow marker
     */
    public onButtonArrowMarker(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create arrow marker
        this.createArrowMarker();
    }
    //</editor-fold>
    //<editor-fold desc="Ellipse Marker">
    /**
     * On button ellipse marker
     */
    public onButtonEllipseMarker(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        //</editor-fold>

        // Create ellipse marker
        this.createEllipseMarker();
    }
    //</editor-fold>
    //<editor-fold desc="Interactive Grouping">
    /**
     * On button interactive grouping
     */
    public onButtonInteractiveGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create interative grouping
        this.createInteractiveGrouping();
    }
    //</editor-fold>
    //<editor-fold desc="Poi Grouping">
    /**
     * On button poi grouping
     */
    public onButtonPoiGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create poi grouping
        this.createPoiGrouping();
    }
    //</editor-fold>
    //<editor-fold desc="Text Grouping">
    /**
     * On button text grouping
     */
    public onButtonTextGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create text grouping
        this.createTextGrouping();
    }
    //</editor-fold>
    //<editor-fold desc="Legend Grouping">
    /**
     * On button legend grouping
     */
    public onButtonLegendGrouping(): void {
        //<editor-fold desc="Elements">
        // -- Menu element
        this._imageObjectMenuElement.hideItem();
        this._groupingMenuElement.hideItem();
        //</editor-fold>

        // Create legend grouping
        this.createLegendGrouping();
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    public controlHideMenu(): void {
        //<editor-fold desc="Elements">
        // -- Select group element
        if (this._selectGroupElement) {
            this._selectGroupElement.controlHideMenu();
        }
        // -- Select item element
        if (this._selectItemElement) {
            this._selectItemElement.controlHideMenu();
        }
        // -- Image object menu element
        if (this._imageObjectMenuElement) {
            this._imageObjectMenuElement.hideItem();
        }
        // -- Grouping menu element
        if (this._groupingMenuElement) {
            this._groupingMenuElement.hideItem();
        }
        //</editor-fold>
    }

    /**
     * Set default mode for elements
     * @param event
     */
    public controlSetDefaultMode(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Set defaut mode
        this.modeInteraction.setDefaultMode(this.modeInteraction.defaultMode);

        // TODO not working this.emit("resetSelection");
        // TODO remove this.emit("selectImageObject", null);
        // TODO remove this.emit("selectLayerGroup", null);
    }

    /**
     * Control update elements
     */
    public controlUpdateElements(): void {
        // Elements
        // -- Select group element
        if (this._selectGroupElement) {
            this._selectGroupElement.controlSetDefaultMode();
        }
        // -- Select item element
        if (this._selectItemElement) {
            this._selectItemElement.controlSetDefaultMode();
        }
    }

    /**
     * Control save color
     * @param color
     * @param colorAlpha
     */
    public controlSaveColor(color: number | undefined, colorAlpha: number | undefined): void {
        // Elements
        // -- Select item element
        this._selectItemElement.saveColor(color, colorAlpha);
    }
    //</editor-fold>

    //<editor-fold desc="Events">

    //<editor-fold desc="Errors">
    /**
     * Send error
     * @param e
     */
    public sendError(e: any): void {
        console.error("-< Send ERROR >-");
        // TODO console.log(e);
        this.emit("sendError", e);
    }
    //</editor-fold>

    //<editor-fold desc="Grouping">
    /**
     * Unselect grouping
     */
    public unselectGrouping(): void {
        if (!this._selectGroupElement.visible && !this._selectItemElement.visible) {
            this.emit("unselectGrouping");
        }
    }
    //</editor-fold>

    //<editor-fold desc="Group Image">
    //</editor-fold>

    //<editor-fold desc="Group Image Layer Group">
    /**
     * Degroup group image layer groups
     */
    public degroupGroupImageLayerGroups(): void {
        const groupImage = this.image.getGroupImageFromGrouping(this._selectGroupElement.grouping);
        if (groupImage) {
            this.emit("degroupGroupImageLayerGroups", groupImage);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Grouping / Layer Group">
    /**
     * Edit image object grouping
     */
    public editImageObjectGrouping(): void {
        if (this._selectGroupElement.grouping) {
            this.emit("editGrouping", this._selectGroupElement.grouping);
        } else {
            this.emit("editLayerGroup", this._selectGroupElement.layerGroup);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Layer Group">
    /**
     * Select layer group
     * @param layerGroup
     */
    public selectLayerGroup(layerGroup: Models.LayerGroup): void {
        // Hide menu
        this.endMoveMenu();

        // Emit
        this.emit("selectLayerGroup", layerGroup);
    }

    /**
     * Select layer group by image object
     */
    public selectLayerGroupByImageObject(imageObject: Models.ImageObject): void {
        // Hide menu
        this.endMoveMenu();

        // Emit
        this.emit("selectLayerGroupByImageObject", imageObject);
    }

    /**
     * Unselect layer group
     */
    public unselectLayerGroup(): void {
        if (!this._selectGroupElement.visible && !this._selectItemElement.visible) {
            this.emit("unselectLayerGroup");
        }
    }

    /**
     * Add layer group
     * @param layerGroup
     * @param image
     */
    public addLayerGroup(layerGroup: Models.LayerGroup, image: Models.Image): void {
        this.emit("addLayerGroup", layerGroup, image);
    }

    /**
     * Add layer group by name
     * @param name
     * @param image
     */
    public addLayerGroupByName(name: string, image: Models.Image): void {
        this.emit("addLayerGroupByName", name, image);
    }

    /**
     * Copy layer group
     * @param layerGroup
     */
    public copyLayerGroup(layerGroup: Models.LayerGroup): void {
        this.emit("copyLayerGroup", layerGroup);
    }

    /**
     * Clipboard layer group
     * @param layerGroup
     */
    public clipboardLayerGroup(layerGroup: Models.LayerGroup): void {
        this.emit("clipboardLayerGroup", layerGroup);
    }

    /**
     * Update layer group
     * @param layerGroup
     */
    public updateLayerGroup(layerGroup: Models.LayerGroup): void {
        this.emit("updateLayerGroup", layerGroup);
    }

    /**
     * Delete layer group
     * @param layerGroup
     */
    public deleteLayerGroup(layerGroup: Models.LayerGroup): void {
        this.emit("deleteLayerGroup", layerGroup);
    }
    //</editor-fold>

    //<editor-fold desc="Layer Group Image Object">
    /**
     * Degroup layer group image object
     * @param imageObject
     */
    public degroupLayerGroupImageObject(imageObject: Models.ImageObject): void {
        this.emit("degroupLayerGroupImageObject", this._selectGroupElement.layerGroup, imageObject);
    }
    //</editor-fold>

    //<editor-fold desc="Image Object">
    /**
     * Select image object
     * @param imageObject
     */
    public selectImageObject(imageObject: Models.ImageObject): void {
        // Hide menu
        this.endMoveMenu();

        // Emit
        this.emit("selectImageObject", imageObject);
    }

    /**
     * Unselect image object
     */
    public unselectImageObject(): void {
        this.emit("unselectImageObject");
    }

    /**
     * Add image object
     * @param imageObject
     * @param layerGroup
     * @param image
     */
    public addImageObject(imageObject: Models.ImageObject, layerGroup: Models.LayerGroup, image: Models.Image): void {
        this.emit("addImageObject", imageObject, layerGroup, image);
    }

    /**
     * Copy image object
     * @param imageObject
     */
    public copyImageObject(imageObject: Models.PositionPoint | Models.Line | Models.FillCollection | Models.LegendCollection): void {
        this.emit("copyImageObject", imageObject);
    }

    /**
     * Clipboard image object
     * @param imageObject
     */
    public clipboardImageObject(imageObject: Models.PositionPoint | Models.Line | Models.FillCollection | Models.LegendCollection): void {
        this.emit("clipboardImageObject", imageObject);
    }

    /**
     * Edit image object
     * @param imageObject
     */
    public editImageObject(imageObject: Models.ImageObject): void {
        this.emit("editImageObject", imageObject);
    }

    /**
     * Update image object
     * @param imageObject
     */
    public updateImageObject(imageObject: Models.ImageObject): void {
        this.emit("updateImageObject", imageObject);
    }

    /**
     * Delete image object
     * @param imageObject
     */
    public deleteImageObject(imageObject: Models.ImageObject): void {
        this.emit("deleteImageObject", imageObject);
    }
    //</editor-fold>

    //<editor-fold desc="Primitive">
    /**
     * Select primitive
     * @param primitive
     * @param positionPoint
     */
    public selectPrimitive(primitive: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive, positionPoint: Models.PositionPoint): void {
        // Hide menu
        this.endMoveMenu();

        // Emit
        this.emit("selectPrimitive", primitive, positionPoint);
    }

    /**
     * Unselect primitive
     */
    public unselectPrimitive(): void {
        this.emit("unselectPrimitive");
    }

    /**
     * Edit primitive
     * @param primitive
     */
    public editPrimitive(primitive: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive): void {
        this.emit("editPrimitive", primitive);
    }

    /**
     * Update primitive
     * @param primitive
     */
    public updatePrimitive(primitive: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive): void {
        this.emit("updatePrimitive", primitive);
    }

    /**
     * Delete primitive
     * @param primitive
     */
    public deletePrimitive(primitive: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive): void {
        this.emit("deletePrimitive", primitive);
    }
    //</editor-fold>

    //<editor-fold desc="Legend column">
    /**
     * Select legend column
     * @param legendColumn
     * @param legend
     */
    public selectLegendColumn(legendColumn: Models.LegendColumn, legend: Models.LegendCollection): void {
        // Hide menu
        this.endMoveMenu();

        // Emit
        this.emit("selectLegendColumn", legendColumn, legend);
    }

    /**
     * Unselect legend column
     */
    public unselectLegendColumn(): void {
        this.emit("unselectLegendColumn");
    }

    /**
     * Edit legend column
     * @param legendColumn
     * @param legendCollection
     */
    public editLegendColumn(legendColumn: Models.LegendColumn, legendCollection: Models.LegendCollection): void {
        this.emit("editLegendColumn", legendColumn, legendCollection);
    }

    /**
     * Update legend column
     * @param legendColumn
     */
    public updateLegendColumn(legendColumn: Models.LegendColumn): void {
        this.emit("updateLegendColumn", legendColumn);
    }

    /**
     * Delete legend column
     * @param legendColumn
     */
    public deleteLegendColumn(legendColumn: Models.LegendColumn): void {
        this.emit("deleteLegendColumn", legendColumn);
    }
    //</editor-fold>

    //<editor-fold desc="Legend row">
    /**
     * Select legend row
     * @param legendRow
     * @param legendColumn
     * @param legend
     */
    public selectLegendRow(legendRow: Models.LegendRow, legendColumn: Models.LegendColumn, legend: Models.LegendCollection): void {
        // Hide menu
        this.endMoveMenu();

        // Emit
        this.emit("selectLegendRow", legendRow, legendColumn, legend);
    }

    /**
     * Unselect legend row
     */
    public unselectLegendRow(): void {
        this.emit("unselectLegendRow");
    }

    /**
     * Edit legend row
     * @param legendRow
     * @param legendColumn
     * @param legendCollection
     */
    public editLegendRow(legendRow: Models.LegendRow, legendColumn: Models.LegendColumn, legendCollection: Models.LegendCollection): void {
        this.emit("editLegendRow", legendRow, legendColumn, legendCollection);
    }

    /**
     * Update legend row
     * @param legendRow
     */
    public updateLegendRow(legendRow: Models.LegendRow): void {
        this.emit("updateLegendRow", legendRow);
    }

    /**
     * Delete legend row
     * @param legendRow
     */
    public deleteLegendRow(legendRow: Models.LegendRow): void {
        this.emit("deleteLegendRow", legendRow);
    }
    //</editor-fold>

    //<editor-fold desc="Generators">
    /**
     * Create polygon
     */
    public createPolygon(): void {
        this.emit("createPolygon");
    }

    /**
     * Create fill
     */
    public createFill(): void {
        this.emit("createFill");
    }

    /**
     * Create interactive area
     */
    public createInteractiveArea(): void {
        this.emit("createInteractiveArea");
    }

    /**
     * Create legend
     */
    public createLegend(): void {
        this.emit("createLegend");
    }

    /**
     * Create text
     */
    public createText(): void {
        this.emit("createText");
    }

    /**
     * Create arrow
     */
    public createArrow(): void {
        this.emit("createArrow");
    }

    /**
     * Create arrow with interactive grouping
     */
    public createArrowWithInteractiveGrouping(): void {
        this.emit("createArrowWithInteractiveGrouping");
    }

    /**
     * Create arrow with poi grouping
     */
    public createArrowWithPoiGrouping(): void {
        this.emit("createArrowWithPoiGrouping");
    }

    /**
     * Create arrow with text grouping
     */
    public createArrowWithTextGrouping(): void {
        this.emit("createArrowWithTextGrouping");
    }

    /**
     * Create ellipse
     */
    public createEllipse(): void {
        this.emit("createEllipse");
    }

    /**
     * Create ellipse with interactive grouping
     */
    public createEllipseWithInteractiveGrouping(): void {
        this.emit("createEllipseWithInteractiveGrouping");
    }

    /**
     * Create ellipse with poi grouping
     */
    public createEllipseWithPoiGrouping(): void {
        this.emit("createEllipseWithPoiGrouping");
    }

    /**
     * Create ellipse with text grouping
     */
    public createEllipseWithTextGrouping(): void {
        this.emit("createEllipseWithTextGrouping");
    }

    /**
     * Create rectangle
     */
    public createRectangle(): void {
        this.emit("createRectangle");
    }

    /**
     * Create rectangle with interactive grouping
     */
    public createRectangleWithInteractiveGrouping(): void {
        this.emit("createRectangleWithInteractiveGrouping");
    }

    /**
     * Create rectangle with poi grouping
     */
    public createRectangleWithPoiGrouping(): void {
        this.emit("createRectangleWithPoiGrouping");
    }

    /**
     * Create rectangle with text grouping
     */
    public createRectangleWithTextGrouping(): void {
        this.emit("createRectangleWithTextGrouping");
    }

    /**
     * Create arrow marker
     */
    public createArrowMarker(): void {
        this.emit("createArrowMarker");
    }

    /**
     * Create ellipse marker
     */
    public createEllipseMarker(): void {
        this.emit("createEllipseMarker");
    }

    /**
     * Create rectangle marker
     */
    public createRectangleMarker(): void {
        this.emit("createRectangleMarker");
    }

    /**
     * Clone from clipboard
     */
    public cloneFromClipboard(): void {
        this.emit("cloneFromClipboard");
    }

    /**
     * Create interactive grouping
     * @param layerGroup
     */
    public createInteractiveGrouping(layerGroup?: Models.LayerGroup): void {
        this.emit("createInteractiveGrouping", layerGroup);
    }

    /**
     * Create poi grouping
     * @param layerGroup
     */
    public createPoiGrouping(layerGroup?: Models.LayerGroup): void {
        this.emit("createPoiGrouping", layerGroup);
    }

    /**
     * Create text grouping
     * @param layerGroup
     */
    public createTextGrouping(layerGroup?: Models.LayerGroup): void {
        this.emit("createTextGrouping", layerGroup);
    }

    /**
     * Create legend grouping
     * @param layerGroup
     */
    public createLegendGrouping(layerGroup?: Models.LayerGroup): void {
        this.emit("createLegendGrouping", layerGroup);
    }
    //</editor-fold>

    //</editor-fold>

    //<editor-fold desc="Extern control">
    /**
     * Extern control select image
     */
    public externControlSelectImage(): void {
        // Setup image
        this.controlSetDefaultMode(new PIXI.InteractionEvent());
    }

    /**
     * Extern control select grouping
     * @param model
     */
    public externControlSelectGrouping(model: Models.Grouping): boolean {
        // Setup grouping
        return this._selectGroupElement.setGrouping(model);
    }

    /**
     * Extern control select layer group
     * @param model
     */
    public externControlSelectLayerGroup(model: Models.LayerGroup): boolean {
        // Setup grouping
        return this._selectGroupElement.setLayerGroup(model);
    }

    /**
     * Extern control select image object
     * @param model
     */
    public externControlSelectImageObject(model: Models.ImageObject): boolean {
        // Get image object
        const imageObject = this.getImageObjects().find((io) => io.name === model.id) as MedsurfDraw.DrawableObject;
        if (imageObject) {
            // Set select item
            imageObject.externControlSelectItem();

            // Item selected
            return true;
        }
        // Item not selected
        return false;
    }

    /**
     * Extern control select primitive
     * @param positionPointModel
     * @param model
     */
    public externControlSelectPrimitive(positionPointModel: Models.PositionPoint, model: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive): boolean {
        // Get position point
        const positionPoint = this.getImageObjects().find((io) => io.name === positionPointModel.id) as MedsurfDraw.PositionPoint;
        if (positionPoint) {
            // Get primitive
            const primitive = positionPoint.children
                                .filter((displayObject) =>
                                    displayObject instanceof MedsurfDraw.ArrowPrimitive
                                    || displayObject instanceof MedsurfDraw.BackgroundRectanglePrimitive
                                    || displayObject instanceof MedsurfDraw.EllipsePrimitive
                                    || displayObject instanceof MedsurfDraw.RectanglePrimitive
                                    || displayObject instanceof MedsurfDraw.TextPrimitive)
                                .find(p => p.name === model.id) as MedsurfDraw.DrawableObject;

            // Set select item
            primitive.externControlSelectItem();

            // Item selected
            return true;
        }
        // Item not selected
        return false;
    }

    /**
     * Extern control select legend column
     * @param legendCollectionModel
     * @param model
     */
    public externControlSelectLegendColumn(legendCollectionModel: Models.LegendCollection, model: Models.LegendColumn): boolean {
        // Get legend collection
        const legendCollection = this.getImageObjects().find((io) => io.name === legendCollectionModel.id) as MedsurfDraw.LegendCollection;
        if (legendCollection) {
            // Get legend column
            const legendColumn = legendCollection.children
                .filter((displayObject) =>
                    displayObject instanceof MedsurfDraw.LegendColumn)
                .find(p => p.name === model.id) as MedsurfDraw.DrawableObject;

            // Set select item
            legendColumn.externControlSelectItem();

            // Item selected
            return true;
        }
        // Item not selected
        return false;
    }

    /**
     * Extern control select legend row
     * @param legendCollectionModel
     * @param legendColumnModel
     * @param model
     */
    public externControlSelectLegendRow(legendCollectionModel: Models.LegendCollection, legendColumnModel: Models.LegendColumn, model: Models.LegendRow): boolean {
        // Get legend collection
        const legendCollection = this.getImageObjects().find((io) => io.name === legendCollectionModel.id) as MedsurfDraw.LegendCollection;
        if (legendCollection) {
            const legendColumn = legendCollection.children
                .filter((displayObject) =>
                    displayObject instanceof MedsurfDraw.LegendColumn)
                .find(p => p.name === model.id) as MedsurfDraw.DrawableObject;

            if (legendColumn) {
                // Get legend row
                const legendRow = legendCollection.children
                    .filter((displayObject) =>
                        displayObject instanceof MedsurfDraw.LegendRow)
                    .find(p => p.name === model.id) as MedsurfDraw.DrawableObject;

                // Set select item
                legendRow.externControlSelectItem();

                // Item selected
                return true;
            }
        }
        // Item not selected
        return false;
    }
    //</editor-fold>

    //<editor-fold desc="Context">
    /**
     * End right
     * @param event
     */
    public endRight(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Show menu
        this.displayMenu();
    }
    //</editor-fold>

    //<editor-fold desc="Resize">
    /**
     * On resize
     */
    public async onResize(): Promise<void> {
        // Update actual values
        await this._setActualValues();

        // Fit to screen
        await this._fitToScreen(this.canvas.width, this.canvas.height);

        // Emit onResize TODO remove from navigator
        debounce(this.emit.bind(this, "onResize"), 50)();
        // TODO remove

        // Promise
        return Promise.resolve();
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    public startMove(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Cursor fix for stage
        this.parent.cursor = "move";
        this.renderer.plugins.interaction.cursorStyles.default = "none";
        this.renderer.plugins.interaction.setCursorMode("move");
    }

    /**
     * On Move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        event.stopPropagation();

        const bounds = this.getBounds();

        const visibleImageX = this.actualWidth / Design.image.visibleImageXAspect;
        const visibleImageY = this.actualHeight / Design.image.visibleImageYAspect;

        const topLeft = new PIXI.Point(this.position.x + visibleImageX, this.position.y + visibleImageY);
        const topRight = new PIXI.Point(this.position.x - visibleImageX + bounds.width, this.position.y + visibleImageY);
        const bottomLeft = new PIXI.Point(this.position.x + visibleImageX, this.position.y - visibleImageY + bounds.height);

        if (topRight.x + dX > 0 && topLeft.x + dX < this.canvas.width) {
            this.position.x += dX;
        }

        if (bottomLeft.y + dY > 0 && topLeft.y + dY < this.canvas.height) {
            this.position.y += dY;
        }

        /* TODO remove
        const name = "asdasfg2";
        let test = this.parent.getChildByName(name) as PIXI.Graphics;
        if (!test) {
            test = new PIXI.Graphics();
            test.name = name;
            this.parent.addChild(test);
        }
        test.clear();
        test.beginFill(0x00FF00, 1);
        test.drawCircle(this.position.x + visibleImageX, this.position.y + visibleImageY, 4);
        test.endFill();

        test.beginFill(0x0000FF, 1);
        test.drawCircle(this.position.x - visibleImageX + bounds.width, this.position.y + visibleImageY, 4);
        test.endFill();

        test.beginFill(0xFF0000, 1);
        test.drawCircle(this.position.x + visibleImageX, this.position.y - visibleImageY + bounds.height, 4);
        test.endFill();

        test.beginFill(0xFFFF00, 1);
        test.drawCircle(this.position.x - visibleImageX + bounds.width, this.position.y - visibleImageY + bounds.height, 4);
        test.endFill();
        */

        // TODO use culling: https://medium.com/javascript-in-plain-english/inside-pixijs-the-ultimate-scene-graph-optimization-35b62bb2153
        // this.setVisibleElements(this.getViewPoint());
    }

    /**
     * End move
     * @param event
     */
    public endMove(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Cursor fix for stage
        this.parent.cursor = "default";
        this.renderer.plugins.interaction.cursorStyles.default = (mode: any) => {
            const cursor = this.parent.getChildByName('cursor') as PIXI.Text;
            cursor.text = "\uf245";
            cursor.anchor.set(0, 0);
            cursor.rotation = 0;
        };
        this.renderer.plugins.interaction.setCursorMode("default");
    }

    /**
     * End move lock
     * @param event
     */
    public endMoveLock(event: PIXI.InteractionEvent): void {
        if (!this._selectGroupElement.menuShown && !this._selectItemElement.menuShown) {
            // Reset selection
            if (['drawing', 'choosing', 'cloning'].indexOf(this.modeInteraction.lastMode) === -1) {
                this.controlSetDefaultMode(event);
            }
        }
    }

    /**
     * End move menu
     */
    public endMoveMenu(): void {
        // Elements
        this.controlHideMenu();

        // Events
        this.image.moveInteraction.on("endMoveLock", this.image.endMoveLock, this.image);
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On Zoom
     * @param event
     * @param x
     * @param y
     * @param factor
     */
    public onZoom(event: PIXI.InteractionEvent, x: number, y: number, factor: number): void {
        event.stopPropagation();

        // Normalize event data
        const point = new PIXI.Point();
        this.renderer.plugins.interaction.mapPositionToPoint(point, x, y);

        const oldPoint = this.toLocal(point);

        // Scale
        this.scale.x *= factor;
        this.scale.y *= factor;

        const newPoint = this.toGlobal(oldPoint);

        // Position
        this.position.x += point.x - newPoint.x;
        this.position.y += point.y - newPoint.y;

        // Reset generator
        this.resetGenerator();

        // Emit
        debounce(this.emit.bind(this, "imageZoom", this.scale.x, this.scale.y), 20)();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get image
     */
    public get image(): MedsurfDraw.Image | MedsurfDraw.DeepZoomImage {
        return this;
    }

    /**
     * Get image canvas
     */
    public get imageCanvas(): HTMLCanvasElement {
        return this.canvas;
    }

    /**
     * Get image scale
     */
    public get imageScale(): PIXI.IPoint {
        return this.scale;
    }

    /**
     * Get image dimensions
     */
    public get imageDimensions(): MedsurfDraw.Dimensions {
        return this.dimensions;
    }
    /**
     * Get control
     */
    /* TODO
    public get control(): ControlModel[] {
        return this._control;
    }
    */

    /**
     * Set control
     * @param value
     */
    /* TODO
    public set control(value: ControlModel[]) {
        this._control = value;
    }
    */

    /**
     * Get renderer
     */
    public get renderer(): PIXI.Renderer | PIXI.AbstractRenderer {
        return this.data.renderer;
    }

    /**
     * Set renderer
     * @param value
     */
    public set renderer(value: PIXI.Renderer | PIXI.AbstractRenderer) {
        this.data.renderer = value;
    }

    /**
     * Get canvas
     */
    public get canvas(): HTMLCanvasElement {
        return this.data.canvas;
    }

    /**
     * Set canvas
     * @param value
     */
    public set canvas(value: HTMLCanvasElement) {
        this.data.canvas = value;
    }

    /**
     * Get loader
     */
    public get loader(): PIXI.Loader {
        return this.data.loader;
    }

    /**
     * Set loader
     * @param value
     */
    public set loader(value: PIXI.Loader) {
        this.data.loader = value;
    }

    /**
     * Get dimensions
     */
    public get dimensions(): Dimensions {
        return this.data.dimensions;
    }

    /**
     * Set dimensions
     * @param value
     */
    public set dimensions(value: Dimensions) {
        this.data.dimensions = value;
    }

    /**
     * Get imageId
     */
    public get imageId(): string {
        return this.data.imageId;
    }

    /**
     * Set imageId
     * @param value
     */
    public set imageId(value: string) {
        this.data.imageId = value;
    }

    /**
     * Get id
     */
    public get id(): string {
        return this.data.id;
    }

    /**
     * Set id
     * @param value
     */
    public set id(value: string) {
        this.data.id = value;
    }

    /**
     * Get url
     */
    public get url(): string {
        return this.data.url;
    }

    /**
     * Set url
     * @param value
     */
    public set url(value: string) {
        this.data.url = value;
    }

    /**
     * Get min zoom factor
     */
    public get minZoomFactor(): number {
        return this.data.minZoomFactor || Design.image.minZoomFactor;
    }

    /**
     * Set min zoom factor
     * @param value
     */
    public set minZoomFactor(value: number) {
        this.data.minZoomFactor = value;
    }

    /**
     * Get max zoom factor
     */
    public get maxZoomFactor(): number {
        return this.data.maxZoomFactor || Design.image.maxZoomFactor;
    }

    /**
     * Set max zoom factor
     * @param value
     */
    public set maxZoomFactor(value: number) {
        this.data.maxZoomFactor = value;
    }

    /**
     * Get image object menu
     */
    public get imageObjectMenu(): MedsurfDraw.MenuEntryModel[] {
        return this._imageObjectMenu;
    }

    /**
     * Set image object menu
     * @param value
     */
    public set imageObjectMenu(value: MedsurfDraw.MenuEntryModel[]) {
        this._imageObjectMenu = value;
    }

    /**
     * Get grouping menu
     */
    public get groupingMenu(): MedsurfDraw.MenuEntryModel[] {
        return this._groupingMenu;
    }

    /**
     * Set grouping menu
     * @param value
     */
    public set groupingMenu(value: MedsurfDraw.MenuEntryModel[]) {
        this._groupingMenu = value;
    }

    /**
     * Get actual width
     */
    public get actualWidth(): number {
        return this._actualWidth;
    }

    /**
     * Get actual height
     */
    public get actualHeight(): number {
        return this._actualHeight;
    }

    /**
     * Get image type
     */
    public get imageType(): ImageTypes {
        return this._imageType;
    }

    /**
     * Get layer groups
     */
    public get layerGroups(): Models.LayerGroup[] {
        return this._layerGroups;
    }

    /**
     * Set layer groups
     * @param value
     */
    public set layerGroups(value: Models.LayerGroup[]) {
        this._layerGroups = value;
    }

    /**
     * Get context interaction
     */
    public get contextInteraction(): MedsurfDraw.ContextInteraction<Image> {
        return this._contextInteraction;
    }

    /**
     * Set context interaction
     * @param value
     */
    public set contextInteraction(value: MedsurfDraw.ContextInteraction<Image>) {
        this._contextInteraction = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<Image> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<Image>) {
        this._moveInteraction = value;
    }

    /**
     * Get zoom interaction
     */
    public get zoomInteraction(): MedsurfDraw.ZoomInteraction<Image> {
        return this._zoomInteraction;
    }

    /**
     * Set zoom interaction
     * @param value
     */
    public set zoomInteraction(value: MedsurfDraw.ZoomInteraction<Image>) {
        this._zoomInteraction = value;
    }
    //</editor-fold>
}