import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import { BaseContainer, BaseContainerModel } from "../../bases/elements/BaseContainer";
import { Design } from "../../config/design";

export class VirtualPointerElementModel extends BaseContainerModel {
    rectangle: PIXI.Rectangle;
}

/**
 * Virtual pointer element
 */
export class VirtualPointerElement extends BaseContainer<VirtualPointerElementModel> {
    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<VirtualPointerElement>;

    /**
     * Elements
     */
    protected _arrowElement: MedsurfDraw.Arrow;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: VirtualPointerElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.virtualPointerElement.zIndex;
        this.rotation = 45 * Math.PI / 180;

        //<editor-fold desc="Interactions">
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this, true);
        this.moveInteraction.on("onMove", this.onMove, this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", (parent: PIXI.Container) => {
            // Events
            this.on("pointerover", this._moveInteraction.onMoveHover, this._moveInteraction);
            this.on("mousedown", this._moveInteraction.startMove, this._moveInteraction);
            this.on("pointermove", this._moveInteraction.onMove, this._moveInteraction);
            this.on("mouseup", this._moveInteraction.endMove, this._moveInteraction);
            this.on("pointerupoutside", this._moveInteraction.endMove, this._moveInteraction);
        });
        // -- Removed
        this.on("removed", (parent: PIXI.Container) => {
            // Events
            this.off("pointerover", this._moveInteraction.onMoveHover, this._moveInteraction);
            this.off("mousedown", this._moveInteraction.startMove, this._moveInteraction);
            this.off("pointermove", this._moveInteraction.onMove, this._moveInteraction);
            this.off("mouseup", this._moveInteraction.endMove, this._moveInteraction);
            this.off("pointerupoutside", this._moveInteraction.endMove, this._moveInteraction);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Setup
        this.interactive = true;

        // Position
        this.position.set(this.rectangle.x, this.rectangle.y);

        //<editor-fold desc="Elements">
        // -- Arrow element
        this._arrowElement = new MedsurfDraw.Arrow({
            rectangle: new PIXI.Rectangle(0, 0, this.rectangle.width, this.rectangle.height),
            wedgePoint: undefined,
            options: {hasLine: true, lineColor: Design.virtualPointerElement.lineColor, lineAlpha: Design.virtualPointerElement.lineAlpha, hasFill: true, fillColor: Design.virtualPointerElement.fillColor, fillAlpha: Design.virtualPointerElement.fillAlpha},
            lineWidth: Design.virtualPointerElement.lineWidth
        });
        this.addChild(this._arrowElement);
        //</editor-fold>
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Arrow element
        this._arrowElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Arrow element
        if (this._arrowElement) {
            this._arrowElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Move">
    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        event.stopPropagation();

        this.position.x += dX;
        this.position.y += dY;
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<VirtualPointerElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<VirtualPointerElement>) {
        this._moveInteraction = value;
    }
    //</editor-fold>
}