import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Caret element model
 */
export class CaretElementModel extends BaseContainerImageModel {
    textElement: MedsurfDraw.EditableText | undefined;
}

/**
 * Caret element
 */
export class CaretElement extends BaseContainer<CaretElementModel> {
    /**
     * Methods
     */
    protected _caretFlashIntervalTimeout: number | null;

    /**
     * Elements
     */
    protected _caretElement: MedsurfDraw.Rectangle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: CaretElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.caretElement.zIndex;
        this.hideItem();
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Caret element
        this._caretElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, Design.caretElement.lineWidth, 0),
            options: {hasLine: false, hasFill: true, fillColor: Design.caretElement.fillColor, fillAlpha: Design.caretElement.fillAlpha},
            lineWidth: Design.caretElement.lineWidth
        });
        this.addChild(this._caretElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Caret element
        this._caretElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Caret element
        if (this._caretElement) {
            this._caretElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Caret">
    /**
     * On caret
     */
    public onCaret(): void {
        if (!this.textElement) {
            return
        }

        let text = this.textElement.text.substring(0, this.textElement.startTextIndex);

        let heightFactor: number = 0;
        let index = text.indexOf("\n");
        while (index > -1) {
            text = text.substring(index + 1);
            heightFactor++;
            index = text.indexOf("\n");
        }

        if (this.textElement.style.wordWrap) {
            let startIndex: number = 0;
            let wordWrapRows: number = 0;
            for (let i = 1; i <= text.length; i++) {
                const tmpText = text.substring(startIndex, i);
                const tmpWidth = this.textElement.context.measureText(tmpText).width;

                if (tmpWidth > this.textElement.style ? this.textElement.style.wordWrapWidth || 0 : 0) {
                    wordWrapRows++;
                    startIndex = (i - 1);
                }
            }

            this._caretElement.position.x = this.textElement.context.measureText(text.substr(startIndex)).width;

            // @ts-ignore TODO ignore
            const localHeight = this.textElement.style.fontSize + this.textElement.style.strokeThickness + this.textElement.style.lineHeight + 6 /*+ 4.4 / 32 * this.targetInteraction.element._textElement.style.fontSize*/; // TODO
            this._caretElement.position.y = wordWrapRows * localHeight + localHeight * heightFactor;
        } else {
            this._caretElement.position.x = this.textElement.context.measureText(text).width;

            // @ts-ignore TODO ignore
            const localHeight = this.textElement.style.fontSize + this.textElement.style.strokeThickness + this.textElement.style.lineHeight + 6 /*+ 4.4 / 32 * this.targetInteraction.element._textElement.style.fontSize*/; // TODO
            this._caretElement.position.y = localHeight * heightFactor;
        }
    }
    //</editor-fold>

    //<editor-fold desc="Base Container">
    /**
     * Show item
     */
    public showItem(): void {
        if (this._caretFlashIntervalTimeout) {
            clearInterval(this._caretFlashIntervalTimeout);
            this._caretFlashIntervalTimeout = null;
        }

        this.emit("debounceDraw");

        // @ts-ignore
        this._caretFlashIntervalTimeout = setInterval(this.toggleItem.bind(this), 500);

        super.showItem();
    }

    /**
     * Hide item
     */
    public hideItem(): void {
        if (this._caretFlashIntervalTimeout) {
            clearInterval(this._caretFlashIntervalTimeout);
            this._caretFlashIntervalTimeout = null;
        }

        super.hideItem();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text element
     */
    public get textElement(): MedsurfDraw.EditableText | undefined {
        return this.data.textElement;
    }

    /**
     * Set text element
     * @param value
     */
    public set textElement(value: MedsurfDraw.EditableText | undefined) {
        this.data.textElement = value;

        if (this.textElement) {
            // Elements
            // -- Caret element
            // @ts-ignore TODO ignore
            this._caretElement.rectangle.height = this.textElement.style.fontSize + this.textElement.style.strokeThickness + this.textElement.style.lineHeight + 2;
        }
    }
    //</editor-fold>
}