import * as MedsurfDraw from "../../public-api";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {BaseGraphics, BaseGraphicsImageModel} from "../../bases/elements/BaseGraphics";
import {Design} from "../../config/design";

/**
 * Selection element model
 */
export class SelectionElementModel extends BaseContainerImageModel {
    textElement: MedsurfDraw.EditableText | undefined;
}

/**
 * Selection element
 */
export class SelectionElement extends BaseContainer<SelectionElementModel> {
    /**
     * Elements
     */
    protected _selectionElement: BaseGraphics<BaseGraphicsImageModel>;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: SelectionElementModel) {
        super(model);

        // Setup
        this.zIndex = Design.selectionElement.zIndex;
        this.hideItem();
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Selection element
        this._selectionElement = new BaseGraphics({image: this.image});
        this.addChild(this._selectionElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        this._selectionElement.cacheAsBitmap = false;

        this._selectionElement.clear();

        if (!this.textElement || !this.textElement.textSelection || !this.textElement.style) {
            return;
        }

        // @ts-ignore TODO ignore
        const localHeight = this.textElement.style.fontSize + this.textElement.style.strokeThickness + this.textElement.style.lineHeight + 6;

        let selectionStart: number;
        let selectionEnd: number;

        if (this.textElement.startTextIndex > this.textElement.endTextIndex) {
            selectionStart = this.textElement.endTextIndex;
            selectionEnd = this.textElement.startTextIndex;
        } else {
            selectionStart = this.textElement.startTextIndex;
            selectionEnd = this.textElement.endTextIndex;
        }

        let previousText = this.textElement.text.substring(0, selectionStart);

        let heightFactor: number = 0;
        let index = previousText.indexOf("\n");
        while (index > -1) {
            previousText = previousText.substring(index + 1);
            heightFactor++;
            index = previousText.indexOf("\n");
        }

        let offset: number = this.textElement.context.measureText(previousText).width;

        let text = this.textElement.text.substring(selectionStart, selectionEnd);
        do {
            index = text.indexOf("\n");

            let displayText: string;
            if (index > -1) {
                displayText = text.substring(0, index);
            } else {
                displayText = text;
            }

            let selectedWidth = this.textElement.context.measureText(displayText).width;

            this._selectionElement.beginFill(Design.selectionElement.fillColor, Design.selectionElement.fillAlpha);
            this._selectionElement.drawRect(offset, localHeight * heightFactor, selectedWidth, localHeight);
            this._selectionElement.endFill();

            text = text.substring(index + 1);
            heightFactor++;

            offset = 0;
        } while (index > -1)

        this._selectionElement.cacheAsBitmap = true;
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Selection element
        if (this._selectionElement) {
            this._selectionElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Selection">
    /**
     * On selection
     */
    public onSelection(): void {
        if (this.textElement && this.textElement.textSelection && this.textElement.startTextIndex !== this.textElement.endTextIndex) {
            this.showItem();
            this.emit("debounceDraw");
        } else {
            this.hideItem();
        }
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text element
     */
    public get textElement(): MedsurfDraw.EditableText | undefined {
        return this.data.textElement;
    }

    /**
     * Set text element
     * @param value
     */
    public set textElement(value: MedsurfDraw.EditableText | undefined) {
        this.data.textElement = value;
    }
    //</editor-fold>
}