import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {BaseElementContainer, BaseElementContainerModel} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";

/**
 * Legend collection
 */
export class LegendCollection extends BaseElementContainer<Models.LegendCollection, BaseElementContainerModel<Models.LegendCollection>> {
    /**
     * Members
     */
    public legendColumns: MedsurfDraw.LegendColumn[];

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<LegendCollection>;
    private _scaleInteraction: MedsurfDraw.ScaleInteraction<LegendCollection>;
    private _selectInteraction: MedsurfDraw.SelectInteraction<LegendCollection>;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: any;
    protected _debounceControlUnselectItemMethod: any;

    /**
     * Elements
     */
    protected _selftestElement: MedsurfDraw.WedgePoint;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.LegendCollection) {
        // Parent constructor
        super({image, model});

        // Members
        this.legendColumns = [];

        // Setup
        this.zIndex = Design.legendCollection.zIndexDefault + this.model.zIndex;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on('select_layer_group', this._modeSelectLayerGroup, this);
        this.modeInteraction.on('remove-select_layer_group', this._removeModeSelectLayerGroup, this);
        this.modeInteraction.on('select_grouping', this._modeSelectGrouping, this);
        this.modeInteraction.on('remove-select_grouping', this._removeModeSelectGrouping, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('select_child', this._modeSelectChild, this);
        this.modeInteraction.on("remove-select_child", this._removeModeSelectChild, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('draw_item', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item", this._removeModeDrawItem, this);
        this.modeInteraction.on('draw_item_new', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item_new", this._removeModeDrawItem, this);
        this.modeInteraction.on('choose_select', this._modeBlockedItem, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeBlockedItem, this);
        this.modeInteraction.on('choose_unselect', this._modeBlockedItem, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeBlockedItem, this);
        this.modeInteraction.on('clone_item', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item", this._removeModeCloneItem, this);
        this.modeInteraction.on('clone_item_new', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item_new", this._removeModeCloneItem, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        this.modeInteraction.on("delete_multiple", this._modeDelete, this);
        this.modeInteraction.on('delete_draw', this._modeDelete, this);
        this.modeInteraction.on('delete_clone', this._modeDelete, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.startMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.moveInteraction.on("endMove", this.endMove, this);
        this.moveInteraction.on("onRelease", this.onRelease, this);
        // -- Scale interaction
        this.scaleInteraction = new MedsurfDraw.ScaleInteraction(this);
        this.scaleInteraction.on("onScale", this.onScale, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelectStart", this.onSelectStart, this);
        this.selectInteraction.on("onSelftest", this.onSelftest, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        this._debounceControlUnselectItemMethod = debounce(this.controlUnselectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.on("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.off("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Get selftest point
        const selftestPoint: Models.Point = this.model.selftestPoint
            ? {x: this.model.selftestPoint.x * imageWidth, y: this.model.selftestPoint.y * imageWidth}
            : {x: 0, y: 0};

        // Position
        this.position.set(
            this.model.rectangle.x * imageWidth,
            this.model.rectangle.y * imageWidth
        );

        //<editor-fold desc="Elements">
        // -- Selftest element
        this._selftestElement = new MedsurfDraw.WedgePoint({
            image: this.image,
            wedgePoint: selftestPoint
        });
        this._selftestElement.moveInteraction.on("onMove", this.onSelftestMove, this);
        this._selftestElement.moveInteraction.on("endMove", this.onSelftestEndMove, this);
        this.addChild(this._selftestElement);
        // -- Legend columns
        for (let col = 0; col < this.model.cols.length; col++) {
            const legendColumn = new MedsurfDraw.LegendColumn(this.image, this.model.cols[col]);
            this.addChild(legendColumn);
            // Push to array
            this.legendColumns.push(legendColumn);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Get col width
        const colWidth = this.model.rectangle.width * imageWidth / this.model.cols.length;

        // Set x values of table
        const heightArray = [];
        let index: number = 1;
        for (let col = 0; col < this.legendColumns.length; col++) {
            const legendColumn = this.legendColumns[col] as MedsurfDraw.LegendColumn;
            for (let row = 0; row < legendColumn.legendRows.length; row++) {
                // Setup data
                const legendRow = legendColumn.legendRows[row];
                legendRow.setPositionX(col * colWidth);
                legendRow.setWordWrapWidth(colWidth);
                legendRow.index = index;

                // Caluclate max height of each row
                const height = legendRow.height;
                if (!heightArray[row]) {
                  heightArray[row] = height;
                } else if (height > heightArray[col]) {
                    heightArray[row] = height;
                }

                // increment
                index++;
            }
        }

        // Set y values of table
        for (let col = 0; col < this.legendColumns.length; col++) {
            for (let row = 0; row < this.legendColumns[col].legendRows.length; row++) {
                this.legendColumns[col].legendRows[row].setPositionY(heightArray[row - 1] || 0);
            }
        }

        //<editor-fold desc="Elements">
        // -- Selftest element
        this._selftestElement.draw();
        // -- Legend columns
        for (let legendColumn of this.legendColumns) {
            legendColumn.draw();
        }
        //</editor-fold>

        // Update hitbox
        this.hitArea = this.getLocalBounds();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        // -- Scale interaction
        if (this.scaleInteraction) {
            this.scaleInteraction.removeAllListeners();
            // TODO remove delete this.scaleInteraction;
        }
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Selftest element
        if (this._selftestElement) {
            this._selftestElement.moveInteraction.off("onMove", this.onSelftestMove, this);
            this._selftestElement.moveInteraction.off("endMove", this.onSelftestEndMove, this);
            this._selftestElement.destroy(options);
        }
        // -- Legend columns
        for (let legendColumn of this.legendColumns) {
            legendColumn.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param imageWidth
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, imageWidth: number): LegendCollection {
        // -- Legend model
        const legendCollectionModel: Models.LegendCollection = {
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDCOLLECTION,
            isStatic: false,
            zIndex: 0,
            rectangle: new PIXI.Rectangle(
                0 / imageWidth,
                0 / imageWidth,
                100 / imageWidth,
                100 / imageWidth
            ),
            cols: [],
            rowDistance: 14 / imageWidth,
            columnDistance: 14 / imageWidth,
            indexWidth: 14 / imageWidth
        };
        // ---- Legend column model
        const legendColumnModel: Models.LegendColumn = {
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDCOLUMN,
            isStatic: false,
            zIndex: 0,
            table: legendCollectionModel,
            rows: []
        };
        legendCollectionModel.cols.push(legendColumnModel);
        // ------ Legend row model
        const legendRowModel: Models.LegendRow = {
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDROW,
            isStatic: false,
            zIndex: 0,
            text: 'Neue Legende',
            style: {
                fill: 0x000000,
                fontSize: 12 / imageWidth,
                fontFamily: 'proxima_nova_altsemibold',
                align: 'left',
                breakWords: false,
            },
            col: legendColumnModel
        };
        legendColumnModel.rows.push(legendRowModel);

        // Element
        // -- Legend primitive
        return new LegendCollection(image, legendCollectionModel);
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Add legend column
     */
    public addLegendColumn(): void {
        // Create new legend column
        const legendColumn = MedsurfDraw.LegendColumn.getInstance(this.image, this.model);
        this.addChild(legendColumn);
        legendColumn.modeInteraction.setDefaultMode("author");
        // Push to array
        this.legendColumns.push(legendColumn);

        // Add legend column
        legendColumn.onButtonAddLegendRow();

        // Draw
        this.emit("debounceDraw");

        // Update model
        // -- LegendCollection
        this.model.cols.push(legendColumn.model);

        // Emits
        this.image.updateImageObject(this.model);

        // Select legend column
        legendColumn.controlSelectItem();
    }

    /**
     * Remove legend column
     * @param model
     */
    public removeLegendColumn(model: Models.LegendColumn): void {
        // Get legend column
        const legendColumnIndex = this.legendColumns.findIndex((lc) => lc.model.id === model.id);
        const legendColumn = this.legendColumns[legendColumnIndex];
        if (!legendColumn) {
            return;
        }

        // Remove column
        this.removeChild(legendColumn);
        // Remove from array
        this.legendColumns.splice(legendColumnIndex, 1);

        // Draw
        (this.parent as MedsurfDraw.LegendCollection).draw();

        // Update model
        // -- LegendCollection
        this.model.cols.splice(legendColumnIndex, 1);

        // Emits
        this.image.updateImageObject(this.model);
    }

    /**
     * Get selftest position
     */
    public getSelftestPosition(): Models.Point {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Get selftest point
        const selftestPoint: Models.Point = this.model.selftestPoint
            ? {x: this.model.selftestPoint.x * imageWidth, y: this.model.selftestPoint.y * imageWidth}
            : {x: 0, y: 0};

        return selftestPoint;
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        if (this.model.isStatic) {
            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        // Show item
        this.showItem();
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        // Show item
        this.showItem();
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
        // TODO this.off("onRelease", this._onSelftestRelease, this);
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        this.interactive = true;

        // Elements
        // -- Position point element
        // TODO this._positionPointElement.circle.radius = Design.positionPoint.radius;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select layer
     * @private
     */
    protected _modeSelectLayerGroup(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select layer
     * @private
     */
    protected _removeModeSelectLayerGroup(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select grouping
     * @private
     */
    protected _modeSelectGrouping(): void {
        // Setup
        this.interactive = true;

        // Elements
        // -- Position point element
        // TODO this._positionPointElement.circle.radius = Design.positionPoint.radius;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select grouping
     * @private
     */
    protected _removeModeSelectGrouping(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Elements
        // -- Selftest element
        this._selftestElement.interactive = true;

        // Events
        this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.on("startMove", this._startMove, this);
        // TODO this.on("endMove", this._endMove, this);
        // TODO this.on("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.on("down_ArrowUp", "onPositionPointMoveUp", this.moveInteraction.moveUp.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowRight", "onPositionPointMoveRight", this.moveInteraction.moveRight.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowDown", "onPositionPointMoveDown", this.moveInteraction.moveDown.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowLeft", "onPositionPointMoveLeft", this.moveInteraction.moveLeft.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowUp", "onPositionPointMoveUpEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowRight", "onPositionPointMoveRightEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowDown", "onPositionPointMoveDownEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowLeft", "onPositionPointMoveLeftEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.off("startMove", this._startMove, this);
        // TODO this.off("endMove", this._endMove, this);
        // TODO this.off("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.remove("down_ArrowUp", "onPositionPointMoveUp");
        MedsurfDraw.Keyboard.events.remove("down_ArrowRight", "onPositionPointMoveRight");
        MedsurfDraw.Keyboard.events.remove("down_ArrowDown", "onPositionPointMoveDown");
        MedsurfDraw.Keyboard.events.remove("down_ArrowLeft", "onPositionPointMoveLeft");
        MedsurfDraw.Keyboard.events.remove("released_ArrowUp", "onPositionPointMoveUpEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowRight", "onPositionPointMoveRightEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowDown", "onPositionPointMoveDownEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowLeft", "onPositionPointMoveLeftEnd");

        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // TODO remove console.log("legend _removeModeSelectItem", mode);
        if (mode === 'author') {
            this.emit("debounceControlUnselectItem", this.image);
            // this.controlUnselectItem();
        }
        /* TODO remove
        if (!mode.startsWith('select_child') && !mode.startsWith('select_item') && !mode.startsWith('choose_item')
            && !mode.startsWith('choose_select') && !mode.startsWith('blocked_item')
            && ['default', 'marker', 'selftest'/*, 'author'* /, 'delete'].indexOf(mode) === -1) {
            // Unselect item
            this.emit("debounceControlUnselectItem");
        }
        */
    }

    /**
     * Mode select child
     * @private
     */
    protected _modeSelectChild(): void {
    }

    /**
     * Remove mode select child
     * @private
     */
    protected _removeModeSelectChild(): void {
    }

    /**
     * Mode select row
     * @private
     */
    protected _modeSelectRow(): void {
    }

    /**
     * Remove mode select row
     * @private
     */
    protected _removeModeSelectRow(): void {
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        this.filters = [];
        */
    }

    /**
     * Mode draw item
     * @private
     */
    protected _modeDrawItem(): void {
    }

    /**
     * Remove mode draw item
     * @private
     */
    protected _removeModeDrawItem(): void {
        this.moveInteraction.reset();
    }

    /**
     * Mode clone item
     * @private
     */
    protected _modeCloneItem(): void {
    }

    /**
     * Remove mode clone item
     * @private
     */
    protected _removeModeCloneItem(): void {
        this.moveInteraction.reset();
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        const image = this.image;

        if (image) {
            // Select layer group
            if (['delete_multiple', 'delete_draw', 'delete_clone'].indexOf(this.modeInteraction.lastMode) === -1) {
                this.controlUnselectItem(image);
            }

            // Delete from parent
            image.removeChild(this);

            // Emits
            image.deleteImageObject(this.model);
        }

        this.destroy();

        // Hide menu
        image.controlHideMenu();

        // Update control elements
        image.controlUpdateElements();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = Design.legendCollection.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = Design.legendCollection.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            // Set mode
            this.modeInteraction.setMode("delete");
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.degroupLayerGroupImageObject(this.model);
        if (unselectElement) {
            this.image.unselectLayerGroup();
            this.image.selectImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        this.image.copyImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        this.image.clipboardImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Selftest">
    /**
     * On button set selftest
     * @param event
     */
    // TODO
    public onButtonSetSelftest(event: PIXI.InteractionEvent): void {
        /* TODO
        this.model.isSelftest = !this.model.isSelftest;
        this._selftestButton.text = this.model.isSelftest ? "\uf00d" : "\uf00c";
        this._lineColor = this.model.isSelftest ? this._lineColorSelftest : this._lineColorSelected;
        this._fillColor = this.model.isSelftest ? this._fillColorSelftest : this._fillColorSelected;
        this.emit("debounceDraw");

        if (this.model.isSelftest) {
            this._selftestContainerElement.addImageObject(this._selftestChooseButton);
            this._selftestContainerElement.draw();
        } else {
            const items = this.selftestItemElements;
            this.selftestItemElements = []; // Remove existing items
            items.forEach((element: MedsurfDraw.ImageObject) => {
                element.modeInteraction.changeMode("blocked", this);
            });
            this._selftestContainerElement.removeImageObject(this._selftestChooseButton);
        }

        this.emit("selftestChanged");

        // Update
        this.getImage().updateImageObject(this.model);
        */
    }

    /**
     * On button choose selftest
     * @param event
     */
    // TODO
    public onButtonChooseSelftest(event: PIXI.InteractionEvent): void {
        /* TODO
        // TODO better implementation: this._configurationElement.hideItem();
        this.removeChild(this._configurationElement);

        const image = this.getImage();

        image.generator = new MedsurfDraw.SelftestGenerator(image, this.parent as MedsurfDraw.PositionPoint, this);
        image.modeInteraction.changeMode("choose");
        */
    }
    //</editor-fold>
    //<editor-fold desc="Static">
    public onButtonStatic(override?: boolean): void {
        this.model.isStatic = override || !this.model.isStatic;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Legend column">
    public onButtonAddLegendColumn(): void {
        this.addLegendColumn();
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectImageObject(this.model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Control unselect item
     * @param image
     */
    public controlUnselectItem(image: MedsurfDraw.Image): void {
        // Set select layer
        if(!this.controlSelectLayerGroup(image)) {
            // Set select image
            image.externControlSelectImage();

            // Emits
            image.unselectImageObject();
        }
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('select_legend');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control select layer group
     * @param image
     */
    public controlSelectLayerGroup(image: MedsurfDraw.Image): boolean {
        // Get layer group
        const layerGroup = image.getLayerGroupByImageObject(this.model);
        if (layerGroup) {
            // Emits
            image.selectLayerGroup(layerGroup);

            // Select layer group
            return image.externControlSelectLayerGroup(layerGroup);
        }

        // Layer not selected
        return false;
    }

    /**
     * Control unselect layer group
     */
    public controlUnselectLayerGroup(): void {
        // Emits
        this.image.unselectLayerGroup();
    }

    /**
     * Control draw item
     * @param isNew
     */
    public controlDrawItem(isNew: boolean): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this && !imageObject.modeInteraction.lastMode.startsWith('draw'))
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('draw_parent');

        // Set draw item
        this.modeInteraction.setModeItem(isNew ? 'draw_item_new' : 'draw_item');
    }

    /**
     * Control clone item
     */
    public controlCloneItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this && !imageObject.modeInteraction.lastMode.startsWith('clone'))
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('clone_parent');

        // Set draw item
        this.modeInteraction.setModeItem('clone_item_new');
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * Start move
     * @private
     */
    public startMove(): void {
        // TODO not needed const radius = this._positionPointElement.circle.radius + MedsurfDraw.MOVE_HITTEST_EXTENSION;

        // TODO better implementation: this._configurationElement.hideItem();
        // TODO this.removeChild(this._configurationElement);

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     * @private
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number) {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Position
        this.position.x += dX;
        this.position.y += dY;
        this.model.rectangle.x = this.position.x / imageWidth;
        this.model.rectangle.y = this.position.y / imageWidth;
    }

    /**
     * End move
     * @param event
     * @private
     */
    public endMove(event: PIXI.InteractionEvent): void {
        // TODO not needed const radius = this._positionPointElement.circle.radius - MedsurfDraw.MOVE_HITTEST_EXTENSION;

        // TODO better implementation: this._configurationElement.showItem();
        // TODO this.addChild(this._configurationElement);

        // Events
        // TODO remove this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);

        // Emits
        this.emit("pointerover", event);
        this.image.updateImageObject(this.model);
    }

    /**
     * On release
     */
    public onRelease(): void {
        this.modeInteraction.setMode(this.modeInteraction.defaultMode);
    }

    /**
     * On selftest move
     * @param event
     * @param dX
     * @param dY
     */
    public onSelftestMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Position
        this._selftestElement.wedgePoint.x += dX;
        this._selftestElement.wedgePoint.y += dY;
        this._selftestElement.draw();

        if (!this.model.selftestPoint) {
            this.model.selftestPoint = {x: 0, y: 0};
        }
        this.model.selftestPoint.x = this._selftestElement.position.x / imageWidth;
        this.model.selftestPoint.y = this._selftestElement.position.x / imageWidth;
    }

    /**
     * On selftest end move
     */
    public onSelftestEndMove(): void {
        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Elements
        // -- Legend columns
        this.legendColumns.forEach((legendColumn) => {
            legendColumn.onHover();
        });

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Elements
        // -- Legend columns
        this.legendColumns.forEach((legendColumn) => {
            legendColumn.onOut();
        });

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_layer_group') {
            this.controlSelectItem();
        } else {
            if (!this.controlSelectLayerGroup(this.image) && !preventSelectItem) {
                this.controlSelectItem();
            }
        }
    }

    /**
     * On selftest
     * @private
     */
    public onSelftest(): void {
        /* TODO no selftest for legend ???
        // Set default mode for image children
        this.getImage().getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.selftestItems && this.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.showItem();
                } else {
                    imageObject.hideItem();
                }
            });

        // Release event
        this.getImage().once('pointerup', this.onSelftestRelease, this);
        */
    }
    //</editor-fold>

    //<editor-fold desc="Scale">
    /**
     * Start scale
     * TODO remove
     */
    public startScale(): void {
        /*
        this._rotateContainer.removeChild(this._wedgePointElement);
        // TODO better implementation: this._rotateElement.hideItem();
        this._rotateContainer.removeChild(this._rotateElement);
        // TODO better implementation: this._configurationElement.hideItem();
        this.removeChild(this._configurationElement);
        */
    }

    /**
     * On scale
     * @param scaleSide
     * @param dX
     * @param dY
     * @param dW
     * @param dH
     */
    public onScale(scaleSide: MedsurfDraw.ScaleSideTypes, dX: number, dY: number, dW: number, dH: number): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        if (this.model.rectangle.width * imageWidth + dW < 1) {
            return;
            /*
            dW = 1 - this._ellipse.width;
            dX = 0;
            dY = 0;
            */
        }
        if (this.model.rectangle.height * imageWidth + dH < 1) {
            return;
            /*
            dH = 1 - this._ellipse.height;
            dX = 0;
            dY = 0;
            */
        }

        // Update model
        this.model.rectangle.width += dW / imageWidth;
        this.model.rectangle.height += dH / imageWidth;

        /* TODO remove
        // Element
        // -- Legend columns
        for (let legendColumn of this.legendColumns) {
            legendColumn.rectangle.width = this.model.rectangle.width / this.model.cols.length;
            legendColumn.rectangle.height = this.model.rectangle.height;
        }
        */

        // Position TODO not needed ??
        // this.position.x += (dX + dW / 2) * Math.cos(this.rotation) - (dY + dH / 2) * Math.sin(this.rotation);
        // this.position.y += (dY + dH / 2) * Math.cos(this.rotation) + (dX + dW / 2) * Math.sin(this.rotation);
        // TODO this.rotateContainer.pivot.x += dW / 2;
        // TODO this.rotateContainer.pivot.y += dH / 2;

        // Draw
        this.draw();

        // TODO needed ???
        // this.parent.emit("onScale", dX, dY, dW, dH);
    }

    /**
     * End scale
     * TODO remove
     */
    public endScale(): void {
        /*
        this._rotateContainer.addChild(this._wedgePointElement);
        this._wedgePointElement.draw();

        // TODO better implementation: this._rotateElement.showItem();
        this._rotateElement.draw();
        this._rotateContainer.addChild(this._rotateElement);

        // TODO better implementation: this._configurationElement.showItem();
        this._configurationElement.draw();
        this.addChild(this._configurationElement);

        this._rotateContainer.sortChildren();
        */

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>

    //<editor-fold desc="Draw">
    /**
     * Start draw
     * @private
     */
    public startDraw(): void {
        this.cursor = "draw";
    }

    /**
     * Start draw scale
     * @private
     */
    public startDrawScale(): void {
        this.cursor = "scale-nwse";
    }

    /**
     * On draw
     * @param event
     */
    public onDraw(event: PIXI.InteractionEvent): void {
        // TODO this.targetImageObject.emit("onDraw", event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<LegendCollection> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<LegendCollection>) {
        this._moveInteraction = value;
    }

    /**
     * Get scale interaction
     */
    public get scaleInteraction(): MedsurfDraw.ScaleInteraction<LegendCollection> {
        return this._scaleInteraction;
    }

    /**
     * Set scale interaction
     * @param value
     */
    public set scaleInteraction(value: MedsurfDraw.ScaleInteraction<LegendCollection>) {
        this._scaleInteraction = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<LegendCollection> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<LegendCollection>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}