import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {BaseElementContainer, BaseElementContainerModel} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";


/**
 * Fill column
 */
export class LegendColumn extends BaseElementContainer<Models.LegendColumn, BaseElementContainerModel<Models.LegendColumn>> {
    /**
     * Members
     */
    public legendRows: MedsurfDraw.LegendRow[];

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: EventListener;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.LegendColumn) {
        // Parent constructor
        super({image, model});

        // Members
        this.legendRows = [];

        // Setup
        this.zIndex = Design.legendColumn.zIndex;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on('select_legend', this._modeSelectLegend, this);
        this.modeInteraction.on("remove-select_legend", this._removeModeSelectLegend, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('select_child', this._modeSelectChild, this);
        this.modeInteraction.on("remove-select_child", this._removeModeSelectChild, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        //<editor-fold desc="Elements">
        // -- Legend rows
        for (let row = 0; row < this.model.rows.length; row++) {
            const legendRow = new MedsurfDraw.LegendRow(this.image, this.model.rows[row]);
            this.addChild(legendRow);
            // Push to array
            this.legendRows.push(legendRow);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Legend rows
        for (let legendRow of this.legendRows) {
            legendRow.draw();
        }
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Legend rows
        for (let legendRow of this.legendRows) {
            legendRow.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param model
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.LegendCollection): LegendColumn {
        // Element
        // -- Legend column
        return new LegendColumn(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDCOLUMN,
            isStatic: false,
            zIndex: 0,
            table: model,
            rows: []
        });
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Add legend row
     * @param model
     */
    public addLegendRow(model: Models.LegendColumn): void {
        // Get image width
        const imageWidth = this.image.imageDimensions.width;

        // Create new legend row
        const legendRow = MedsurfDraw.LegendRow.getInstance(this.image, model, imageWidth);
        this.addChild(legendRow);
        // Push to array
        this.legendRows.push(legendRow);

        // Draw
        (this.parent as MedsurfDraw.LegendCollection).draw();

        // Update model
        // -- LegendColumn
        this.model.rows.push(legendRow.model);

        // Emits
        this.image.updateLegendColumn(this.model);

        // Select legend row
        legendRow.controlSelectItem();
    }

    /**
     * Remove legend row
     * @param model
     */
    public removeLegendRow(model: Models.LegendRow): void {
        // Get legend row
        const legendRowIndex = this.legendRows.findIndex((lr) => lr.model.id === model.id);
        const legendRow = this.legendRows[legendRowIndex];
        if (!legendRow) {
            return;
        }

        // Remove row
        this.removeChild(legendRow);
        // Remove from array
        this.legendRows.splice(legendRowIndex, 1);

        // Draw
        (this.parent as MedsurfDraw.LegendCollection).draw();

        // Update model
        // -- LegendColumn
        this.model.rows.splice(legendRowIndex, 1);

        // Emits
        this.image.updateLegendColumn(this.model);
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode select parent
     * @private
     */
    protected _modeSelectLegend(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.onHover, this);
        this.on("pointerout", this.onOut, this);
        this.on("mousedown", this.onPointerDown, this);
        this.on("mouseup", this.onSelectStart, this);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select parent
     * @private
     */
    protected _removeModeSelectLegend(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.onHover, this);
        this.off("pointerout", this.onOut, this);
        this.off("mousedown", this.onPointerDown, this);
        this.off("mouseup", this.onSelectStart, this);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select iten
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        if (mode !== 'select_child' && !mode.startsWith('select_item')) {
            // Set select legend collection
            (this.parent as MedsurfDraw.LegendCollection).emit("debounceControlSelectItem");
        }
    }

    /**
     * Mode select child
     * @private
     */
    protected _modeSelectChild(): void {
    }

    /**
     * Remove mode select child
     * @private
     */
    protected _removeModeSelectChild(): void {
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        this.filters = [];
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editLegendColumn(this.model, (this.parent as MedsurfDraw.LegendCollection).model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateLegendColumn(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateLegendColumn(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            (this.parent as MedsurfDraw.LegendCollection).removeLegendColumn(this.model);
            /*
            // Set mode
            this.modeInteraction.setMode("delete");
            */
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Degroup image object
        (this.parent as MedsurfDraw.LegendCollection).onButtonDegroup(unselectElement);
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        (this.parent as MedsurfDraw.LegendCollection).onButtonDuplicate();
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Clipboard image object
        (this.parent as MedsurfDraw.LegendCollection).onButtonClipboard();
    }
    //</editor-fold>
    //<editor-fold desc="LegendRow">
    public onButtonAddLegendRow(): void {
        this.addLegendRow(this.model);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectLegendColumn(this.model, (this.parent as MedsurfDraw.LegendCollection).model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this.parent)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set default mode for parent
        (this.parent as MedsurfDraw.LegendCollection).modeInteraction.setModeItem('select_child');

        // Set default mode for siblings
        (this.parent as MedsurfDraw.LegendCollection).children
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                (imageObject as MedsurfDraw.ImageObject).modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('select_column');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        this.legendRows.forEach((legendRow) => {
            legendRow.onHover();
        });
    }

    /**
     * On out
     */
    public onOut(): void {
        this.legendRows.forEach((legendRow) => {
            legendRow.onOut();
        });
    }

    /**
     * On pointer down
     * @param event
     */
    public onPointerDown(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_legend') {
            this.controlSelectItem();
        }
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        const rect = super.getRectangle();
        const position = this.parent.position;
        return new PIXI.Rectangle(position.x + rect.x, position.y + rect.y, rect.width, rect.height);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    //</editor-fold>
}