import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {
    BaseElementContainer,
    BaseElementContainerModel,
    ColorMode,
    NumberMode
} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";

/**
 * Fill line
 */
export class LegendRow extends BaseElementContainer<Models.LegendRow, BaseElementContainerModel<Models.LegendRow>> {
    /**
     * Members
     */
    private _index: number;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: EventListener;

    /**
     * Elements
     */
    protected _indexElement: MedsurfDraw.Text;
    protected _textElement: MedsurfDraw.Text;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.LegendRow) {
        // Parent constructor
        super({image, model});

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on('default', this._modeDefault, this);
        this.modeInteraction.on("marker", this._modeDefault, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("author", this._modeDefault, this);
        this.modeInteraction.on('select_column', this._modeSelectColumn, this);
        this.modeInteraction.on("remove-select_column", this._removeModeSelectColumn, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('choose_select', this._modeChooseSelect, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeChooseSelect, this);
        this.modeInteraction.on('choose_unselect', this._modeChooseUnselect, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeChooseUnselect, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // style
        const style = this._getCorrectStyle();

        //<editor-fold desc="Elements">
        // -- Index elememt
        this._indexElement = new MedsurfDraw.Text({
            text: "",
            style: style
        });
        this.addChild(this._indexElement);
        // -- Text element
        this._textElement = new MedsurfDraw.Text({
            text: this.model.text,
            style: style
        });
        this._textElement.style.breakWords = true;
        this._textElement.style.wordWrap = true;
        this._textElement.position.x += Design.legendRow.enumerationPadding;
        this.addChild(this._textElement);
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: 6,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>
    }

    /**
     * Draw
     */
    public draw(): void {
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Index elememt
        this._indexElement.destroy(options);
        // -- Text element
        this._textElement.destroy(options);
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param model
     * @param imageWidth
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.LegendColumn,
                              imageWidth: number): LegendRow {
        // Element
        // -- Legend row
        return new LegendRow(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDROW,
            isStatic: false,
            zIndex: 0,
            col: model,
            text: "Neue Reihe",
            style: {
                fontSize: 12 / imageWidth,
                fill: 0xffffff,
                align: 'left',
                breakWords: true,
                fontFamily: 'proxima_nova_altsemibold'
            },
        });
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Set position x
     * @param x
     */
    public setPositionX(x: number): void {
        // Elements
        // -- Index element
        this._indexElement.position.x = x;
        // -- Text element
        this._textElement.position.x = x + Design.legendRow.enumerationPadding;
    }

    /**
     * Set position y
     * @param y
     */
    public setPositionY(y: number): void {
        // Elements
        // -- Index element
        this._indexElement.position.y = y;
        // -- Text element
        this._textElement.position.y = y;
    }

    /**
     * Set word wrap width
     * @param width
     */
    public setWordWrapWidth(width: number): void {
        // Elements
        // -- Text element
        this._textElement.style.wordWrapWidth = width - Design.legendRow.enumerationPadding;
    }

    /**
     * Edit new text
     */
    public editNewText(): void {
        // Reset text
        this._textElement.text = '';

        // Edit Text
        this.editText();
    }

    /**
     * Edit text
     */
    public editText(): void {
        // Select item
        this.controlSelectItem();

        // Change element mode
        this.modeInteraction.setModeItem("select_child");
        this._textElement.modeInteraction.setMode("select_item");
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Get correct style
     */
    private _getCorrectStyle(): PIXI.TextStyle {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Defaults
        let fontSize: number = 1;
        if (this.model.style.fontSize) {
            if (typeof this.model.style.fontSize === 'string') {
                fontSize = parseFloat(this.model.style.fontSize);
            } else {
                fontSize = this.model.style.fontSize;
            }
            fontSize *= imageWidth;
        }

        // Return model
        return new PIXI.TextStyle({
            align: this.model.style.align as PIXI.TextStyleAlign,
            fontSize: fontSize,
            fill: this.model.style.fill,
            fontFamily: this.model.style.fontFamily,
            breakWords: this.model.style.breakWords,
        });
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        // Elements
        // -- Text element
        this._textElement.onImageZoom(this.imageScale.x, this.imageScale.y);

        // Show item
        this.showItem();
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        // Get position
        const postition = (this.parent.parent as MedsurfDraw.LegendCollection).getSelftestPosition();

        // Setup position
        this.setPositionX(postition.x);
        this.setPositionY(postition.y);

        // Hide item
        this.hideItem();
    }

    /**
     * Mode select child
     * @private
     */
    protected _modeSelectColumn(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.onHover, this);
        this.on("pointerout", this.onOut, this);
        this.on("mousedown", this.onPointerDown, this);
        this.on("mouseup", this.onSelectStart, this);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select child
     * @private
     */
    protected _removeModeSelectColumn(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.onHover, this);
        this.off("pointerout", this.onOut, this);
        this.off("mousedown", this.onPointerDown, this);
        this.off("mouseup", this.onSelectStart, this);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select iten
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // TODO check old: if (mode !== 'select_child' && !mode.startsWith('select_item')) {
        if (!mode.startsWith('select_parent') && mode !== 'select_child' && !mode.startsWith('select_item')
            && !mode.startsWith('choose_select') && !mode.startsWith('choose_parent') && !mode.startsWith('blocked_item')) {
            // Set select legend column
            (this.parent as MedsurfDraw.LegendColumn).emit("debounceControlSelectItem");
        }
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
        // Element
        // -- Index element
        this._indexElement.alpha = Design.blocked.fillAlpha;
        // -- Text element
        this._textElement.alpha = Design.blocked.fillAlpha;
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        this.filters = [];
        */
        // Element
        // -- Index element
        this._indexElement.alpha = 1;
        // -- Text element
        this._textElement.alpha = 1;
    }

    /**
     * Mode choose select
     * @private
     */
    protected _modeChooseSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Index element
        this._indexElement.style.fill = Design.selftest.lineColorSelect;
        // -- Text element
        this._textElement.style.fill = Design.selftest.lineColorSelect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose select
     * @private
     */
    protected _removeModeChooseSelect(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Defaults
        const style = Object.assign({}, this.model.style);
        if (style.fontSize) {
            if (typeof style.fontSize === 'string') {
                style.fontSize = parseFloat(style.fontSize);
            }
            style.fontSize *= imageWidth;
        }

        // Setup
        this.interactive = false;
        // -- Index element
        // @ts-ignore TODO ignore
        this._indexElement.style = new PIXI.TextStyle(style);
        // -- Text element
        // @ts-ignore TODO ignore
        this._textElement.style = new PIXI.TextStyle(style);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose unselect
     * @private
     */
    protected _modeChooseUnselect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Index element
        this._indexElement.style.fill = Design.selftest.lineColorUnselect;
        // -- Text element
        this._textElement.style.fill = Design.selftest.lineColorUnselect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose unselect
     * @private
     */
    protected _removeModeChooseUnselect(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Defaults
        const style = Object.assign({}, this.model.style);
        if (style.fontSize) {
            if (typeof style.fontSize === 'string') {
                style.fontSize = parseFloat(style.fontSize);
            }
            style.fontSize *= imageWidth;
        }

        // Setup
        this.interactive = false;
        // -- Index element
        // @ts-ignore TODO ignore
        this._indexElement.style = new PIXI.TextStyle(style);
        // -- Text element
        // @ts-ignore TODO ignore
        this._textElement.style = new PIXI.TextStyle(style);

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editLegendRow(this.model, (this.parent as MedsurfDraw.LegendColumn).model, (this.parent.parent as MedsurfDraw.LegendCollection).model);

        // TODO remove this.editText();
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateLegendRow(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateLegendRow(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            (this.parent as MedsurfDraw.LegendColumn).removeLegendRow(this.model);
            /* TODO remove
            // Set mode
            this.modeInteraction.setMode("delete");
            */
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Degroup image object
        (this.parent as MedsurfDraw.LegendColumn).onButtonDegroup(unselectElement);
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        (this.parent as MedsurfDraw.LegendColumn).onButtonDuplicate();
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        (this.parent as MedsurfDraw.LegendColumn).onButtonClipboard();
    }
    //</editor-fold>
    //<editor-fold desc="Color">
    /**
     * On color
     * @param mode
     * @param has
     * @param color
     * @param colorAlpha
     */
    public onColor(mode: ColorMode, has: boolean, color: number | undefined, colorAlpha: number | undefined): void {
        if (!color) {
            return;
        }

        if (mode === ColorMode.TEXT) {
            // Update model
            this.model.style.fill = color;

            // Elements
            // -- Text element
            this._textElement.style.fill = color;

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updateLegendRow(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Number">
    /**
     * On number
     * @param mode
     * @param number
     */
    public onNumber(mode: NumberMode, number: number): void {
        if (mode === NumberMode.SIZE) {
            // Round
            number = Math.round(number);

            // Update model
            this.model.style.fontSize = number / this.imageDimensions.width;

            // Elements
            // -- Text element
            this._textElement.style.fontSize = number;

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.emit("onBackgroundNumber");
            this.image.updateLegendRow(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Static">
    public onButtonStatic(override?: boolean): void {
        this.model.isStatic = override || !this.model.isStatic;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updateLegendRow(this.model);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectLegendRow(this.model, (this.parent as MedsurfDraw.LegendColumn).model, (this.parent.parent as MedsurfDraw.LegendCollection).model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this.parent.parent)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set default mode for parent
        (this.parent as MedsurfDraw.LegendColumn).modeInteraction.setModeItem('select_child');

        // Set default mode for parent.parent
        (this.parent.parent as MedsurfDraw.LegendCollection).modeInteraction.setModeItem('select_child');

        // Set default mode for siblings
        (this.parent as MedsurfDraw.LegendColumn).children
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                (imageObject as MedsurfDraw.ImageObject).modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('select_row');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control choose item
     */
    public controlChooseItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                /* TODO selftest
                if (this.model.selftestItems && this.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else {
                */
                    imageObject.modeInteraction.setMode('choose_unselect');
                /* TODO selftest
                }
                */
            });

        // Set children mode
        this.modeInteraction.setModeChildren('choose_parent');

        // Set draw item
        this.modeInteraction.setModeItem('choose_item');
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Elements
        // -- Index element
        this._indexElement.style.fill = Design.legendRow.fillColorHover;
        // -- Text element
        this._textElement.style.fill = Design.legendRow.fillColorHover;
    }

    /**
     * On out
     */
    public onOut(): void {
        // style
        const style = this._getCorrectStyle();

        // Elements
        // -- Index element
        this._indexElement.style = style;
        // -- Text element
        this._textElement.style = style;
    }

    /**
     * On pointer down
     * @param event
     */
    public onPointerDown(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_column') {
            this.controlSelectItem();
        }
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        const rect = super.getRectangle();
        const position = (this.parent as MedsurfDraw.LegendColumn).getRectangle();
        return new PIXI.Rectangle(position.x + rect.x, position.y + rect.y, rect.width, rect.height);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get index
     */
    public get index(): number {
        return this._index;
    }

    /**
     * Set index
     * @param value
     */
    public set index(value: number) {
        this._index = value;

        // Elements
        // -- Index element
        this._indexElement.text = this._index.toString();
    }
    //</editor-fold>
}