import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Menu entry model
 */
export class MenuEntryModel {
    element: MedsurfDraw.MenuObject | MedsurfDraw.RoundSubMenuElement | MedsurfDraw.SubMenuElement; // TODO change to this: MedsurfDraw.ButtonContainerElement | MedsurfDraw.ButtonElement | MedsurfDraw.NumberElement | MedsurfDraw.ColorElement;
    order: number;
}

/**
 * Menu element model
 */
export class MenuElementModel extends BaseContainerImageModel {
    menuElements: MenuEntryModel[] | undefined;
    circle?: PIXI.Circle;
}

/**
 * Menu element
 */
export class MenuElement extends BaseContainer<MenuElementModel> {
    /**
     * Members
     */
    protected _point: PIXI.IPoint;

    /**
     * Elements
     */
    protected _buttonTextElement: MedsurfDraw.Text;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: MenuElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.menu.zIndex;

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Move
            this.image.moveInteraction.on("onMove", this.onMove, this);
            // -- Zoom
            this.image.zoomInteraction.on("onZoom", this.onZoom, this);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Move
            this.image.moveInteraction.off("onMove", this.onMove, this);
            // -- Zoom
            this.image.zoomInteraction.off("onZoom", this.onZoom, this);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Button text element
        this._buttonTextElement = new MedsurfDraw.Text({
            text: "",
            style: new PIXI.TextStyle({
                fontSize: Design.menu.fontSize,
                fill: 0xffffff,
                align: 'center',
                breakWords: true,
                fontFamily: 'proxima_nova_altsemibold'
            })
        });
        this._buttonTextElement.anchor.set(0.5, 0.5);
        this._buttonTextElement.hideItem();
        this.addChild(this._buttonTextElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Menu elements
        if (this.menuElements) {
            this.menuElements.forEach((menuElement: MenuEntryModel) => {
                // Set position
                const angle = Design.menu.anglePositions[menuElement.order];
                const x = this.circle.radius * Math.cos(angle * Math.PI / 180);
                const y = this.circle.radius * Math.sin(angle * Math.PI / 180);
                menuElement.element.position.set(x, y);

                if (menuElement.element instanceof MedsurfDraw.IndexedRoundButtonElement
                    || menuElement.element instanceof MedsurfDraw.RoundSubMenuElement
                    || menuElement.element instanceof MedsurfDraw.SubMenuElement) {
                    menuElement.element.angle = angle;
                }

                // Draw
                menuElement.element.draw();
            });
        }
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Menu elements
        if (this.menuElements && this.menuElements.length > 0) {
            this._unsetMenu(this.menuElements);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Set mouse position
     * @param point
     */
    public setMousePosition(point: PIXI.IPoint): void {
        // Member
        this._point = point;

        // Position
        this.position.set(
            this.image.position.x + point.x * this.imageScale.x,
            this.image.position.y + point.y * this.imageScale.y
        );
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    /**
     * Set menu
     * @param menuElements
     * @private
     */
    protected _setMenu(menuElements: MenuEntryModel[] | undefined): void {
        if (menuElements && menuElements.length > 0) {
            menuElements.forEach((menuElement: MenuEntryModel) => {
                this.addChild(menuElement.element);
                menuElement.element.modeInteraction.setDefaultMode('default');

                // Events
                menuElement.element.on("onHover", this._showText.bind(this, menuElement.element.description, menuElement.element.descriptionFontSize), this);
                menuElement.element.on("onOut", this._hideText, this);
                if (menuElement.element instanceof MedsurfDraw.RoundSubMenuElement || menuElement.element instanceof MedsurfDraw.SubMenuElement) {
                    menuElement.element.menuElements.forEach((me) => {
                        me.element.on("onHover", this._showText.bind(this, me.element.description, me.element.descriptionFontSize), this);
                        me.element.on("onOut", this._hideText, this);
                    })
                }
            });

            // Draw
            this.emit("debounceDraw");
        }
    }

    /**
     * Unset menu
     * @param menuElements
     * @private
     */
    protected _unsetMenu(menuElements: MenuEntryModel[] | undefined): void {
        if (menuElements && menuElements.length > 0) {
            menuElements.forEach((menuElement: MenuEntryModel) => {
                this.removeChild(menuElement.element);

                // Events
                menuElement.element.off("onHover", this._showText.bind(this, menuElement.element.description, menuElement.element.descriptionFontSize), this);
                menuElement.element.off("onOut", this._hideText, this);
                if (menuElement.element instanceof MedsurfDraw.RoundSubMenuElement || menuElement.element instanceof MedsurfDraw.SubMenuElement) {
                    menuElement.element.menuElements.forEach((me) => {
                        me.element.off("onHover", this._showText.bind(this, me.element.description, me.element.descriptionFontSize), this);
                        me.element.off("onOut", this._hideText, this);
                    })
                }
            });
        }
    }
    //</editor-fold>

    //<editor-fold desc="Text">
    /**
     * Show text
     * @param text
     * @param fontSize
     * @protected
     */
    protected _showText(text: string, fontSize: number): void {
        // Set text
        // @ts-ignore
        this._buttonTextElement.text = text;

        // Elements
        // -- Button text element
        if (fontSize < Design.menu.minFontSize) {
            this._buttonTextElement.hideItem();
        } else {
            this._buttonTextElement.style.fontSize = fontSize;
            this._buttonTextElement.showItem();
        }
    }

    /**
     * Hide text
     * @protected
     */
    protected _hideText(): void {
        // Elements
        // -- Button text element
        this._buttonTextElement.hideItem();
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        if (!this._point) {
            return;
        }

        // Position
        this.position.set(
            this.image.position.x + this._point.x * this.imageScale.x,
            this.image.position.y + this._point.y * this.imageScale.y
        );
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On zoom
     */
    public onZoom(): void {
        if (!this._point) {
            return;
        }

        // Position
        this.position.set(
            this.image.position.x + this._point.x * this.imageScale.x,
            this.image.position.y + this._point.y * this.imageScale.y
        );
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    public hideItem() {
        super.hideItem();

        // Hide submenus
        this.menuElements?.filter((me) => me.element instanceof MedsurfDraw.RoundSubMenuElement || me.element instanceof MedsurfDraw.SubMenuElement).forEach((me) => {
            // @ts-ignore
            me.element.hideMenu(new PIXI.InteractionEvent());
        });
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get circle
     */
    public get circle(): PIXI.Circle {
        return this.data.circle || new PIXI.Circle(0, 0, Design.menu.radius);
    }

    /**
     * Set circle
     * @param value
     */
    public set circle(value: PIXI.Circle) {
        this.data.circle = value;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get menu elements
     */
    public get menuElements(): MenuEntryModel[] | undefined {
        return this.data.menuElements;
    }

    /**
     * Set menu elements
     * @param value
     */
    public set menuElements(value: MenuEntryModel[] | undefined) {
        // Remove menu elemenst
        this._unsetMenu(this.menuElements);

        // Set new value
        this.data.menuElements = value;

        // Add menu elemenst
        this._setMenu(this.menuElements);
    }
    //</editor-fold>
}