import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Round sub menu entry model
 */
export class RoundSubMenuEntryModel {
    element: MedsurfDraw.MenuObject; // TODO change to this: MedsurfDraw.ButtonContainerElement | MedsurfDraw.ButtonElement | MedsurfDraw.NumberElement | MedsurfDraw.ColorElement;
    order: number;
}

/**
 * Round sub menu element model
 */
export class RoundSubMenuElementModel extends BaseContainerImageModel {
    text: string;
    description: string;
    descriptionFontSize: number;
    menuElements: RoundSubMenuEntryModel[];
    circle: PIXI.Circle;
    fontSize?: number;
    fontColor?: number;
}

/**
 * Round Sub menu element
 */
export class RoundSubMenuElement extends BaseContainer<RoundSubMenuElementModel> {
    /**
     * Elements
     */
    protected _buttonElement: MedsurfDraw.RoundButtonElement;
    protected _buttonIndicatorElement: MedsurfDraw.RoundButtonIndicatorElement;
    protected _containerElement: MedsurfDraw.Container;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RoundSubMenuElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.subMenu.zIndex;
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Button element
        this._buttonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: this.text,
            description: this.description,
            descriptionFontSize: this.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, this.circle.radius),
            fontSize: this.fontSize,
            fontColor: this.fontColor,
        });
        this._buttonElement.on('onHover', this.onHover, this);
        this._buttonElement.on('onOut', this.onOut, this);
        this._buttonElement.on('button', this.onRelease, this);
        this.addChild(this._buttonElement);
        // -- Button indicator element
        this._buttonIndicatorElement = new MedsurfDraw.RoundButtonIndicatorElement({
            image: this.image,
            circle: new PIXI.Circle(0, 0, this.circle.radius),
        });
        this._buttonIndicatorElement.angle += 90;
        this.addChild(this._buttonIndicatorElement);
        // -- Container element
        this._containerElement = new MedsurfDraw.Container({image: this.image});
        this._containerElement.position.x = this._buttonElement.width + Design.subMenu.buttonGab;
        this._containerElement.hideItem();
        this.addChild(this._containerElement);
        // -- Menu elements
        if (this.menuElements && this.menuElements.length > 0) {
            this.menuElements.forEach((menuElement: RoundSubMenuEntryModel) => {
                this._containerElement.addChild(menuElement.element);
            });
        }
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Button element
        this._buttonElement.setIconAngle(360 - this.angle);
        this._buttonElement.draw();
        // -- Button indicator element
        this._buttonIndicatorElement.draw();
        // -- Container element
        this._containerElement.position.x = this._buttonElement.width + Design.subMenu.buttonGab;
        // -- Menu elements
        if (this.menuElements) {
            let prevElement: RoundSubMenuEntryModel;
            this.menuElements.forEach((menuElement: RoundSubMenuEntryModel) => {
                if (prevElement) {
                    menuElement.element.position.x = prevElement.element.position.x + prevElement.element.width + menuElement.order * Design.subMenuItems.buttonGab;
                } else {
                    menuElement.element.position.x = menuElement.order * Design.subMenuItems.buttonGab;
                }
                prevElement = menuElement;

                // Set icon angle
                if (menuElement.element instanceof MedsurfDraw.IndexedRoundButtonElement || menuElement.element instanceof MedsurfDraw.RoundButtonElement) {
                    menuElement.element.setIconAngle(360 - this.angle);
                }

                // Draw
                menuElement.element.draw();
            });
        }
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Menu elements
        if (this.menuElements && this.menuElements.length > 0) {
            this.menuElements.forEach((menuElement: RoundSubMenuEntryModel) => {
                if (this._containerElement) {
                    this._containerElement.removeChild(menuElement.element);
                }
            });
        }
        // -- Button element
        if (this._buttonElement) {
            this._buttonElement.destroy(options);
        }
        // -- Button indicaotr element
        if (this._buttonIndicatorElement) {
            this._buttonIndicatorElement.destroy(options);
        }
        // -- Container element
        if (this._containerElement) {
            this._containerElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Show menu
     * @param event
     */
    public showMenu(event: PIXI.InteractionEvent): void {
        this.toggleMenu(event, true);
    }

    /**
     * Hide menu
     * @param event
     */
    public hideMenu(event: PIXI.InteractionEvent): void {
        this.toggleMenu(event, false);
    }

    /**
     * Toggle menu
     * @param event
     * @param override
     */
    public toggleMenu(event: PIXI.InteractionEvent, override: boolean | undefined = undefined): void {
        event.stopPropagation();

        // Toggle menu
        if (this._containerElement) {
            if (override !== undefined) {
                this._containerElement.visible = override;
            } else {
                this._containerElement.visible = !this._containerElement.visible;
            }

            // Emits
            this.emit("toggleMenu", event, this._containerElement.visible);

            // Reset indicator
            if (!this._containerElement.visible) {
                //<editor-fold desc="Elements">
                // -- Button indicator element
                this._buttonIndicatorElement.visible = true;
                //</editor-fold>
            }
        } else {
            // Emits
            this.emit("toggleMenu", event, false);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * Emit on hover event
     * @param event
     */
    protected onHover(event: PIXI.InteractionEvent): void {
        // Element
        // -- Button indicator element
        this._buttonIndicatorElement.visible = false;

        // Emits
        this.emit("onHover", event);
    }

    /**
     * Emit on out event
     * @param event
     */
    protected onOut(event: PIXI.InteractionEvent): void {
        // Emits
        this.emit("onOut", event);
    }

    /**
     * Emit button event
     * @param event
     */
    protected onRelease(event: PIXI.InteractionEvent): void {
        // Hide sub menu
        this.hideMenu(event);

        // Emits
        this.emit("button", event);
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text
     */
    public get text(): string {
        return this.data.text;
    }

    /**
     * Set text
     * @param value
     */
    public set text(value: string) {
        this.data.text = value;
    }

    /**
     * Get description
     */
    public get description(): string {
        return this.data.description;
    }

    /**
     * Set description
     * @param value
     */
    public set description(value: string) {
        this.data.description = value;
    }

    /**
     * Get description font size
     */
    public get descriptionFontSize(): number {
        return this.data.descriptionFontSize;
    }

    /**
     * Set description font size
     * @param value
     */
    public set descriptionFontSize(value: number) {
        this.data.descriptionFontSize = value;
    }

    /**
     * Get menu elements
     */
    public get menuElements(): RoundSubMenuEntryModel[] {
        return this.data.menuElements;
    }

    /**
     * Set menu elements
     * @param value
     */
    public set menuElements(value: RoundSubMenuEntryModel[]) {
        this.data.menuElements = value;
    }

    /**
     * Get circle
     */
    public get circle(): PIXI.Circle {
        return this.data.circle;
    }

    /**
     * Set circle
     * @param value
     */
    public set circle(value: PIXI.Circle) {
        this.data.circle = value;

        this.emit("debounceDraw");
    }

    /**
     * Get fontSize
     */
    public get fontSize(): number {
        return this.data.fontSize || Design.subMenu.fontSize;
    }

    /**
     * Set fontSize
     * @param value
     */
    public set fontSize(value: number) {
        this.data.fontSize = value;

        // Elements
        // -- Button element
        this._buttonElement.fontSize = this.fontSize;
    }

    /**
     * Get fontColor
     */
    public get fontColor(): number {
        return this.data.fontColor || Design.subMenu.fontColor;
    }

    /**
     * Set fontColor
     * @param value
     */
    public set fontColor(value: number) {
        this.data.fontColor = value;

        // Elements
        // -- Button element
        this._buttonElement.fontColor = this.fontColor;
    }
    //</editor-fold>
}