import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {debounce} from "debounce";
import {RoundSubMenuEntryModel} from "../../public-api";

/**
 * Sub menu entry model
 */
export class SubMenuEntryModel {
    element: MedsurfDraw.MenuObject; // TODO change to this: MedsurfDraw.ButtonContainerElement | MedsurfDraw.ButtonElement | MedsurfDraw.NumberElement | MedsurfDraw.ColorElement;
    order: number;
}

/**
 * Sub menu element model
 */
export class SubMenuElementModel extends BaseContainerImageModel {
    text: string;
    description: string;
    descriptionFontSize: number;
    menuElements: SubMenuEntryModel[];
    rectangle: PIXI.Rectangle;
    fontSize?: number;
    fontColor?: number;
}

/**
 * Sub menu element
 */
export class SubMenuElement extends BaseContainer<SubMenuElementModel> {
    /**
     * Methods
     */
    protected _debounceToggleMenuMethod: any;

    /**
     * Elements
     */
    protected _buttonElement: MedsurfDraw.ButtonElement;
    protected _containerElement: MedsurfDraw.Container;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: SubMenuElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.subMenu.zIndex;

        //<editor-fold desc="Methods">
        this._debounceToggleMenuMethod = debounce(this.toggleMenu.bind(this), 50).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceToggleMenu", this._debounceToggleMenuMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceToggleMenu", this._debounceToggleMenuMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Button element
        this._buttonElement = new MedsurfDraw.ButtonElement({
            image: this.image,
            text: this.text,
            description: this.description,
            descriptionFontSize: this.descriptionFontSize,
            rectangle: new PIXI.Rectangle(0, 0, this.rectangle.width, this.rectangle.height),
            fontSize: this.fontSize,
            fontColor: this.fontColor,
        });
        this._buttonElement.on('button', this.toggleMenu, this);
        this._buttonElement.on('button', this._emitButtonEvent, this);
        this.addChild(this._buttonElement);
        // -- Container element
        this._containerElement = new MedsurfDraw.Container({image: this.image});
        this._containerElement.position.x = this._buttonElement.width + Design.subMenu.buttonGab;
        this._containerElement.hideItem();
        this.addChild(this._containerElement);
        // -- Menu elements
        if (this.menuElements && this.menuElements.length > 0) {
            this.menuElements.forEach((menuElement: SubMenuEntryModel) => {
                this._containerElement.addChild(menuElement.element);
            });
        }
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Button element
        this._buttonElement.setIconAngle(360 - this.angle);
        // -- Container element
        this._containerElement.position.x = this._buttonElement.width + Design.subMenu.buttonGab;
        // -- Menu elements
        if (this.menuElements) {
            let prevElement: RoundSubMenuEntryModel;
            this.menuElements.forEach((menuElement: SubMenuEntryModel) => {
                if (prevElement) {
                    menuElement.element.position.x = prevElement.element.position.x + prevElement.element.width + menuElement.order * Design.subMenuItems.buttonGab;
                } else {
                    menuElement.element.position.x = menuElement.order * Design.subMenuItems.buttonGab;
                }
                prevElement = menuElement;

                // Set icon angle
                if (menuElement.element instanceof MedsurfDraw.ButtonElement || menuElement.element instanceof MedsurfDraw.IndexedRoundButtonElement || menuElement.element instanceof MedsurfDraw.RoundButtonElement) {
                    menuElement.element.setIconAngle(360 - this.angle);
                }

                // Draw
                menuElement.element.draw();
            });
        }
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Menu elements
        if (this.menuElements && this.menuElements.length > 0) {
            this.menuElements.forEach((menuElement: SubMenuEntryModel) => {
                if (this._containerElement) {
                    this._containerElement.removeChild(menuElement.element);
                }
            });
        }
        // -- Button element
        if (this._buttonElement) {
            this._buttonElement.off('button', this.toggleMenu, this);
            this._buttonElement.off('button', this._emitButtonEvent, this);
            this._buttonElement.destroy(options);
        }
        // -- Container element
        if (this._containerElement) {
            this._containerElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Show menu
     * @param event
     */
    public showMenu(event: PIXI.InteractionEvent): void {
        this.toggleMenu(event, true);
    }

    /**
     * Hide menu
     * @param event
     */
    public hideMenu(event: PIXI.InteractionEvent): void {
        this.toggleMenu(event, false);
    }

    /**
     * Toggle menu
     * @param event
     * @param override
     */
    public toggleMenu(event: PIXI.InteractionEvent, override: boolean | undefined = undefined): void {
        event.stopPropagation();

        // Toggle menu
        if (this._containerElement) {
            if (override !== undefined) {
                this._containerElement.visible = override;
            } else {
                this._containerElement.visible = !this._containerElement.visible;
            }
        }
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Emit button event
     * @param event
     * @private
     */
    protected _emitButtonEvent(event: PIXI.InteractionEvent): void {
        // Hide sub menu
        this.hideMenu(event);

        // Emits
        this.emit("debounceToggleMenu", event, this._containerElement.visible);
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get text
     */
    public get text(): string {
        return this.data.text;
    }

    /**
     * Set text
     * @param value
     */
    public set text(value: string) {
        this.data.text = value;
    }

    /**
     * Get description
     */
    public get description(): string {
        return this.data.description;
    }

    /**
     * Set description
     * @param value
     */
    public set description(value: string) {
        this.data.description = value;
    }

    /**
     * Get description font size
     */
    public get descriptionFontSize(): number {
        return this.data.descriptionFontSize;
    }

    /**
     * Set description font size
     * @param value
     */
    public set descriptionFontSize(value: number) {
        this.data.descriptionFontSize = value;
    }

    /**
     * Get menu elements
     */
    public get menuElements(): SubMenuEntryModel[] {
        return this.data.menuElements;
    }

    /**
     * Set menu elements
     * @param value
     */
    public set menuElements(value: SubMenuEntryModel[]) {
        this.data.menuElements = value;
    }

    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get fontSize
     */
    public get fontSize(): number {
        return this.data.fontSize || Design.subMenu.fontSize;
    }

    /**
     * Set fontSize
     * @param value
     */
    public set fontSize(value: number) {
        this.data.fontSize = value;

        // Elements
        // -- Button element
        this._buttonElement.fontSize = this.fontSize;
    }

    /**
     * Get fontColor
     */
    public get fontColor(): number {
        return this.data.fontColor || Design.subMenu.fontColor;
    }

    /**
     * Set fontColor
     * @param value
     */
    public set fontColor(value: number) {
        this.data.fontColor = value;

        // Elements
        // -- Button element
        this._buttonElement.fontColor = this.fontColor;
    }
    //</editor-fold>
}