import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import { BaseContainer, BaseContainerImageModel } from "../../bases/elements/BaseContainer";
import { BaseSprite, BaseSpriteImageModel } from "../../bases/elements/BaseSprite";
import { Design } from "../../config/design";

/**
 * Navigator element model
 */
export class NavigatorElementModel extends BaseContainerImageModel {
    renderer: PIXI.Renderer | PIXI.CanvasRenderer;
    canvas: HTMLCanvasElement;
    app: PIXI.Application;
}

/**
 * Image navigator element
 * TODO not finished
 *  overrider behaviour -> too many functions -> merge them
 *  Extract Rectangle to separate file
 *  Exctract visible area to separate file
 */
export class NavigatorElement extends BaseContainer<NavigatorElementModel> {
    /**
     * Members
     */
    private _navigationHeight: number; // TODO

    private _baseTexture: PIXI.BaseRenderTexture;
    private _texture: PIXI.RenderTexture;

    private _scale: PIXI.ObservablePoint;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<NavigatorElement>;
    private _zoomInteraction: MedsurfDraw.ZoomInteraction<NavigatorElement>;

    /**
     * Methods
     */
    protected _wheelMethod: any;

    /**
     * Elements
     */
    protected _backgroundElement: MedsurfDraw.Rectangle;
    protected _spriteElement: BaseSprite<BaseSpriteImageModel>;
    protected _visibleAreaElement: MedsurfDraw.Rectangle;
    protected _toggleButtonElement: MedsurfDraw.ButtonElement;
    protected _zoomInButtonElement: MedsurfDraw.ButtonElement;
    protected _zoomOutButtonElement: MedsurfDraw.ButtonElement;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: NavigatorElementModel) {
        // Parent constructor
        super(model);

        // Members
        this.navigationHeight = Math.round(this.canvas.height / 6); // TODO

        // Setup
        this.zIndex = Design.imageNavigator.zIndex;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("hidden", this._modeHidden, this);
        this.modeInteraction.on("remove-hidden", this._removeModeHidden, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this, true);
        this.moveInteraction.on("onMove", this._onMove, this);
        // -- Zoom interaction
        this.zoomInteraction = new MedsurfDraw.ZoomInteraction(this, false);
        this.zoomInteraction.on("onZoom", this._onZoom, this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            this.image.on("onResize", this._onResize, this);

            // Ensure wheel Method works
            this._wheelMethod = this.zoomInteraction.onZoom.bind(this.zoomInteraction);
            this.canvas.addEventListener("wheel", this._wheelMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            this.image.off("onResize", this._onResize, this);

            // Ensure removeEventListener works
            this.canvas.removeEventListener("wheel", this._wheelMethod);
            // TODO remove delete this._wheelMethod;
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Texture
        // -- Base Texture
        // TODO
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.dimensions.width, height: this.image.dimensions.height});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.texture.width, height: this.image.texture.height});
        this._baseTexture = new PIXI.BaseRenderTexture({width: this.canvas.width, height: this.canvas.height});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.width, height: this.image.height});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.actualWidth, height: this.image.actualHeight});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this._app.stage.width, height: this._app.stage.height});
        // -- Texture
        this._texture = new PIXI.RenderTexture(this._baseTexture);

        const buttonSize = this.navigationHeight / Design.imageNavigator.buttonSizeAspect;
        const buttonFontSize = Math.round(buttonSize / Design.imageNavigator.buttonTextAspect);
        //<editor-fold desc="Elements">
        // -- Background element
        this._backgroundElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, this.canvas.width / this.canvas.height * this.navigationHeight, this.navigationHeight),
            options: {hasLine: false, hasFill: true, fillColor: this.renderer.backgroundColor, fillAlpha: 1},
            lineWidth: undefined
        });
        this._backgroundElement .zIndex = 10;
        this.addChild(this._backgroundElement);
        // -- Image Element
        this._spriteElement = new BaseSprite({
            image: this.image,
            texture: this._texture
        });
        this._spriteElement.interactive = true;
        this._spriteElement.zIndex = 20;
        // this._spriteElement.alpha = 1;
        this.addChild(this._spriteElement);
        // -- Visible area element
        this._visibleAreaElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(),
            options: {hasLine: true, lineColor: Design.imageNavigator.lineColor, lineAlpha: Design.imageNavigator.lineAlpha, hasFill: true, fillColor: Design.imageNavigator.fillColor, fillAlpha: Design.imageNavigator.fillAlpha},
            lineWidth: Design.imageNavigator.lineWidth
        });
        this._visibleAreaElement.zIndex = 30;
        this.addChild(this._visibleAreaElement);
        // -- Zoom in button element
        this._zoomInButtonElement = new MedsurfDraw.ButtonElement({
            image: this.image,
            text: "\uf00e",
            description: "Hineinzoomen",
            descriptionFontSize: 6,
            rectangle: new PIXI.Rectangle(0, 0, buttonSize, buttonSize),
            fontSize: buttonFontSize
        });
        this._zoomInButtonElement.on("button", this._onZoomIn, this);
        this.addChild(this._zoomInButtonElement);
        // -- Toggle button element
        this._toggleButtonElement = new MedsurfDraw.ButtonElement({
            image: this.image,
            text: this._spriteElement.visible ? "\uf070" : "\uf06e",
            description: "Sichtbar",
            descriptionFontSize: 6,
            rectangle: new PIXI.Rectangle(0, 0, buttonSize, buttonSize),
            fontSize: buttonFontSize
        }); // f06e
        this._toggleButtonElement.on("button", this._onToggle, this);
        this.addChild(this._toggleButtonElement);
        // -- Zoom out button element
        this._zoomOutButtonElement = new MedsurfDraw.ButtonElement({
            image: this.image,
            text: "\uf010",
            description: "Herrauszoomen",
            descriptionFontSize: 6,
            rectangle: new PIXI.Rectangle(0, 0, buttonSize, buttonSize),
            fontSize: buttonFontSize
        });
        this._zoomOutButtonElement.on("button", this._onZoomOut, this);
        this.addChild(this._zoomOutButtonElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.draw();
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Background element
        this._backgroundElement.draw();
        // -- Visible area element
        this._visibleAreaElement.draw();
        // -- Zoom in button element
        this._zoomInButtonElement.draw();
        // -- Toggle button element
        this._toggleButtonElement.draw();
        // -- Zoom out button element
        this._zoomOutButtonElement.draw();
        //</editor-fold>

        const bounds = this.getLocalBounds();
        this.position.set(this.canvas.width - bounds.width, 0);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        // -- Zoom interaction
        if (this.zoomInteraction) {
            this.zoomInteraction.removeAllListeners();
            // TODO remove delete this.zoomInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Background element
        if (this._backgroundElement) {
            this._backgroundElement.destroy(options);
        }
        // -- Sprite Element
        if (this._spriteElement) {
            this._spriteElement.destroy(options);
        }
        // -- Visible area element
        if (this._visibleAreaElement) {
            this._visibleAreaElement.destroy(options);
        }
        // -- Zoom in button element
        if (this._zoomInButtonElement) {
            this._zoomInButtonElement.off("button", this._onZoomIn, this);
            this._zoomInButtonElement.destroy(options);
        }
        // -- Toggle button element
        if (this._toggleButtonElement) {
            this._toggleButtonElement.off("button", this._onToggle, this);
            this._toggleButtonElement.destroy(options);
        }
        // -- Zoom out button element
        if (this._zoomOutButtonElement) {
            this._zoomOutButtonElement.off("button", this._onZoomOut, this);
            this._zoomOutButtonElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    /**
     * Update texture
     */
    public updateTexture(): void {
        // const tmpTransform = this.image.transform;
        // this.image.setTransform();
        // @ts-ignore TODO ignore
        this.renderer.render(this.image, this._texture);
        // this.renderer.render(this._app.stage, this._texture);
        // this.image.transform = tmpTransform;
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * On resize
     * @private
     */
    protected _onResize(): void {
        this._scale = this.image.scale;
        this.navigationHeight = Math.round(this.canvas.height / 6); // TODO

        const buttonSize = this.navigationHeight / Design.imageNavigator.buttonSizeAspect;
        const buttonFontSize = Math.round(buttonSize / Design.imageNavigator.buttonTextAspect);

        // Update base texture
        // TODO
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.dimensions.width, height: this.image.dimensions.height});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.texture.width, height: this.image.texture.height});
        this._baseTexture = new PIXI.BaseRenderTexture({width: this.canvas.width, height: this.canvas.height});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.width, height: this.image.height});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this.image.actualWidth, height: this.image.actualHeight});
        // this._baseTexture = new PIXI.BaseRenderTexture({width: this._app.stage.width, height: this._app.stage.height});

        this.updateTexture();

        // Elements
        // -- Background element
        this._backgroundElement.rectangle = new PIXI.Rectangle(0, 0, this.canvas.width / this.canvas.height * this.navigationHeight, this.navigationHeight);
        // -- Image Element
        this._spriteElement.width = this.canvas.width / this.canvas.height * this.navigationHeight;
        this._spriteElement.height = this.navigationHeight;
        // -- Visible area element
        this._visibleAreaElement.rectangle = new PIXI.Rectangle(0, 0, this.canvas.width / this.canvas.height * this.navigationHeight, this.navigationHeight);
        // -- Zoom in button element
        this._zoomInButtonElement.rectangle = new PIXI.Rectangle(0, 0, buttonSize, buttonSize);
        this._zoomInButtonElement.fontSize = buttonFontSize;
        this._zoomInButtonElement.position.set(this._spriteElement.width, 0);
        // -- Toggle button element
        this._toggleButtonElement.rectangle = new PIXI.Rectangle(0, 0, buttonSize, buttonSize);
        this._toggleButtonElement.fontSize = buttonFontSize;
        this._toggleButtonElement.position.set(this._spriteElement.width, buttonSize);
        // -- Zoom out button element
        this._zoomOutButtonElement.rectangle = new PIXI.Rectangle(0, 0, buttonSize, buttonSize);
        this._zoomOutButtonElement.fontSize = buttonFontSize;
        this._zoomOutButtonElement.position.set(this._spriteElement.width, 2 * buttonSize);

        this.draw();

        this._onImage(); // TODO merge with above???
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        this.interactive = true;

        if (this._spriteElement) {
            this._spriteElement.on("pointerover", this._onHover, this);
            this._spriteElement.on("pointerout", this._onRelease, this);
        }
        this.on("mousedown", this._moveInteraction.startMove, this._moveInteraction);
        this.on("pointermove", this._moveInteraction.onMove, this._moveInteraction);
        this.on("mouseup", this._moveInteraction.endMove, this._moveInteraction);
        this.on("pointerupoutside", this._moveInteraction.endMove, this._moveInteraction);

        if (this.image) {
            this.image.moveInteraction.on("onMove", this._onImage, this);
            this.image.zoomInteraction.on("onZoom", this._onImage, this);
            this.image.on("onSetPosition", this._onImage, this);
            this.image.on("onSetZoom", this._onImage, this);
        }
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
        this.interactive = false;

        if (this._spriteElement) {
            this._spriteElement.off("pointerover", this._onHover, this);
            this._spriteElement.off("pointerout", this._onRelease, this);
        }
        this.off("mousedown", this._moveInteraction.startMove, this._moveInteraction);
        this.off("pointermove", this._moveInteraction.onMove, this._moveInteraction);
        this.off("mouseup", this._moveInteraction.endMove, this._moveInteraction);
        this.off("pointerupoutside", this._moveInteraction.endMove, this._moveInteraction);

        if (this.image) {
            this.image.moveInteraction.off("onMove", this._onImage, this);
            this.image.zoomInteraction.off("onZoom", this._onImage, this);
            this.image.off("onSetPosition", this._onImage, this);
            this.image.off("onSetZoom", this._onImage, this);
        }
    }

    /**
     * Mode hidden
     * @private
     */
    protected _modeHidden(): void {
        this._backgroundElement.hideItem();
        this._spriteElement.hideItem();
        this._visibleAreaElement.hideItem();

        this._toggleButtonElement.text = "\uf06e";
    }

    /**
     * Remove mode hidden
     * @private
     */
    protected _removeModeHidden(): void {
        this._backgroundElement.showItem();
        this._spriteElement.showItem();
        this._visibleAreaElement.showItem();

        this._toggleButtonElement.text = "\uf070";

        this._onImage();
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     * @private
     */
    protected _onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        event.stopPropagation();

        const scale = this.imageScale;
        const scaledX = -1 * dX / this._spriteElement.width * this.canvas.width * scale.x;
        const scaledY = -1 * dY / this._spriteElement.height * this.canvas.height * scale.y;

        this.image.moveInteraction.emit("onMove", event, scaledX, scaledY);
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On zoom
     * @param event
     * @param x
     * @param y
     * @param factor
     * @private
     */
    protected _onZoom(event: PIXI.InteractionEvent, x: number, y: number, factor: number): void {
        event.stopPropagation();

        // TODO console.log((x - this.position.x) / this._spriteElement.width * this.canvas.width);
        // TODO console.log(y / this._spriteElement.height * this.canvas.height);

        const scale = this.imageScale;
        const scaledX = (x - this.position.x) / this._spriteElement.width * this.canvas.width / scale.x;
        const scaledY = y / this._spriteElement.height * this.canvas.height / scale.y;

        this.image.zoomInteraction.emit("onZoom", event, scaledX, scaledY, factor);
    }
    //</editor-fold>

    //<editor-fold desc="Elements">
    /**
     * On toggle
     * @param event
     * @private
     */
    protected _onToggle(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        this.modeInteraction.setMode(this._spriteElement.visible ? "hidden" : "default");
    }

    /**
     * On zoom in
     * @param event
     * @private
     */
    protected _onZoomIn(event: PIXI.InteractionEvent) {
        const factor = (1 + Design.imageNavigator.zoomStep);

        const scaledX = (this._visibleAreaElement.rectangle.x + this._visibleAreaElement.rectangle.width / 2) / this._spriteElement.width * this.canvas.width;
        const scaledY = (this._visibleAreaElement.rectangle.y + this._visibleAreaElement.rectangle.height / 2) / this._spriteElement.height * this.canvas.height;

        this.image.zoomInteraction.emit("onZoom", event, scaledX, scaledY, factor);
    }

    /**
     * On zoom out
     * @param event
     * @private
     */
    protected _onZoomOut(event: PIXI.InteractionEvent) {
        const factor = (1 + -1 * Design.imageNavigator.zoomStep);

        const scaledX = (this._visibleAreaElement.rectangle.x + this._visibleAreaElement.rectangle.width / 2) / this._spriteElement.width * this.canvas.width;
        const scaledY = (this._visibleAreaElement.rectangle.y + this._visibleAreaElement.rectangle.height / 2) / this._spriteElement.height * this.canvas.height;

        this.image.zoomInteraction.emit("onZoom", event, scaledX, scaledY, factor);
    }

    /**
     * On hover
     * @param event
     * @private
     */
    protected _onHover(event: PIXI.InteractionEvent): void {
        this.image.zoomInteraction.active = false;
        this.zoomInteraction.active = true;
    }

    /**
     * On release
     * @param event
     * @private
     */
    protected _onRelease(event: PIXI.InteractionEvent): void {
        this.image.zoomInteraction.active = true;
        this.zoomInteraction.active = false;
    }

    /**
     * On image
     * @private
     */
    protected _onImage(): void {
        if (!this._visibleAreaElement.rectangle) {
            return;
        }

        // TODO correct calculation
        if (this.image.imageType === MedsurfDraw.ImageTypes.DEEPZOOM) {
            // TODO const scale = this.getImageScale();
            /*
            this._visibleAreaElement.rectangle.x = -1 * this.image.position.x / scale.x * (this._spriteElement.width / this.image.actualWidth);
            this._visibleAreaElement.rectangle.y = -1 * this.image.position.y / scale.y * (this._spriteElement.height / this.image.actualHeight);

            this._visibleAreaElement.rectangle.width = this._spriteElement.width * (this.canvas.width / this.image.actualWidth / scale.x);
            this._visibleAreaElement.rectangle.height = this._spriteElement.height * (this.canvas.height / this.image.actualHeight / scale.y);
            */
        } else {
            const scaleX = 1 + this._scale.x - this.imageScale.x;
            const scaleY = 1 + this._scale.y - this.imageScale.y;
            const widthAspect = this._backgroundElement.width / this.canvas.width;
            const heightAspect = this._backgroundElement.height / this.canvas.height;
            this._visibleAreaElement.rectangle = new PIXI.Rectangle(
                (this.canvas.width / 2 - this.image.width / 2 - this.image.position.x) * scaleX * widthAspect,
                (this.canvas.height / 2 - this.image.height / 2 - this.image.position.y) * scaleY * heightAspect,
                this._backgroundElement.width,
                this._backgroundElement.height
            );
        }

        /**/
        if (this._visibleAreaElement.rectangle.x < 0) {
            this._visibleAreaElement.rectangle.width += this._visibleAreaElement.rectangle.x;
            this._visibleAreaElement.rectangle.x = 0;
        } else if (this._visibleAreaElement.rectangle.x + this._visibleAreaElement.rectangle.width > this._spriteElement.width) {
            this._visibleAreaElement.rectangle.width = this._spriteElement.width - this._visibleAreaElement.rectangle.x;
        }

        if (this._visibleAreaElement.rectangle.y < 0) {
            this._visibleAreaElement.rectangle.height += this._visibleAreaElement.rectangle.y;
            this._visibleAreaElement.rectangle.y = 0;
        } else if (this._visibleAreaElement.rectangle.y + this._visibleAreaElement.rectangle.height > this._spriteElement.height) {
            this._visibleAreaElement.rectangle.height = this._spriteElement.height - this._visibleAreaElement.rectangle.y;
        }

        if (this._visibleAreaElement.rectangle.width > this._backgroundElement.width) {
            this._visibleAreaElement.rectangle.width = this._backgroundElement.width;
        }
        if (this._visibleAreaElement.rectangle.height > this._backgroundElement.height) {
            this._visibleAreaElement.rectangle.height = this._backgroundElement.height;
        }
        /**/

        this.draw();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<NavigatorElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<NavigatorElement>) {
        this._moveInteraction = value;
    }

    /**
     * Get zoom interaction
     */
    public get zoomInteraction(): MedsurfDraw.ZoomInteraction<NavigatorElement> {
        return this._zoomInteraction;
    }

    /**
     * Set zoom interaction
     * @param value
     */
    public set zoomInteraction(value: MedsurfDraw.ZoomInteraction<NavigatorElement>) {
        this._zoomInteraction = value;
    }

    /**
     * Get renderer
     */
    public get renderer(): PIXI.Renderer | PIXI.CanvasRenderer {
        return this.data.renderer;
    }

    /**
     * Set renderer
     * @param value
     */
    public set renderer(value: PIXI.Renderer | PIXI.CanvasRenderer) {
        this.data.renderer = value;
    }

    /**
     * Get canvas
     */
    public get canvas(): HTMLCanvasElement {
        return this.data.canvas;
    }

    /**
     * Set canvas
     * @param value
     */
    public set canvas(value: HTMLCanvasElement) {
        this.data.canvas = value;
    }

    /**
     * Get image
     */
    public get image(): MedsurfDraw.Image {
        return this.data.image;
    }

    /**
     * Set image
     * @param value
     */
    public set image(value: MedsurfDraw.Image) {
        this.data.image = value;
        /*
        this.modeInteraction.changeMode('remove-default');
        this._image = value;

        // Elements
        if (this._spriteElement && this._spriteElement !== undefined) {
            gsap.to(this._spriteElement, {
                duration: 0.3,
                alpha: 0,
                onComplete: () => {
                    this.removeChild(this._spriteElement);
                    this._spriteElement.destroy({ children: true });

                    this._spriteElement = PIXI.Image.from(this._image.getTexture());
                    // this._spriteElement = PIXI.Image.from(texture); TODO
                    this._spriteElement.interactive = true;
                    this._spriteElement.zIndex = 10;
                    this._spriteElement.alpha = 0;
                    this.addChild(this._spriteElement);
                    this.sortChildren();

                    gsap.to(this._spriteElement, {
                        duration: 0.3,
                        alpha: 1
                    });
                    gsap.to(this._visibleAreaElement, {
                        duration: 0.3,
                        alpha: 1,
                    });

                    this._spriteElement.texture.baseTexture.once("loaded", this._onLoaded, this);

                    /**
                     * Texture already loaded
                     * /
                    if (this._spriteElement.texture.orig.width > 1) {
                        this._onLoaded();
                    }

                    this.modeInteraction.changeMode('default');
                }
            });
            gsap.to(this._visibleAreaElement, {
                duration: 0.3,
                alpha: 0,
            });
        } else {
            this._spriteElement = PIXI.Image.from(this._image.getTexture());
            // this._spriteElement = PIXI.Image.from(texture); TODO
            this._spriteElement.interactive = true;
            this._spriteElement.zIndex = 10;
            this.addChild(this._spriteElement);
            this.sortChildren();

            this._spriteElement.texture.baseTexture.once("loaded", this._onLoaded, this);

            /**
             * Texture already loaded
             * /
            if (this._spriteElement.texture.orig.width > 1) {
                this._onLoaded();
            }

            this.modeInteraction.changeMode('default');
        }
        */
    }

    /**
     * Get navigation height
     */
    public get navigationHeight(): number {
        return this._navigationHeight;
    }

    /**
     * Set navigation height
     * @param value
     */
    public set navigationHeight(value: number) {
        this._navigationHeight = value;
    }
    //</editor-fold>
}