import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

export class NumberElementModel extends BaseContainerImageModel {
    rectangle: PIXI.Rectangle = new PIXI.Rectangle(0, 0, 20, 20);
    value: number | undefined;
    minValue: number;
    maxValue: number;
    gab: number;
}

/**
 * Number element
 */
export class NumberElement extends BaseContainer<NumberElementModel> {
    /**
     * Elements
     */
    protected _backgroundElement: PIXI.Graphics;
    protected _valueElement: MedsurfDraw.Text;
    protected _numberSliderElement: MedsurfDraw.NumberSliderElement;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: NumberElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.numberElement.zIndex;

        // Position
        this.position.set(this.data.rectangle.x, this.data.rectangle.y);
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        //<editor-fold desc="Elements">
        // -- Background element
        this._backgroundElement = new PIXI.Graphics();
        this._backgroundElement.cursor = "select";
        this._backgroundElement.interactive = true;
        // this._backgroundElement.on("mousedown", this._onPointerDown, this);
        // this._backgroundElement.on("mouseup", this._onPointerUp, this);
        this.addChild(this._backgroundElement);
        // -- Value element
        this._valueElement = new MedsurfDraw.Text({
            text: (this.value || 'NAN').toString(),
            // @ts-ignore TODO ignore
            style: new PIXI.TextStyle({ fill: Design.numberElement.textFillColor, fontSize: Design.numberElement.textFontSize, fontFamily: Design.numberElement.fontFamilyDefault, align: Design.numberElement.alignDefault, wordWrap: Design.numberElement.wordWrapDefault })
        });
        this.addChild(this._valueElement);
        // -- Number slider element
        this._numberSliderElement = new MedsurfDraw.NumberSliderElement({
            image: this.image,
            value: this.value,
            minValue: this.minValue,
            maxValue: this.maxValue,
            squareWidth: Design.numberElement.buttonWidth * Design.numberSliderElement.squareWidthAspect,
            squareHeight: Design.numberElement.buttonWidth,
            gab: this.gab
        });
        this._numberSliderElement.on("number", this.onNumber, this);
        this.addChild(this._numberSliderElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Value element
        this._valueElement.position.set(this.rectangle.x + this.rectangle.width, this.rectangle.y - this.rectangle.height / 2);
        // -- Number slider element
        this._numberSliderElement.position.set(this._valueElement.x + this._valueElement.width + 5, this._valueElement.y);
        this._numberSliderElement.draw();
        //</editor-fold>

        // Pivot
        // this.pivot.set(0, this.height / 4);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Background element
        if (this._backgroundElement) {
            this._backgroundElement.destroy(options);
        }
        // -- Value element
        if (this._valueElement) {
            this._valueElement.destroy(options);
        }
        // -- Number slider element
        if (this._numberSliderElement) {
            this._numberSliderElement.off("number", this.onNumber, this);
            this._numberSliderElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Number">
    /**
     * On number
     * @param number
     */
    public onNumber(number: number): void {
        // Elements
        // -- Value element
        this._valueElement.text = number.toString();

        // Emit number
        this.emit("number", number);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get value
     */
    public get value(): number | undefined {
        return this.data.value;
    }

    /**
     * Set value
     * @param value
     */
    public set value(value: number | undefined) {
        this.data.value = value;

        // Elements
        // -- Value element
        this._valueElement.text = (this.value || "N/A").toString();
        // -- Number slider element
        this._numberSliderElement.value = this.value;
    }

    /**
     * Get min Value
     */
    public get minValue(): number {
        return this.data.minValue;
    }

    /**
     * Set min value
     * @param value
     */
    public set minValue(value: number) {
        this.data.minValue = value;

        // Elements
        // -- Number slider element
        this._numberSliderElement.minValue = this.minValue;
    }

    /**
     * Get max value
     */
    public get maxValue(): number {
        return this.data.maxValue;
    }

    /**
     * Set max value
     * @param value
     */
    public set maxValue(value: number) {
        this.data.maxValue = value;

        // Elements
        // -- Number slider element
        this._numberSliderElement.maxValue = this.maxValue;
    }

    /**
     * Get gab
     */
    public get gab(): number {
        return this.data.gab;
    }

    /**
     * Set gab
     * @param value
     */
    public set gab(value: number) {
        this.data.gab = value;

        // Elements
        // -- Number slider element
        this._numberSliderElement.gab = this.gab;
    }
    //</editor-fold>
}