import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

export class NumberSliderElementModel extends BaseContainerImageModel {
    value: number | undefined;
    minValue: number;
    maxValue: number;
    squareWidth: number = 20;
    squareHeight: number = 20;
    gab: number = 1;
}

/**
 * Number slider element
 */
export class NumberSliderElement extends BaseContainer<NumberSliderElementModel> {
    /**
     * Members
     */
    protected _valueRange: number;
    protected _minDisplayValue: number;
    protected _maxDisplayValue: number;
    protected _gabWidth: number;
    protected _x: number;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<NumberSliderElement>;

    /**
     * Elements
     */
    protected _backgroundElement: PIXI.Graphics;
    protected _sliderElement: MedsurfDraw.Rectangle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: NumberSliderElementModel) {
        // Parent constructor
        super(model);

        // Members
        this._valueRange = this.maxValue - this.minValue;
        this._minDisplayValue = (this.value || this.minValue) - Design.numberSliderElement.maxVisibleSnapPoints / 2;
        if (this._minDisplayValue < this.minValue) {
            this._minDisplayValue = this.minValue;
        }
        this._maxDisplayValue = (this.value || this.minValue) + Design.numberSliderElement.maxVisibleSnapPoints / 2;
        if (this._maxDisplayValue > this.maxValue) {
            this._maxDisplayValue = this.maxValue;
        }
        this._gabWidth = this._valueRange < Design.numberSliderElement.maxVisibleSnapPoints ? this.squareWidth / this._valueRange / this.gab : this.squareWidth / Design.numberSliderElement.maxVisibleSnapPoints / this.gab;

        // Setup
        this.zIndex = Design.numberSliderElement.zIndex;

        //<editor-fold desc="Interactions">
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.onStartMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        this.on("pointerout", this.moveInteraction.endMove, this.moveInteraction);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        //<editor-fold desc="Elements">
        // -- Background element
        this._backgroundElement = new PIXI.Graphics();
        this.addChild(this._backgroundElement);
        // -- Slider element
        this._sliderElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, Design.numberSliderElement.sliderWidth, this.squareHeight),
            options: {hasLine: false, hasFill: true, fillColor: Design.numberSliderElement.fillColor, fillAlpha: Design.numberSliderElement.fillAlpha},
            lineWidth: undefined
        });
        this.addChild(this._sliderElement);
        //</editor-fold>

        // Setup background
        this.createSliderBackground();
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Slider element
        this._sliderElement.rectangle = new PIXI.Rectangle(0, 0, Design.numberSliderElement.sliderWidth, this.squareHeight);
        this._sliderElement.draw();
        //</editor-fold>

        // Set position
        this._sliderElement.position.x = ((this.value || this.minValue) - this._minDisplayValue) * this._gabWidth;
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
            this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
            this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
            this.off("pointerout", this.moveInteraction.endMove, this.moveInteraction);
            this.moveInteraction.off("startMove", this.onStartMove, this);
            this.moveInteraction.off("onMove", this.onMove, this);
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Background element
        if (this._backgroundElement) {
            this._backgroundElement.destroy(options);
        }
        // -- Slider element
        if (this._sliderElement) {
            this._sliderElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Public functions">
    public createSliderBackground(): void {
        this._backgroundElement.clear();
        this._backgroundElement.lineStyle(Design.numberSliderElement.borderLineWidth, Design.numberSliderElement.borderLineColor, Design.numberSliderElement.borderLineAlpha);
        this._backgroundElement.beginFill(Design.numberSliderElement.borderLineColor, 0.5);
        this._backgroundElement.drawRect(0, 0, this.squareWidth, this.squareHeight);
        this._backgroundElement.endFill();

        for (let i = 1; i === i; i) {
            this._backgroundElement.lineStyle(Design.numberSliderElement.borderLineWidth, Design.numberSliderElement.borderLineColor, Design.numberSliderElement.borderLineAlpha);
            this._backgroundElement.moveTo(this._gabWidth * i, 0);
            this._backgroundElement.lineTo(this._gabWidth * i, 4);
            this._backgroundElement.endFill();

            if (this._gabWidth * ++i > this.squareWidth) {
                break;
            }
        }
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On start move
     * @param event
     */
    public onStartMove(event: PIXI.InteractionEvent): void {
        // @ts-ignore TODO ignore
        const globalPosition = this.getGlobalPosition();

        this._x = Math.round(event.data.global.x - globalPosition.x) / this.imageScale.x;
        const modX = this._x % this._gabWidth;

        // Get correct posX
        let posX: number = 0;
        if (modX < this._gabWidth / 2) {
            posX = this._x - modX;
        } else {
            posX = this._x + (this._gabWidth - modX);
        }

        if (posX < 0) {
            posX = 0;
        }
        if (posX >= this.squareWidth) {
            posX = this.squareWidth - 1;
        }

        // pos on slider
        const pos = Math.round(posX / this._gabWidth);

        // get diffX
        const diffX = (this._minDisplayValue + pos) - (this.value || this.minValue);

        // Set values
        if (!this.data.value) {
            this.data.value = this.data.minValue;
        }
        this.data.value += diffX;
        if ((this.value || this.minValue) < this.minValue) {
            this.data.value = this.minValue;
        }
        if ((this.value || this.minValue) > this.maxValue) {
            this.data.value = this.maxValue;
        }

        // Set boundaries
        this._minDisplayValue = (this.value || this.minValue) - Design.numberSliderElement.maxVisibleSnapPoints / 2;
        this._maxDisplayValue = (this.value || this.minValue) + Design.numberSliderElement.maxVisibleSnapPoints / 2;

        // Adjust min/max values
        if (this._minDisplayValue <= this.minValue) {
            this._minDisplayValue = this.minValue;
            this._maxDisplayValue = this.minValue + (Design.numberSliderElement.maxVisibleSnapPoints + 1);
        }
        if (this._maxDisplayValue >= this.maxValue) {
            this._minDisplayValue = this.maxValue - (Design.numberSliderElement.maxVisibleSnapPoints + 1);
            this._maxDisplayValue = this.maxValue;
        }

        // Check boundaries
        if (this._minDisplayValue < this.minValue) {
            this._minDisplayValue = this.minValue;
        }
        if (this._maxDisplayValue > this.maxValue) {
            this._maxDisplayValue = this.maxValue;
        }

        // Elements
        // -- Slider element
        this._sliderElement.position.x = ((this.value || this.minValue) - this._minDisplayValue) * this._gabWidth;

        // Emit number
        this.emit("number", this.value);
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        this._x += dX;
        const modX = this._x % this._gabWidth;

        // Get correct posX
        let posX: number = 0;
        if (modX < this._gabWidth / 2) {
            posX = this._x - modX;
        } else {
            posX = this._x + (this._gabWidth - modX);
        }

        if (posX < 0) {
            posX = 0;
        }
        if (posX >= this.squareWidth) {
            posX = this.squareWidth - 1;
        }

        // pos on slider
        const pos = Math.round(posX / this._gabWidth);

        // get diffX
        const diffX = (this._minDisplayValue + pos) - (this.value || this.minValue);

        // Set values
        if (!this.data.value) {
            this.data.value = this.minValue;
        }
        this.data.value += diffX;
        if ((this.value || this.minValue) < this.minValue) {
            this.data.value = this.minValue;
        }
        if ((this.value || this.minValue) > this.maxValue) {
            this.data.value = this.maxValue;
        }

        // Set boundaries
        this._minDisplayValue = (this.value || this.minValue) - Design.numberSliderElement.maxVisibleSnapPoints / 2;
        this._maxDisplayValue = (this.value || this.minValue) + Design.numberSliderElement.maxVisibleSnapPoints / 2;

        // Adjust min/max values
        if (this._minDisplayValue <= this.minValue) {
            this._minDisplayValue = this.minValue;
            this._maxDisplayValue = this.minValue + (Design.numberSliderElement.maxVisibleSnapPoints + 1);
        }
        if (this._maxDisplayValue >= this.maxValue) {
            this._minDisplayValue = this.maxValue - (Design.numberSliderElement.maxVisibleSnapPoints + 1);
            this._maxDisplayValue = this.maxValue;
        }

        // Check boundaries
        if (this._minDisplayValue < this.minValue) {
            this._minDisplayValue = this.minValue;
        }
        if (this._maxDisplayValue > this.maxValue) {
            this._maxDisplayValue = this.maxValue;
        }

        // Elements
        // -- Slider element
        this._sliderElement.position.x = ((this.value || this.minValue) - this._minDisplayValue) * this._gabWidth;

        /// Emit number
        this.emit("number", this.value);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get value
     */
    public get value(): number | undefined {
        return this.data.value;
    }

    /**
     * Set value
     * @param value
     */
    public set value(value: number | undefined) {
        this.data.value = value;

        // Update members
        // -- Min display value
        this._minDisplayValue = (this.value || this.minValue) - Design.numberSliderElement.maxVisibleSnapPoints / 2;
        if (this._minDisplayValue < this.minValue) {
            this._minDisplayValue = this.minValue;
        }
        // -- Max display value
        this._maxDisplayValue = (this.value || this.minValue) + Design.numberSliderElement.maxVisibleSnapPoints / 2;
        if (this._maxDisplayValue > this.maxValue) {
            this._maxDisplayValue = this.maxValue;
        }

        // Update background
        this.createSliderBackground();
    }

    /**
     * Get min value
     */
    public get minValue(): number {
        return this.data.minValue;
    }

    /**
     * Set min value
     * @param value
     */
    public set minValue(value: number) {
        // Set value
        this.data.minValue = value;

        // Update members
        // -- Value range
        this._valueRange = this.maxValue - this.minValue;
        // -- Min display value
        this._minDisplayValue = (this.value || this.minValue) - Design.numberSliderElement.maxVisibleSnapPoints / 2;
        if (this._minDisplayValue < this.minValue) {
            this._minDisplayValue = this.minValue;
        }
        // -- Gab width
        this._gabWidth = this._valueRange < Design.numberSliderElement.maxVisibleSnapPoints ? this.squareWidth / this._valueRange / this.gab : this.squareWidth / Design.numberSliderElement.maxVisibleSnapPoints / this.gab;

        // Update background
        this.createSliderBackground();
    }

    /**
     * Get max value
     */
    public get maxValue(): number {
        return this.data.maxValue;
    }

    /**
     * Set max value
     * @param value
     */
    public set maxValue(value: number) {
        this.data.maxValue = value;

        // Update members
        // -- Value range
        this._valueRange = this.maxValue - this.minValue;
        // -- Max display value
        this._maxDisplayValue = (this.value || this.minValue) + Design.numberSliderElement.maxVisibleSnapPoints / 2;
        if (this._maxDisplayValue > this.maxValue) {
            this._maxDisplayValue = this.maxValue;
        }
        // -- Gab width
        this._gabWidth = this._valueRange < Design.numberSliderElement.maxVisibleSnapPoints ? this.squareWidth / this._valueRange / this.gab : this.squareWidth / Design.numberSliderElement.maxVisibleSnapPoints / this.gab;

        // Update backgrouns
        this.createSliderBackground();
    }

    /**
     * Get squareWidth
     */
    public get squareWidth(): number {
        return this.data.squareWidth;
    }

    /**
     * Set squareWidth
     * @param value
     */
    public set squareWidth(value: number) {
        this.data.squareWidth = value;

        // Update members
        // -- Gab width
        this._gabWidth = this._valueRange < Design.numberSliderElement.maxVisibleSnapPoints ? this.squareWidth / this._valueRange / this.gab : this.squareWidth / Design.numberSliderElement.maxVisibleSnapPoints / this.gab;

        // Update backgrouns
        this.createSliderBackground();
    }

    /**
     * Get squareHeight
     */
    public get squareHeight(): number {
        return this.data.squareHeight;
    }

    /**
     * Set squareHeight
     * @param value
     */
    public set squareHeight(value: number) {
        this.data.squareHeight = value;
    }

    /**
     * Get gab
     */
    public get gab(): number {
        return this.data.gab;
    }

    /**
     * Set gab
     * @param value
     */
    public set gab(value: number) {
        this.data.gab = value;

        // Update members
        // -- Gab width
        this._gabWidth = this._valueRange < Design.numberSliderElement.maxVisibleSnapPoints ? this.squareWidth / this._valueRange / this.gab : this.squareWidth / Design.numberSliderElement.maxVisibleSnapPoints / this.gab;

        // Update backgrouns
        this.createSliderBackground();
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<NumberSliderElement> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<NumberSliderElement>) {
        this._moveInteraction = value;
    }
    //</editor-fold>
}