import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {BaseElementContainer, BaseElementContainerModel} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";

/**
 * Position point
 */
export class PositionPoint extends BaseElementContainer<Models.PositionPoint, BaseElementContainerModel<Models.PositionPoint>> {
    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<PositionPoint>;
    private _selectInteraction: MedsurfDraw.SelectInteraction<PositionPoint>;
    private _chooseInteraction: MedsurfDraw.ChooseInteraction<PositionPoint>;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: any;
    protected _debounceControlUnselectItemMethod: any;
    protected _debounceDeleteItemMethod: any;

    /**
     * Elements
     */
    protected _positionPointElement: MedsurfDraw.Circle;
    protected _lineElement: MedsurfDraw.Rectangle;
    // TODO selftest indicator: protected _killboxElement: MedsurfDraw.Killbox;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.PositionPoint) {
        // Parent constructor
        super({image, model});

        // Setup
        this.zIndex = Design.positionPoint.zIndexDefault + this.model.zIndex || 0;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on('select_layer_group', this._modeSelectLayerGroup, this);
        this.modeInteraction.on('remove-select_layer_group', this._removeModeSelectLayerGroup, this);
        this.modeInteraction.on('select_grouping', this._modeSelectGrouping, this);
        this.modeInteraction.on('remove-select_grouping', this._removeModeSelectGrouping, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('select_child', this._modeSelectChild, this);
        this.modeInteraction.on("remove-select_child", this._removeModeSelectChild, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('select_line', this._modeSelectLine, this);
        this.modeInteraction.on("remove-select_line", this._removeModeSelectLine, this);
        this.modeInteraction.on('draw_item', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item", this._removeModeDrawItem, this);
        this.modeInteraction.on('draw_item_new', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item_new", this._removeModeDrawItem, this);
        this.modeInteraction.on('draw_select', this._modeDrawSelect, this);
        this.modeInteraction.on('remove-draw_select', this._removeModeDrawSelect, this);
        this.modeInteraction.on('draw_select_new', this._modeDrawSelect, this);
        this.modeInteraction.on('remove-draw_select_new', this._removeModeDrawSelect, this);
        this.modeInteraction.on('draw_clickable', this._modeDrawSelect, this);
        this.modeInteraction.on('remove-draw_clickable', this._removeModeDrawSelect, this);
        this.modeInteraction.on('choose_item', this._modeChooseItem, this);
        this.modeInteraction.on("remove-choose_item", this._removeModeChooseItem, this);
        this.modeInteraction.on('choose_select', this._modeChooseSelect, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeChooseSelect, this);
        this.modeInteraction.on('choose_unselect', this._modeChooseUnselect, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeChooseUnselect, this);
        this.modeInteraction.on('clone_item', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item", this._removeModeCloneItem, this);
        this.modeInteraction.on('clone_item_new', this._modeCloneItem, this);
        this.modeInteraction.on("remove-clone_item_new", this._removeModeCloneItem, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        this.modeInteraction.on("delete_multiple", this._modeDelete, this);
        this.modeInteraction.on('delete_draw', this._modeDelete, this);
        this.modeInteraction.on('delete_clone', this._modeDelete, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.startMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.moveInteraction.on("endMove", this.endMove, this);
        this.moveInteraction.on("onRelease", this.onRelease, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelectStart", this.onSelectStart, this);
        this.selectInteraction.on("onSelftest", this.onSelftest, this);
        // -- Choose interaction
        this.chooseInteraction = new MedsurfDraw.ChooseInteraction<PositionPoint>(this, MedsurfDraw.ChooseTypes.Collections);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        this._debounceControlUnselectItemMethod = debounce(this.controlUnselectItem.bind(this), 10).bind(this);
        this._debounceDeleteItemMethod = debounce(this._deleteItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.on("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
            this.on("debounceDeleteItem", this._debounceDeleteItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.off("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
            this.off("debounceDeleteItem", this._debounceDeleteItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Position
        this.position.set(
            this.model.position.x * imageWidth,
            this.model.position.y * imageWidth
        );

        //<editor-fold desc="Elements">
        // -- Position point element
        this._positionPointElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.positionPoint.radius),
            options: {hasLine: true, lineColor: Design.positionPoint.lineColorNormal, lineAlpha: Design.positionPoint.lineAlpha, hasFill: true, fillColor: Design.positionPoint.fillColorNormal, fillAlpha: Design.positionPoint.fillAlpha},
            lineWidth: Design.positionPoint.lineWidth
        });
        this.addChild(this._positionPointElement);
        // -- Line element
        this._lineElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(-1 * Design.positionPoint.lineRectangleWidth / 2, -1 * Design.positionPoint.lineRectangleHeight / 2, Design.positionPoint.lineRectangleWidth, Design.positionPoint.lineRectangleHeight),
            options: {hasLine: true, lineColor: 0xffffff, lineAlpha: 1, hasFill: true, fillColor: 0xffffff, fillAlpha: 0.8},
            lineWidth: Design.positionPoint.lineWidth
        });
        this._lineElement.zIndex = Design.positionPoint.lineRectangleZIndex;
        this._lineElement.hideItem();
        this.addChild(this._lineElement);

        /* TODO selftest indicator
        // -- Killbox element
        this._killboxElement = new MedsurfDraw.Killbox({
            rectangle: new PIXI.Rectangle(0, 0, 0, 0)
        });
        this._killboxElement.visible = this.model.isSelftest;
        this.parent.addChild(this._killboxElement);
        */
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // Elements
        let scale = this.imageScale.x * Design.positionPoint.scaleAspect;
        let lineWidth = Math.floor(Design.positionPoint.lineWidth || 1 / scale);
        if (lineWidth < 1) {
            lineWidth = 1;
        }
        // -- Position point element
        if (this._positionPointElement.visible) {
            let radius = Design.positionPoint.radius;
            // TODO if (this.modeInteraction.lastMode === 'select' || this.modeInteraction.lastMode === 'selftest') {
            radius = Math.floor(radius / scale);
            /* TODO } else if (scale < 1) {
                scale = 1;
            }*/
            this._positionPointElement.circle.radius = radius;
            this._positionPointElement.lineWidth = lineWidth;
            this._positionPointElement.draw();
        }
        // -- Line element
        if (this._lineElement.visible) {
            let width = Design.positionPoint.lineRectangleWidth;
            width = Math.floor(width / scale);
            let height = Design.positionPoint.lineRectangleHeight;
            height = Math.floor(height / scale);
            this._lineElement.rectangle = new PIXI.Rectangle(-1 * width / 2, -1 * height / 2, width, height);
            this._lineElement.lineWidth = lineWidth;
            this._lineElement.draw();
        }
        /* TODO selftest indicator
        // -- Killbox element
        const rectangle = this.getRectangle();
        rectangle.pad(Design.selftestElement.gab);
        this._killboxElement.rectangle = rectangle;
        this._killboxElement.draw();
        */
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        // -- Choose interaction
        if (this.chooseInteraction) {
            this.chooseInteraction.removeAllListeners();
            // TODO remove delete this.chooseInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Position point element
        if (this._positionPointElement) {
            this._positionPointElement.destroy(options);
        }
        // -- Line element
        if (this._lineElement) {
            this._lineElement.destroy(options);
        }
        /* TODO selftest indicator
        // -- Killbox element
        if (this._killboxElement) {
            this._killboxElement.destroy(options);
        }
        */
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage): PositionPoint {
        // Element
        // -- Position point
        return new PositionPoint(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.POSITIONPOINT,
            isStatic: false,
            zIndex: 0,
            position: {
                x: 0,
                y: 0,
            },
            isSelftest: false
        });
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Get lines
     */
    public getLines(): MedsurfDraw.Line[] {
        return this.image.getLinesByPositionPoint(this);
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        // Hide item
        this.hideItem();
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        if (this.model.primitives && this.model.primitives.length > 0) {
            // Elements
            // -- Position point element
            this._positionPointElement.alpha = 0;

            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
        // Elements
        // -- Position point element
        this._positionPointElement.alpha = 1;
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        if (this.model.isSelftest) {
            // Setup
            this.interactive = true;
            this.cursor = "select";

            // TODO not used: if (this.model.primitives && this.model.primitives.length > 0) {
                // Elements
                // -- Position point element
                this._positionPointElement.alpha = 0;
            /* TODO not used:
            } else {
                // Elements
                // -- Position point element
                this._positionPointElement.circle.radius = Design.positionPoint.radius;

                // Events
                this.selectInteraction.reset();
                this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
                this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
                this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
                this.on("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
                // TODO this.on("onRelease", this._onSelftestRelease, this);
            }
            */

            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Setup
        this.interactive = false;

        // Elements
        // -- Position point element
        this._positionPointElement.alpha = 1;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
        // TODO this.off("onRelease", this._onSelftestRelease, this);
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        this.interactive = true;

        // Elements
        // -- Position point element
        // TODO this._positionPointElement.circle.radius = Design.positionPoint.radius;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Hide element when primitives exists
        if (this.model.primitives && this.model.primitives.length > 0) {
            // Elements
            // -- Position point element
            this._positionPointElement.alpha = 0;
        }

        // Show item
        this.showItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        this.interactive = false;

        // Elements
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorNormal;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorNormal;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select layer
     * @private
     */
    protected _modeSelectLayerGroup(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select layer
     * @private
     */
    protected _removeModeSelectLayerGroup(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select grouping
     * @private
     */
    protected _modeSelectGrouping(): void {
        // Setup
        this.interactive = true;

        // Elements
        // -- Position point element
        // TODO this._positionPointElement.circle.radius = Design.positionPoint.radius;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select grouping
     * @private
     */
    protected _removeModeSelectGrouping(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Elements
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorNormal;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorNormal;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.on("startMove", this._startMove, this);
        // TODO this.on("endMove", this._endMove, this);
        // TODO this.on("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.on("down_ArrowUp", "onPositionPointMoveUp", this.moveInteraction.moveUp.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowRight", "onPositionPointMoveRight", this.moveInteraction.moveRight.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowDown", "onPositionPointMoveDown", this.moveInteraction.moveDown.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowLeft", "onPositionPointMoveLeft", this.moveInteraction.moveLeft.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowUp", "onPositionPointMoveUpEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowRight", "onPositionPointMoveRightEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowDown", "onPositionPointMoveDownEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowLeft", "onPositionPointMoveLeftEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.off("startMove", this._startMove, this);
        // TODO this.off("endMove", this._endMove, this);
        // TODO this.off("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.remove("down_ArrowUp", "onPositionPointMoveUp");
        MedsurfDraw.Keyboard.events.remove("down_ArrowRight", "onPositionPointMoveRight");
        MedsurfDraw.Keyboard.events.remove("down_ArrowDown", "onPositionPointMoveDown");
        MedsurfDraw.Keyboard.events.remove("down_ArrowLeft", "onPositionPointMoveLeft");
        MedsurfDraw.Keyboard.events.remove("released_ArrowUp", "onPositionPointMoveUpEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowRight", "onPositionPointMoveRightEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowDown", "onPositionPointMoveDownEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowLeft", "onPositionPointMoveLeftEnd");

        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // TODO remove console.log("pp _removeModeSelectItem", mode);
        if (mode === 'author') {
            this.emit("debounceControlUnselectItem", this.image);
        }
        /* TODO remove
        if (!mode.startsWith('select_child') && !mode.startsWith('select_item') && !mode.startsWith('choose_item')
            && !mode.startsWith('choose_select') && !mode.startsWith('blocked_item')
            && ['default', 'marker', 'selftest'/*, 'author'* /, 'delete'].indexOf(mode) === -1) {
            // Unselect item
            this.emit("debounceControlUnselectItem");
        }
        */
    }

    /**
     * Mode select child
     * @private
     */
    protected _modeSelectChild(): void {
    }

    /**
     * Remove mode select child
     * @private
     */
    protected _removeModeSelectChild(): void {
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        // Filters
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
        // Element
        // -- Position point element
        this._positionPointElement.options.lineAlpha = Design.blocked.lineAlpha;
        this._positionPointElement.options.fillAlpha = Design.blocked.fillAlpha;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        // Filters
        this.filters = [];
        */

        // Element
        // -- Position point element
        this._positionPointElement.options.lineAlpha = Design.positionPoint.lineAlpha;
        this._positionPointElement.options.fillAlpha = Design.positionPoint.fillAlpha;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode select line
     * @private
     */
    protected _modeSelectLine(): void {
        // Setup
        this.interactive = true;

        // Elements
        // -- Position point element
        this._positionPointElement.hideItem();
        // -- Line element
        this._lineElement.showItem();

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.on("startMove", this._startMove, this);
        // TODO this.on("endMove", this._endMove, this);
        // TODO this.on("onRelease", this._onRelease, this);

        // Sort children
        this.sortChildren();

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select line
     * @private
     */
    protected _removeModeSelectLine(): void {
        // Setup
        this.interactive = false;

        // Elements
        // -- Position point element
        this._positionPointElement.showItem();
        // -- Line element
        this._lineElement.hideItem();

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.off("startMove", this._startMove, this);
        // TODO this.off("endMove", this._endMove, this);
        // TODO this.off("onRelease", this._onRelease, this);
    }

    /**
     * Mode draw item
     * @private
     */
    protected _modeDrawItem(): void {
    }

    /**
     * Remove mode draw item
     * @private
     */
    protected _removeModeDrawItem(): void {
        this.moveInteraction.reset();
    }

    /**
     * Mode draw select
     * @private
     */
    protected _modeDrawSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorGroup;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorGroup;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode draw select
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeDrawSelect(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorNormal;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorNormal;

        // Draw
        if (!mode.startsWith('delete')) {
            this.emit("debounceDraw");
        }
    }

    /**
     * Mode choose item
     * @private
     */
    protected _modeChooseItem(): void {
        // Interactions
        // -- Choose interaction
        this.chooseInteraction.on("chooseElement", this.chooseElement, this);
    }

    /**
     * Remove mode choose item
     * @private
     */
    protected _removeModeChooseItem(): void {
        // Interactions
        // -- Choose interaction
        this.chooseInteraction.off("chooseElement", this.chooseElement, this);
    }

    /**
     * Mode choose select
     * @private
     */
    protected _modeChooseSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.selftest.lineColorSelect;
        this._positionPointElement.options.fillColor = Design.selftest.fillColorSelect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose select
     * @private
     */
    protected _removeModeChooseSelect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorNormal;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorNormal;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose unselect
     * @private
     */
    protected _modeChooseUnselect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.selftest.lineColorUnselect;
        this._positionPointElement.options.fillColor = Design.selftest.fillColorUnselect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose unselect
     * @private
     */
    protected _removeModeChooseUnselect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorNormal;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorNormal;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode clone item
     * @private
     */
    protected _modeCloneItem(): void {
    }

    /**
     * Remove mode clone item
     * @private
     */
    protected _removeModeCloneItem(): void {
        this.moveInteraction.reset();
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        const image = this.image;

        if (image) {
            // Select layer group
            if (['delete_multiple', 'delete_draw', 'delete_clone'].indexOf(this.modeInteraction.lastMode) === -1) {
                this.controlUnselectItem(image);
            }

            // Delete from parent
            image.removeChild(this);

            // Emits
            image.deleteImageObject(this.model);
        }

        this.destroy();

        // Hide menu
        image.controlHideMenu();

        // Update control elements
        image.controlUpdateElements();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = Design.positionPoint.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = Design.positionPoint.zIndexDefault + this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            // Set mode
            this.modeInteraction.setMode("delete");
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.degroupLayerGroupImageObject(this.model);
        if (unselectElement) {
            this.image.unselectLayerGroup();
            this.image.selectImageObject(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        this.image.copyImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        this.image.clipboardImageObject(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Selftest">
    /**
     * On button set selftest
     * @param override
     */
    public onButtonSetSelftest(override?: boolean): void {
        this.model.isSelftest = override || !this.model.isSelftest;

        /* TODO no color defined
        this._lineColor = this.model.isSelftest ? this._lineColorSelftest : this._lineColorSelected;
        this._fillColor = this.model.isSelftest ? this._fillColorSelftest : this._fillColorSelected;
        this.draw();
        */

        if (this.model.isSelftest) {
            /* TODO not needed
            this._selftestContainerElement.addImageObject(this._selftestChooseButton);
            this._selftestContainerElement.draw();
            */
        } else {
            const imageObjects = this.image.getImageObjects();
            const items = this.model.selftestItems;
            this.model.selftestItems = []; // Remove existing items
            if (items) {
                items.forEach((item) => {
                    const element = imageObjects.find((io) => io.name === item) as MedsurfDraw.DrawableObject;
                    if (element) {
                        element.modeInteraction.setMode('blocked_item');
                    }
                });
            }
            // TODO not needed this._selftestContainerElement.removeImageObject(this._selftestChooseButton);
        }

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.emit("selftestChanged");

        // Update
        this.image.updateImageObject(this.model);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectImageObject(this.model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Control unselect item
     * @param image
     */
    public controlUnselectItem(image: MedsurfDraw.Image): void {
        // Set select layer
        if(!this.controlSelectLayerGroup(image)) {
            // Set select image
            image.externControlSelectImage();

            // Emits
            image.unselectImageObject();
        }
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.selftestItems && this.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else if (this.model.selftestItems && imageObject instanceof MedsurfDraw.LegendCollection) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => { return {name: lr.model.id, element: lr}; });
                    for (let legendRow of legendRows) {
                        if (this.model.selftestItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.modeInteraction.setMode('choose_select');
                        } else {
                            legendRow.element.modeInteraction.setMode('blocked_item');
                        }
                    }
                } else {
                    imageObject.modeInteraction.setMode('blocked_item');
                }
            });

        // Set children mode
        this.modeInteraction.setModeChildren('select_parent');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control select layer group
     * @param image
     */
    public controlSelectLayerGroup(image: MedsurfDraw.Image): boolean {
        // Get layer group
        const layerGroup = image.getLayerGroupByImageObject(this.model);
        if (layerGroup) {
            // Emits
            image.selectLayerGroup(layerGroup);

            // Select layer group
            return image.externControlSelectLayerGroup(layerGroup);
        }

        // Layer not selected
        return false;
    }

    /**
     * Control unselect layer group
     */
    public controlUnselectLayerGroup(): void {
        // Emits
        this.image.unselectLayerGroup();
    }

    /**
     * Control draw item
     * @param isNew
     */
    public controlDrawItem(isNew: boolean): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this && !imageObject.modeInteraction.lastMode.startsWith('draw'))
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('draw_parent');

        // Set draw item
        this.modeInteraction.setModeItem(isNew ? 'draw_item_new' : 'draw_item');
    }

    /**
     * Control choose item
     */
    public controlChooseItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                if (this.model.selftestItems && this.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else if (this.model.selftestItems && imageObject instanceof MedsurfDraw.LegendCollection) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => { return {name: lr.model.id, element: lr}; });
                    for (let legendRow of legendRows) {
                        if (this.model.selftestItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.modeInteraction.setMode('choose_select');
                        } else {
                            legendRow.element.modeInteraction.setMode('choose_unselect');
                        }
                    }
                } else {
                    imageObject.modeInteraction.setMode('choose_unselect');
                }
            });

        // Set children mode
        this.modeInteraction.setModeChildren('choose_parent');

        // Set draw item
        this.modeInteraction.setModeItem('choose_item');
    }

    /**
     * Control clone item
     */
    public controlCloneItem(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this && !imageObject.modeInteraction.lastMode.startsWith('clone'))
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode('blocked_item');
            });

        // Set children mode
        this.modeInteraction.setModeChildren('clone_parent');

        // Set draw item
        this.modeInteraction.setModeItem('clone_item_new');
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * Start move
     * @private
     */
    public startMove(): void {
        // TODO not needed const radius = this._positionPointElement.circle.radius + MedsurfDraw.MOVE_HITTEST_EXTENSION;

        // TODO better implementation: this._configurationElement.hideItem();
        // TODO this.removeChild(this._configurationElement);

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     * @private
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number) {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Position
        this.position.x += dX;
        this.position.y += dY;
        this.model.position.x = this.position.x / imageWidth;
        this.model.position.y = this.position.y / imageWidth;

        /* TODO selftest indicator
        // Elements
        // -- Killbox element
        const rectangle = this.getRectangle();
        rectangle.pad(Design.selftestElement.gab);
        this._killboxElement.rectangle = rectangle;
        this._killboxElement.draw();
        */

        // Emit on position point move
        this.emit("onPositionPointMove");
    }

    /**
     * End move
     * @param event
     * @private
     */
    public endMove(event: PIXI.InteractionEvent): void {
        // TODO not needed const radius = this._positionPointElement.circle.radius - MedsurfDraw.MOVE_HITTEST_EXTENSION;

        // TODO better implementation: this._configurationElement.showItem();
        // TODO this.addChild(this._configurationElement);

        // Events
        // TODO remove this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);

        // Emits
        this.emit("pointerover", event);
        this.image.updateImageObject(this.model);

        // Emit end position point move
        this.emit("endPositionPointMove");
    }

    /**
     * On release
     */
    public onRelease(): void {
        this.modeInteraction.setMode(this.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorHover;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorHover;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Element
        // -- Position point element
        this._positionPointElement.options.lineColor = Design.positionPoint.lineColorNormal;
        this._positionPointElement.options.fillColor = Design.positionPoint.fillColorNormal;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_layer_group') {
            this.controlSelectItem();
        } else {
            if (!this.controlSelectLayerGroup(this.image) && !preventSelectItem) {
                this.controlSelectItem();
            }
        }
    }

    /**
     * On selftest
     * @private
     */
    public onSelftest(): void {
        // Set default mode for image children
        this.image.getImageObjects()
            .forEach((imageObject) => {
                if (imageObject === this
                    || (this.model.selftestItems
                        && this.model.selftestItems.indexOf(imageObject.name as string) > -1)) {
                    if (imageObject instanceof MedsurfDraw.PositionPoint) {
                        imageObject.children.forEach((child) => {
                            if (child instanceof MedsurfDraw.BackgroundRectanglePrimitive) {
                                child.showItem();
                                // @ts-ignore
                                child._rectangleElement.showItem();
                            }
                            else if (child instanceof MedsurfDraw.ArrowPrimitive) {
                                child.showItem();
                                // @ts-ignore
                                child._arrowElement.showItem();
                                // @ts-ignore
                                child._selftestElement.hideItem();
                            } else if (child instanceof MedsurfDraw.EllipsePrimitive) {
                                child.showItem();
                                // @ts-ignore
                                child._ellipseElement.showItem();
                                // @ts-ignore
                                child._selftestElement.hideItem();
                            } else if (child instanceof MedsurfDraw.RectanglePrimitive) {
                                child.showItem();
                                // @ts-ignore
                                child._rectangleElement.showItem();
                                // @ts-ignore
                                child._selftestElement.hideItem();
                            } else if (child instanceof MedsurfDraw.TextPrimitive) {
                                child.showItem();
                                // @ts-ignore
                                child._textElement.showItem();
                                // @ts-ignore
                                child._selftestElement.hideItem();
                            }
                        });
                    }
                    imageObject.showItem();
                } else if (imageObject instanceof MedsurfDraw.LegendCollection && this.model.selftestItems) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => {
                            return {name: lr.model.id, element: lr};
                        });
                    for (let legendRow of legendRows) {
                        if (this.model.selftestItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.showItem();
                        }
                    }
                } else {
                    imageObject.hideItem();
                }
            });

        // Get image
        const image = this.image;

        // Release event
        image.parent.once('pointerup', this.onSelftestRelease, this);
    }

    /**
     * On selftest release
     * @param event
     */
    public onSelftestRelease(event: PIXI.InteractionEvent): void {
        // Reset select interaction
        this.selectInteraction.onRelease(event);

        // Set default mode for image children
        const defaultMode = this.image.modeInteraction.defaultMode;
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== this)
            .forEach((imageObject) => {
                imageObject.modeInteraction.setMode(defaultMode);
            });

        /* TODO
        setTimeout(() => {
            // Elements
            // -- Position point element
            this._positionPointElement.circle.radius = Design.positionPoint.radiusSelected;

            // Draw
            this.draw();
        }, 2000);
        */

        // Get image
        const image = this.image;

        // Prevent Image move
        image.moveInteraction.endMove(new PIXI.InteractionEvent());
    }
    //</editor-fold>

    //<editor-fold desc="Draw">
    /**
     * Start draw
     */
    public startDraw(): void {
        this.cursor = "draw";
    }

    /**
     * Start draw scale
     */
    public startDrawScale(): void {
        this.cursor = "scale-nwse";
    }

    /**
     * On draw
     * @param event
     */
    public onDraw(event: PIXI.InteractionEvent): void {
        // TODO this.targetImageObject.emit("onDraw", event);
    }
    //</editor-fold>

    //<editor-fold desc="Choose">
    /**
     * Choose element
     * @param event
     * @param element
     */
    public chooseElement(event: PIXI.InteractionEvent, element: MedsurfDraw.DrawableObject): void {
        if (!this.model.selftestItems) {
            this.model.selftestItems = [];
        }

        const index = this.model.selftestItems.findIndex((item: string) => item === element.name);
        if (index > -1) {
            // Remove from selftest items
            this.model.selftestItems.splice(index, 1);

            // Set element mode unselect
            element.modeInteraction.setMode('choose_unselect');
        } else {
            // Add to selftest items
            this.model.selftestItems.push(element.name as string);

            // Set element mode select
            element.modeInteraction.setMode('choose_select');
        }
    }
    //</editor-fold>

    //<editor-fold desc="Delete">
    /**
     * Delete Element
     * @param image
     * @protected
     */
    protected _deleteItem(image: MedsurfDraw.Image): void {
        // Check for primitives
        if ((this.model.primitives || []).length > 0) {
            return;
        }

        // Check for line
        if (image.getLinesByPositionPoint(this).length > 0) {
            return;
        }

        this.modeInteraction.setMode("delete");
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        // TODO
        // if (this.modeInteraction.lastMode === 'select' || this.modeInteraction.lastMode === 'selftest') {
            this.emit("debounceDraw");
        // }
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get element rectangle
     */
    public getElementRectangle(): PIXI.Rectangle {
        // Elements
        let scale = this.imageScale.x * Design.positionPoint.scaleAspect;
        let radius = Design.positionPoint.radius;
        // TODO if (this.modeInteraction.lastMode === 'select' || this.modeInteraction.lastMode === 'selftest') {
        radius = Math.floor(radius / scale);
        /* TODO } else if (scale < 1) {
            scale = 1;
        }*/
        return new PIXI.Rectangle(this.position.x - radius, this.position.y - radius, radius * 2, radius * 2);
    }

    /**
     * Show item
     */
    /* TODO selftest indicator
    public showItem(): void {
        this.visible = true;

        // Elements
        // -- Killbox elemnt
        if (this.model.isSelftest && this._killboxElement) {
            this._killboxElement.visible = true;
        }
    }
    */

    /**
     * Hide item
     */
    /* TODO selftest indicator
    public hideItem(): void {
        this.visible = false;

        // Elements
        // -- Killbox elemnt
        if (this._killboxElement) {
            this._killboxElement.visible = false;
        }
    }
    */
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<PositionPoint> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<PositionPoint>) {
        this._moveInteraction = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<PositionPoint> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<PositionPoint>) {
        this._selectInteraction = value;
    }

    /**
     * Get choose interaction
     */
    public get chooseInteraction(): MedsurfDraw.ChooseInteraction<PositionPoint> {
        return this._chooseInteraction;
    }

    /**
     * Set choose interaction
     * @param value
     */
    public set chooseInteraction(value: MedsurfDraw.ChooseInteraction<PositionPoint>) {
        this._chooseInteraction = value;
    }
    //</editor-fold>
}