import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Selftest point model
 */
export class SelftestPointModel extends BaseContainerImageModel {
    isSelftest: boolean;
    selftestPoint: Models.Point;
    options: Models.ElementOptions = {hasLine: true, lineColor: 0xFFFFFF, lineAlpha: 1, hasFill: false};
}

/**
 * Selftest point
 */
export class SelftestPoint extends BaseContainer<SelftestPointModel> {
    /**
     * Interactions
     */
    private _selectInteraction: MedsurfDraw.SelectInteraction<SelftestPoint>;

    /**
     * Elements
     */
    protected _selftestPointElement: MedsurfDraw.Circle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: SelftestPointModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.wedgePoint.zIndex;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelftest", this.onSelftest, this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Move
            this.image.moveInteraction.on("onMove", this.onMove, this);
            // -- Zoom
            this.image.zoomInteraction.on("onZoom", this.onZoom, this);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Move
            this.image.moveInteraction.off("onMove", this.onMove, this);
            // -- Zoom
            this.image.zoomInteraction.off("onZoom", this.onZoom, this);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Position
        this.position.set(
            this.image.position.x + this.selftestPoint.x * this.imageScale.x,
            this.image.position.y + this.selftestPoint.y * this.imageScale.y
        );

        //<editor-fold desc="Elements">
        // -- Selftest point element
        this._selftestPointElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.selftest.selftestRadius),
            options: Object.assign({}, this.options),
            lineWidth: Design.selftest.lineWidth
        });
        this.addChild(this._selftestPointElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Selftest point element
        this._selftestPointElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Selftest point element
        if (this._selftestPointElement) {
            this._selftestPointElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        // Hide item
        this.hideItem();
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        if (this.data.isSelftest) {
            // Setup
            this.interactive = true;
            this.cursor = "select";

            // Events
            this.selectInteraction.reset();
            this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
            this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
            this.on("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);

            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode selftest
     * @param parent
     * @param mode
     * @protected
     */
    protected _removeModeSelftest(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);


        if (mode !== 'selftest') {
            //<editor-fold desc="Elements">
            // -- Selftest ppint element
            this._selftestPointElement.circle.radius = Design.selftest.selftestRadius;
            //</editor-fold>
        }

        // Hide item
        this.hideItem();
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * On move
     */
    public onMove(): void {
        if (!this.selftestPoint) {
            return;
        }

        // Position
        this.position.set(
            this.image.position.x + this.selftestPoint.x * this.imageScale.x,
            this.image.position.y + this.selftestPoint.y * this.imageScale.y
        );
    }

    /**
     * On selftest
     * @private
     */
    public onSelftest(): void {
        //<editor-fold desc="Elements">
        // -- Selftest ppint element
        this._selftestPointElement.circle.radius = Design.selftest.selftestRadiusSelected;
        //</editor-fold>

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On zoom
     */
    public onZoom(): void {
        if (!this.selftestPoint) {
            return;
        }

        // Position
        this.position.set(
            this.image.position.x + this.selftestPoint.x * this.imageScale.x,
            this.image.position.y + this.selftestPoint.y * this.imageScale.y
        );
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Setup
        this._selftestPointElement.options.lineColor = Design.primitive.lineColorHover;
        this._selftestPointElement.options.fillColor = Design.primitive.fillColorHover;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Setup
        this._selftestPointElement.options = Object.assign({}, this.options);

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get isSelftest
     */
    public get isSelftest(): boolean {
        return this.data.isSelftest;
    }

    /**
     * Set isSelftest
     * @param value
     */
    public set isSelftest(value: boolean) {
        this.data.isSelftest = value;
    }

    /**
     * Get selftest point
     */
    public get selftestPoint(): Models.Point {
        return this.data.selftestPoint;
    }

    /**
     * Set selftest point
     * @param value
     */
    public set selftestPoint(value: Models.Point) {
        this.data.selftestPoint = value;

        // Position
        this.position.set(
            this.image.position.x + this.selftestPoint.x * this.imageScale.x,
            this.image.position.y + this.selftestPoint.y * this.imageScale.y
        );
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;

        // Elements
        this._selftestPointElement.options = Object.assign({}, this.data.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<SelftestPoint> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<SelftestPoint>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}