import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Wedge point model
 */
export class WedgePointModel extends BaseContainerImageModel {
    wedgePoint: Models.Point;
}

/**
 * Wedge point
 */
export class WedgePoint extends BaseContainer<WedgePointModel> {
    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<WedgePoint>;

    /**
     * Elements
     */
    protected _wedgePointElement: MedsurfDraw.Circle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: WedgePointModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.wedgePoint.zIndex;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeDefault, this);
        this.modeInteraction.on("remove-marker", this._removeModeDefault, this);
        this.modeInteraction.on("selftest", this._modeDefault, this);
        this.modeInteraction.on("remove-selftest", this._removeModeDefault, this);
        this.modeInteraction.on("author", this._modeDefault, this);
        this.modeInteraction.on("remove-author", this._removeModeDefault, this);
        this.modeInteraction.on('select_parent', this._modeSelectParent, this);
        this.modeInteraction.on("remove-select_parent", this._removeModeSelectParent, this);
        this.modeInteraction.on('select_legend', this._modeSelectParent, this);
        this.modeInteraction.on("remove-select_legend", this._removeModeSelectParent, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Position
        this.position.set(this.wedgePoint.x, this.wedgePoint.y);

        //<editor-fold desc="Elements">
        // -- Wedge point element
        this._wedgePointElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.wedgePoint.radius),
            options: {hasLine: true, lineColor: Design.wedgePoint.lineColorNormal, lineAlpha: Design.wedgePoint.lineAlpha, hasFill: true, fillColor: Design.wedgePoint.fillColorNormal, fillAlpha: Design.wedgePoint.fillAlpha},
            lineWidth: Design.wedgePoint.lineWidth
        });
        this.addChild(this._wedgePointElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // Position
        this.position.set(this.wedgePoint.x, this.wedgePoint.y);

        // Elements
        let scale = this.imageScale.x * Design.wedgePoint.scaleAspect;
        let radius = Design.wedgePoint.radius;
        // TODO if (this.modeInteraction.lastMode === 'select' || this.modeInteraction.lastMode === 'selftest') {
        radius = Math.floor(radius / scale);
        /* TODO } else if (scale < 1) {
            scale = 1;
        }*/
        let lineWidth = Math.floor(Design.wedgePoint.lineWidth || 1 / scale);
        if (lineWidth < 1) {
            lineWidth = 1;
        }

        //<editor-fold desc="Elements">
        // -- Wedge point element
        this._wedgePointElement.circle.radius = radius;
        this._wedgePointElement.lineWidth = lineWidth;
        this._wedgePointElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        //</editor-fold>

        // Elements
        // -- Wedge point element
        if (this._wedgePointElement) {
            this._wedgePointElement.destroy(options);
        }

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        // Hide item
        this.hideItem();
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode select parent
     * @private
     */
    protected _modeSelectParent(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.onHover, this);
        this.on("pointerout", this.onOut, this);
        this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO check this: this.on("mouseupoutside", this.moveInteraction.endMove, this.moveInteraction);

        // Show item
        this.showItem();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode select parent
     * @private
     */
    protected _removeModeSelectParent(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.onHover, this);
        this.off("pointerout", this.onOut, this);
        this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO check this: this.off("mouseupoutside", this.moveInteraction.endMove, this.moveInteraction);

        // Hide item
        this.hideItem();
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Element
        // -- Wedge point element
        this._wedgePointElement.options.lineColor = Design.wedgePoint.lineColorHover;
        this._wedgePointElement.options.fillColor = Design.wedgePoint.fillColorHover;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Element
        // -- Wedge point element
        this._wedgePointElement.options.lineColor = Design.wedgePoint.lineColorNormal;
        this._wedgePointElement.options.fillColor = Design.wedgePoint.fillColorNormal;

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get wedge point
     */
    public get wedgePoint(): Models.Point {
        return this.data.wedgePoint;
    }

    /**
     * Set wedge point
     * @param value
     */
    public set wedgePoint(value: Models.Point) {
        this.data.wedgePoint = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<WedgePoint> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<WedgePoint>) {
        this._moveInteraction = value;
    }
    //</editor-fold>
}