import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {
    BaseElementContainer,
    BaseElementContainerModel,
    ColorMode,
    NumberMode
} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";

/**
 * Arrow primitive
 */
export class ArrowPrimitive extends BaseElementContainer<Models.ArrowPrimitive, BaseElementContainerModel<Models.ArrowPrimitive>> {
    /**
     * Members
     */
    protected _wedgePoint: Models.Point | null;
    protected _selftestSelected = false;

    /**
     * Interactions
     */
    private _moveInteraction: MedsurfDraw.MoveInteraction<ArrowPrimitive>;
    private _scaleInteraction: MedsurfDraw.ScaleInteraction<ArrowPrimitive>;
    private _selectInteraction: MedsurfDraw.SelectInteraction<ArrowPrimitive>;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: any;
    protected _debounceControlUnselectItemMethod: any;

    /**
     * Elements
     */
    protected _arrowElement: MedsurfDraw.Arrow;
    protected _selftestElement: MedsurfDraw.SelftestPoint;
    // protected _wedgePointElement: MedsurfDraw.WedgePoint;

    /**
     * Menu elements
     */
    // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Mirror
    protected _horizontalMirrorButtonElement: MedsurfDraw.RoundButtonElement;
    protected _verticalMirrorButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, model: Models.ArrowPrimitive) {
        // Parent constructor
        super({image, model});

        // Setup
        this.zIndex = Design.arrowPrimitive.zIndexDefault + this.model.zIndex;
        this.rotation = this.model.angle || 0;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on('select_layer_group', this._modeSelectLayerGroup, this);
        this.modeInteraction.on('remove-select_layer_group', this._removeModeSelectLayerGroup, this);
        this.modeInteraction.on('select_grouping', this._modeSelectGrouping, this);
        this.modeInteraction.on('remove-select_grouping', this._removeModeSelectGrouping, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('select_item_color_line', this._modeSelectItemColor, this);
        this.modeInteraction.on('remove-select_item_color_line', this._removeModeSelectItemColorLine, this);
        this.modeInteraction.on('select_item_color_fill', this._modeSelectItemColor, this);
        this.modeInteraction.on("remove-select_item_color_fill", this._removeModeSelectItemColorFill, this);
        this.modeInteraction.on('select_item_number_line_width', this._modeSelectItemNumber, this);
        this.modeInteraction.on("remove-select_item_number_line_width", this._removeModeSelectItemNumber, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('draw_item', this._modeDrawItem, this);
        this.modeInteraction.on("remove-draw_item", this._removeModeDrawItem, this);
        this.modeInteraction.on('draw_clickable', this._modeDrawSelect, this);
        this.modeInteraction.on('remove-draw_clickable', this._removeModeDrawSelect, this);
        this.modeInteraction.on('choose_select', this._modeChooseSelect, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeChooseSelect, this);
        this.modeInteraction.on('choose_unselect', this._modeChooseUnselect, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeChooseUnselect, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        this.modeInteraction.on("delete_draw", this._modeDelete, this);
        // -- Move interaction
        this.moveInteraction = new MedsurfDraw.MoveInteraction(this);
        this.moveInteraction.on("startMove", this.startMove, this);
        this.moveInteraction.on("onMove", this.onMove, this);
        this.moveInteraction.on("endMove", this.endMove, this);
        // -- Scale interaction
        this.scaleInteraction = new MedsurfDraw.ScaleInteraction(this, true);
        this.scaleInteraction.on("onScale", this.onScale, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelectStart", this.onSelectStart, this);
        this.selectInteraction.on("onSelftest", this.onSelftest, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        this._debounceControlUnselectItemMethod = debounce(this.controlUnselectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", (parent: MedsurfDraw.PositionPoint) => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.on("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
            // -- Move
            parent.on("endPositionPointMove", this.endPositionPointMove, this);
            // -- Selftest
            parent.on("selftestChanged", this.selftestChanged, this);
            this._selftestElement.selectInteraction.on("onSelftest", this.onSelftest, this);
        });
        // -- Removed
        this.on("removed", (parent: MedsurfDraw.PositionPoint) => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.off("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
            // -- Move
            parent.off("endPositionPointMove", this.endPositionPointMove, this);
            // -- Selftest
            parent.off("selftestChanged", this.selftestChanged, this);
            this._selftestElement.selectInteraction.off("onSelftest", this.onSelftest, this);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Get wedge point
        this.wedgePoint = this.model.wedgePoint
            ? {x: this.model.wedgePoint.x * imageWidth, y: this.model.wedgePoint.y * imageWidth}
            : {x: this.model.rectangle.width * imageWidth, y: this.model.rectangle.height / 2 * imageWidth};

        // Position
        this.position.set(
            this.model.rectangle.x * imageWidth,
            this.model.rectangle.y * imageWidth
        );

        //<editor-fold desc="Elements">
        // -- Arrow element
        this._arrowElement = new MedsurfDraw.Arrow({
            rectangle: new PIXI.Rectangle(0, 0, this.model.rectangle.width * imageWidth, this.model.rectangle.height * imageWidth),
            wedgePoint: this.wedgePoint,
            options: Object.assign({}, this.model.options),
            lineWidth: this.model.strokeWidth ? Math.round(this.model.strokeWidth * imageWidth * 100) / 100 : undefined
        });
        this.addChild(this._arrowElement);
        // -- Selftest element
        this._selftestElement = new MedsurfDraw.SelftestPoint({
            image: this.image,
            isSelftest: (this.parent as MedsurfDraw.PositionPoint).model.isSelftest,
            selftestPoint: new PIXI.Point(
                this.parent.x + this.model.rectangle.x * imageWidth,
                this.parent.y + this.model.rectangle.y * imageWidth
            ),
            options: Object.assign({}, this.model.options),
        });
        this.image.parent.addChild(this._selftestElement);
        // -- Wedge point element
        /* TODO remove
        this._wedgePointElement = new MedsurfDraw.WedgePoint({
            wedgePoint: wedgePoint
        });
        this._wedgePointElement.moveInteraction.on("onMove", this._onWedgeMove, this);
        this._wedgePointElement.moveInteraction.on("endMove", this._onWedgeEndMove, this);
        this.addChild(this._wedgePointElement);
        */
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        this._moveToFrontButtonElement.on("toggleMenu", this.onMenuLayer, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Mirror">
        // -- Horizontal mirror button element
        this._horizontalMirrorButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf89c',
            description: 'Horizontal\nspiegeln',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._horizontalMirrorButtonElement.on("button", this.onButtonHorizontalMirror, this);
        // -- Vertical mirror button element
        this._verticalMirrorButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf89d',
            description: 'Vertikal\nspiegeln',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._horizontalMirrorButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._verticalMirrorButtonElement.on("onHover", this._verticalMirrorButtonElement.showMenu, this._verticalMirrorButtonElement);
        this._verticalMirrorButtonElement.on("button", this.onButtonVerticalMirror, this);
        this._verticalMirrorButtonElement.on("toggleMenu", this.onMenuMirror, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._verticalMirrorButtonElement,
            order: 3,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>

        // Rotate with line
        if (this.model.rotateWithLine) {
            this.parent.on("onRotateWithLine", this.onRotateWithLine, this);
            this.onRotateWithLine(new PIXI.InteractionEvent());
        }

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Arrow element
        this._arrowElement.draw();
        this._arrowElement.pivot.set(this._arrowElement.rectangle.width / 2, this._arrowElement.rectangle.height / 2);
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // Events
        if (this.parent) {
            this.parent.off("onRotateWithLine", this.onRotateWithLine, this);
        }

        //<editor-fold desc="Interactions">
        // -- Move interaction
        if (this.moveInteraction) {
            this.moveInteraction.removeAllListeners();
            // TODO remove delete this.moveInteraction;
        }
        // -- Scale interaction
        if (this.scaleInteraction) {
            this.scaleInteraction.removeAllListeners();
            // TODO remove delete this._scaleInteraction;
        }
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Arrow element
        if (this._arrowElement) {
            this._arrowElement.destroy(options);
        }
        // -- Selftest element
        if (this._selftestElement) {
            // this.image.parent.removeChild(this._selftestElement);
            this._selftestElement.destroy(options);
        }
        // -- Wedge point element
        /* TODO remove
        if (this._wedgePointElement) {
            this._wedgePointElement.moveInteraction.off("onMove", this._onWedgeMove, this);
            this._wedgePointElement.moveInteraction.off("endMove", this._onWedgeEndMove, this);
            this._wedgePointElement.destroy(options);
        }
        */
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Mirror">
        // -- Horizontal mirror button element
        if (this._horizontalMirrorButtonElement) {
            this._horizontalMirrorButtonElement.destroy(options);
        }
        // -- Vertical mirror button element
        if (this._verticalMirrorButtonElement) {
            this._verticalMirrorButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param hasWedgePoint
     * @param rotateWithLine
     * @param imageWidth
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, hasWedgePoint: boolean,
                              rotateWithLine: boolean, imageWidth: number): ArrowPrimitive {
        // Element
        // -- Arrow primitive
        return new ArrowPrimitive(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.ARROW,
            isStatic: false,
            zIndex: 0,
            rotateWithLine: rotateWithLine,
            wedgePoint: (hasWedgePoint ? {x: 34 / imageWidth, y: 20 / imageWidth} : undefined),
            rectangle: {
                x: 0,
                y: 0,
                width: (hasWedgePoint ? 50 : 20) / imageWidth,
                height: (hasWedgePoint ? 40 : 20) / imageWidth
            },
            strokeWidth: 1 / imageWidth,
            dashFill: 2,
            dashEmpty: 2,
            options: {
                hasLine: true,
                lineColor: 0xffffff,
                lineAlpha: 1,
                hasFill: true,
                fillColor: 0xffffff,
                fillAlpha: 0.3,
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        // Setup
        this._selftestSelected = false;

        if (this.model.isStatic) {
            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        // Setup
        this._selftestSelected = false;

        // Show item
        this.showItem();
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        // Get position point
        const positionPoint = this.parent as MedsurfDraw.PositionPoint;

        if (this.model.isStatic) {
            // Setup
            this.interactive = true;
            this.cursor = "select";

            // Events
            /* TODO remove
            this.selectInteraction.reset();
            this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
            this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
            this.on("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
            */

            // Show item
            this.showItem();

            // Draw
            this.emit("debounceDraw");
        } else if (positionPoint.model.isSelftest) {
            // Setup
            this.interactive = true;
            this.cursor = "select";

            // Elements
            // -- Arrow element
            this._arrowElement.hideItem();

            // Events
            /* TODO remove
            this.selectInteraction.reset();
            this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
            this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
            this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
            this.on("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
            */

            // Show item
            this.showItem();

            // Draw
            this.emit("debounceDraw");
        } else {
            // Hide Item
            this.hideItem();
        }
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Setupthis
        this.interactive = false;

        // Elements
        // -- Arrow element
        this._arrowElement.showItem();

        // Events
        /* TODO remove
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelftest, this.selectInteraction);
        */
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        this.interactive = true;
        this._selftestSelected = false;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select layer group
     * @private
     */
    protected _modeSelectLayerGroup(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select layer group
     * @private
     */
    protected _removeModeSelectLayerGroup(): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select grouping
     * @private
     */
    protected _modeSelectGrouping(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select grouping
     * @private
     */
    protected _removeModeSelectGrouping(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mousedown", this.selectInteraction.onPointerDown, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Elements
        // -- Wedge point element
        /* TODO remove
        this._wedgePointElement.interactive = true
        */

        // Events
        this.on("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.on("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.on("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.on("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.on("startMove", this._startMove, this);
        // TODO this.on("endMove", this._endMove, this);
        // TODO this.on("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.on("down_ArrowUp", "onPositionPointMoveUp", this.moveInteraction.moveUp.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowRight", "onPositionPointMoveRight", this.moveInteraction.moveRight.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowDown", "onPositionPointMoveDown", this.moveInteraction.moveDown.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("down_ArrowLeft", "onPositionPointMoveLeft", this.moveInteraction.moveLeft.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowUp", "onPositionPointMoveUpEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowRight", "onPositionPointMoveRightEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowDown", "onPositionPointMoveDownEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));
        MedsurfDraw.Keyboard.events.on("released_ArrowLeft", "onPositionPointMoveLeftEnd", this.moveInteraction.moveRelease.bind(this.moveInteraction));

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        // TODO remove this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        // TODO remove this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("pointerover", this.moveInteraction.onMoveHover, this.moveInteraction);
        this.off("mousedown", this.moveInteraction.startMove, this.moveInteraction);
        this.off("pointermove", this.moveInteraction.onMove, this.moveInteraction);
        this.off("mouseup", this.moveInteraction.endMove, this.moveInteraction);
        // TODO this.off("startMove", this._startMove, this);
        // TODO this.off("endMove", this._endMove, this);
        // TODO this.off("onRelease", this._onRelease, this);

        MedsurfDraw.Keyboard.events.remove("down_ArrowUp", "onPositionPointMoveUp");
        MedsurfDraw.Keyboard.events.remove("down_ArrowRight", "onPositionPointMoveRight");
        MedsurfDraw.Keyboard.events.remove("down_ArrowDown", "onPositionPointMoveDown");
        MedsurfDraw.Keyboard.events.remove("down_ArrowLeft", "onPositionPointMoveLeft");
        MedsurfDraw.Keyboard.events.remove("released_ArrowUp", "onPositionPointMoveUpEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowRight", "onPositionPointMoveRightEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowDown", "onPositionPointMoveDownEnd");
        MedsurfDraw.Keyboard.events.remove("released_ArrowLeft", "onPositionPointMoveLeftEnd");

        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Set select layer
        // TODO remove console.log("arrow _removeModeSelectItem", mode);
        if (mode === 'author') {
            this.emit("debounceControlUnselectItem", this.parent);
        }
        /* TODO remove
        if (!mode.startsWith('select_parent') && mode !== 'select_child' && !mode.startsWith('select_item')
            && !mode.startsWith('choose_select') && !mode.startsWith('choose_parent') && !mode.startsWith('blocked_item')
            && ['default', 'marker', 'selftest'/*, 'author'* /, 'delete'].indexOf(mode) === -1) {
            // Unselect item
            this.controlUnselectItem();
            // this.emit("debounceControlUnselectItem");
        }
        */
    }

    /**
     * Mode select item color
     * @private
     */
    protected _modeSelectItemColor(): void {
        this.onOut();

        // Control Update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item color line
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemColorLine(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Control save color line
        this.image.controlSaveColor(this.model.options.lineColor, this.model.options.lineAlpha);

        // Remove mode select item color
        this._removeModeSelectItemColor(parent, mode);
    }

    /**
     * Remove mode select item color fill
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemColorFill(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Control save color fill
        this.image.controlSaveColor(this.model.options.fillColor, this.model.options.fillAlpha);

        // Remove mode select item color
        this._removeModeSelectItemColor(parent, mode);
    }

    /**
     * Remove mode select item color
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemColor(parent: MedsurfDraw.ImageObject, mode: string): void {
        if (mode !== 'select_child' && !mode.startsWith('select_item_color')) {
            // Set select item
            this.emit("debounceControlSelectItem");
        }
    }

    /**
     * Mode select item number
     * @private
     */
    protected _modeSelectItemNumber(): void {
        this.onOut();

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item number
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemNumber(parent: MedsurfDraw.ImageObject, mode: string): void {
        if (mode !== 'select_child' && !mode.startsWith('select_item_number')) {
            // Set select item
            this.emit("debounceControlSelectItem");
        }
    }

    /**
     * Mode blocked item
     * @private
     */
    protected _modeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        // Filters
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
        // Element
        // -- Arrow element
        this._arrowElement.options.lineAlpha = Design.blocked.lineAlpha;
        this._arrowElement.options.fillAlpha = Design.blocked.fillAlpha;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode blocked item
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeBlockedItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        /* TODO Not working for canvas renderer
        // Filters
        this.filters = [];
        */
        // Element
        // -- Arrow element
        this._arrowElement.options = Object.assign({}, this.model.options);

        // Draw
        if (!mode.startsWith('delete')) {
            this.emit("debounceDraw");
        }
    }

    /**
     * Mode draw item
     * @private
     */
    protected _modeDrawItem(): void {
    }

    /**
     * Remove mode draw item
     * @private
     */
    protected _removeModeDrawItem(): void {
        this.scaleInteraction.reset();
    }

    /**
     * Mode draw select
     * @private
     */
    protected _modeDrawSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Arrow element
        this._arrowElement.options.lineColor = Design.positionPoint.lineColorGroup;
        this._arrowElement.options.fillColor = Design.positionPoint.fillColorGroup;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode draw select
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeDrawSelect(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Arrow element
        this._arrowElement.options = Object.assign({}, this.model.options);

        // Draw
        if (!mode.startsWith('delete')) {
            this.emit("debounceDraw");
        }
    }

    /**
     * Mode choose select
     * @private
     */
    protected _modeChooseSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Arrow element
        this._arrowElement.options.lineColor = Design.selftest.lineColorSelect;
        this._arrowElement.options.fillColor = Design.selftest.fillColorSelect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose select
     * @private
     */
    protected _removeModeChooseSelect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Arrow element
        this._arrowElement.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose unselect
     * @private
     */
    protected _modeChooseUnselect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Arrow element
        this._arrowElement.options.lineColor = Design.selftest.lineColorUnselect;
        this._arrowElement.options.fillColor = Design.selftest.fillColorUnselect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose unselect
     * @private
     */
    protected _removeModeChooseUnselect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Arrow element
        this._arrowElement.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        // Get parent
        const parent = this.parent as MedsurfDraw.PositionPoint;

        // Setup parent
        if (parent) {
            // Get image
            const image = parent.image;

            if (parent.model.primitives) {
                let index = parent.model.primitives.findIndex((primitive: any) => primitive.id === this.model.id);
                if (index > -1) {
                    // Select layer group
                    if (['delete_draw'].indexOf(this.modeInteraction.lastMode) === -1) {
                        this.emit("debounceControlUnselectItem", parent);
                    }

                    // Delete from parent
                    parent.model.primitives.splice(index, 1);
                    parent.removeChild(this);

                    // Check if position point has/contains in any other elements
                    parent.emit("debounceDeleteItem", image);

                    // Emits
                    image.deletePrimitive(this.model);
                }
            }

            // Hide menu
            image.controlHideMenu();

            // Control update elements
            image.controlUpdateElements();
        }

        // Destroy
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
        this._verticalMirrorButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editPrimitive(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On menu layer
     * @param event
     * @param visible
     */
    public onMenuLayer(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._verticalMirrorButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            // Set mode
            this.modeInteraction.setMode("delete");
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Degroup image object
        (this.parent as MedsurfDraw.PositionPoint).onButtonDegroup(false);

        // Emit
        if (unselectElement) {
            this.image.unselectLayerGroup();
            this.image.selectPrimitive(this.model, (this.parent as MedsurfDraw.PositionPoint).model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Mirror">
    /**
     * On button vertical mirror
     */
    public onButtonVerticalMirror(): void {
        // Prevent rotate with line
        if (this.model.rotateWithLine) {
            return;
        }

        // Get rotation
        let rotation = (Math.round(((this.model.angle || 0) / Math.PI * 180) * 100) / 100) % 360;
        if (rotation < 0) {
            rotation = 360 + rotation;
        }

        // Model
        this.rotation = Math.round(180 - rotation) % 360 * (2 * Math.PI / 360);
        this.model.angle = this.rotation;

        // Control update elements
        this.image.controlUpdateElements();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On button horizontal mirror
     */
    public onButtonHorizontalMirror(): void {
        // Prevent rotate with line
        if (this.model.rotateWithLine) {
            return;
        }

        // Get rotation
        let rotation = (Math.round(((this.model.angle || 0) / Math.PI * 180) * 100) / 100) % 360;
        if (rotation < 0) {
            rotation = 360 + rotation;
        }

        // Model
        this.rotation = Math.round(360 - rotation) % 360 * (2 * Math.PI / 360);
        this.model.angle = this.rotation;

        // Control update elements
        this.image.controlUpdateElements();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On menu mirror
     * @param event
     * @param visible
     */
    public onMenuMirror(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._moveToFrontButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        (this.parent as MedsurfDraw.PositionPoint).onButtonDuplicate();
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        (this.parent as MedsurfDraw.PositionPoint).onButtonClipboard();
    }
    //</editor-fold>
    //<editor-fold desc="Color">
    /**
     * On color
     * @param mode
     * @param has
     * @param color
     * @param colorAlpha
     */
    public onColor(mode: ColorMode, has: boolean, color: number | undefined, colorAlpha: number | undefined): void {
        if (mode === ColorMode.LINE) {
            // Update model
            this.model.options.hasLine = has;
            this.model.options.lineColor = color;
            this.model.options.lineAlpha = colorAlpha;

            // Elements
            // -- Arrow element
            this._arrowElement.options = Object.assign({}, this.model.options);
            // -- Selftest element
            this._selftestElement.options = Object.assign({}, this.model.options);

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updatePrimitive(this.model);
        } else if (mode === ColorMode.FILL) {
            // Update model
            this.model.options.hasFill = has;
            this.model.options.fillColor = color;
            this.model.options.fillAlpha = colorAlpha;

            // Elements
            // -- Arrow element
            this._arrowElement.options = Object.assign({}, this.model.options);
            // -- Selftest element
            this._selftestElement.options = Object.assign({}, this.model.options);

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updatePrimitive(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Number">
    /**
     * On number
     * @param mode
     * @param number
     */
    public onNumber(mode: NumberMode, number: number): void {
        if (mode === NumberMode.WIDTH) {
            // Update model
            this.model.strokeWidth = number / this.imageDimensions.width;

            // Elements
            // -- Arrow element
            this._arrowElement.lineWidth = this.model.strokeWidth ? Math.round(this.model.strokeWidth * this.imageDimensions.width * 100) / 100 : undefined;

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.emit("onBackgroundNumber");
            this.image.updatePrimitive(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Rotate with line">
    public onButtonRotateWithLine(override?: boolean): void {
        this.model.rotateWithLine = override || !this.model.rotateWithLine;

        // On rotate wiht line
        this.onRotateWithLine(new PIXI.InteractionEvent());

        // Events
        if (this.model.rotateWithLine) {
            this.parent.on("onRotateWithLine", this.onRotateWithLine, this);
        } else {
            this.parent.off("onRotateWithLine", this.onRotateWithLine, this);
        }

        // Emits
        this.image.updatePrimitive(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Static">
    public onButtonStatic(override?: boolean): void {
        this.model.isStatic = override || !this.model.isStatic;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updatePrimitive(this.model);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectPrimitive(this.model, (this.parent as MedsurfDraw.PositionPoint).model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Control unselect item
     * @param parent
     */
    public controlUnselectItem(parent: MedsurfDraw.PositionPoint): void {
        // Get image
        const image = parent.image;

        // Set select layer
        if(!parent.controlSelectLayerGroup(image)) {
            // Set select image
            image.externControlSelectImage();

            // Emits
            image.unselectPrimitive();
        }
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Get parent
        const parent = this.parent as MedsurfDraw.PositionPoint;

        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== parent)
            .forEach((imageObject) => {
                if (parent.model.selftestItems && parent.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else if (parent.model.selftestItems && imageObject instanceof MedsurfDraw.LegendCollection) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => { return {name: lr.model.id, element: lr}; });
                    for (let legendRow of legendRows) {
                        if (parent.model.selftestItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.modeInteraction.setMode('choose_select');
                        } else {
                            legendRow.element.modeInteraction.setMode('blocket_item');
                        }
                    }
                } else {
                    imageObject.modeInteraction.setMode('blocked_item');
                }
            });

        // Set default mode for parent
        parent.modeInteraction.setModeItem('select_child');

        // Set children mode
        this.modeInteraction.setModeChildren('select_parent');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control draw item
     */
    public controlDrawItem(): void {
        // Set children mode
        this.modeInteraction.setModeChildren('draw_parent');

        // Set draw item
        this.modeInteraction.setModeItem('draw_item');
    }
    //</editor-fold>

    //<editor-fold desc="Move">
    /**
     * Start move
     * @param event
     */
    public startMove(event: PIXI.InteractionEvent): void {
        (this.parent as MedsurfDraw.PositionPoint).moveInteraction.emit("startMove", event);
    }

    /**
     * On move
     * @param event
     * @param dX
     * @param dY
     */
    public onMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Check if "Ctrl"-Key pressed
        if (MedsurfDraw.Keyboard.isKeyDown("ControlLeft") || MedsurfDraw.Keyboard.isKeyDown("ControlRight")) {
            this.position.x += dX;
            this.position.y += dY;
            this.model.rectangle.x = this.position.x / imageWidth;
            this.model.rectangle.y = this.position.y / imageWidth;

            // Elements
            // -- Selftest element
            this._selftestElement.selftestPoint = new PIXI.Point(
                this.parent.x + this.model.rectangle.x * this.imageDimensions.width,
                this.parent.y + this.model.rectangle.y * this.imageDimensions.width
            );

            this.parent.emit("childMove", event);
        } else {
            (this.parent as MedsurfDraw.PositionPoint).moveInteraction.emit("onMove", event, dX, dY);
        }
    }

    /**
     * End move
     * @param event
     */
    public endMove(event: PIXI.InteractionEvent): void {
        // Emits
        (this.parent as MedsurfDraw.PositionPoint).moveInteraction.emit("endMove", event);
    }

    /**
     * End position point move
     */
    public endPositionPointMove(): void {
        this._selftestElement.selftestPoint = new PIXI.Point(
            this.parent.x + this.model.rectangle.x * this.imageDimensions.width,
            this.parent.y + this.model.rectangle.y * this.imageDimensions.width
        );
    }

    /**
     * On wedge move
     * @param event
     * @param dX
     * @param dY
     * @private
     */
    protected _onWedgeMove(event: PIXI.InteractionEvent, dX: number, dY: number): void {
        /* TODO remove
        // Get image width
        const imageWidth = this.getImageDimensions().width;

        // Elements
        // -- Arrow element
        if (!this._arrowElement.wedgePoint) {
            this._arrowElement.wedgePoint = this.model.wedgePoint
                ? {x: this.model.wedgePoint.x * imageWidth, y: this.model.wedgePoint.y * imageWidth}
                : {x: (this.model.rectangle.x + this.model.rectangle.width) * imageWidth, y: (this.model.rectangle.y + this.model.rectangle.height / 2) * imageWidth};
        }
        this._arrowElement.wedgePoint.x += dX * Math.cos(this.rotation) + dY * Math.sin(this.rotation);
        // -- Wedge point element
        this._wedgePointElement.wedgePoint.x += dX * Math.cos(this.rotation) + dY * Math.sin(this.rotation);

        // Model
        if (!this.model.wedgePoint) {
            this.model.wedgePoint = new PIXI.Point();
        }
        this.model.wedgePoint.x = this._wedgePointElement.wedgePoint.x / imageWidth;

        // Draw
        this.draw();
        */
    }

    /**
     * On wedge end move
     * @param event
     * @private
     */
    protected _onWedgeEndMove(event: PIXI.InteractionEvent): void {
        // Emits
        this.image.updatePrimitive(this.model);
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    /**
     * On hover
     */
    public onHover(): void {
        // Setup
        this._arrowElement.options.lineColor = Design.primitive.lineColorHover;
        this._arrowElement.options.fillColor = Design.primitive.fillColorHover;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On out
     */
    public onOut(): void {
        // Setup
        this._arrowElement.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * On select start
     * @param event
     * @param preventSelectItem
     */
    public onSelectStart(event: PIXI.InteractionEvent, preventSelectItem: boolean = false): void {
        event.stopPropagation();

        // Get last mode
        const lastMode = this.modeInteraction.lastMode;

        // Set current mode
        if (lastMode === 'select_layer_group' || lastMode === 'select_parent') {
            this.controlSelectItem();
        } else {
            if (!(this.parent as MedsurfDraw.PositionPoint).controlSelectLayerGroup(this.image) && !preventSelectItem) {
                this.controlSelectItem();
            }
        }
    }

    /**
     * On selftest
     * @private
     */
    public onSelftest(): void {
        this._selftestSelected = true;
        (this.parent as MedsurfDraw.PositionPoint).onSelftest();
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Scale">
    /**
     * Start scale
     * TODO remove
     */
    public startScale(): void {
        this.interactive = false;
        /*
        this._rotateContainer.removeChild(this._wedgePointElement);
        // TODO better implementation: this._rotateElement.hideItem();
        this._rotateContainer.removeChild(this._rotateElement);
        // TODO better implementation: this._configurationElement.hideItem();
        this.removeChild(this._configurationElement);
        */
    }

    /**
     * On scale
     * @param scaleSide
     * @param dX
     * @param dY
     * @param dW
     * @param dH
     */
    public onScale(scaleSide: MedsurfDraw.ScaleSideTypes, dX: number, dY: number, dW: number, dH: number): void {
        if (this._arrowElement.rectangle.width + dW < 1) {
            return;
            /*
            dW = 1 - this._ellipse.width;
            dX = 0;
            dY = 0;
            */
        }
        if (this._arrowElement.rectangle.height + dH < 1) {
            return;
            /*
            dH = 1 - this._ellipse.height;
            dX = 0;
            dY = 0;
            */
        }

        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Calc wedge adjust
        /* TODO remove
        const wedgeXAdjust = this._arrowElement.rectangle.width / (this._wedgePointElement.wedgePoint.x);
        const wedgeYAdjust = this._arrowElement.rectangle.height / (this._wedgePointElement.wedgePoint.y);
        */
        let wedgeXAdjust = 1;
        let wedgeYAdjust = 1;
        if (this.wedgePoint) {
            wedgeXAdjust = this._arrowElement.rectangle.width / this.wedgePoint.x;
            wedgeYAdjust = this._arrowElement.rectangle.height / this.wedgePoint.y;
        }


        // Element
        // -- Arrow element
        this._arrowElement.rectangle.width += dW;
        this._arrowElement.rectangle.height += dH;
        // -- Wedge point element
        /* TODO remove
        this._wedgePointElement.wedgePoint.x += dW / wedgeXAdjust;
        this._wedgePointElement.wedgePoint.y += dH / wedgeYAdjust;
        */
        if (this.wedgePoint) {
            this.wedgePoint.x += dW / wedgeXAdjust;
            this.wedgePoint.y += dH / wedgeYAdjust;
        }


        // Update model
        this.model.rectangle.width = this._arrowElement.rectangle.width / imageWidth;
        this.model.rectangle.height = this._arrowElement.rectangle.height / imageWidth;
        if (!this.model.wedgePoint) {
            this.model.wedgePoint = new PIXI.Point();
        }
        /* TODO remove
        this.model.wedgePoint.x = this._wedgePointElement.wedgePoint.x / imageWidth;
        this.model.wedgePoint.y = this._wedgePointElement.wedgePoint.y / imageWidth;
        */
        if (this.wedgePoint) {
            this.model.wedgePoint.x = this.wedgePoint.x / imageWidth;
            this.model.wedgePoint.y = this.wedgePoint.y / imageWidth;
        }

        // Draw
        this.draw();

        // TODO needed ???
        // this.parent.emit("onScale", dX, dY, dW, dH);
    }

    /**
     * End scale
     * TODO remove
     */
    public endScale(): void {
        this.interactive = true;
        /*
        this._rotateContainer.addChild(this._wedgePointElement);
        this._wedgePointElement.draw();

        // TODO better implementation: this._rotateElement.showItem();
        this._rotateElement.draw();
        this._rotateContainer.addChild(this._rotateElement);

        // TODO better implementation: this._configurationElement.showItem();
        this._configurationElement.draw();
        this.addChild(this._configurationElement);

        this._rotateContainer.sortChildren();
        */

        // Emits
        this.image.updatePrimitive(this.model);
    }
    //</editor-fold>

    //<editor-fold desc="Rotate">
    public startRotate(): void {
        this.rotation = (Math.round(this.rotation / Math.PI * 180) * Math.PI / 180);
    }

    /**
     * On rotate
     * @param degree
     */
    public onRotate(degree: number): void {
        // Setup
        this.rotation += degree;

        // Update model
        this.model.angle = this.rotation;

        // Emit on background rotate
        this.emit("onBackgroundRotate", degree);
    }

    /**
     * End rotate
     */
    public endRotate(): void {
        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On rotate with line
     * @param event
     */
    public onRotateWithLine(event: PIXI.InteractionEvent): void {
        const parent = this.parent as MedsurfDraw.PositionPoint;

        let lineStartX: number = 0;
        let lineStartY: number = 0;
        let lineEndX: number = 0;
        let lineEndY: number = 0;

        const lines = parent.getLines();
        if (lines.length === 0) {
            return;
        }
        for (let line of lines) {
            if (line.start === parent) {
                if (!line.start || !line.end) {
                    return;
                }
                lineStartX += line.start.x;
                lineStartY += line.start.y;
                lineEndX += line.end.x;
                lineEndY += line.end.y;
            } else {
                if (!line.start || !line.end) {
                    return;
                }
                lineStartX += line.end.x;
                lineStartY += line.end.y;
                lineEndX += line.start.x;
                lineEndY += line.start.y;
            }
        }

        lineStartX /= lines.length;
        lineStartY /= lines.length;
        lineEndX /= lines.length;
        lineEndY /= lines.length;

        const degree = Math.atan2(lineEndY - lineStartY, lineEndX - lineStartX) - (90 * Math.PI / 180) || 0;

        // Setup
        this.rotation = degree + Math.PI / 2;

        // Update model
        this.model.angle = this.rotation;

        // Emit on background rotate
        this.emit("onBackgroundRotate", degree);
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Selftest">
    /**
     * Selftest changed
     */
    public selftestChanged(): void {
        this._selftestElement.isSelftest = (this.parent as MedsurfDraw.PositionPoint).model.isSelftest;
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        const rect = super.getRectangle();
        const position = this.parent.position;
        return new PIXI.Rectangle(position.x + rect.x , position.y + rect.y, rect.width, rect.height);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get wedge point
     */
    public get wedgePoint(): Models.Point | null {
        return this._wedgePoint;
    }

    /**
     * Set wedge point
     * @param value
     */
    public set wedgePoint(value: Models.Point | null) {
        this._wedgePoint = value;
    }

    /**
     * Get move interaction
     */
    public get moveInteraction(): MedsurfDraw.MoveInteraction<ArrowPrimitive> {
        return this._moveInteraction;
    }

    /**
     * Set move interaction
     * @param value
     */
    public set moveInteraction(value: MedsurfDraw.MoveInteraction<ArrowPrimitive>) {
        this._moveInteraction = value;
    }

    /**
     * Get scale interaction
     */
    public get scaleInteraction(): MedsurfDraw.ScaleInteraction<ArrowPrimitive> {
        return this._scaleInteraction;
    }

    /**
     * Set scale interaction
     * @param value
     */
    public set scaleInteraction(value: MedsurfDraw.ScaleInteraction<ArrowPrimitive>) {
        this._scaleInteraction = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<ArrowPrimitive> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<ArrowPrimitive>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}