import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {Design} from "../../config/design";
import {
    BaseElementContainer,
    BaseElementContainerModel,
    ColorMode,
    NumberMode
} from "../../bases/elements/BaseElementContainer";
import {debounce} from "debounce";

/**
 * Background rectangle element
 */
export class BackgroundRectanglePrimitive extends BaseElementContainer<Models.BackgroundRectanglePrimitive, BaseElementContainerModel<Models.BackgroundRectanglePrimitive>> {
    /**
     * Members
     */
    protected _primitive: MedsurfDraw.PrimitiveObject | null;

    /**
     * Interactions
     */
    private _selectInteraction: MedsurfDraw.SelectInteraction<BackgroundRectanglePrimitive>;

    /**
     * Methods
     */
    protected _debounceControlSelectItemMethod: any;
    protected _debounceControlUnselectItemMethod: any;

    /**
     * Elements
     */
    private _rectangleElement: MedsurfDraw.Rectangle;

    /**
     * Menu elements
     */
        // -- Edit
    protected _editButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Layer
    protected _moveToBackButtonElement: MedsurfDraw.RoundButtonElement;
    protected _moveToFrontButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Delete
    protected _deleteButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Degroup
    protected _degroupButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Mirror
    protected _horizontalMirrorButtonElement: MedsurfDraw.RoundButtonElement;
    protected _verticalMirrorButtonElement: MedsurfDraw.RoundSubMenuElement;
    // -- Duplicate
    protected _duplicateButtonElement: MedsurfDraw.RoundButtonElement;
    // -- Clipboard
    protected _clipboardButtonElement: MedsurfDraw.RoundButtonElement;

    /**
     * Constructor
     * @param image
     * @param model
     */
    public constructor(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage,
                       model: Models.BackgroundRectanglePrimitive) {
        // Parent constructor
        super({image, model});

        // Members
        this.primitive = null;

        // Setup
        this.zIndex = Design.backgroundRectangleElement.zIndexDefault + this.model.zIndex;
        this.rotation = this.model.angle || 0;

        //<editor-fold desc="Interactions">
        // -- Mode interaction
        this.modeInteraction.on("default", this._modeDefault, this);
        this.modeInteraction.on("remove-default", this._removeModeDefault, this);
        this.modeInteraction.on("marker", this._modeMarker, this);
        this.modeInteraction.on("remove-marker", this._removeModeMarker, this);
        this.modeInteraction.on("selftest", this._modeSelftest, this);
        this.modeInteraction.on("remove-selftest", this._removeModeSelftest, this);
        this.modeInteraction.on("author", this._modeAuthor, this);
        this.modeInteraction.on("remove-author", this._removeModeAuthor, this);
        this.modeInteraction.on('select_layer_group', this._modeSelectLayerGroup, this);
        this.modeInteraction.on('remove-select_layer_group', this._removeModeSelectLayerGroup, this);
        this.modeInteraction.on('select_grouping', this._modeSelectGrouping, this);
        this.modeInteraction.on('remove-select_grouping', this._removeModeSelectGrouping, this);
        this.modeInteraction.on('select_item', this._modeSelectItem, this);
        this.modeInteraction.on("remove-select_item", this._removeModeSelectItem, this);
        this.modeInteraction.on('select_item_color_line', this._modeSelectItemColor, this);
        this.modeInteraction.on('remove-select_item_color_line', this._removeModeSelectItemColorLine, this);
        this.modeInteraction.on('select_item_color_fill', this._modeSelectItemColor, this);
        this.modeInteraction.on("remove-select_item_color_fill", this._removeModeSelectItemColorFill, this);
        this.modeInteraction.on('select_item_number_line_width', this._modeSelectItemNumber, this);
        this.modeInteraction.on("remove-select_item_number_line_width", this._removeModeSelectItemNumber, this);
        this.modeInteraction.on('select_item_number_pad', this._modeSelectItemNumber, this);
        this.modeInteraction.on('remove-select_item_number_pad', this._removeModeSelectItemNumber, this);
        this.modeInteraction.on('blocked_item', this._modeBlockedItem, this);
        this.modeInteraction.on("remove-blocked_item", this._removeModeBlockedItem, this);
        this.modeInteraction.on('choose_select', this._modeChooseSelect, this);
        this.modeInteraction.on('remove-choose_select', this._removeModeChooseSelect, this);
        this.modeInteraction.on('choose_unselect', this._modeChooseUnselect, this);
        this.modeInteraction.on('remove-choose_unselect', this._removeModeChooseUnselect, this);
        this.modeInteraction.on("delete", this._modeDelete, this);
        this.modeInteraction.on("delete_draw", this._modeDelete, this);
        // -- Select interaction
        this.selectInteraction = new MedsurfDraw.SelectInteraction(this);
        this.selectInteraction.on("onHover", this.onHover, this);
        this.selectInteraction.on("onOut", this.onOut, this);
        this.selectInteraction.on("onSelectStart", this.onSelectStart, this);
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceControlSelectItemMethod = debounce(this.controlSelectItem.bind(this), 10).bind(this);
        this._debounceControlUnselectItemMethod = debounce(this.controlUnselectItem.bind(this), 10).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.on("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceControlSelectItem", this._debounceControlSelectItemMethod);
            this.off("debounceControlUnselectItem", this._debounceControlUnselectItemMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Get primitive
        this.primitive = this.image.getImageObjects()
            .filter((child: MedsurfDraw.ImageObject) => child instanceof MedsurfDraw.PositionPoint)
            .map((positionPoint) => (positionPoint as MedsurfDraw.PositionPoint).getChildByName(this.model.primitiveId))
            // TODO: .map((positionPoint: MedsurfDraw.PositionPoint) => positionPoint.children.find((c) => c.name === this.model.primitiveId))
            .filter((result) => result)[0] as MedsurfDraw.PrimitiveObject;

        // Get rectangle
        const rectangle = this.getBackgroundRectangle();

        //<editor-fold desc="Elements">
        // -- Rectangle element
        this._rectangleElement = new MedsurfDraw.Rectangle({
            rectangle: rectangle,
            options: Object.assign({}, this.model.options),
            lineWidth: this.model.strokeWidth ? Math.round(this.model.strokeWidth * imageWidth * 100) / 100 : undefined
        });
        this.addChild(this._rectangleElement);
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        this._editButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf3f0',
            description: 'Bearbeiten',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._editButtonElement.on("onHover", this.collapseAllMenus, this);
        this._editButtonElement.on("button", this.onButtonEdit, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._editButtonElement,
            order: 0,
        });
        //</editor-fold>
        //<editor-fold desc="Layer">
        // -- Move to back button element
        this._moveToBackButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf87f',
            description: 'Nach Hinten\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._moveToBackButtonElement.on("button", this.onButtonMoveToBack, this);
        // -- Move to front button element
        this._moveToFrontButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf856',
            description: 'Nach Vorne\nverschieben',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._moveToBackButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._moveToFrontButtonElement.on("onHover", this._moveToFrontButtonElement.showMenu, this._moveToFrontButtonElement);
        this._moveToFrontButtonElement.on("button", this.onButtonMoveToFront, this);
        this._moveToFrontButtonElement.on("toggleMenu", this.onMenuLayer, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._moveToFrontButtonElement,
            order: 7,
        });
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        this._deleteButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf1f8',
            description: 'Löschen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._deleteButtonElement.on("onHover", this.collapseAllMenus, this);
        this._deleteButtonElement.on("button", this.onButtonDelete, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._deleteButtonElement,
            order: 6,
        });
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        this._degroupButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf248',
            description: 'Aus Gruppe\nentfernen',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._degroupButtonElement.on("onHover", this.collapseAllMenus, this);
        this._degroupButtonElement.on("button", this.onButtonDegroup, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._degroupButtonElement,
            order: 4,
        });
        //</editor-fold>
        //<editor-fold desc="Mirror">
        // -- Horizontal mirror button element
        this._horizontalMirrorButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf89c',
            description: 'Horizontal\nspiegeln',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._horizontalMirrorButtonElement.on("button", this.onButtonHorizontalMirror, this);
        // -- Vertical mirror button element
        this._verticalMirrorButtonElement = new MedsurfDraw.RoundSubMenuElement({
            image: this.image,
            text: '\uf89d',
            description: 'Vertikal\nspiegeln',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            menuElements: [
                {
                    element: this._horizontalMirrorButtonElement,
                    order: 0,
                }
            ],
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius)
        });
        this._verticalMirrorButtonElement.on("onHover", this._verticalMirrorButtonElement.showMenu, this._verticalMirrorButtonElement);
        this._verticalMirrorButtonElement.on("button", this.onButtonVerticalMirror, this);
        this._verticalMirrorButtonElement.on("toggleMenu", this.onMenuMirror, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._verticalMirrorButtonElement,
            order: 3,
        });
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        this._duplicateButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf24d',
            description: 'Duplizieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._duplicateButtonElement.on("onHover", this.collapseAllMenus, this);
        this._duplicateButtonElement.on("button", this.onButtonDuplicate, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._duplicateButtonElement,
            order: 2,
        });
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        this._clipboardButtonElement = new MedsurfDraw.RoundButtonElement({
            image: this.image,
            text: '\uf0c5',
            description: 'Kopieren',
            descriptionFontSize: Design.subMenu.descriptionFontSize,
            circle: new PIXI.Circle(0, 0, Design.subMenu.buttonRadius),
            fontSize: Design.subMenu.fontSize,
            fontColor: Design.subMenu.fontColor
        });
        this._clipboardButtonElement.on("onHover", this.collapseAllMenus, this);
        this._clipboardButtonElement.on("button", this.onButtonClipboard, this);
        // Add to menu
        this.imageObjectMenu.push({
            element: this._clipboardButtonElement,
            order: 1,
        });
        //</editor-fold>
        //</editor-fold>

        //<editor-fold desc="Events">
        this.primitive.modeInteraction.on("delete", this._modeDelete, this);
        this.primitive.moveInteraction.on("onMove", this._updateRectangle, this);
        this.primitive.scaleInteraction.on("onScale", this._updateRectangle, this);
        this.primitive.on("onBackgroundRotate", this.onRotate, this);
        this.primitive.on("onBackgroundText", this._updateRectangle, this);
        this.primitive.on("onBackgroundNumber", this._updateRectangle, this);
        //</editor-fold>

        // Sort parent items
        this.parent.sortChildren();

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Rectangle element
        this._rectangleElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Select interaction
        if (this.selectInteraction) {
            this.selectInteraction.removeAllListeners();
            // TODO remove delete this.selectInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Rectangle element
        if (this._rectangleElement) {
            this._rectangleElement.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Menu elements">
        //<editor-fold desc="Edit">
        // -- Edit button element
        if (this._editButtonElement) {
            this._editButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Layer">
        // Move to back button element
        if (this._moveToBackButtonElement) {
            this._moveToBackButtonElement.destroy(options);
        }
        // -- Move to front button element
        if (this._moveToFrontButtonElement) {
            this._moveToFrontButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Delete">
        // -- Delete button element
        if (this._deleteButtonElement) {
            this._deleteButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Degroup">
        // -- Degroup button element
        if (this._degroupButtonElement) {
            this._degroupButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Mirror">
        // -- Horizontal mirror button element
        if (this._horizontalMirrorButtonElement) {
            this._horizontalMirrorButtonElement.destroy(options);
        }
        // -- Vertical mirror button element
        if (this._verticalMirrorButtonElement) {
            this._verticalMirrorButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Duplicate">
        // -- Duplicate button element
        if (this._duplicateButtonElement) {
            this._duplicateButtonElement.destroy(options);
        }
        //</editor-fold>
        //<editor-fold desc="Clipboard">
        // -- Clipboard button element
        if (this._clipboardButtonElement) {
            this._clipboardButtonElement.destroy(options);
        }
        //</editor-fold>
        //</editor-fold>

        //<editor-fold desc="Events">
        if (this.primitive) {
            this.primitive.modeInteraction.off("delete", this._modeDelete, this);
            this.primitive.moveInteraction.off("onMove", this._updateRectangle, this);
            this.primitive.scaleInteraction.off("onScale", this._updateRectangle, this);
            this.primitive.off("onBackgroundRotate", this.onRotate, this);
            this.primitive.off("onBackgroundText", this._updateRectangle, this);
            this.primitive.off("onBackgroundNumber", this._updateRectangle, this);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Static functions">
    /**
     * Get instance
     * @param image
     * @param primitiveId
     * @param rotateWithLine
     * @param imageWidth
     */
    public static getInstance(image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage, primitiveId: string,
                              rotateWithLine: boolean, imageWidth: number): BackgroundRectanglePrimitive {
        // Element
        // -- Background rectangle primitive
        return new BackgroundRectanglePrimitive(image, {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.BACKGROUNDRECTANGLE,
            isStatic: false,
            zIndex: 0,
            primitiveId: primitiveId,
            rotateWithLine: rotateWithLine,
            strokeWidth: 1 / imageWidth,
            dashFill: 2,
            dashEmpty: 2,
            options: {
                hasLine: true,
                lineColor: 0xffffff,
                lineAlpha: 1,
                hasFill: true,
                fillColor: 0xffffff,
                fillAlpha: 0.3,
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Public function">
    /**
     * Get background rectangle
     */
    public getBackgroundRectangle(): PIXI.Rectangle {
        if (!this.primitive) {
            return new PIXI.Rectangle(0, 0, 0, 0);
        }

        // Get image width
        const imageWidth = this.imageDimensions.width;

        // Get pad
        const pad = this.model.pad ? Math.round(this.model.pad * imageWidth * 100) / 100 : Design.backgroundRectangleElement.padDefault;

        // Get element rectangle
        const elementRect = this.primitive.getElementRectangle();

        // Calc rectangle
        const rectangle = new PIXI.Rectangle(
            0,
            0,
            elementRect.width + pad,
            elementRect.height + pad
        );

        // Pivot
        this.pivot.set(rectangle.width / 2, rectangle.height / 2);

        // Position
        this.position.set(this.primitive.position.x, this.primitive.position.y);

        // Return rect
        return rectangle;
    }
    //</editor-fold>

    //<editor-fold desc="Private function">
    /**
     * Get rectangle
     */
    protected _updateRectangle(): void {
        const rectangle = this.getBackgroundRectangle();

        // Elements
        // -- Rectangle element
        this._rectangleElement.rectangle = rectangle;

        // Draw
        this.draw();
    }
    //</editor-fold>

    //<editor-fold desc="Mode">
    /**
     * Mode default
     * @private
     */
    protected _modeDefault(): void {
        if (this.model.isStatic) {
            // Show item
            this.showItem();
        } else {
            // Hide item
            this.hideItem();
        }
    }

    /**
     * Remove mode default
     * @private
     */
    protected _removeModeDefault(): void {
    }

    /**
     * Mode marker
     * @private
     */
    protected _modeMarker(): void {
        // Show item
        this.showItem();
    }

    /**
     * Remove mode marker
     * @private
     */
    protected _removeModeMarker(): void {
    }

    /**
     * Mode selftest
     * @private
     */
    protected _modeSelftest(): void {
        // Get position point
        const positionPoint = this.parent as MedsurfDraw.PositionPoint;

        if (this.model.isStatic) {
            // Setup
            this.interactive = true;
            this.cursor = "select";

            // Events
            positionPoint.selectInteraction.reset();
            this.on("pointerover", positionPoint.selectInteraction.onHover, positionPoint.selectInteraction);
            this.on("pointerout", positionPoint.selectInteraction.onOut, positionPoint.selectInteraction);
            this.on("mousedown", positionPoint.selectInteraction.onPointerDown, positionPoint.selectInteraction);
            this.on("mouseup", positionPoint.selectInteraction.onSelftest, positionPoint.selectInteraction);

            // Show item
            this.showItem();
        } else {
            // Hide Item
            this.hideItem();
        }
    }

    /**
     * Remove mode selftest
     * @private
     */
    protected _removeModeSelftest(): void {
        // Get position point
        const positionPoint = this.parent as MedsurfDraw.PositionPoint;

        // Setup
        this.interactive = false;

        // Elements
        // -- Rectangle element
        this._rectangleElement.showItem();

        // Events
        this.off("pointerover", positionPoint.selectInteraction.onHover, positionPoint.selectInteraction);
        this.off("pointerout", positionPoint.selectInteraction.onOut, positionPoint.selectInteraction);
        this.off("mousedown", positionPoint.selectInteraction.onPointerDown, positionPoint.selectInteraction);
        this.off("mouseup", positionPoint.selectInteraction.onSelftest, positionPoint.selectInteraction);
    }

    /**
     * Mode author
     * @private
     */
    protected _modeAuthor(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode author
     * @private
     */
    protected _removeModeAuthor(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select layer
     * @private
     */
    protected _modeSelectLayerGroup(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select layer
     * @private
     */
    protected _removeModeSelectLayerGroup(): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select grouping
     * @private
     */
    protected _modeSelectGrouping(): void {
        // Setup
        this.interactive = true;

        // Events
        this.on("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.on("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);

        // Show item
        this.showItem();
    }

    /**
     * Remove mode select grouping
     * @private
     */
    protected _removeModeSelectGrouping(): void {
        // Setup
        this.interactive = false;

        // Events
        this.off("pointerover", this.selectInteraction.onHover, this.selectInteraction);
        this.off("pointerout", this.selectInteraction.onOut, this.selectInteraction);
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
    }

    /**
     * Mode select item
     * @private
     */
    protected _modeSelectItem(): void {
        this.onOut();

        // Setup
        this.interactive = true;

        // TODO this.setAsFrontElement();
        // TODO this._showInteractions();

        // Events
        this.on("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
        // TODO this.on("onRelease", this._onRelease, this);

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select iten
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Setup
        this.interactive = false;

        // TODO this.setAsNormalElement();
        // TODO this._hideInteractions();

        // Events
        this.off("mouseup", this.selectInteraction.onSelect, this.selectInteraction);
        // TODO this.off("onRelease", this._onRelease, this);

        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Set select layer
        // TODO remove console.log("backRect _removeModeSelectItem", mode);
        if (mode === 'author') {
            this.emit("debounceControlUnselectItem", this.parent);
        }
        /* TODO remove
        if (!mode.startsWith('select_parent') && mode !== 'select_child' && !mode.startsWith('select_item')
            && !mode.startsWith('choose_select') && !mode.startsWith('choose_parent') && !mode.startsWith('blocked_item')
            && ['default', 'marker', 'selftest'/*, 'author'* /, 'delete'].indexOf(mode) === -1) {
            // Unselect item
            this.controlUnselectItem();
            // this.emit("debounceControlUnselectItem");
        }
        */
    }

    /**
     * Mode select item color
     * @private
     */
    protected _modeSelectItemColor(): void {
        this.onOut();

        // Control Update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item color line
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemColorLine(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Control save color line
        this.image.controlSaveColor(this.model.options.lineColor, this.model.options.lineAlpha);

        // Remove mode select item color
        this._removeModeSelectItemColor(parent, mode);
    }

    /**
     * Remove mode select item color fill
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemColorFill(parent: MedsurfDraw.ImageObject, mode: string): void {
        // Control save color fill
        this.image.controlSaveColor(this.model.options.fillColor, this.model.options.fillAlpha);

        // Remove mode select item color
        this._removeModeSelectItemColor(parent, mode);
    }

    /**
     * Remove mode select item color
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemColor(parent: MedsurfDraw.ImageObject, mode: string): void {
        if (mode !== 'select_child' && !mode.startsWith('select_item_color')) {
            // Set select item
            this.emit("debounceControlSelectItem");
        }
    }

    /**
     * Mode select item number
     * @private
     */
    protected _modeSelectItemNumber(): void {
        this.onOut();

        // Control update elements
        this.image.controlUpdateElements();

        // Show item
        this.showItem();

        // Emit
        this.emit("pointerover", new PIXI.InteractionEvent());
    }

    /**
     * Remove mode select item number
     * @param parent
     * @param mode
     * @private
     */
    protected _removeModeSelectItemNumber(parent: MedsurfDraw.ImageObject, mode: string): void {
        if (mode !== 'select_child' && !mode.startsWith('select_item_number')) {
            // Set select item
            this.emit("debounceControlSelectItem");
        }
    }

    /**
     * Mode blocked item
     * @param parent
     * @param mode
     * @private
     */
    protected _modeBlockedItem(parent: MedsurfDraw.ImageObject, mode: string): void {
        /* TODO Not working for canvas renderer
        // Filters
        const blockedFilter = new PIXI.filters.ColorMatrixFilter();
        blockedFilter.greyscale(0, true);

        this.filters = [blockedFilter];
        */
        // Element
        // -- Rectangle element
        this._rectangleElement.options.lineAlpha = Design.blocked.lineAlpha;
        this._rectangleElement.options.fillAlpha = Design.blocked.fillAlpha;

        // Draw
        if (!mode.startsWith('delete')) {
            this.emit("debounceDraw");
        }
    }

    /**
     * Remove mode blocked item
     * @private
     */
    protected _removeModeBlockedItem(): void {
        /* TODO Not working for canvas renderer
        // Filters
        this.filters = [];
        */
        // Element
        // -- Rectangle element
        this._rectangleElement.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose select
     * @private
     */
    protected _modeChooseSelect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Rectangle element
        this._rectangleElement.options.lineColor = Design.selftest.lineColorSelect;
        this._rectangleElement.options.fillColor = Design.selftest.fillColorSelect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose select
     * @private
     */
    protected _removeModeChooseSelect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Rectangle element
        this._rectangleElement.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode choose unselect
     * @private
     */
    protected _modeChooseUnselect(): void {
        // Setup
        this.interactive = true;
        this.cursor = "select";

        // Element
        // -- Rectangle element
        this._rectangleElement.options.lineColor = Design.selftest.lineColorUnselect;
        this._rectangleElement.options.fillColor = Design.selftest.fillColorUnselect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Remove mode choose unselect
     * @private
     */
    protected _removeModeChooseUnselect(): void {
        // Setup
        this.interactive = false;

        // Element
        // -- Rectangle element
        this._rectangleElement.options = Object.assign({}, this.model.options);

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Mode delete
     * @private
     */
    protected _modeDelete(): void {
        // Get parent
        const parent = this.parent as MedsurfDraw.PositionPoint;

        // Setup parent
        if (parent) {
            // Get image
            const image = parent.image;

            if (parent.model.primitives) {
                let index = parent.model.primitives.findIndex((primitive: any) => primitive.id === this.model.id);
                if (index > -1) {
                    // Select layer group
                    if (['delete_draw'].indexOf(this.modeInteraction.lastMode) === -1) {
                        this.emit("debounceControlUnselectItem", parent);
                    }

                    // Delete from parent
                    parent.model.primitives.splice(index, 1);
                    parent.removeChild(this);

                    // Emits
                    image.deletePrimitive(this.model);
                }
            }

            // Hide menu
            image.controlHideMenu();

            // Control update elements
            image.controlUpdateElements();
        }

        // Destroy
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Menu">
    //<editor-fold desc="Defaults">
    public collapseAllMenus(event: PIXI.InteractionEvent): void {
        this._moveToFrontButtonElement.toggleMenu(event, false);
        this._verticalMirrorButtonElement.toggleMenu(event, false);
    }
    //</editor-fold>
    //<editor-fold desc="Edit">
    /**
     * On button edit
     */
    public onButtonEdit(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Emits
        this.image.editPrimitive(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Layer">
    /**
     * On button move to front
     */
    public onButtonMoveToFront(): void {
        // Model
        this.model.zIndex++;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On button move to back
     */
    public onButtonMoveToBack(): void {
        // Model
        this.model.zIndex--;
        this.zIndex = this.model.zIndex;

        this.image.sortChildren();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On menu layer
     * @param event
     * @param visible
     */
    public onMenuLayer(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._verticalMirrorButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Delete">
    /**
     * On button delete
     * @param event
     */
    public onButtonDelete(event: PIXI.InteractionEvent): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        setTimeout(() => {
            // Set mode
            this.modeInteraction.setMode("delete");
        }, 300);
    }
    //</editor-fold>
    //<editor-fold desc="Degroup">
    /**
     * On button degroup
     *
     * @param unselectElement
     */
    public onButtonDegroup(unselectElement: boolean = true): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Degroup image object
        (this.parent as MedsurfDraw.PositionPoint).onButtonDegroup(false);

        // Emit
        if (unselectElement) {
            this.image.unselectLayerGroup();
            this.image.selectPrimitive(this.model, (this.parent as MedsurfDraw.PositionPoint).model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Mirror">
    /**
     * On button vertical mirror
     */
    public onButtonVerticalMirror(): void {
        // Prevent rotate with line
        if (this.model.rotateWithLine) {
            return;
        }

        // Get rotation
        let rotation = (Math.round(((this.model.angle || 0) / Math.PI * 180) * 100) / 100) % 360;
        if (rotation < 0) {
            rotation = 360 + rotation;
        }

        // Model
        this.rotation = Math.round(180 - rotation) % 360 * (2 * Math.PI / 360);
        this.model.angle = this.rotation;

        // Control update elements
        this.image.controlUpdateElements();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On button horizontal mirror
     */
    public onButtonHorizontalMirror(): void {
        // Prevent rotate with line
        if (this.model.rotateWithLine) {
            return;
        }

        // Get rotation
        let rotation = (Math.round(((this.model.angle || 0) / Math.PI * 180) * 100) / 100) % 360;
        if (rotation < 0) {
            rotation = 360 + rotation;
        }

        // Model
        this.rotation = Math.round(360 - rotation) % 360 * (2 * Math.PI / 360);
        this.model.angle = this.rotation;

        // Control update elements
        this.image.controlUpdateElements();

        // Emits
        this.image.updatePrimitive(this.model);
    }

    /**
     * On menu mirror
     * @param event
     * @param visible
     */
    public onMenuMirror(event: PIXI.InteractionEvent, visible: boolean): void {
        if (visible) {
            this._moveToFrontButtonElement.toggleMenu(event, false);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Duplicate">
    /**
     * On button duplicate
     */
    public onButtonDuplicate(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Copy image object
        (this.parent as MedsurfDraw.PositionPoint).onButtonDuplicate();
    }
    //</editor-fold>
    //<editor-fold desc="Clipboard">
    /**
     * On button clipboard
     */
    public onButtonClipboard(): void {
        // Menu
        this.collapseAllMenus(new PIXI.InteractionEvent());

        // Hide menu
        this.image.controlHideMenu();

        // Clipboard image object
        (this.parent as MedsurfDraw.PositionPoint).onButtonClipboard();
    }
    //</editor-fold>
    //<editor-fold desc="Color">
    /**
     * On color
     * @param mode
     * @param has
     * @param color
     * @param colorAlpha
     */
    public onColor(mode: ColorMode, has: boolean, color: number | undefined, colorAlpha: number | undefined): void {
        if (mode === ColorMode.LINE) {
            // Update model
            this.model.options.hasLine = has;
            this.model.options.lineColor = color;
            this.model.options.lineAlpha = colorAlpha;

            // Elements
            // -- Rectangle element
            this._rectangleElement.options = Object.assign({}, this.model.options);

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updatePrimitive(this.model);
        } else if (mode === ColorMode.FILL) {
            // Update model
            this.model.options.hasFill = has;
            this.model.options.fillColor = color;
            this.model.options.fillAlpha = colorAlpha;

            // Elements
            // -- Rectangle element
            this._rectangleElement.options = Object.assign({}, this.model.options);

            // Draw
            this.emit("debounceDraw");

            // Emits
            this.image.updatePrimitive(this.model);
        }
    }
    //</editor-fold>
    //<editor-fold desc="Number">
    /**
     * On number
     * @param mode
     * @param number
     */
    public onNumber(mode: NumberMode, number: number): void {
        if (mode === NumberMode.WIDTH) {
            // Update model
            this.model.strokeWidth = number / this.imageDimensions.width;

            // Elements
            // -- Rectangle element
            this._rectangleElement.lineWidth = this.model.strokeWidth ? Math.round(this.model.strokeWidth * this.imageDimensions.width * 100) / 100 : undefined;

            // Draw
            this.emit("debounceDraw");
        } else if (mode === NumberMode.PAD) {
            // Update model
            this.model.pad = number / this.imageDimensions.width;

            // Elements
            // -- Rectangle element
            this._rectangleElement.rectangle.pad(this.model.pad, this.model.pad); // TODO test this

            // Draw
            this.emit("debounceDraw");
        }
    }
    //</editor-fold>
    //<editor-fold desc="Rotate with line">
    public onButtonRotateWithLine(override?: boolean): void {
        this.model.rotateWithLine = override || !this.model.rotateWithLine;

        // On rotate wiht line
        this.onRotateWithLine(new PIXI.InteractionEvent());

        // Events
        if (this.model.rotateWithLine) {
            this.parent.on("onRotateWithLine", this.onRotateWithLine, this);
        } else {
            this.parent.off("onRotateWithLine", this.onRotateWithLine, this);
        }

        // Emits
        this.image.updatePrimitive(this.model);
    }
    //</editor-fold>
    //<editor-fold desc="Static">
    public onButtonStatic(override?: boolean): void {
        this.model.isStatic = override || !this.model.isStatic;

        // Draw
        this.emit("debounceDraw");

        // Emits
        this.image.updatePrimitive(this.model);
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Control">
    /**
     * Control select item
     */
    public controlSelectItem(): void {
        // Emits
        this.image.selectPrimitive(this.model, (this.parent as MedsurfDraw.PositionPoint).model);

        // Select item
        this.externControlSelectItem();
    }

    /**
     * Control unselect item
     * @param parent
     */
    public controlUnselectItem(parent: MedsurfDraw.PositionPoint): void {
        // Get image
        const image = parent.image;

        // Set select layer
        if(!parent.controlSelectLayerGroup(image)) {
            // Set select image
            image.externControlSelectImage();

            // Emits
            image.unselectPrimitive();
        }
    }

    /**
     * Extern control select item
     */
    public externControlSelectItem(): void {
        // Get parent
        const parent = this.parent as MedsurfDraw.PositionPoint;

        // Set default mode for image children
        this.image.getImageObjects()
            .filter((imageObject) => imageObject !== parent)
            .forEach((imageObject) => {
                if (parent.model.selftestItems && parent.model.selftestItems.indexOf(imageObject.name as string) > -1) {
                    imageObject.modeInteraction.setMode('choose_select');
                } else if (parent.model.selftestItems && imageObject instanceof MedsurfDraw.LegendCollection) {
                    const legendRows = imageObject.legendColumns.map(lc => lc.legendRows)
                        .reduce((cur, acc: any) => [...cur, ...acc], [])
                        .map((lr) => { return {name: lr.model.id, element: lr}; });
                    for (let legendRow of legendRows) {
                        if (parent.model.selftestItems.indexOf(legendRow.name) > -1) {
                            legendRow.element.modeInteraction.setMode('choose_select');
                        } else {
                            legendRow.element.modeInteraction.setMode('blocked_item');
                        }
                    }
                } else {
                    imageObject.modeInteraction.setMode('blocked_item');
                }
            });

        // Set default mode for parent
        parent.modeInteraction.setModeItem('select_child');

        // Set children mode
        this.modeInteraction.setModeChildren('select_parent');

        // Set select item
        this.modeInteraction.setModeItem('select_item');
    }

    /**
     * Control draw item
     */
    public controlDrawItem(): void {
        // Set children mode
        this.modeInteraction.setModeChildren('draw_parent');

        // Set draw item
        this.modeInteraction.setModeItem('draw_item');
    }
    //</editor-fold>

    //<editor-fold desc="Select">
    public onHover(): void {
        (this.parent.children.filter(
            (element) => element instanceof BaseElementContainer && element.model.id === this.model.primitiveId
        )[0] as MedsurfDraw.TextPrimitive).onHover();
        // TODO this.onHover();
    }

    public onOut(): void {
        (this.parent.children.filter(
            (element) => element instanceof BaseElementContainer && element.model.id === this.model.primitiveId
        )[0] as MedsurfDraw.TextPrimitive).onOut();
        // TODO this.onOut();
    }

    public onSelectStart(event: PIXI.InteractionEvent): void {
        (this.parent.children.filter(
            (element) => element instanceof BaseElementContainer && element.model.id === this.model.primitiveId
        )[0] as MedsurfDraw.TextPrimitive).onSelectStart(event)
    }
    //</editor-fold>

    //<editor-fold desc="Rotate">
    public startRotate(): void {
        this.rotation = (Math.round(this.rotation / Math.PI * 180) * Math.PI / 180);
    }

    /**
     * On rotate
     * @param degree
     */
    public onRotate(degree: number): void {
        // Setup
        this.rotation += degree;

        // Update model
        this.model.angle = this.rotation;

        // Update rectangle
        this._updateRectangle();
    }

    /**
     * On rotate with line
     * @param event
     */
    public onRotateWithLine(event: PIXI.InteractionEvent): void {
        const parent = this.parent as MedsurfDraw.PositionPoint;

        let lineStartX: number = 0;
        let lineStartY: number = 0;
        let lineEndX: number = 0;
        let lineEndY: number = 0;

        const lines = parent.getLines();
        if (lines.length === 0) {
            return;
        }
        for (let line of lines) {
            if (line.start === parent) {
                if (!line.start || !line.end) {
                    return;
                }
                lineStartX += line.start.x;
                lineStartY += line.start.y;
                lineEndX += line.end.x;
                lineEndY += line.end.y;
            } else {
                if (!line.start || !line.end) {
                    return;
                }
                lineStartX += line.end.x;
                lineStartY += line.end.y;
                lineEndX += line.start.x;
                lineEndY += line.start.y;
            }
        }

        lineStartX /= lines.length;
        lineStartY /= lines.length;
        lineEndX /= lines.length;
        lineEndY /= lines.length;

        const degree = Math.atan2(lineEndY - lineStartY, lineEndX - lineStartX) - (90 * Math.PI / 180) || 0;

        // Setup
        this.rotation = degree + Math.PI / 2;

        // Update model
        this.model.angle = this.rotation;

        // Emit on background rotate
        this.emit("onBackgroundRotate", degree);
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        const rect = super.getRectangle();
        const position = this.parent.position;
        return new PIXI.Rectangle(position.x + rect.x - rect.width / 2, position.y + rect.y - rect.height / 2, rect.width, rect.height);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get primitive
     */
    public get primitive(): MedsurfDraw.PrimitiveObject | null {
        return this._primitive;
    }

    /**
     * Set primitive
     * @param value
     */
    public set primitive(value: MedsurfDraw.PrimitiveObject | null) {
        this._primitive = value;
    }

    /**
     * Get select interaction
     */
    public get selectInteraction(): MedsurfDraw.SelectInteraction<BackgroundRectanglePrimitive> {
        return this._selectInteraction;
    }

    /**
     * Set select interaction
     * @param value
     */
    public set selectInteraction(value: MedsurfDraw.SelectInteraction<BackgroundRectanglePrimitive>) {
        this._selectInteraction = value;
    }
    //</editor-fold>
}