import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Rotate element model
 */
export class RotateElementModel extends BaseContainerImageModel {
    rectangle: PIXI.Rectangle;
}

/**
 * Rotate element
 */
export class RotateElement extends BaseContainer<RotateElementModel> {
    /**
     * Interactions
     */
    private _rotateInteraction: MedsurfDraw.RotateInteraction<RotateElement>;

    /**
     * Elements
     */
    protected _rotateCircleElement: MedsurfDraw.Circle;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RotateElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.rotate.zIndex;

        //<editor-fold desc="Interactions">
        // -- Rotate interaction
        this._rotateInteraction = new MedsurfDraw.RotateInteraction(this);
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        let scale = this.imageScale.x * Design.positionPoint.scaleAspect;
        let radius = Design.positionPoint.radius;
        // TODO if (this.modeInteraction.lastMode === 'select' || this.modeInteraction.lastMode === 'selftest') {
        radius = Math.floor(radius / scale);
        /* TODO } else if (scale < 1) {
            scale = 1;
        }*/
        const lineWidth = Design.rotate.lineWidth / scale;
        const cornerRadius = Design.rotate.cornerRadius / scale / 2;
        // const rotationDY = Design.rotate.rotationDY / scale;

        //<editor-fold desc="Elements">
        // -- Rotate circle element
        this._rotateCircleElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(this.rectangle.x, this.rectangle.y, cornerRadius),
            options: {hasLine: true, lineColor: Design.rotate.lineColor, lineAlpha: Design.rotate.lineAlpha, hasFill: true, fillColor: Design.rotate.fillColor, fillAlpha: Design.rotate.fillAlpha},
            lineWidth: lineWidth
        });
        this._rotateCircleElement.interactive = true;
        this._rotateCircleElement.cursor = "rotate";
        this._rotateCircleElement.on("mousedown", this._rotateInteraction.startRotate, this._rotateInteraction);
        this._rotateCircleElement.on("pointermove", this._rotateInteraction.onRotate, this._rotateInteraction);
        this._rotateCircleElement.on("mouseup", this._rotateInteraction.endRotate, this._rotateInteraction);
        this._rotateCircleElement.on("pointerupoutside", this._rotateInteraction.endRotate, this._rotateInteraction);
        this.addChild(this._rotateCircleElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        let scale = this.imageScale.x * Design.rotate.scaleAspect;
        if (scale > 1) {
            scale = 1;
        }
        // TODO rectangle.pad(this._pad, this._pad);
        const lineWidth = Design.rotate.lineWidth / scale;
        const cornerRadius = Design.rotate.cornerRadius / scale / 2;
        // const rotationDY = Design.rotate.rotationDY / scale;

        //<editor-fold desc="Elements">
        // -- Rotate cirlce element
        this._rotateCircleElement.lineWidth = lineWidth;
        this._rotateCircleElement.circle = new PIXI.Circle(this.rectangle.x, this.rectangle.y - this.rectangle.height / 2 - Design.rotate.rotationDY, cornerRadius);
        this._rotateCircleElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Interactions">
        // -- Rotate interaction
        if (this.rotateInteraction) {
            this.rotateInteraction.removeAllListeners();
            // TODO remove delete this.rotateInteraction;
        }
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Rotate circle element
        if (this._rotateCircleElement) {
            this._rotateCircleElement.off("mousedown", this._rotateInteraction.startRotate, this._rotateInteraction);
            this._rotateCircleElement.off("pointermove", this._rotateInteraction.onRotate, this._rotateInteraction);
            this._rotateCircleElement.off("mouseup", this._rotateInteraction.endRotate, this._rotateInteraction);
            this._rotateCircleElement.off("pointerupoutside", this._rotateInteraction.endRotate, this._rotateInteraction);
            this._rotateCircleElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rotate interaction
     */
    public get rotateInteraction(): MedsurfDraw.RotateInteraction<RotateElement> {
        return this._rotateInteraction;
    }

    /**
     * Set rotate interaction
     * @param value
     */
    public set rotateInteraction(value: MedsurfDraw.RotateInteraction<RotateElement>) {
        this._rotateInteraction = value;
    }

    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }
    //</editor-fold>
}