import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {BaseGraphics, BaseGraphicsImageModel} from "../../bases/elements/BaseGraphics";

/**
 * Degree element model
 */
export class DegreeElementModel extends BaseGraphicsImageModel {
    rectangle: PIXI.Rectangle;
    options: Models.ElementOptions = {hasLine: true, lineColor: 0xFFFFFF, lineAlpha: 1, hasFill: false};
    lineWidth: number | undefined = 1;
}

export class RotateScaleElementModel extends BaseContainerImageModel {
    rotateElement: MedsurfDraw.RotateElement;
}

/**
 * Rotate Scale element
 */
export class RotateScaleElement extends BaseContainer<RotateScaleElementModel> {
    /**
     * Elements
     */
    protected _degreeGraphicElement: BaseGraphics<DegreeElementModel>;
    protected _degreeTextElement: MedsurfDraw.Text;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RotateScaleElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.rotateScale.zIndex;
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Degree element
        this._degreeGraphicElement = new BaseGraphics<DegreeElementModel>({
            image: this.image,
            rectangle: this.rotateElement.rectangle,
            options: { hasLine: true, lineColor: 0xFFFFAA, lineAlpha: 1, hasFill: false},
            lineWidth: 15
        });
        this.addChild(this._degreeGraphicElement);
        // -- Degree text
        this._degreeTextElement = new MedsurfDraw.Text({
            text: this._formatDegree(),
            style: new PIXI.TextStyle({
                fill: Design.rotateScale.textColor,
                fontSize: Design.rotateScale.textFontSize,
            })
        });
        this._degreeTextElement.anchor.set(1, 1);
        this.addChild(this._degreeTextElement);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Rotate
        this.rotateElement.rotateInteraction.on("onRotate", this.onRotate, this);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        let scale = this.imageScale.x * Design.rotateScale.scaleAspect;
        if (scale > 1) {
            scale = 1;
        }

        const lineWidth = Design.rotateScale.lineWidth / scale;
        const lineLength = Design.rotateScale.lineLength / scale;

        // const radius = this._radiusDelta - Math.min(rectangle.width, rectangle.height);
        let radius = this.rotateElement.rectangle.y - this.rotateElement.rectangle.height / 2 - Design.rotate.rotationDY;
        /*
        if (!this._isEllipse) {
            radius -= this.rectangle.height / 2
        }
        */

        //<editor-fold desc="Elements">
        // -- Degree graphic
        this._degreeGraphicElement.clear();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.beginFill(Design.rotateScale.fillColor, Design.rotateScale.fillAlpha);
        this._degreeGraphicElement.drawCircle(0, 0, Math.abs(radius));
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(0, radius);
        this._degreeGraphicElement.lineTo(0, radius + lineLength);
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(-radius * Math.cos(45 * Math.PI / 180), radius * Math.cos(45 * Math.PI / 180));
        this._degreeGraphicElement.lineTo((-radius - lineLength / 2) * Math.cos(45 * Math.PI / 180), (radius + lineLength / 2) * Math.cos(45 * Math.PI / 180));
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(-radius, 0);
        this._degreeGraphicElement.lineTo(-radius - lineLength, 0);
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(radius * Math.cos(135 * Math.PI / 180), radius * Math.cos(135 * Math.PI / 180));
        this._degreeGraphicElement.lineTo((radius + lineLength / 2) * Math.cos(135 * Math.PI / 180), (radius + lineLength / 2) * Math.cos(135 * Math.PI / 180));
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(0, -radius);
        this._degreeGraphicElement.lineTo(0, -radius - lineLength);
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(radius * Math.cos(45 * Math.PI / 180), -radius * Math.cos(45 * Math.PI / 180));
        this._degreeGraphicElement.lineTo((radius + lineLength / 2) * Math.cos(45 * Math.PI / 180), (-radius - lineLength / 2) * Math.cos(45 * Math.PI / 180));
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(radius, 0);
        this._degreeGraphicElement.lineTo(radius + lineLength, 0);
        this._degreeGraphicElement.endFill();

        this._degreeGraphicElement.lineStyle(lineWidth, Design.rotateScale.lineColor, Design.rotateScale.lineAlpha);
        this._degreeGraphicElement.moveTo(-radius * Math.cos(135 * Math.PI / 180), -radius * Math.cos(135 * Math.PI / 180));
        this._degreeGraphicElement.lineTo((-radius - lineLength / 2) * Math.cos(135 * Math.PI / 180), (-radius - lineLength / 2) * Math.cos(135 * Math.PI / 180));
        this._degreeGraphicElement.endFill();
        // -- Degree Text
        this._degreeTextElement.text = this._formatDegree();
        this._degreeTextElement.position.set(radius, radius);
        this._degreeTextElement.style.fontSize = Design.rotateScale.textFontSize / scale;
        // TODO remove this._degreeTextElement.resolution = this._textResolution;
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Degree graphic element
        if (this._degreeGraphicElement) {
            this._degreeGraphicElement.destroy(options);
        }
        // -- Degree text element
        if (this._degreeTextElement) {
            this._degreeTextElement.destroy(options);
        }
        //</editor-fold>

        //<editor-fold desc="Events">
        this.rotateElement.rotateInteraction.off("onRotate", this.onRotate, this);
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Private functions">
    protected _formatDegree(): string {
        let rotation = (Math.round((this.rotateElement.rotation / Math.PI * 180))) % 360;
        if (rotation < 0) {
            rotation = 360 + rotation;
        }
        return rotation + "°";
    }
    //</editor-fold>

    //<editor-fold desc="Rotate">
    /**
     * On rotate
     * @param degree
     */
    public onRotate(degree: number): void {
        this._degreeTextElement.text = this._formatDegree();
    }
    //</editor-fold>

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    public get rotateElement(): MedsurfDraw.RotateElement {
        return this.data.rotateElement;
    }

    public set rotateElement(value: MedsurfDraw.RotateElement) {
        this.data.rotateElement = value;
    }
    //</editor-fold>
}