import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";
import {debounce} from "debounce";

/**
 * Scale element type
 */
export enum ScaleElementType {
    NOT_SCALEABLE,
    SCALEABLE,
    TEXT
}

/**
 * Scale element model
 */
export class ScaleElementModel extends BaseContainerImageModel {
    type: ScaleElementType;
    rectangle: PIXI.Rectangle
}

/**
 * Scale element
 */
export class ScaleElement extends BaseContainer<ScaleElementModel> {
    /**
     * Methods
     */
    protected _debounceTypeChangedMethod: EventListener;

    /**
     * Elements
     */
    protected _rectangleElement: MedsurfDraw.Rectangle;
    protected _topLeftCircleElement: MedsurfDraw.Circle;
    protected _topRightCircleElement: MedsurfDraw.Circle;
    protected _bottomRightCircleElement: MedsurfDraw.Circle;
    protected _bottomLeftCircleElement: MedsurfDraw.Circle;
    protected _topLineElement: MedsurfDraw.StraightLine;
    protected _rightLineElement: MedsurfDraw.StraightLine;
    protected _bottomLineElement: MedsurfDraw.StraightLine;
    protected _leftLineElement: MedsurfDraw.StraightLine;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: ScaleElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.scale.zIndex;

        //<editor-fold desc="Methods">
        this._debounceTypeChangedMethod = debounce(this._typeChanged.bind(this), 20).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceTypeChanged", this._debounceTypeChangedMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceTypeChanged", this._debounceTypeChangedMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        //<editor-fold desc="Elements">
        // -- Rectangle element
        this._rectangleElement = new MedsurfDraw.Rectangle({
            rectangle: new PIXI.Rectangle(0, 0, 0, 0),
            options: {hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: false},
            lineWidth: Design.scale.lineWidth
        });
        this.addChild(this._rectangleElement);
        // -- Top left point element
        this._topLeftCircleElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.scale.cornerRadius),
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth
        });
        this._topLeftCircleElement.interactive = true;
        this._topLeftCircleElement.cursor = "scale-nwse";
        this._topLeftCircleElement.on("pointerover", this.onPointerOver, this);
        this._topLeftCircleElement.on("mousedown", this.onPointerDownXY, this);
        this._topLeftCircleElement.on("pointermove", this.onScale, this);
        this._topLeftCircleElement.on("mouseup", this.endScale, this);
        this._topLeftCircleElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._topLeftCircleElement);
        // -- Top right point element
        this._topRightCircleElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.scale.cornerRadius),
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth
        });
        this._topRightCircleElement.interactive = true;
        this._topRightCircleElement.cursor = "scale-nesw";
        this._topRightCircleElement.on("pointerover", this.onPointerOver, this);
        this._topRightCircleElement.on("mousedown", this.onPointerDownWidthY, this);
        this._topRightCircleElement.on("pointermove", this.onScale, this);
        this._topRightCircleElement.on("mouseup", this.endScale, this);
        this._topRightCircleElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._topRightCircleElement);
        // -- Bottom right point element
        this._bottomRightCircleElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.scale.cornerRadius),
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth
        });
        this._bottomRightCircleElement.interactive = true;
        this._bottomRightCircleElement.cursor = "scale-nwse";
        this._bottomRightCircleElement.on("pointerover", this.onPointerOver, this);
        this._bottomRightCircleElement.on("mousedown", this.onPointerDownWidthHeight, this);
        this._bottomRightCircleElement.on("pointermove", this.onScale, this);
        this._bottomRightCircleElement.on("mouseup", this.endScale, this);
        this._bottomRightCircleElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._bottomRightCircleElement);
        // -- Bottom left point element
        this._bottomLeftCircleElement = new MedsurfDraw.Circle({
            circle: new PIXI.Circle(0, 0, Design.scale.cornerRadius),
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth
        });
        this._bottomLeftCircleElement.interactive = true;
        this._bottomLeftCircleElement.cursor = "scale-nesw";
        this._bottomLeftCircleElement.on("pointerover", this.onPointerOver, this);
        this._bottomLeftCircleElement.on("mousedown", this.onPointerDownXHeight, this);
        this._bottomLeftCircleElement.on("pointermove", this.onScale, this);
        this._bottomLeftCircleElement.on("mouseup", this.endScale, this);
        this._bottomLeftCircleElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._bottomLeftCircleElement);
        // -- Top line element
        this._topLineElement = new MedsurfDraw.StraightLine({
            points: {
                start: new PIXI.Point(0, 0),
                end: new PIXI.Point(0, 0),
            },
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth,
        });
        this._topLineElement.interactive = true;
        this._topLineElement.cursor = "scale-v";
        this._topLineElement.on("pointerover", this.onPointerOver, this);
        this._topLineElement.on("mousedown", this.onPointerDownY, this);
        this._topLineElement.on("pointermove", this.onScale, this);
        this._topLineElement.on("mouseup", this.endScale, this);
        this._topLineElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._topLineElement);
        // -- Right line element
        this._rightLineElement = new MedsurfDraw.StraightLine({
            points: {
                start: new PIXI.Point(0, 0),
                end: new PIXI.Point(0, 0),
            },
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth,
        });
        this._rightLineElement.interactive = true;
        this._rightLineElement.cursor = "scale-h";
        this._rightLineElement.on("pointerover", this.onPointerOver, this);
        this._rightLineElement.on("mousedown", this.onPointerDownWidth, this);
        this._rightLineElement.on("pointermove", this.onScale, this);
        this._rightLineElement.on("mouseup", this.endScale, this);
        this._rightLineElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._rightLineElement);
        // -- Bottom line element
        this._bottomLineElement = new MedsurfDraw.StraightLine({
            points: {
                start: new PIXI.Point(0, 0),
                end: new PIXI.Point(0, 0),
            },
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth,
        });
        this._bottomLineElement.interactive = true;
        this._bottomLineElement.cursor = "scale-v";
        this._bottomLineElement.on("pointerover", this.onPointerOver, this);
        this._bottomLineElement.on("mousedown", this.onPointerDownHeight, this);
        this._bottomLineElement.on("pointermove", this.onScale, this);
        this._bottomLineElement.on("mouseup", this.endScale, this);
        this._bottomLineElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._bottomLineElement);
        // -- Left line element
        this._leftLineElement = new MedsurfDraw.StraightLine({
            points: {
                start: new PIXI.Point(0, 0),
                end: new PIXI.Point(0, 0),
            },
            options: { hasLine: true, lineColor: Design.scale.lineColorNormal, lineAlpha: Design.scale.lineAlpha, hasFill: true, fillColor: Design.scale.fillColorNormal, fillAlpha: Design.scale.fillAlpha},
            lineWidth: Design.scale.lineWidth,
        });
        this._leftLineElement.interactive = true;
        this._leftLineElement.cursor = "scale-h";
        this._leftLineElement.on("pointerover", this.onPointerOver, this);
        this._leftLineElement.on("mousedown", this.onPointerDownX, this);
        this._leftLineElement.on("pointermove", this.onScale, this);
        this._leftLineElement.on("mouseup", this.endScale, this);
        this._leftLineElement.on("pointerupoutside", this.endScale, this);
        this.addChild(this._leftLineElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        //<editor-fold desc="Elements">
        // -- Rectangle element
        if (this._rectangleElement) {
            this._rectangleElement.position.set(this.rectangle.x, this.rectangle.y);
            this._rectangleElement.rectangle = new PIXI.Rectangle(0, 0, this.rectangle.width, this.rectangle.height);
            this._rectangleElement.draw();
        }
        // -- Top left point element
        if (this._topLeftCircleElement) {
            this._topLeftCircleElement.position.set(this.rectangle.x, this.rectangle.y);
            this._topLeftCircleElement.draw();
        }
        // -- Top right point element
        if (this._topRightCircleElement) {
            this._topRightCircleElement.position.set(this.rectangle.x + this.rectangle.width, this.rectangle.y);
            this._topRightCircleElement.draw();
        }
        // -- Bottom right point element
        if (this._bottomRightCircleElement) {
            this._bottomRightCircleElement.position.set(this.rectangle.x + this.rectangle.width, this.rectangle.y + this.rectangle.height);
            this._bottomRightCircleElement.draw();
        }
        // -- Bottom left point element
        if (this._bottomLeftCircleElement) {
            this._bottomLeftCircleElement.position.set(this.rectangle.x, this.rectangle.y + this.rectangle.height);
            this._bottomLeftCircleElement.draw();
        }
        // -- Top line element
        if (this._topLineElement) {
            this._topLineElement.points = {
                start: new PIXI.Point(this._topLeftCircleElement.position.x + Design.scale.cornerRadius, this._topLeftCircleElement.position.y),
                end: new PIXI.Point(this._topRightCircleElement.position.x - Design.scale.cornerRadius, this._topRightCircleElement.position.y),
            };
            this._topLineElement.draw();
        }
        // -- Right line element
        if (this._rightLineElement) {
            this._rightLineElement.points = {
                start: new PIXI.Point(this._topRightCircleElement.position.x, this._topRightCircleElement.position.y + Design.scale.cornerRadius),
                end: new PIXI.Point(this._bottomRightCircleElement.position.x, this._bottomRightCircleElement.position.y - Design.scale.cornerRadius),
            };
            this._rightLineElement.draw();
        }
        // -- Bottom line element
        if (this._bottomLineElement) {
            this._bottomLineElement.points = {
                start: new PIXI.Point(this._bottomRightCircleElement.position.x - Design.scale.cornerRadius, this._bottomRightCircleElement.position.y),
                end: new PIXI.Point(this._bottomLeftCircleElement.position.x + Design.scale.cornerRadius, this._bottomLeftCircleElement.position.y),
            };
            this._bottomLineElement.draw();
        }
        // -- Left line element
        if (this._leftLineElement) {
            this._leftLineElement.points = {
                start: new PIXI.Point(this._bottomLeftCircleElement.position.x, this._bottomLeftCircleElement.position.y - Design.scale.cornerRadius),
                end: new PIXI.Point(this._topLeftCircleElement.position.x, this._topLeftCircleElement.position.y + Design.scale.cornerRadius),
            };
            this._leftLineElement.draw();
        }
        //</editor-fold>

        // Pivot
        this.pivot.set(this.rectangle.x + this.rectangle.width / 2, this.rectangle.y + this.rectangle.height / 2);
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Rectangle element
        if (this._rectangleElement) {
            this._rectangleElement.destroy(options);
        }
        // -- Top left point element
        if (this._topLeftCircleElement) {
            this._topLeftCircleElement.destroy(options);
        }
        // -- Top right point element
        if (this._topRightCircleElement) {
            this._topRightCircleElement.destroy(options);
        }
        // -- Bottom right point element
        if (this._bottomRightCircleElement) {
            this._bottomRightCircleElement.destroy(options);
        }
        // -- Bottom left point element
        if (this._bottomLeftCircleElement) {
            this._bottomLeftCircleElement.destroy(options);
        }
        // -- Top line element
        if (this._topLineElement) {
            this._topLineElement.destroy(options);
        }
        // -- Right line element
        if (this._rightLineElement) {
            this._rightLineElement.destroy(options);
        }
        // -- Bottom line element
        if (this._bottomLineElement) {
            this._bottomLineElement.destroy(options);
        }
        // -- Left line element
        if (this._leftLineElement) {
            this._leftLineElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Private functions">
    /**
     * Set lock aspect ratio
     * @private
     */
    protected _setLockAspectRatio(): void {
        // Elements
        // -- Rectangle element
        this._rectangleElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._rectangleElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Top left point element
        this._topLeftCircleElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._topLeftCircleElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Top right point element
        this._topRightCircleElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._topRightCircleElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Bottom right point element
        this._bottomRightCircleElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._bottomRightCircleElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Bottom left point element
        this._bottomLeftCircleElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._bottomLeftCircleElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Top line element
        this._topLineElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._topLineElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Right line element
        this._rightLineElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._rightLineElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Bottom line element
        this._bottomLineElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._bottomLineElement.options.fillColor = Design.scale.fillColorLockAspect;
        // -- Left line element
        this._leftLineElement.options.lineColor = Design.scale.lineColorLockAspect;
        this._leftLineElement.options.fillColor = Design.scale.fillColorLockAspect;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Unset lock aspect ratio
     * @private
     */
    protected _unsetLockAspectRatio(): void {
        // Elements
        // -- Rectangle element
        this._rectangleElement.options.lineColor = Design.scale.lineColorNormal;
        this._rectangleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Top left point element
        this._topLeftCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._topLeftCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Top right point element
        this._topRightCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._topRightCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Bottom right point element
        this._bottomRightCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._bottomRightCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Bottom left point element
        this._bottomLeftCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._bottomLeftCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Top line element
        this._topLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._topLineElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Right line element
        this._rightLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._rightLineElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Bottom line element
        this._bottomLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._bottomLineElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Left line element
        this._leftLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._leftLineElement.options.fillColor = Design.scale.fillColorNormal;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Set move relative
     * @private
     */
    protected _setMoveRelative(): void {
        // Elements
        // -- Rectangle element
        this._rectangleElement.options.lineColor = Design.scale.lineColorRelative;
        this._rectangleElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Top left point element
        this._topLeftCircleElement.options.lineColor = Design.scale.lineColorRelative;
        this._topLeftCircleElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Top right point element
        this._topRightCircleElement.options.lineColor = Design.scale.lineColorRelative;
        this._topRightCircleElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Bottom right point element
        this._bottomRightCircleElement.options.lineColor = Design.scale.lineColorRelative;
        this._bottomRightCircleElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Bottom left point element
        this._bottomLeftCircleElement.options.lineColor = Design.scale.lineColorRelative;
        this._bottomLeftCircleElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Top line element
        this._topLineElement.options.lineColor = Design.scale.lineColorRelative;
        this._topLineElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Right line element
        this._rightLineElement.options.lineColor = Design.scale.lineColorRelative;
        this._rightLineElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Bottom line element
        this._bottomLineElement.options.lineColor = Design.scale.lineColorRelative;
        this._bottomLineElement.options.fillColor = Design.scale.fillColorRelative;
        // -- Left line element
        this._leftLineElement.options.lineColor = Design.scale.lineColorRelative;
        this._leftLineElement.options.fillColor = Design.scale.fillColorRelative;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Unset move relative
     * @private
     */
    protected _unsetMoveRelative(): void {
        // Elements
        // -- Rectangle element
        this._rectangleElement.options.lineColor = Design.scale.lineColorNormal;
        this._rectangleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Top left point element
        this._topLeftCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._topLeftCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Top right point element
        this._topRightCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._topRightCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Bottom right point element
        this._bottomRightCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._bottomRightCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Bottom left point element
        this._bottomLeftCircleElement.options.lineColor = Design.scale.lineColorNormal;
        this._bottomLeftCircleElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Top line element
        this._topLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._topLineElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Right line element
        this._rightLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._rightLineElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Bottom line element
        this._bottomLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._bottomLineElement.options.fillColor = Design.scale.fillColorNormal;
        // -- Left line element
        this._leftLineElement.options.lineColor = Design.scale.lineColorNormal;
        this._leftLineElement.options.fillColor = Design.scale.fillColorNormal;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Type changed
     * @private
     */
    protected _typeChanged(): void {
        // -- Keyboard
        MedsurfDraw.Keyboard.events.remove("pressed_ShiftLeft", "onScaleElementLockAspectLeft");
        MedsurfDraw.Keyboard.events.remove("released_ShiftLeft", "onScaleElementUnlockAspectLeft");
        MedsurfDraw.Keyboard.events.remove("pressed_ControlLeft", "onScaleElementMoveRelativeLeft");
        MedsurfDraw.Keyboard.events.remove("released_ControlLeft", "onScaleElementUnmoveRelativeLeft");
        MedsurfDraw.Keyboard.events.remove("pressed_ShiftRight", "onScaleElementLockAspectRight");
        MedsurfDraw.Keyboard.events.remove("released_ShiftRight", "onScaleElementUnlockAspectRight");
        MedsurfDraw.Keyboard.events.remove("pressed_ControlLeft", "onScaleElementMoveRelativeRight");
        MedsurfDraw.Keyboard.events.remove("released_ControlLeft", "onScaleElementUnmoveRelativeRight");

        // Type
        if (this.type === ScaleElementType.TEXT) {
            // Elements
            // -- Rectangle element
            this._rectangleElement.showItem();
            // -- Top left point element
            this._topLeftCircleElement.hideItem();
            // -- Top right point element
            this._topRightCircleElement.hideItem();
            // -- Bottom right point element
            this._bottomRightCircleElement.hideItem();
            // -- Bottom left point element
            this._bottomLeftCircleElement.hideItem();
            // -- Top line element
            this._topLineElement.hideItem();
            // -- Right line element
            this._rightLineElement.hideItem();
            // -- Bottom line element
            this._bottomLineElement.hideItem();
            // -- Left line element
            this._leftLineElement.hideItem();
        } else {
            if (this.type === ScaleElementType.SCALEABLE) {
                // Elements
                // -- Rectangle element
                this._rectangleElement.hideItem();
                // -- Top left point element
                this._topLeftCircleElement.showItem();
                // -- Top right point element
                this._topRightCircleElement.showItem();
                // -- Bottom right point element
                this._bottomRightCircleElement.showItem();
                // -- Bottom left point element
                this._bottomLeftCircleElement.showItem();
                // -- Top line element
                this._topLineElement.showItem();
                // -- Right line element
                this._rightLineElement.showItem();
                // -- Bottom line element
                this._bottomLineElement.showItem();
                // -- Left line element
                this._leftLineElement.showItem();
            } else {
                // Elements
                // -- Rectangle element
                this._rectangleElement.showItem();
                // -- Top left point element
                this._topLeftCircleElement.hideItem();
                // -- Top right point element
                this._topRightCircleElement.hideItem();
                // -- Bottom right point element
                this._bottomRightCircleElement.hideItem();
                // -- Bottom left point element
                this._bottomLeftCircleElement.hideItem();
                // -- Top line element
                this._topLineElement.hideItem();
                // -- Right line element
                this._rightLineElement.hideItem();
                // -- Bottom line element
                this._bottomLineElement.hideItem();
                // -- Left line element
                this._leftLineElement.hideItem();
            }

            // -- Keyboard
            MedsurfDraw.Keyboard.events.on("pressed_ShiftLeft", "onScaleElementLockAspectLeft", this._setLockAspectRatio.bind(this));
            MedsurfDraw.Keyboard.events.on("released_ShiftLeft", "onScaleElementUnlockAspectLeft", this._unsetLockAspectRatio.bind(this));
            MedsurfDraw.Keyboard.events.on("pressed_ControlLeft", "onScaleElementMoveRelativeLeft", this._setMoveRelative.bind(this));
            MedsurfDraw.Keyboard.events.on("released_ControlLeft", "onScaleElementUnmoveRelativeLeft", this._unsetMoveRelative.bind(this));
            MedsurfDraw.Keyboard.events.on("pressed_ShiftRight", "onScaleElementLockAspectRight", this._setLockAspectRatio.bind(this));
            MedsurfDraw.Keyboard.events.on("released_ShiftRight", "onScaleElementUnlockAspectRight", this._unsetLockAspectRatio.bind(this));
            MedsurfDraw.Keyboard.events.on("pressed_ControlRight", "onScaleElementMoveRelativeRight", this._setMoveRelative.bind(this));
            MedsurfDraw.Keyboard.events.on("released_ControlRight", "onScaleElementUnmoveRelativeRight", this._unsetMoveRelative.bind(this));
        }
    }
    //</editor-fold>

    //<editor-fold desc="Scale">
    /**
     * On pointer over
     * @param event
     */
    public onPointerOver(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
    }

    /**
     * On pointer down y
     * @param event
     */
    public onPointerDownY(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.Y);
    }

    /**
     * On pointer down width
     * @param event
     */
    public onPointerDownWidth(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.WIDTH);
    }

    /**
     * On pointer down height
     * @param event
     */
    public onPointerDownHeight(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.HEIGHT);
    }

    /**
     * On pointer down x
     * @param event
     */
    public onPointerDownX(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.X);
    }

    /**
     * On pointer down x y
     * @param event
     */
    public onPointerDownXY(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.X_Y);
    }

    /**
     * On pointer down width y
     * @param event
     */
    public onPointerDownWidthY(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.WIDTH_Y);
    }

    /**
     * On pointer down width height
     * @param event
     */
    public onPointerDownWidthHeight(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.WIDTH_HEIGHT);
    }

    /**
     * On pointer down x height
     * @param event
     */
    public onPointerDownXHeight(event: PIXI.InteractionEvent): void {
        this.emit("startScale", event, MedsurfDraw.ScaleSideTypes.X_HEIGHT);
    }

    /**
     * On scale
     * @param event
     */
    public onScale(event: PIXI.InteractionEvent): void {
        this.emit("onScale", event);
    }

    /**
     * End scale
     */
    public endScale(event: PIXI.InteractionEvent): void {
        this.emit("endScale", event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get type
     */
    public get type(): ScaleElementType {
        return this.data.type;
    }

    /**
     * Set type
     * @param value
     */
    public set type(value: ScaleElementType) {
        this.data.type = value;

        // Type changed
        this.emit("debounceTypeChanged");
    }

    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }
    //</editor-fold>
}