import * as MedsurfDraw from "../../public-api";
import {BaseGenerator} from "../../bases/generators/BaseGenerator";
import {FillGeneratorModel} from "../fills/FillGenerator";

/**
 * Interactive area generator states
 */
export enum InteractiveAreaGeneratorStates {
    FILL,
    TEXT
}

/**
 * Interactive area generator model
 */
export class InteractiveAreaGeneratorModel extends FillGeneratorModel {
}

/**
 * Interactive area generator
 */
export class InteractiveAreaGenerator extends BaseGenerator<MedsurfDraw.Image, InteractiveAreaGeneratorModel> {
    /**
     * Members
     */
    protected _state: InteractiveAreaGeneratorStates;

    /**
     * Generators
     */
    protected _fillGenerator: MedsurfDraw.FillGenerator;
    protected _textPrimitiveGenerator: MedsurfDraw.TextPrimitiveGenerator;

    /**
     * Elements
     */
    protected _fillCollectionElement: MedsurfDraw.FillCollection;
    protected _lineElement: MedsurfDraw.Line;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: InteractiveAreaGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._state = InteractiveAreaGeneratorStates.FILL;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        // Setup generator
        this._setupGenerator();
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Generators">
        // -- Interactive area generator
        if (this._fillGenerator) {
            this._fillGenerator.removeAllListeners();
        }
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Line element
        if (this._lineElement && this._lineElement.modeInteraction.lastMode.endsWith('_new')) {
            this._lineElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Interactive area generator
        if (this._fillGenerator) {
            this._fillGenerator.destroy();
        }
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Setup interactive area
        this._fillCollectionElement.onButtonSetInteractive(true);
        this._fillCollectionElement.model.interactiveItems = this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_select'))
            .map((imageObject) => imageObject.name) as string[];

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Setup generator
     */
    protected _setupGenerator(positionPoint?: MedsurfDraw.PositionPoint): void {
        if (this._state === InteractiveAreaGeneratorStates.FILL) {
            // Generator
            // -- interactive area generator
            this._fillGenerator = new MedsurfDraw.FillGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: this.stickMode,
                bezierCourve: this.bezierCourve,
                startElement: positionPoint
            });
            this._fillGenerator.once("startGenerator", this.start, this);
            this._fillGenerator.once("endGenerator", this.endInteractiveAreaGenerator, this);
            this._fillGenerator.once("abortGenerator", this.abortInteractiveAreaGenerator, this);
        } else {
            // Generator
            // -- Text primitive generator
            this._textPrimitiveGenerator = new MedsurfDraw.TextPrimitiveGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                startElement: positionPoint,
                rotateWithLine: false,
                withScaling: false,
                triggerImmediateEndDraw: true
            });
            this._textPrimitiveGenerator.once("startGenerator", this.startTextPrimitiveGenerator, this);
            this._textPrimitiveGenerator.once("endGenerator", this.endTextPrimitiveGenerator, this);
            this._textPrimitiveGenerator.once("abortGenerator", this.abortTextPrimitiveGenerator, this);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Destroy text
     */
    protected destroyText(): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Line element
        if (this._lineElement) {
            this._lineElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.destroy();
        }
        //</editor-fold>
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    /**
     * End interactive area generator
     */
    public endInteractiveAreaGenerator(): void {
        // Elements
        if (!this._fillCollectionElement) {
            this._fillCollectionElement = this._fillGenerator.fillCollectionElement;
        }

        // Members
        this._state = InteractiveAreaGeneratorStates.TEXT;

        // Setup generator
        this._setupGenerator();
    }

    /**
     * Abort interactive area generator
     */
    public abortInteractiveAreaGenerator(): void {
        // Destroy
        this.destroy();

        // Emit end generator
        this.emit("abortGenerator");
    }

    /**
     * Start text primitive generator
     */
    public startTextPrimitiveGenerator(): void {
        //<editor-fold desc="Elements">
        // -- Line element
        this._lineElement = MedsurfDraw.Line.getInstance(this.target,
            this._fillGenerator.fillCollectionElement.name as string,
            this._textPrimitiveGenerator.positionPointElement.name as string, false,
            this.target.dimensions.width);
        this.target.addChild(this._lineElement);
        this._lineElement.controlDrawItem(true);
        this._lineElement.showItem();
        //</editor-fold>

        // Add Model
        this.target.addImageObject(this._lineElement.model, this.data.layerGroup, this.data.image);
    }

    /**
     * End text primitive generator
     */
    public endTextPrimitiveGenerator(): void {
        // Set mode
        this._lineElement.modeInteraction.setModeItem('draw_select_new');

        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort text primitive generator
     */
    public abortTextPrimitiveGenerator(): void {
        // Destroy
        this.destroyText();

        // Emit end generator
        this.emit("endGenerator");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection | MedsurfDraw.TextPrimitive> {
        if (this._state === InteractiveAreaGeneratorStates.FILL) {
            return this._fillGenerator.drawInteraction;
        } else {
            return this._textPrimitiveGenerator.drawInteraction;
        }
    }

    /**
     * Get bezier courve
     */
    public get bezierCourve(): boolean | undefined {
        return this.data.bezierCourve;
    }

    /**
     * Set bezier courve
     * @param value
     */
    public set bezierCourve(value: boolean | undefined) {
        this.data.bezierCourve = value;
    }

    /**
     * Get stick mode
     */
    public get stickMode(): MedsurfDraw.StickMode {
        return this.data.stickMode;
    }

    /**
     * Set stick mode
     * @param value
     */
    public set stickMode(value: MedsurfDraw.StickMode) {
        this.data.stickMode = value;
    }
    //</editor-fold>
}