import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";
import {debounce} from "debounce";

/**
 * Marker generator states
 */
export enum MarkerGeneratorStates {
    LINE,
    TEXT
}

/**
 * Marker generator model
 */
export class MarkerGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    layerGroup: Models.LayerGroup;
    image: Models.Image;
    startElement?: MedsurfDraw.PositionPoint
}

/**
 * Marker generator
 */
export class MarkerGenerator extends BaseGenerator<MedsurfDraw.Image, MarkerGeneratorModel> {
    /**
     * Members
     */
    protected _enableKeyEvents: boolean;
    protected _state: MarkerGeneratorStates;

    /**
     * Generators
     */
    protected _lineGenerator: MedsurfDraw.LineGenerator;
    protected _textPrimitiveGenerator: MedsurfDraw.TextPrimitiveGenerator;

    /**
     * Methods
     */
    protected _debounceSetupGeneratorMethod: any;

    /**
     * Elements
     */
    protected _startPositionPointElement: MedsurfDraw.PositionPoint;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: MarkerGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._enableKeyEvents = false;
        this._state = MarkerGeneratorStates.LINE;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        //<editor-fold desc="Methods">
        this._debounceSetupGeneratorMethod = debounce(this._setupGenerator.bind(this), 50).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Keyboard
        MedsurfDraw.Keyboard.events.on("pressed_ControlLeft", "onMarkerGeneratorControlLeft", this.onKeyPressed.bind(this));
        MedsurfDraw.Keyboard.events.on("pressed_ControlRight", "onMarkerGeneratorControlRight", this.onKeyPressed.bind(this));
        MedsurfDraw.Keyboard.events.on("released_ControlLeft", "onMarkerGeneratorControlLeftEnd", this.onKeyReleased.bind(this));
        MedsurfDraw.Keyboard.events.on("released_ControlRight", "onMarkerGeneratorControlRightEnd", this.onKeyReleased.bind(this));
        // -- Debounce
        this.on("debounceSetupGenerator", this._debounceSetupGeneratorMethod);
        //</editor-fold>

        // Setup generator
        this.emit("debounceSetupGenerator", this.startElement);
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Keyboard
        MedsurfDraw.Keyboard.events.remove("pressed_ControlLeft", "onMarkerGeneratorControlLeft");
        MedsurfDraw.Keyboard.events.remove("released_ControlLeft", "onMarkerGeneratorControlLeftEnd");
        MedsurfDraw.Keyboard.events.remove("pressed_ControlRight", "onMarkerGeneratorControlRight");
        MedsurfDraw.Keyboard.events.remove("released_ControlRight", "onMarkerGeneratorControlRightEnd");
        // -- Debounce
        this.off("debounceSetupGenerator", this._debounceSetupGeneratorMethod);
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Line generator
        if (this._lineGenerator) {
            this._lineGenerator.removeAllListeners();
        }
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Generators">
        // -- Line generator
        if (this._lineGenerator) {
            this._lineGenerator.destroy();
        }
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Setup selftest
        this._startPositionPointElement.onButtonSetSelftest(true);
        this._startPositionPointElement.model.selftestItems = this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_select'))
            .map((imageObject) => imageObject.name) as string[];

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Setup generator
     */
    protected _setupGenerator(positionPoint?: MedsurfDraw.PositionPoint): void {
        if (this._state === MarkerGeneratorStates.LINE) {
            // Generator
            // -- Line generator
            this._lineGenerator = new MedsurfDraw.LineGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                bezierCourve: false,
                stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                startElement: positionPoint
            });
            this._lineGenerator.once("startGenerator", this.start, this);
            this._lineGenerator.once("endGenerator", this.endLineGenerator, this);
            this._lineGenerator.once("abortGenerator", this.abortLineGenerator, this);
        } else {
            // Generator
            // -- Text primitive generator
            this._textPrimitiveGenerator = new MedsurfDraw.TextPrimitiveGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                startElement: positionPoint,
                rotateWithLine: false,
                withScaling: false,
                triggerImmediateEndDraw: false
            });
            this._textPrimitiveGenerator.once("endGenerator", this.endTextPrimitiveGenerator, this);
            this._textPrimitiveGenerator.once("abortGenerator", this.abortTextPrimitiveGenerator, this);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    /**
     * End line generator
     */
    public endLineGenerator(): void {
        // Elements
        if (!this._startPositionPointElement) {
            this._startPositionPointElement = this._lineGenerator.startPositionPointElement as MedsurfDraw.PositionPoint;
        }

        const primitives = (this._lineGenerator.endPositionPointElement as MedsurfDraw.PositionPoint).model.primitives || [] as any [];
        if (primitives.filter((p: any) => p.type === Models.ImageObjectType.TEXT).length > 0) {
            // Emit end generator
            this.emit("endGenerator");

            // End
            this.end();
        } else {
            if ((this._lineGenerator.endPositionPointElement as MedsurfDraw.PositionPoint).modeInteraction.lastMode !== 'draw_select') {
                // Check if key pressed
                if (!(MedsurfDraw.Keyboard.isKeyDown("ControlLeft") || MedsurfDraw.Keyboard.isKeyDown("ControlRight"))) {
                    // Members
                    this._enableKeyEvents = true;

                    // Members
                    this._state = MarkerGeneratorStates.TEXT;
                }
            }

            // Setup generator
            this.emit("debounceSetupGenerator", this._lineGenerator.endPositionPointElement);
        }
    }

    /**
     * Abort line generator
     */
    public abortLineGenerator(): void {
        // Emit end generator
        this.emit("abortGenerator");

        // Destroy
        this.destroy();
    }

    /**
     * End text primitive generator
     */
    public endTextPrimitiveGenerator(): void {
        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort text primitive generator
     */
    public abortTextPrimitiveGenerator(): void {
        // Emit end generator
        this.emit("abortGenerator");

        // Destroy
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Keyboard">
    /**
     * On key pressed
     * @param keyCode
     * @param event
     */
    public onKeyPressed(keyCode: string, event: KeyboardEvent): void {
        event.preventDefault();

        if (!this._enableKeyEvents || this._state === MarkerGeneratorStates.LINE) {
            return;
        }

        // Get position point
        const positionPoint = this._textPrimitiveGenerator.startElement as MedsurfDraw.PositionPoint;

        // Generator
        // -- Text primitive generator
        this._textPrimitiveGenerator.destroy();

        // Members
        this._state = MarkerGeneratorStates.LINE;

        // Setup generator
        this.emit("debounceSetupGenerator", positionPoint);
    }

    /**
     * On key released
     * @param keyCode
     * @param event
     */
    public onKeyReleased(keyCode: string, event: KeyboardEvent): void {
        event.preventDefault();

        if (!this._enableKeyEvents || this._state === MarkerGeneratorStates.TEXT) {
            return;
        }

        // Get position point
        const positionPoint = this._lineGenerator.startElement as MedsurfDraw.PositionPoint;

        // Generator
        // -- Line generator
        this._lineGenerator.destroyEnd();

        // Members
        this._state = MarkerGeneratorStates.TEXT;

        // Setup generator
        this.emit("debounceSetupGenerator", positionPoint);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection | MedsurfDraw.TextPrimitive> {
        if (this._state === MarkerGeneratorStates.LINE) {
            return this._lineGenerator.drawInteraction;
        } else {
            return this._textPrimitiveGenerator.drawInteraction;
        }
    }

    /**
     * Get start element
     */
    public get startElement(): MedsurfDraw.PositionPoint | undefined {
        return this.data.startElement;
    }

    /**
     * Set start element
     * @param value
     */
    public set startElement(value: MedsurfDraw.PositionPoint | undefined) {
        this.data.startElement = value;
    }
    //</editor-fold>
}