import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseGenerator} from "../../bases/generators/BaseGenerator";
import {MarkerGeneratorModel} from "./MarkerGenerator";

/**
 * Primitive Marker generator states
 */
export enum PrimitiveMarkerGeneratorStates {
    PRIMITIVE,
    MARKER
}

/**
 * Primitive Marker generator model
 */
export class PrimitiveMarkerGeneratorModel extends MarkerGeneratorModel {
    type: Models.ImageObjectType.ARROW | Models.ImageObjectType.ELLIPSE | Models.ImageObjectType.RECTANGLE | Models.ImageObjectType.TEXT;
    rotateWithLine: boolean;
}

/**
 * Arrow Marker generator model
 */
export class ArrowMarkerGeneratorModel extends PrimitiveMarkerGeneratorModel {
    hasWedgePoint: boolean;
}

/**
 * Primitive Marker generator
 */
export class PrimitiveMarkerGenerator extends BaseGenerator<MedsurfDraw.Image, PrimitiveMarkerGeneratorModel> {
    /**
     * Members
     */
    protected _enableKeyEvents: boolean;
    protected _state: PrimitiveMarkerGeneratorStates;

    /**
     * Generators
     */
    protected _primitiveGenerator: MedsurfDraw.ArrowPrimitiveGenerator | MedsurfDraw.EllipsePrimitiveGenerator | MedsurfDraw.RectanglePrimitiveGenerator | MedsurfDraw.TextPrimitiveGenerator;
    protected _markerGenerator: MedsurfDraw.MarkerGenerator;

    /**
     * Elements
     */
    protected _startPositionPointElement: MedsurfDraw.PositionPoint;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: PrimitiveMarkerGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._enableKeyEvents = false;
        this._state = PrimitiveMarkerGeneratorStates.PRIMITIVE;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Keyboard
        /* TODO remove
        MedsurfDraw.Keyboard.events.on("pressed_ControlLeft", "onMarkerGeneratorControlLeft", this.onKeyPressed.bind(this));
        MedsurfDraw.Keyboard.events.on("pressed_ControlRight", "onMarkerGeneratorControlRight", this.onKeyPressed.bind(this));
        MedsurfDraw.Keyboard.events.on("released_ControlLeft", "onMarkerGeneratorControlLeftEnd", this.onKeyReleased.bind(this));
        MedsurfDraw.Keyboard.events.on("released_ControlRight", "onMarkerGeneratorControlRightEnd", this.onKeyReleased.bind(this));
        */
        //</editor-fold>

        // Setup generator
        this._setupGenerator();
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Keyboard
        /* TODO remove
        MedsurfDraw.Keyboard.events.remove("pressed_ControlLeft", "onPrimitiveMarkerGeneratorControlLeft");
        MedsurfDraw.Keyboard.events.remove("released_ControlLeft", "onPrimitiveMarkerGeneratorControlLeftEnd");
        MedsurfDraw.Keyboard.events.remove("pressed_ControlRight", "onPrimitiveMarkerGeneratorControlRight");
        MedsurfDraw.Keyboard.events.remove("released_ControlRight", "onPrimitiveMarkerGeneratorControlRightEnd");
        */
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Primitive generator
        if (this._primitiveGenerator) {
            this._primitiveGenerator.removeAllListeners();
        }
        // -- Marker generator
        if (this._markerGenerator) {
            this._markerGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Generators">
        // -- Primitive generator
        if (this._primitiveGenerator) {
            this._primitiveGenerator.destroy();
        }
        // -- Marker generator
        if (this._markerGenerator) {
            this._markerGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Setup selftest
        this._startPositionPointElement.onButtonSetSelftest(true);
        this._startPositionPointElement.model.selftestItems = this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_select'))
            .map((imageObject) => imageObject.name) as string[];

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Setup generator
     */
    protected _setupGenerator(positionPoint?: MedsurfDraw.PositionPoint): void {
        if (this._state === PrimitiveMarkerGeneratorStates.PRIMITIVE) {
            // Generator
            // -- Primitive generator
            switch (this.type) {
                case Models.ImageObjectType.ARROW:
                    this._primitiveGenerator = new MedsurfDraw.ArrowPrimitiveGenerator({
                        target: this.target,
                        layerGroup: this.data.layerGroup,
                        image: this.data.image,
                        stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                        startElement: positionPoint,
                        hasWedgePoint: (this.data as ArrowMarkerGeneratorModel).hasWedgePoint,
                        rotateWithLine: this.data.rotateWithLine,
                        withScaling: false,
                        triggerImmediateEndDraw: true
                    });
                    break;
                case Models.ImageObjectType.ELLIPSE:
                    this._primitiveGenerator = new MedsurfDraw.EllipsePrimitiveGenerator({
                        target: this.target,
                        layerGroup: this.data.layerGroup,
                        image: this.data.image,
                        stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                        startElement: positionPoint,
                        rotateWithLine: this.data.rotateWithLine,
                        withScaling: false,
                        triggerImmediateEndDraw: true
                    });
                    break;
                case Models.ImageObjectType.RECTANGLE:
                    this._primitiveGenerator = new MedsurfDraw.RectanglePrimitiveGenerator({
                        target: this.target,
                        layerGroup: this.data.layerGroup,
                        image: this.data.image,
                        stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                        startElement: positionPoint,
                        rotateWithLine: this.data.rotateWithLine,
                        withScaling: false,
                        triggerImmediateEndDraw: true
                    });
                    break;
                case Models.ImageObjectType.TEXT:
                    this._primitiveGenerator = new MedsurfDraw.TextPrimitiveGenerator({
                        target: this.target,
                        layerGroup: this.data.layerGroup,
                        image: this.data.image,
                        stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                        startElement: positionPoint,
                        rotateWithLine: this.data.rotateWithLine,
                        withScaling: false,
                        triggerImmediateEndDraw: true
                    });
                    break;
                default:
                    this._state = PrimitiveMarkerGeneratorStates.MARKER;
                    return this._setupGenerator(positionPoint);
            }
            this._primitiveGenerator.once("startGenerator", this.start, this);
            this._primitiveGenerator.once("endGenerator", this.endPrimitiveGenerator, this);
            this._primitiveGenerator.once("abortGenerator", this.abortPrimitiveGenerator, this);
        } else {
            // Generator
            // -- Marker generator
            this._markerGenerator = new MedsurfDraw.MarkerGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                startElement: this._startPositionPointElement
            });
            this._markerGenerator.once("endGenerator", this.endMarkerGenerator, this);
            this._markerGenerator.once("abortGenerator", this.abortMarkerGenerator, this);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    /**
     * End line generator
     */
    public endPrimitiveGenerator(): void {
        // Elements
        if (!this._startPositionPointElement) {
            this._startPositionPointElement = this._primitiveGenerator.positionPointElement;
        }

        // Check if key pressed
        if (!(MedsurfDraw.Keyboard.isKeyDown("ControlLeft") || MedsurfDraw.Keyboard.isKeyDown("ControlRight"))) {
            // Members
            this._enableKeyEvents = true;

            // Members
            this._state = PrimitiveMarkerGeneratorStates.MARKER;
        }

        // Setup generator
        this._setupGenerator(this._primitiveGenerator.positionPointElement);
    }

    /**
     * Abort line generator
     */
    public abortPrimitiveGenerator(): void {
        // Emit end generator
        this.emit("abortGenerator");

        // Destroy
        this.destroy();
    }

    /**
     * End text primitive generator
     */
    public endMarkerGenerator(): void {
        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort text primitive generator
     */
    public abortMarkerGenerator(): void {
        // Emit end generator
        this.emit("abortGenerator");

        // Destroy
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Keyboard">
    /**
     * On key pressed
     * @param keyCode
     * @param event
     */
    public onKeyPressed(keyCode: string, event: KeyboardEvent): void {
        /* TODO remove
        event.preventDefault();

        if (!this._enableKeyEvents || this._state === PMarkerGeneratorStates.LINE) {
            return;
        }

        // Get position point
        const positionPoint = this._textPrimitiveGenerator.startElement as MedsurfDraw.PositionPoint;

        // Generator
        // -- Text primitive generator
        this._textPrimitiveGenerator.destroy();

        // Members
        this._state = MarkerGeneratorStates.LINE;

        // Setup generator
        debounce(this._setupGenerator.bind(this, positionPoint), 50)();
        */
    }

    /**
     * On key released
     * @param keyCode
     * @param event
     */
    public onKeyReleased(keyCode: string, event: KeyboardEvent): void {
        /* TODO remove
        event.preventDefault();

        if (!this._enableKeyEvents || this._state === MarkerGeneratorStates.TEXT) {
            return;
        }

        // Get position point
        const positionPoint = this._lineGenerator.startElement as MedsurfDraw.PositionPoint;

        // Generator
        // -- Line generator
        this._lineGenerator.destroyEnd();

        // Members
        this._state = MarkerGeneratorStates.TEXT;

        // Setup generator
        debounce(this._setupGenerator.bind(this, positionPoint), 50)();
        */
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection | MedsurfDraw.TextPrimitive> {
        if (this._state === PrimitiveMarkerGeneratorStates.PRIMITIVE) {
            // @ts-ignore TODO
            return this._primitiveGenerator.drawInteraction; // as MedsurfDraw.DrawInteraction<unknown>;
        } else {
            return this._markerGenerator.drawInteraction;
        }
    }

    /**
     * Get type
     */
    public get type(): Models.ImageObjectType.ARROW | Models.ImageObjectType.ELLIPSE | Models.ImageObjectType.RECTANGLE | Models.ImageObjectType.TEXT {
        return this.data.type;
    }

    /**
     * Set type
     * @param value
     */
    public set type(value: Models.ImageObjectType.ARROW | Models.ImageObjectType.ELLIPSE | Models.ImageObjectType.RECTANGLE | Models.ImageObjectType.TEXT) {
        this.data.type = value;
    }
    //</editor-fold>
}