import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";

/**
 * Rectangle generator states
 */
export enum RectangleGeneratorStates {
    RECTANGLE,
    TEXT
}

/**
 * Rectangle generator model
 */
export class RectangleGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    layerGroup: Models.LayerGroup;
    image: Models.Image;
}

/**
 * Rectangle generator
 */
export class RectangleGenerator extends BaseGenerator<MedsurfDraw.Image, RectangleGeneratorModel> {
    /**
     * Members
     */
    protected _state: RectangleGeneratorStates;

    /**
     * Generators
     */
    protected _rectanglePrimitiveGenerator: MedsurfDraw.RectanglePrimitiveGenerator;
    protected _textPrimitiveGenerator: MedsurfDraw.TextPrimitiveGenerator;

    /**
     * Elements
     */
    protected _startPositionPointElement: MedsurfDraw.PositionPoint;
    protected _lineElement: MedsurfDraw.Line;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RectangleGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._state = RectangleGeneratorStates.RECTANGLE;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        // Setup generator
        this._setupGenerator();
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Generators">
        // -- Rectangle primitive generator
        if (this._rectanglePrimitiveGenerator) {
            this._rectanglePrimitiveGenerator.removeAllListeners();
        }
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Line element
        if (this._lineElement && this._lineElement.modeInteraction.lastMode.endsWith('new')) {
            this._lineElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Rectangle primitive generator
        if (this._rectanglePrimitiveGenerator) {
            this._rectanglePrimitiveGenerator.destroy();
        }
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Setup selftest
        this._startPositionPointElement.onButtonSetSelftest(true);
        this._startPositionPointElement.model.selftestItems = this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_select'))
            .map((imageObject) => imageObject.name) as string[];

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Setup generator
     */
    protected _setupGenerator(positionPoint?: MedsurfDraw.PositionPoint): void {
        if (this._state === RectangleGeneratorStates.RECTANGLE) {
            // Generator
            // -- Rectangle primitive generator
            this._rectanglePrimitiveGenerator = new MedsurfDraw.RectanglePrimitiveGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                startElement: positionPoint,
                rotateWithLine: false,
                withScaling: true,
                triggerImmediateEndDraw: true
            });
            this._rectanglePrimitiveGenerator.once("startGenerator", this.start, this);
            this._rectanglePrimitiveGenerator.once("endGenerator", this.endRectanglePrimitiveGenerator, this);
            this._rectanglePrimitiveGenerator.once("abortGenerator", this.abortRectanglePrimitiveGenerator, this);
        } else {
            // Generator
            // -- Text primitive generator
            this._textPrimitiveGenerator = new MedsurfDraw.TextPrimitiveGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: MedsurfDraw.StickMode.NOELEMENTS,
                startElement: positionPoint,
                rotateWithLine: false,
                withScaling: false,
                triggerImmediateEndDraw: true
            });
            this._textPrimitiveGenerator.once("startGenerator", this.startTextPrimitiveGenerator, this);
            this._textPrimitiveGenerator.once("endGenerator", this.endTextPrimitiveGenerator, this);
            this._textPrimitiveGenerator.once("abortGenerator", this.abortTextPrimitiveGenerator, this);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Destroy text
     */
    protected destroyText(): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Line element
        if (this._lineElement) {
            this._lineElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Text primitive generator
        if (this._textPrimitiveGenerator) {
            this._textPrimitiveGenerator.destroy();
        }
        //</editor-fold>
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    /**
     * End rectangle primitive generator
     */
    public endRectanglePrimitiveGenerator(): void {
        // Elements
        if (!this._startPositionPointElement) {
            this._startPositionPointElement = this._rectanglePrimitiveGenerator.positionPointElement;
        }

        // Members
        this._state = RectangleGeneratorStates.TEXT;

        // Setup generator
        this._setupGenerator();
    }

    /**
     * Abort rectangle primitive generator
     */
    public abortRectanglePrimitiveGenerator(): void {
        // Destroy
        this.destroy();

        // Emit end generator
        this.emit("abortGenerator");
    }

    /**
     * Start text primitive generator
     */
    public startTextPrimitiveGenerator(): void {
        //<editor-fold desc="Elements">
        // -- Line element
        this._lineElement = MedsurfDraw.Line.getInstance(this.target,
            this._rectanglePrimitiveGenerator.positionPointElement.name as string,
            this._textPrimitiveGenerator.positionPointElement.name as string, false,
            this.target.dimensions.width);
        this.target.addChild(this._lineElement);
        this._lineElement.controlDrawItem(true);
        this._lineElement.showItem();
        //</editor-fold>

        // Add Model
        this.target.addImageObject(this._lineElement.model, this.data.layerGroup, this.data.image);
    }

    /**
     * End text primitive generator
     */
    public endTextPrimitiveGenerator(): void {
        // Set mode
        this._lineElement.modeInteraction.setModeItem('draw_select_new');

        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort text primitive generator
     */
    public abortTextPrimitiveGenerator(): void {
        // Destroy
        this.destroyText();

        // Emit end generator
        this.emit("endGenerator");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.RectanglePrimitive | MedsurfDraw.TextPrimitive> {
        if (this._state === RectangleGeneratorStates.RECTANGLE) {
            return this._rectanglePrimitiveGenerator.drawInteraction;
        } else {
            return this._textPrimitiveGenerator.drawInteraction;
        }
    }
    //</editor-fold>
}