import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";

/**
 * Line generator states
 */
export enum LineGeneratorStates {
    START,
    END,
    FINISH
}

/**
 * Line generator model
 */
export class LineGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    layerGroup: Models.LayerGroup;
    image: Models.Image;
    stickMode: MedsurfDraw.StickMode;
    bezierCourve?: boolean;
    startElement?: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
}

/**
 * Line generator
 */
export class LineGenerator extends BaseGenerator<MedsurfDraw.Image, LineGeneratorModel> {
    /**
     * Members
     */
    protected _enableKeyEvents: boolean;
    protected _state: LineGeneratorStates;

    /**
     * Generators
     */
    protected _startPositionPointGenerator: MedsurfDraw.PositionPointGenerator;
    protected _endPositionPointGenerator: MedsurfDraw.PositionPointGenerator;

    /**
     * Elements
     */
    protected _lineElement: MedsurfDraw.Line;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: LineGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._enableKeyEvents = false;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Keyboard
        MedsurfDraw.Keyboard.events.on("pressed_ShiftLeft", "onScaleElementSetBezierLeft", this.onPressedShiftKey.bind(this));
        MedsurfDraw.Keyboard.events.on("released_ShiftLeft", "onScaleElementUnsetBezierLeft", this.onReleaseShiftKey.bind(this));
        MedsurfDraw.Keyboard.events.on("pressed_ShiftRight", "onScaleElementSetBezierRight", this.onPressedShiftKey.bind(this));
        MedsurfDraw.Keyboard.events.on("released_ShiftRight", "onScaleElementUnsetBezierRight", this.onReleaseShiftKey.bind(this));
        //</editor-fold>

        //<editor-fold desc="Generators">
        if (!this.startElement) {
            // -- Start position point generator
            // Members
            this._state = LineGeneratorStates.START;

            // Setup generator
            this._setupGenerator();
        } else {
            // -- End position point generator
            // Members
            this._state = LineGeneratorStates.END;

            // Setup generator
            this._setupGenerator();
        }
        //</editor-fold>
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Keyboard
        MedsurfDraw.Keyboard.events.remove("pressed_ShiftLeft", "onScaleElementSetBezierLeft");
        MedsurfDraw.Keyboard.events.remove("released_ShiftLeft", "onScaleElementUnsetBezierLeft");
        MedsurfDraw.Keyboard.events.remove("pressed_ShiftRight", "onScaleElementSetBezierRight");
        MedsurfDraw.Keyboard.events.remove("released_ShiftRight", "onScaleElementUnsetBezierRight");
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Start position point generator
        if (this._startPositionPointGenerator) {
            this._startPositionPointGenerator.removeAllListeners();
        }
        // -- End position point generator
        if (this._endPositionPointGenerator) {
            this._endPositionPointGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();

    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Line element
        if (this.lineElement && this.lineElement.modeInteraction.lastMode.endsWith('_new')) {
            this.lineElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Start position point generator
        if (this._startPositionPointGenerator) {
            this._startPositionPointGenerator.destroy();
        }
        // -- End position point generator
        if (this._endPositionPointGenerator) {
            this._endPositionPointGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Setup generator
     * @private
     */
    protected _setupGenerator(): void {
        if (this._state === LineGeneratorStates.START) {
            // Generator
            // -- Start position point generator
            this._startPositionPointGenerator = new MedsurfDraw.PositionPointGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: this.stickMode
            });
            this._startPositionPointGenerator.once("startGenerator", this.start, this);
            this._startPositionPointGenerator.once("endGenerator", this.endStartPositionPointGenerator, this);
            this._startPositionPointGenerator.once("updatePoint", this.endStartPositionPointGenerator, this);
            this._startPositionPointGenerator.once("abortGenerator", this.abortStartPositionPointGenerator, this);
        } else {
            // Generator
            // -- End position point generator
            this._endPositionPointGenerator = new MedsurfDraw.PositionPointGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: this.stickMode
            });
            if (this.startElement) {
                this._endPositionPointGenerator.once("startGenerator", this.start, this);
            }
            this._endPositionPointGenerator.once("startGenerator", this.startEndPositionPointGenerator, this);
            this._endPositionPointGenerator.once("endGenerator", this.endEndPositionPointGenerator, this);
            this._endPositionPointGenerator.once("updatePoint", this.updatePointEndPositionPointGenerator, this);
            this._endPositionPointGenerator.once("abortGenerator", this.abortEndPositionPointGenerator, this);
        }
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Destroy end
     */
    public destroyEnd(): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Line element
        if (this.lineElement && this.lineElement.modeInteraction.lastMode.endsWith('new')) {
            this.lineElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- End position point generator
        if (this._endPositionPointGenerator) {
            this._endPositionPointGenerator.destroy();
        }
        //</editor-fold>
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    //<editor-fold desc="Start">
    /**
     * End start position point generator
     */
    public endStartPositionPointGenerator(): void {
        // Members
        this._enableKeyEvents = true;

        // Members
        this._state = LineGeneratorStates.END;

        // Setup generator
        this._setupGenerator();
    }

    /**
     * Abort start position point generator
     */
    public abortStartPositionPointGenerator(): void {
        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="End">
    /**
     * Start end position point generator
     */
    public startEndPositionPointGenerator(): void {
        const startElement = this.startPositionPointElement;

        const sharedLine = this.target.getLineForElements(startElement, this.endPositionPointElement);
        if (sharedLine) {
            this._lineElement = sharedLine;
            this.lineElement.controlDrawItem(false);
        } else {
            // Elements
            // -- Line element
            this._lineElement = MedsurfDraw.Line.getInstance(this.target, startElement.name as string,
                this.endPositionPointElement.name as string, this.bezierCourve, this.target.dimensions.width);
            this.target.addChild(this.lineElement);
            this.lineElement.controlDrawItem(true);

            // Add Model
            this.target.addImageObject(this.lineElement.model, this.data.layerGroup, this.data.image);
        }
    }

    /**
     * End end position point generator
     */
    public endEndPositionPointGenerator(): void {
        // Members
        this._state = LineGeneratorStates.FINISH;

        /* TODO remove
        // Update Model
        // -- Image
        if (!this.data.image.imageObjects) {
            this.data.image.imageObjects = [];
        }
        if (this.data.image.imageObjects.map(io => io.id).indexOf(this.lineElement.model.id) === -1) {
            this.data.image.imageObjects.push(this.lineElement.model);
        }
        // -- Layer group
        if (!this.data.layerGroup.layerImageObjects) {
            this.data.layerGroup.layerImageObjects = [];
        }
        const layerImageObject: Models.LayerImageObject = {
            dirty: true,
            id: uuid.v4(),
            order: this.data.layerGroup.layerImageObjects.length,
            layerGroup: this.data.layerGroup,
            imageObject: this.lineElement.name as string
        };
        this.data.layerGroup.layerImageObjects.push(layerImageObject);
        */

        // Set mode
        this.lineElement.modeInteraction.setModeItem('draw_select_new');

        // Emit end generator
        this.emit("endGenerator", this.endPositionPointElement);

        // End
        this.end();
    }

    /**
     * Update point end position point generator
     */
    public updatePointEndPositionPointGenerator(): void {
        // Set mode
        this.lineElement.modeInteraction.setMode("delete_draw");

        // Check for same position point
        if (this.startPositionPointElement !== this.endPositionPointElement) {
            // Create new line element
            this.startEndPositionPointGenerator();

            // Members
            this._state = LineGeneratorStates.FINISH;

            /* TODO
            // Update Model
            // -- Image
            if (!this.data.image.imageObjects) {
                this.data.image.imageObjects = [];
            }
            if (this.data.image.imageObjects.map(io => io.id).indexOf(this.lineElement.model.id) === -1) {
                this.data.image.imageObjects.push(this.lineElement.model);
            }
            // -- Layer group
            if (!this.data.layerGroup.layerImageObjects) {
                this.data.layerGroup.layerImageObjects = [];
            }
            const layerImageObject: Models.LayerImageObject = {
                dirty: true,
                id: uuid.v4(),
                order: this.data.layerGroup.layerImageObjects.length,
                layerGroup: this.data.layerGroup,
                imageObject: this.lineElement.name as string
            };
            this.data.layerGroup.layerImageObjects.push(layerImageObject);
            */

            // Set mode
            this.lineElement.modeInteraction.setModeItem('draw_select');

            // Emit end generator
            this.emit("endGenerator", this.endPositionPointElement);

            // End
            this.end();
        } else {
            // Setup generator
            this._setupGenerator();
        }
    }

    /**
     * Abort end position point generator
     */
    public abortEndPositionPointGenerator(): void {
        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>
    //</editor-fold>

    //<editor-fold desc="Keyboard">
    public onPressedShiftKey(keyCode: string, event: KeyboardEvent): void {
        event.preventDefault();

        // TODO remove Finish -> have to fix remove keyboard events
        if (this._state === LineGeneratorStates.START || this._state === LineGeneratorStates.FINISH || !this.lineElement) {
            return;
        }

        // Elements
        // -- Line element
        if (this.lineElement) {
            this.lineElement.model.isBezier = true;
            this.lineElement.emit("debounceDraw");
        }
    }

    public onReleaseShiftKey(keyCode: string, event: KeyboardEvent): void {
        event.preventDefault();

        // TODO remove Finish -> have to fix remove keyboard events
        if (this._state === LineGeneratorStates.START || this._state === LineGeneratorStates.FINISH || !this.lineElement) {
            return;
        }

        // Elements
        // -- Line element
        if (this.lineElement) {
            this.lineElement.model.isBezier = false;
            this.lineElement.emit("debounceDraw");
        }
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection> {
        if (this._state === LineGeneratorStates.START) {
            return this._startPositionPointGenerator.drawInteraction;
        } else {
            return this._endPositionPointGenerator.drawInteraction;
        }
    }

    /**
     * Get start position point element6
     */
    public get startPositionPointElement(): MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection {
        return this.startElement ? this.startElement : this._startPositionPointGenerator.positionPointElement;
    }

    /**
     * Get end position point element6
     */
    public get endPositionPointElement(): MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection {
        return this._endPositionPointGenerator.positionPointElement;
    }

    /**
     * Get line element
     */
    public get lineElement(): MedsurfDraw.Line {
        return this._lineElement;
    }

    /**
     * Get bezier courve
     */
    public get bezierCourve(): boolean {
        if (this.data.bezierCourve) {
            return this.data.bezierCourve;
        } else {
            return MedsurfDraw.Keyboard.isKeyDown('ShiftLeft') || MedsurfDraw.Keyboard.isKeyDown('ShiftRight');
        }
    }

    /**
     * Set bezier courve
     * @param value
     */
    public set bezierCourve(value: boolean) {
        this.data.bezierCourve = value;
    }

    /**
     * Get stick mode
     */
    public get stickMode(): MedsurfDraw.StickMode {
        return this.data.stickMode;
    }

    /**
     * Set stick mode
     * @param value
     */
    public set stickMode(value: MedsurfDraw.StickMode) {
        this.data.stickMode = value;
    }

    /**
     * Get start element
     */
    public get startElement(): MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection | undefined {
        return this.data.startElement;
    }

    /**
     * Set start element
     * @param value
     */
    public set startElement(value: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection | undefined) {
        this.data.startElement = value;
    }
    //</editor-fold>
}