import * as MedsurfDraw from "../../public-api";
import {BaseGenerator} from "../../bases/generators/BaseGenerator";
import {LineGeneratorModel} from "./LineGenerator";

/**
 * Polygon generator model
 */
export class PolygonGeneratorModel extends LineGeneratorModel {
}

/**
 * Polygon generator
 */
export class PolygonGenerator extends BaseGenerator<MedsurfDraw.Image, PolygonGeneratorModel> {
    /**
     * Generators
     */
    protected _lineGenerator: MedsurfDraw.LineGenerator;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: PolygonGeneratorModel) {
        // Parent constructor
        super(model);

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Line generator
        this._lineGenerator = new MedsurfDraw.LineGenerator(this.data);
        this._lineGenerator.once("startGenerator", this.start, this);
        this._lineGenerator.once("endGenerator", this.endLineGenerator, this);
        this._lineGenerator.once("abortGenerator", this.abortLineGenerator, this);
        //</editor-fold>
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Generators">
        // -- Line generator
        if (this._lineGenerator) {
            this._lineGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();

    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Generators">
        // -- Line generator
        if (this._lineGenerator) {
            this._lineGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Setup generator
     * @private
     */
    protected _setupGenerator(positionPoint: MedsurfDraw.PositionPoint): void {
        // Generator
        // -- Line generator
        this._lineGenerator = new MedsurfDraw.LineGenerator({
            target: this.target,
            layerGroup: this.data.layerGroup,
            image: this.data.image,
            bezierCourve: this.bezierCourve,
            stickMode: this.stickMode,
            startElement: positionPoint
        });
        this._lineGenerator.once("endGenerator", this.endLineGenerator, this);
        this._lineGenerator.once("abortGenerator", this.abortLineGenerator, this);
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    /**
     * End line generator
     */
    public endLineGenerator(): void {
        // Change element mode
        this._lineGenerator.startPositionPointElement.modeInteraction.setMode("draw_select");
        this._lineGenerator.endPositionPointElement.modeInteraction.setMode("draw_select");
        this._lineGenerator.lineElement.modeInteraction.setMode("draw_select");

        // Setup generator
        this._setupGenerator(this._lineGenerator.endPositionPointElement as MedsurfDraw.PositionPoint);
    }

    /**
     * Abort line generator
     */
    public abortLineGenerator(): void {
        // Emit end generator
        this.emit("abortGenerator");

        // Destroy
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection> {
        return this._lineGenerator.drawInteraction;
    }

    /**
     * Get bezier courve
     */
    public get bezierCourve(): boolean | undefined {
        return this.data.bezierCourve;
    }

    /**
     * Set bezier courve
     * @param value
     */
    public set bezierCourve(value: boolean | undefined) {
        this.data.bezierCourve = value;
    }

    /**
     * Get stick mode
     */
    public get stickMode(): MedsurfDraw.StickMode {
        return this.data.stickMode;
    }

    /**
     * Set stick mode
     * @param value
     */
    public set stickMode(value: MedsurfDraw.StickMode) {
        this.data.stickMode = value;
    }
    //</editor-fold>
}