import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";

/**
 * Copy image object mapping
 */
export class CopyImageObjectMapping {
    old: Models.ImageObject;
    new: Models.ImageObject;
}

/**
 * Clone item generator model
 */
export class CloneItemGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    imageObjects: Models.ImageObject[];
    layerGroup: Models.LayerGroup;
    image: Models.Image;
    element: Models.PositionPoint | Models.Line | Models.FillCollection | Models.LegendCollection;
    imageObjectMappings: CopyImageObjectMapping[];
    relativePoint?: Models.Point | undefined;
}

/**
 * Clone item generator
 */
export class CloneItemGenerator extends BaseGenerator<MedsurfDraw.Image, CloneItemGeneratorModel> {
    /**
     * Interactions
     */
    protected _cloneInteraction: MedsurfDraw.CloneInteraction;

    /**
     * Elements
     */
    protected _cloneElement: MedsurfDraw.PositionPoint | MedsurfDraw.Line | MedsurfDraw.FillCollection | MedsurfDraw.LegendCollection;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: CloneItemGeneratorModel) {
        // Parent constructor
        super(model);

        // Image objects mappings
        if (!this.data.imageObjectMappings) {
            this.data.imageObjectMappings = [];
        }

        //<editor-fold desc="Interactions">
        // -- Clone interaction
        this._cloneInteraction = new MedsurfDraw.CloneInteraction();
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("cloning");
        // -- Clone element
        const data = this.getCloneElement(this.data.element, this.data.relativePoint);
        this.data.relativePoint = data.relativePoint;
        this._cloneElement = data.element;
        this.cloneElement.controlCloneItem();
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Image
        this.target.on('imageZoom', this.onZoom, this);
        this.target.parent.on("mousedown", this.cloneInteraction.startClone, this.cloneInteraction);
        this.target.parent.on("rightup", this.cloneInteraction.abortClone, this.cloneInteraction);
        this.target.parent.on("pointermove", this.cloneInteraction.onClone, this.cloneInteraction);
        this.target.parent.on("mouseup", this.cloneInteraction.endClone, this.cloneInteraction);
        // -- Clone
        this.cloneInteraction.on("endClone", this.endClone, this);
        this.cloneInteraction.on("abortClone", this.abortClone, this);
        //</editor-fold>

        // Check pointerover
        // -- Check if mouse is in image
        const globalPoint = this.target.renderer.plugins.interaction.mouse.global;
        const hitTest = this.target.renderer.plugins.interaction.hitTest(globalPoint, this.target.parent);
        if (!hitTest) {
            // -- Not hit test result -> mouse outside image
            this.target.parent.once("pointerover", this.start, this);
        } else {
            // -- Hit rest result -> mouse inside image
            const event = new PIXI.InteractionEvent();
            event.data = new PIXI.InteractionData();
            event.data.global = globalPoint;
            this.emit("debounceStart", event);
        }
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        // Update position
        this.onZoom();

        // Start clone
        this.cloneInteraction.startClone(event);

        // Parent
        super.start(event);

        // Fix to center mouse
        this.onZoom();
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.off('imageZoom', this.onZoom, this);
        this.target.parent.off("mousedown", this.cloneInteraction.startClone, this.cloneInteraction);
        this.target.parent.off("rightup", this.cloneInteraction.abortClone, this.cloneInteraction);
        this.target.parent.off("pointermove", this.cloneInteraction.onClone, this.cloneInteraction);
        this.target.parent.off("mouseup", this.cloneInteraction.endClone, this.cloneInteraction);
        // -- Clone
        this.cloneInteraction.removeAllListeners();
        // -- Generator
        // TODO what's this -> needed
        this.removeAllListeners("onZoom");
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Get image objects
        const imageObjects = this.target.getImageObjects();

        // Remove new elements
        imageObjects.filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('clone')
            && imageObject.modeInteraction.lastMode.endsWith('new'))
            .sort((a, b) => { // Prio for line elements
                const aModel = (a as any).model;
                const bModel = (b as any).model;
                if (aModel.type === Models.ImageObjectType.LINE) {
                    if (bModel.type === Models.ImageObjectType.LINE) {
                        return 0;
                    } else {
                        return -1;
                    }
                } else {
                    if (bModel.type === Models.ImageObjectType.LINE) {
                        return 1;
                    }
                }
                return 0;
            }).forEach((imageObject) => {
            // Remove element
            imageObject.modeInteraction.setMode("delete_clone");
        });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Get clone element
     * @param model
     * @param relativePoint
     */
    public getCloneElement(model: Models.PositionPoint | Models.Line | Models.FillCollection | Models.LegendCollection,
                           relativePoint: Models.Point | undefined = undefined): { element: MedsurfDraw.PositionPoint | MedsurfDraw.Line | MedsurfDraw.FillCollection | MedsurfDraw.LegendCollection, relativePoint: Models.Point } {
        // Check for type
        switch(model.type) {
            case Models.ImageObjectType.POSITIONPOINT:
                return this.clonePositionPoint(model as Models.PositionPoint, relativePoint);
            case Models.ImageObjectType.LINE:
                return this.cloneLine(model as Models.Line, relativePoint);
            case Models.ImageObjectType.FILLCOLLECTION:
                return this.cloneFillCollection(model as Models.FillCollection, relativePoint);
            case Models.ImageObjectType.LEGENDCOLLECTION:
                return this.cloneLegendCollection(model as Models.LegendCollection, relativePoint);
            default:
                throw "Not implemented";
        }
        throw "Not implemented";
    }

    /**
     * Clone position point
     * @param model
     * @param relativePoint
     */
    public clonePositionPoint(model: Models.PositionPoint,
                              relativePoint: Models.Point | undefined = undefined): {element: MedsurfDraw.PositionPoint, relativePoint: Models.Point} {
        // Relative point
        if (!relativePoint) {
            relativePoint = model.position;
        }

        // Check for mapping
        let positionPoint: MedsurfDraw.PositionPoint;
        const positionPointMapping = this.data.imageObjectMappings.find((mapping) => mapping.old.id === model.id);
        if (positionPointMapping) {
            positionPoint = this.target.getImageObjects().find((imageObject) => imageObject.name === positionPointMapping.new.id) as MedsurfDraw.PositionPoint;
        } else {
            positionPoint = this.createPositionPointInstance(model, relativePoint);
        }

        // Return position point
        return { element: positionPoint, relativePoint: relativePoint };
    }

    /**
     * Clone line
     * @param model
     * @param relativePoint
     */
    public cloneLine(model: Models.Line,
                     relativePoint: Models.Point | undefined = undefined): { element: MedsurfDraw.Line, relativePoint: Models.Point } {
        // Get image objects
        const imageObjects = this.target.getImageObjects();

        // Check for mapping
        let line: MedsurfDraw.Line;
        const lineMapping = this.data.imageObjectMappings.find((mapping) => mapping.old.id === model.id);
        if (lineMapping) {
            // Get line
            line = imageObjects.find((imageObject) => imageObject.name === lineMapping.new.id) as MedsurfDraw.Line;

            // Create start position point
            let start = imageObjects.find((imageObject) => imageObject.name === model.start) as MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
            if (start instanceof MedsurfDraw.PositionPoint) {
                // Relative point
                if (!relativePoint) {
                    relativePoint = start.model.position;
                }
            }
        } else {
            let start: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
            const startMapping = this.data.imageObjectMappings.find((mapping) => mapping.old.id === model.start);
            if (startMapping) {
                // Get start position point
                start = imageObjects.find((imageObject) => imageObject.name === startMapping.new.id) as MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
            } else {
                // Create start position point
                const startModel = this.data.imageObjects.find((imageObject) => imageObject.id === model.start) as Models.PositionPoint | Models.FillCollection;
                if (startModel.type === Models.ImageObjectType.POSITIONPOINT) {
                    // Create start instance
                    const data = this.clonePositionPoint(startModel as Models.PositionPoint, relativePoint);
                    start = data.element;
                    relativePoint = data.relativePoint;

                    // Set mode
                    start.modeInteraction.setMode('clone_item_new');
                } else {
                    throw 'Fill collection not implemented';
                }
                /* TODO test or remove
                else {
                    // Set mode
                    start.modeInteraction.setMode('clone_select');

                    // Create mapping
                    this.data.imageObjectMappings.push({
                        new: start.model,
                        old: start.model
                    });
                }
                */
            }

            let end: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
            const endMapping = this.data.imageObjectMappings.find((mapping) => mapping.old.id === model.end);
            if (endMapping) {
                // Get end position point
                end = imageObjects.find((imageObjects) => imageObjects.name === endMapping.new.id) as MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
            } else {
                // Create end position point
                const endModel = this.data.imageObjects.find((imageObject) => imageObject.id === model.end) as Models.PositionPoint | Models.FillCollection;
                if (endModel.type === Models.ImageObjectType.POSITIONPOINT) {
                    // Create end instance
                    const data = this.clonePositionPoint(endModel as Models.PositionPoint, relativePoint);
                    end = data.element;
                    relativePoint = data.relativePoint;

                    // Set mode
                    end.modeInteraction.setMode('clone_item_new');
                } else {
                    throw 'Fill collection not implemented';
                }
                /* TODO test or remove
                else {
                    // Set mode
                    end.modeInteraction.setMode('clone_select');

                    // Create mapping
                    this.data.imageObjectMappings.push({
                        new: end.model,
                        old: end.model,
                    });
                }
                */
            }

            line = this.createLineInstance(model, start.model, end.model);
        }

        // Relative point
        if (!relativePoint) {
            relativePoint = { x: 0, y: 0 };
        }

        // Return line
        return { element: line, relativePoint: relativePoint };
    }

    /**
     * Clone fill collection
     * @param model
     * @param relativePoint
     */
    public cloneFillCollection(model: Models.FillCollection,
                               relativePoint: Models.Point | undefined = undefined): { element: MedsurfDraw.FillCollection, relativePoint: Models.Point } {
        // Check for mapping
        let fillCollection: MedsurfDraw.FillCollection;
        const fillCollectionMapping = this.data.imageObjectMappings.find((mapping) => mapping.old.id === model.id);
        if (fillCollectionMapping) {
            // Get fill collection
            fillCollection = this.target.getImageObjects().find((imageObject) => imageObject.name === fillCollectionMapping.new.id) as MedsurfDraw.FillCollection;
        } else {
            // Get lines
            const lineModels = this.data.imageObjects.filter((imageObject) => model.lines.indexOf(imageObject.id as string) > -1) as Models.Line[];

            // Create line
            const fillLines: string[] = [];
            for (let lineModel of lineModels) {
                const data = this.cloneLine(lineModel, relativePoint);
                const line = data.element;
                relativePoint = data.relativePoint;

                // Set mode
                line.modeInteraction.setMode('clone_item_new');

                // Fill lines
                fillLines.push(line.name as string);
            }

            fillCollection = this.createFillCollectionInstance(model, fillLines);
        }

        // Relative point
        if (!relativePoint) {
            relativePoint = { x: 0, y: 0 };
        }

        // Return fill collection
        return { element: fillCollection, relativePoint: relativePoint };
    }

    /**
     * Clone legend collection
     * @param model
     * @param relativePoint
     */
    public cloneLegendCollection(model: Models.LegendCollection,
                                 relativePoint: Models.Point | undefined = undefined): { element: MedsurfDraw.LegendCollection, relativePoint: Models.Point } {
        // Relative point
        if (!relativePoint) {
            relativePoint = model.rectangle;
        }

        // Check for mapping
        let legendCollection: MedsurfDraw.LegendCollection;
        const legendCollectionMapping = this.data.imageObjectMappings.find((mapping) => mapping.old.id === model.id);
        if (legendCollectionMapping) {
            // Get legend collection
            legendCollection = this.target.getImageObjects().find((imageObject) => imageObject.name === legendCollectionMapping.new.id) as MedsurfDraw.LegendCollection;
        } else {
            // Create legend collection
            legendCollection = this.createLegendInstance(model, relativePoint);
        }

        // Return legend collection
        return { element: legendCollection, relativePoint: relativePoint };
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Position point on zoom
     * @param model
     * @param positionPoint
     * @param relativePoint
     * @protected
     */
    protected _positionPointOnReset(model: Models.PositionPoint,
                                    positionPoint: MedsurfDraw.PositionPoint,
                                    relativePoint: Models.Point): void {
        const imageWidth = this.target.imageDimensions.width;
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        positionPoint.position.x = (model.position.x - relativePoint.x) * imageWidth + (pos.x - rect.x) / this.target.scale.x;
        positionPoint.position.y = (model.position.y - relativePoint.y) * imageWidth + (pos.y - rect.y) / this.target.scale.y;

        // Reset draw
        const event = new PIXI.InteractionEvent();
        event.data = new PIXI.InteractionData();
        event.data.global = pos;
        this.cloneInteraction.resetClone(event);

        // Emit on position point move
        positionPoint.emit("onPositionPointMove");
    }

    /**
     * Position point on zoom
     * @param model
     * @param positionPoint
     * @param relativePoint
     * @protected
     */
    protected _positionPointOnZoom(model: Models.PositionPoint,
                                   positionPoint: MedsurfDraw.PositionPoint,
                                   relativePoint: Models.Point): void {
        const imageWidth = this.target.imageDimensions.width;
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        positionPoint.position.x = (model.position.x - relativePoint.x) * imageWidth + (pos.x - rect.x) / this.target.scale.x;
        positionPoint.position.y = (model.position.y - relativePoint.y) * imageWidth + (pos.y - rect.y) / this.target.scale.y;

        // Reset draw
        const event = new PIXI.InteractionEvent();
        event.data = new PIXI.InteractionData();
        event.data.global = pos;
        this.cloneInteraction.resetClone(event);

        // Emit on position point move
        positionPoint.emit("onPositionPointMove");
    }

    /**
     * Legend collection on zoom
     * @param model
     * @param legendCollection
     * @param relativePoint
     * @protected
     */
    protected _legendCollectionOnZoom(model: Models.LegendCollection,
                                      legendCollection: MedsurfDraw.LegendCollection,
                                      relativePoint: Models.Point): void {
        const imageWidth = this.target.imageDimensions.width;
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        legendCollection.position.x = (model.rectangle.x - relativePoint.x) * imageWidth + (pos.x - rect.x) / this.target.scale.x;
        legendCollection.position.y = (model.rectangle.y - relativePoint.y) * imageWidth + (pos.y - rect.y) / this.target.scale.y;

        // Reset draw
        const event = new PIXI.InteractionEvent();
        event.data = new PIXI.InteractionData();
        event.data.global = pos;
        this.cloneInteraction.resetClone(event);
    }
    //</editor-fold>

    //<editor-fold desc="Instances">
    /**
     * Create position point instance
     * @param model
     * @param relativePoint
     */
    public createPositionPointInstance(model: Models.PositionPoint, relativePoint: Models.Point): MedsurfDraw.PositionPoint {
        // Get new model
        const positionPointModel = this.createPositionPointModel(model, relativePoint);

        // Create new instance
        const positionPoint = new MedsurfDraw.PositionPoint(this.target, positionPointModel);
        // TODO remove positionPoint.hideItem();

        // Add Model
        this.target.addImageObject(positionPoint.model, this.data.layerGroup, this.data.image);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: positionPoint.model,
            old: model
        });

        // Methods
        const onZoomMethod: EventListener = this._positionPointOnZoom.bind(this, model, positionPoint, relativePoint);

        //<editor-fold desc="Events">
        // -- Generator
        this.on("onZoom", onZoomMethod);
        // -- Move
        this.cloneInteraction.on("startClone", positionPoint.moveInteraction.startMove, positionPoint.moveInteraction);
        this.cloneInteraction.on("resetClone", positionPoint.moveInteraction.resetMove, positionPoint.moveInteraction);
        this.cloneInteraction.on("onClone", positionPoint.moveInteraction.onMove, positionPoint.moveInteraction);
        this.cloneInteraction.once("endClone", positionPoint.moveInteraction.endMove, positionPoint.moveInteraction);
        //</editor-fold>

        // Add to image
        this.target.addChild(positionPoint);

        // Create primitives
        if (!positionPoint.model.primitives) {
            positionPoint.model.primitives = [];
        }
        for (let primitive of model.primitives || []) {
            switch (primitive.type) {
                case Models.ImageObjectType.ARROW:
                    // Create instance
                    const arrowPrimitive = this.createArrowPrimitiveInstance(primitive as Models.ArrowPrimitive, positionPoint.model);
                    // Model
                    // @ts-ignore
                    positionPoint.model.primitives.push(arrowPrimitive.model);

                    // Add to position point
                    positionPoint.addChild(arrowPrimitive);
                    arrowPrimitive.showItem();
                    break;
                case Models.ImageObjectType.ELLIPSE:
                    // Create instance
                    const ellipsePrimitive = this.createEllipsePrimitiveInstance(primitive as Models.EllipsePrimitive, positionPoint.model);
                    // Model
                    // @ts-ignore
                    positionPoint.model.primitives.push(ellipsePrimitive.model);

                    // Add to position point
                    positionPoint.addChild(ellipsePrimitive);
                    ellipsePrimitive.showItem();
                    break;
                case Models.ImageObjectType.RECTANGLE:
                    // Create instance
                    const rectanglePrimitive = this.createRectanglePrimitiveInstance(primitive as Models.RectanglePrimitive, positionPoint.model);
                    // Model
                    // @ts-ignore
                    positionPoint.model.primitives.push(rectanglePrimitive.model);

                    // Add to position point
                    positionPoint.addChild(rectanglePrimitive);
                    rectanglePrimitive.showItem();
                    break;
                case Models.ImageObjectType.TEXT:
                    // Create instance
                    const textPrimitive = this.createTextPrimitiveInstance(primitive as Models.TextPrimitive, positionPoint.model);
                    // Model
                    // @ts-ignore
                    positionPoint.model.primitives.push(textPrimitive.model);

                    // Add to position point
                    positionPoint.addChild(textPrimitive);
                    textPrimitive.showItem();
                    break;
            }
        }

        // Return position point
        return positionPoint;
    }

    /**
     * Create line instance
     * @param model
     * @param start
     * @param end
     */
    public createLineInstance(model: Models.Line, start: Models.PositionPoint | Models.FillCollection, end: Models.PositionPoint | Models.FillCollection): MedsurfDraw.Line {
        // Check position points
        if (!start || !end) {
            throw '1.1: position point not cloned';
        }

        // Get new model
        const lineModel = this.createLineModel(model, start, end);

        // Create new instance
        const line = new MedsurfDraw.Line(this.target, lineModel);
        // TODO remove line.hideItem();

        // Add Model
        this.target.addImageObject(line.model, this.data.layerGroup, this.data.image);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: line.model,
            old: model
        });

        // Add to image
        this.target.addChild(line);

        // Return line
        return line;
    }

    /**
     * Create fill collection instance
     * @param model
     * @param lines
     */
    public createFillCollectionInstance(model: Models.FillCollection, lines: string[]): MedsurfDraw.FillCollection {
        // Check lines
        if (!lines) {
            throw '3.1: lines not cloned';
        }

        // Get new model
        const fillCollectionModel = this.createFillCollectionModel(model, lines);

        // Create new instance
        const fillCollection = new MedsurfDraw.FillCollection(this.target, fillCollectionModel);
        // TODO remove fillCollection.hideItem();

        // Add Model
        this.target.addImageObject(fillCollection.model, this.data.layerGroup, this.data.image);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: fillCollection.model,
            old: model
        });

        // Add to image
        this.target.addChild(fillCollection);

        // Return fill collection
        return fillCollection;
    }

    /**
     * Create legend instance
     * @param model
     * @param relativePoint
     */
    public createLegendInstance(model: Models.LegendCollection, relativePoint: Models.Point): MedsurfDraw.LegendCollection {
        // Get new model
        const legendCollectionModel = this.createLegendCollectionModel(model, relativePoint);

        // Create column models
        for (let col of model.cols || []) {
            // Create instance
            const legendColumnModel = this.createLegendColumnModel(col, legendCollectionModel);

            // Create row models
            for (let row of col.rows || []) {
                // Create instance
                const legendRowModel = this.createLegendRowModel(row, legendColumnModel);

                // Model
                legendColumnModel.rows.push(legendRowModel);
            }

            // Model
            legendCollectionModel.cols.push(legendColumnModel);
        }

        // Create new instance
        const legendCollection = new MedsurfDraw.LegendCollection(this.target, legendCollectionModel);
        // TODO remove legendCollection.hideItem();

        // Add Model
        this.target.addImageObject(legendCollection.model, this.data.layerGroup, this.data.image);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: legendCollection.model,
            old: model
        });

        // Methods
        const onZoomMethod: EventListener = this._legendCollectionOnZoom.bind(this, model, legendCollection, relativePoint);

        //<editor-fold desc="Events">
        // -- Generator
        this.on("onZoom", onZoomMethod);
        // -- Move
        this.cloneInteraction.on("startClone", legendCollection.moveInteraction.startMove, legendCollection.moveInteraction);
        this.cloneInteraction.on("resetClone", legendCollection.moveInteraction.resetMove, legendCollection.moveInteraction);
        this.cloneInteraction.on("onClone", legendCollection.moveInteraction.onMove, legendCollection.moveInteraction);
        this.cloneInteraction.once("endClone", legendCollection.moveInteraction.endMove, legendCollection.moveInteraction);
        //</editor-fold>

        // Add to image
        this.target.addChild(legendCollection);

        // Return legend collection
        return legendCollection;
    }

    /**
     * Create arrow primitive instance
     * @param model
     * @param positionPoint
     */
    public createArrowPrimitiveInstance(model: Models.ArrowPrimitive, positionPoint: Models.PositionPoint): MedsurfDraw.ArrowPrimitive {
        if (!positionPoint) {
            throw '1.2: position point not cloned';
        }

        // Get new model
        const arrowPrimitiveModel = this.createArrowPrimitiveModel(model, positionPoint);

        // Create new instance
        const arrowPrimitive = new MedsurfDraw.ArrowPrimitive(this.target, arrowPrimitiveModel);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: arrowPrimitive.model,
            old: model
        });

        // Return arrow
        return arrowPrimitive;
    }

    /**
     * Create ellipse primitive instance
     * @param model
     * @param positionPoint
     */
    public createEllipsePrimitiveInstance(model: Models.EllipsePrimitive, positionPoint: Models.PositionPoint): MedsurfDraw.EllipsePrimitive {
        if (!positionPoint) {
            throw '1.3: position point not cloned';
        }

        // Get new model
        const ellipsePrimitiveModel = this.createEllipsePrimitiveModel(model, positionPoint);

        // Create new instance
        const ellipsePrimitive = new MedsurfDraw.EllipsePrimitive(this.target, ellipsePrimitiveModel);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: ellipsePrimitive.model,
            old: model
        });

        // Return arrow
        return ellipsePrimitive;
    }

    /**
     * Create rectangle primitive instance
     * @param model
     * @param positionPoint
     */
    public createRectanglePrimitiveInstance(model: Models.RectanglePrimitive, positionPoint: Models.PositionPoint): MedsurfDraw.RectanglePrimitive {
        if (!positionPoint) {
            throw '1.4: position point not cloned';
        }

        // Get new model
        const rectanglePrimitiveModel = this.createRectanglePrimitiveModel(model, positionPoint);

        // Create new instance
        const rectanglePrimitive = new MedsurfDraw.RectanglePrimitive(this.target, rectanglePrimitiveModel);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: rectanglePrimitive.model,
            old: model
        });

        // Return arrow
        return rectanglePrimitive;
    }

    /**
     * Create text primitive instance
     * @param model
     * @param positionPoint
     */
    public createTextPrimitiveInstance(model: Models.TextPrimitive, positionPoint: Models.PositionPoint): MedsurfDraw.TextPrimitive {
        if (!positionPoint) {
            throw '1.5: position point not cloned';
        }

        // Get new model
        const textPrimitiveModel = this.createTextPrimitiveModel(model, positionPoint);

        // Create new instance
        const textPrimitive = new MedsurfDraw.TextPrimitive(this.target, textPrimitiveModel);

        // Create mapping
        this.data.imageObjectMappings.push({
            new: textPrimitive.model,
            old: model
        });

        // Return arrow
        return textPrimitive;
    }
    //</editor-fold>

    //<editor-fold desc="Models">
    /**
     * Create position point model
     * @param model
     * @param relativePoint
     */
    public createPositionPointModel(model: Models.PositionPoint,
                                    relativePoint: Models.Point): Models.PositionPoint {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.POSITIONPOINT,
            image: model.image,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            isSelftest: model.isSelftest,
            selftestItems: [], // TODO remove model.selftestItems,
            position: {
                x: model.position.x - relativePoint.x,
                y: model.position.y - relativePoint.y,
            },
            primitives: []
        }
        // Object.assign(data, {id: uuid.v4()})
    }

    /**
     * Create line model
     * @param model
     * @param start
     * @param end
     */
    public createLineModel(model: Models.Line,
                           start: Models.PositionPoint | Models.FillCollection,
                           end: Models.PositionPoint | Models.FillCollection): Models.Line {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LINE,
            image: model.image,
            start: start.id as string,
            end: end.id as string,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            isBezier: model.isBezier,
            interactWithLines: model.interactWithLines,
            strokeWidth: model.strokeWidth,
            options: JSON.parse(JSON.stringify(model.options)),
            dashEmpty: model.dashEmpty,
            dashFill: model.dashFill,
        }
        // Object.assign(data, {id: uuid.v4(), start: start.new.id, end: end.new.id})
    }

    /**
     * Create fill collection model
     * @param model
     * @param lines
     */
    public createFillCollectionModel(model: Models.FillCollection,
                                     lines: string[]): Models.FillCollection {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.FILLCOLLECTION,
            image: model.image,
            lines: lines,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            isInteractive: model.isInteractive,
            interactiveItems: [], // TODO remove model.interactiveItems,
            options: JSON.parse(JSON.stringify(model.options)),
            linkedSite: model.linkedSite,
            linkedSlideId: model.linkedSlideId,
        }
        // Object.assign(data, {id: uuid.v4(), lines: lines})
    }

    /**
     * Create legend collection model
     * @param model
     * @param relativePoint
     */
    public createLegendCollectionModel(model: Models.LegendCollection,
                                       relativePoint:Models.Point): Models.LegendCollection {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDCOLLECTION,
            image: model.image,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            selftestPoint: model.selftestPoint ?
                { x: model.selftestPoint.x, y: model.selftestPoint.y} :
                undefined,
            indexWidth: model.indexWidth,
            columnDistance: model.columnDistance,
            rowDistance: model.rowDistance,
            rectangle: {
                x: model.rectangle.x - relativePoint.x,
                y: model.rectangle.y - relativePoint.y,
                width: model.rectangle.width,
                height: model.rectangle.height,
            },
            cols: [],
        };
    }

    /**
     * Create legend column model
     * @param model
     * @param legendCollection
     */
    public createLegendColumnModel(model: Models.LegendColumn,
                                   legendCollection: Models.LegendCollection): Models.LegendColumn {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDCOLUMN,
            image: model.image,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            table: legendCollection,
            rows: [],
        };
    }

    /**
     * Create legend row model
     * @param model
     * @param legendColumn
     */
    public createLegendRowModel(model: Models.LegendRow,
                                legendColumn: Models.LegendColumn): Models.LegendRow {
        // Get image width
        const imageWidth = this.target.imageDimensions.width;

        const legendRowModel = {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.LEGENDROW,
            image: model.image,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            col: legendColumn,
            text: model.text,
            style: {
                align: model.style.align,
                fontSize: model.style.fontSize || 0 / imageWidth,
                fill: model.style.fill,
                fontFamily: model.style.fontFamily,
                breakWords: model.style.breakWords,
            },
        };

        return legendRowModel;
    }

    /**
     * Create arrow primitive model
     * @param model
     * @param positionPoint
     */
    public createArrowPrimitiveModel(model: Models.ArrowPrimitive,
                                     positionPoint: Models.PositionPoint): Models.ArrowPrimitive {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.ARROW,
            image: model.image,
            positionPoint: positionPoint,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            wedgePoint: model.wedgePoint ?
                { x: model.wedgePoint.x, y: model.wedgePoint.y} :
                undefined,
            rotateWithLine: model.rotateWithLine,
            rectangle: JSON.parse(JSON.stringify(model.rectangle)),
            angle: model.angle,
            strokeWidth: model.strokeWidth,
            options: JSON.parse(JSON.stringify(model.options)),
            dashEmpty: model.dashEmpty,
            dashFill: model.dashFill,
        };
    }

    /**
     * Create ellipse primitive model
     * @param model
     * @param positionPoint
     */
    public createEllipsePrimitiveModel(model: Models.EllipsePrimitive,
                                       positionPoint: Models.PositionPoint): Models.EllipsePrimitive {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.ELLIPSE,
            image: model.image,
            positionPoint: positionPoint,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            rotateWithLine: model.rotateWithLine,
            ellipse: JSON.parse(JSON.stringify(model.ellipse)),
            angle: model.angle,
            strokeWidth: model.strokeWidth,
            options: JSON.parse(JSON.stringify(model.options)),
            dashEmpty: model.dashEmpty,
            dashFill: model.dashFill,
        };
    }

    /**
     * Create rectangle primitive model
     * @param model
     * @param positionPoint
     */
    public createRectanglePrimitiveModel(model: Models.RectanglePrimitive,
                                         positionPoint: Models.PositionPoint): Models.RectanglePrimitive {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.RECTANGLE,
            image: model.image,
            positionPoint: positionPoint,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            rotateWithLine: model.rotateWithLine,
            rectangle: JSON.parse(JSON.stringify(model.rectangle)),
            angle: model.angle,
            strokeWidth: model.strokeWidth,
            options: JSON.parse(JSON.stringify(model.options)),
            dashEmpty: model.dashEmpty,
            dashFill: model.dashFill,
        };
    }

    /**
     * Create text primitive model
     * @param model
     * @param positionPoint
     */
    public createTextPrimitiveModel(model: Models.TextPrimitive,
                                    positionPoint: Models.PositionPoint): Models.TextPrimitive {
        return {
            dirty: true,
            id: uuidv4(),
            type: Models.ImageObjectType.TEXT,
            image: model.image,
            positionPoint: positionPoint,
            zIndex: model.zIndex,
            isStatic: model.isStatic,
            rotateWithLine: model.rotateWithLine,
            rectangle: JSON.parse(JSON.stringify(model.rectangle)),
            angle: model.angle,
            text: model.text,
            style: {
                align: model.style.align,
                fontSize: model.style.fontSize,
                fill: model.style.fill,
                fontFamily: model.style.fontFamily,
                breakWords: model.style.breakWords,
            },
        };
    }
    //</editor-fold>

    //<editor-fold desc="Clone">
    /**
     * End clone
     * @param event
     */
    public endClone(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit end generator
        this.emit("endGenerator", event);

        // End
        this.end();
    }

    /**
     * Abort clone
     * @param event
     */
    public abortClone(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On zoom
     */
    public onZoom(): void {
        this.emit("onZoom");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get clone interaction
     */
    public get cloneInteraction(): MedsurfDraw.CloneInteraction {
        return this._cloneInteraction;
    }

    /**
     * Get clone element
     */
    public get cloneElement(): MedsurfDraw.PositionPoint | MedsurfDraw.Line | MedsurfDraw.FillCollection
        | MedsurfDraw.LegendCollection {
        return this._cloneElement;
    }
    //</editor-fold>
}