import * as PIXI from 'pixi.js-legacy';
import * as MedsurfDraw from '../../public-api';
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { ImageObjectType } from '@ascii-dev-user/medsurf-lib/models';
import { v4 as uuidv4 } from 'uuid';
import { BaseGenerator, BaseGeneratorModel } from '../../bases/generators/BaseGenerator';
import { debounce } from 'debounce';

/**
 * Copy layer group mapping
 */
export class CopyLayerGroupMapping {
    old: Models.LayerGroup;
    new: Models.LayerGroup;
}

/**
 * Clone layer generator model
 */
export class CloneLayerGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    imageObjects: Models.ImageObject[];
    // TODO layerGroup: Models.LayerGroup;
    image: Models.Image;
    cloneLayerGroup: Models.LayerGroup;
    imageObjectMappings: MedsurfDraw.CopyImageObjectMapping[];
    layerGroupMappings: CopyLayerGroupMapping[];
    relativePoint?: Models.Point | undefined;
}

/**
 * Clone layer generator
 */
export class CloneLayerGenerator extends BaseGenerator<MedsurfDraw.Image, CloneLayerGeneratorModel> {
    /**
     * Members
     */
    protected _generators: MedsurfDraw.CloneItemGenerator[];

    /**
     * Methods
     */
    public _debounceEndGeneratorMethod: any;

    /**
     * Elements
     */
    protected _cloneLayerGroup: Models.LayerGroup;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: CloneLayerGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._generators = [];

        //<editor-fold desc="Methods">
        this._debounceEndGeneratorMethod = debounce(this.endClone.bind(this), 20).bind(this);
        //</editor-fold>

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("cloning");
        // -- Clone layer
        this._cloneLayerGroup = this.cloneLayerGroup(this.data.cloneLayerGroup, this.data.relativePoint);
        //</editor-fold>

        // Check pointerover
        // -- Check if mouse is in image
        const globalPoint = this.target.renderer.plugins.interaction.mouse.global;
        const hitTest = this.target.renderer.plugins.interaction.hitTest(globalPoint, this.target.parent);
        if (!hitTest) {
            // -- Not hit test result -> mouse outside image
            this.target.parent.once("pointerover", this.start, this);
        } else {
            // -- Hit rest result -> mouse inside image
            const event = new PIXI.InteractionEvent();
            event.data = new PIXI.InteractionData();
            event.data.global = globalPoint;
            this.emit("debounceStart", event);
        }
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.parent.on("rightup", this.abortClone, this);
        //</editor-fold>

        //<editor-fold desc="Generators">
        for (let generator of this._generators) {
            generator.emit("debounceStart", event);
            generator.once('endGenerator', this._debounceEndGeneratorMethod, this);
        }
        //</editor-fold>

        // Parent
        super.start(event);
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.parent.off("rightup", this.abortClone, this);
        //</editor-fold>

        //<editor-fold desc="Generators">
        for (let generator of this._generators) {
            generator.emit("debounceEnd");
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Clone selftest and interactive items
        this.cloneRelations();

        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Get image objects
        const imageObjects = this.target.getImageObjects();

        // Remove new elements
        imageObjects.filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('clone')
            && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .sort((a, b) => { // Prio for line elements
                const aModel = (a as any).model;
                const bModel = (b as any).model;
                if (aModel.type === Models.ImageObjectType.LINE) {
                    if (bModel.type === Models.ImageObjectType.LINE) {
                        return 0;
                    } else {
                        return -1;
                    }
                } else {
                    if (bModel.type === Models.ImageObjectType.LINE) {
                        return 1;
                    }
                }
                return 0;
            }).forEach((imageObject) => {
            // Remove element
            imageObject.modeInteraction.setMode("delete_clone");
        });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Clone layer group
     * @param model
     * @param relativePoint
     */
    public cloneLayerGroup(model: Models.LayerGroup,
                           relativePoint: Models.Point | undefined = undefined): Models.LayerGroup {
        // Check for mapping
        let layer: Models.LayerGroup;
        const layerMapping = this.data.layerGroupMappings.find((mapping) => mapping.old.id === model.id);
        if (layerMapping) {
            layer = layerMapping.new;
        } else {
            // Get new model
            layer = this.createLayerGroupInstance(model, relativePoint);
        }

        // Return layer
        return layer;
    }

    /**
     * Clone layer image object
     * @param layerImageObject
     * @param layerModel
     * @param relativePoint
     */
    public cloneLayerImageObject(layerImageObject: Models.LayerImageObject,
                                 layerModel: Models.LayerGroup,
                                 relativePoint: Models.Point | undefined = undefined): {element: Models.LayerImageObject, relativePoint: Models.Point} {
        // Clone image object
        let imageObject: MedsurfDraw.PositionPoint | MedsurfDraw.Line | MedsurfDraw.FillCollection | MedsurfDraw.LegendCollection;
        const imageObjectModel = this.data.imageObjects.find((imageObject) => imageObject.id === layerImageObject.imageObject) as Models.PositionPoint | Models.Line | Models.FillCollection | Models.LegendCollection;
        if (imageObjectModel) {
            const generator = new MedsurfDraw.CloneItemGenerator({
                imageObjects: this.data.imageObjects,
                layerGroup: layerModel,
                image: this.data.image,
                target: this.target,
                element: imageObjectModel,
                imageObjectMappings: this.data.imageObjectMappings,
                relativePoint: relativePoint,
            });
            imageObject = generator.cloneElement;
            this.data.imageObjectMappings.push(...generator.data.imageObjectMappings);
            relativePoint = generator.data.relativePoint;

            this._generators.push(generator);
        } else {
            throw 'Image object not created';
        }

        const layerImageObjectModel = this.createLayerImageObjectModel(layerImageObject, layerModel, imageObject.name as string);

        // Relative point
        if (!relativePoint) {
            relativePoint = { x: 0, y: 0 };
        }

        // Return layer image object model
        return { element: layerImageObjectModel, relativePoint: relativePoint };
    }

    /**
     * Clone relations
     */
    public cloneRelations(): void {
        // Iterate through all elements
        this.data.imageObjectMappings.forEach((imageObjectMapping) => {
            const oldItem = imageObjectMapping.old;
            const newItem = imageObjectMapping.new;
            switch (oldItem.type) {
                case ImageObjectType.POSITIONPOINT:
                    ((oldItem as Models.PositionPoint).selftestItems || []).forEach((item) => {
                        const reference = this.data.imageObjectMappings.find((mapping) => mapping.old.id === item);
                        // Add only existing mappings -> remove all other references
                        if (reference && reference.new) {
                            // @ts-ignore
                            (newItem as Models.PositionPoint).selftestItems.push(reference.new.id);
                        }
                    });
                    break;
                case ImageObjectType.FILLCOLLECTION:
                    ((oldItem as Models.FillCollection).interactiveItems || []).forEach((item) => {
                        const reference = this.data.imageObjectMappings.find((mapping) => mapping.old.id === item);
                        // Add only existing mappings -> remove all other references
                        if (reference && reference.new) {
                            // @ts-ignore
                            (newItem as Models.FillCollection).interactiveItems.push(reference.new.id);
                        }
                    });
                    break;
            }
        });
    }
    //</editor-fold>

    //<editor-fold desc="Instances">
    /**
     * Create layer group instance
     * @param model
     * @param relativePoint
     */
    public createLayerGroupInstance(model: Models.LayerGroup,
                                    relativePoint: Models.Point | undefined = undefined): Models.LayerGroup {
        // Get new model
        const layerModel = this.createLayerGroupModel(model);

        //<editor-fold desc="Layer group contents">
        // Prevent error when copy to same layer group contents
        let tmpLayerGroupContents: Models.LayerGroupContent[] = [];
        Object.assign(tmpLayerGroupContents, model.groupContents || []);

        // Create layer group contents
        const layerGroupContents: Models.LayerGroupContent[] = [];
        for (let layerGroupContent of tmpLayerGroupContents) {
            const layerGroupContentModel = this.createLayerGroupContentModel(layerGroupContent);

            // Layer image object
            layerGroupContents.push(layerGroupContentModel);
        }
        layerModel.groupContents = layerGroupContents;
        //</editor-fold>

        //<editor-fold desc="Layer group image objects">
        // Prevent error when copy to same layer group image objects
        let tmpLayerImageObjects: Models.LayerImageObject[] = [];
        Object.assign(tmpLayerImageObjects, model.layerImageObjects || []);

        // Create layer group image objects
        for (let layerImageObject of tmpLayerImageObjects) {
            const data = this.cloneLayerImageObject(layerImageObject, layerModel, relativePoint);
            layerImageObject = data.element;
            relativePoint = data.relativePoint;
        }
        //</editor-fold>

        // Create mapping
        this.data.layerGroupMappings.push({
            new: layerModel,
            old: model
        });

        // Get image objects
        this.target.layerGroups.push(layerModel);

        // Return layer group
        return layerModel;
    }
    //</editor-fold>

    //<editor-fold desc="Models">
    /**
     * Create layer group model
     * @param model
     */
    public createLayerGroupModel(model: Models.LayerGroup): Models.LayerGroup {
        return {
            dirty: true,
            id: uuidv4(),
            image: model.image,
            name: model.name,
            order: model.order,
            groupContents: [],
            layerImageObjects: [],
        };
        // Object.assign(data, {id: uuid.v4(), lines: lines})
    }

    /**
     * Create layer group content model
     * @param model
     */
    public createLayerGroupContentModel(model: Models.LayerGroupContent): Models.LayerGroupContent {
        return {
            dirty: true,
            id: uuidv4(),
            order: model.order,
            content: model.content,
        };
    }

    /**
     * Create layer image object model
     * @param model
     * @param layerGroup
     * @param imageObject
     */
    public createLayerImageObjectModel(model: Models.LayerImageObject, layerGroup: Models.LayerGroup, imageObject: string): Models.LayerImageObject {
        return {
            dirty: true,
            id: uuidv4(),
            layerGroup: layerGroup,
            order: model.order,
            imageObject: imageObject,
        };
        // Object.assign(data, {id: uuid.v4(), lines: lines})
    }
    //</editor-fold>

    //<editor-fold desc="Clone">
    /**
     * End clone
     * @param event
     */
    public endClone(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort clone
     * @param event
     */
    public abortClone(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>
}