import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";

/**
 * Legend generator model
 */
export class LegendGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    layerGroup: Models.LayerGroup;
    image: Models.Image;
}

/**
 * Legend generator
 */
export class LegendGenerator extends BaseGenerator<MedsurfDraw.Image, LegendGeneratorModel> {
    /**
     * Members
     */
    private _isImageMoving: boolean;

    /**
     * Interactions
     */
    protected _drawInteraction: MedsurfDraw.DrawInteraction<MedsurfDraw.LegendCollection>;

    /**
     * Elements
     */
    protected _legendElement: MedsurfDraw.LegendCollection;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: LegendGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._isImageMoving = false;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        // -- legend element
        this._legendElement = MedsurfDraw.LegendCollection.getInstance(this.target, this.target.dimensions.width);
        this.legendElement.hideItem();
        this.target.addChild(this.legendElement);
        this.legendElement.controlDrawItem(true);
        //</editor-fold>

        //<editor-fold desc="Interactions">
        // -- Draw interaction
        this._drawInteraction = new MedsurfDraw.DrawInteraction<MedsurfDraw.LegendCollection>(this.legendElement, MedsurfDraw.StickMode.NOELEMENTS);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Image
        this.target.on('imageZoom', this.onZoom, this);
        this.target.parent.on("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
        this.target.parent.on("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
        this.target.parent.on("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
        this.target.parent.on("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
        // ---- Move
        this.drawInteraction.on("startMove", this.legendElement.moveInteraction.startMove, this.legendElement.moveInteraction);
        this.drawInteraction.on("resetMove", this.legendElement.moveInteraction.resetMove, this.legendElement.moveInteraction);
        this.drawInteraction.on("onMove", this.legendElement.moveInteraction.onMove, this.legendElement.moveInteraction);
        this.drawInteraction.on("endMove", this.legendElement.moveInteraction.endMove, this.legendElement.moveInteraction);
        // -- Draw
        this.drawInteraction.on("startDraw", this.legendElement.startDraw, this.legendElement);
        this.drawInteraction.on("startDrawScale", this.legendElement.startDrawScale, this.legendElement);
        this.drawInteraction.on("onDraw", this.legendElement.onDraw, this.legendElement);
        this.drawInteraction.on("onScale", this.onScale, this);
        this.drawInteraction.on("endDraw", this.endDraw, this);
        this.drawInteraction.on("abortDraw", this.abortDraw, this);
        //</editor-fold>

        // Check pointerover
        // -- Check if mouse is in image
        const globalPoint = this.target.renderer.plugins.interaction.mouse.global;
        const hitTest = this.target.renderer.plugins.interaction.hitTest(globalPoint, this.target.parent);
        if (!hitTest) {
            // -- Not hit test result -> mouse outside image
            this.target.parent.once("pointerover", this.start, this);
        } else {
            // -- Hit rest result -> mouse inside image
            const event = new PIXI.InteractionEvent();
            event.data = new PIXI.InteractionData();
            event.data.global = globalPoint;
            this.emit("debounceStart", event);
        }
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        this.legendElement.position.x += (pos.x - rect.x) / this.target.scale.x;
        this.legendElement.position.y += (pos.y - rect.y) / this.target.scale.y;

        // Start draw
        this.drawInteraction.startDraw(event);

        //<editor-fold desc="Elements">
        // -- Legend element
        this.legendElement.showItem();
        //</editor-fold>

        // Parent
        super.start(event);

        // Add Model
        this.target.addImageObject(this.legendElement.model, this.data.layerGroup, this.data.image);
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.off('imageZoom', this.onZoom, this);
        this.target.parent.off("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
        this.target.parent.off("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
        this.target.parent.off("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
        this.target.parent.off("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
        // ---- Move
        this.drawInteraction.off("startMove", this.legendElement.moveInteraction.startMove, this.legendElement.moveInteraction);
        this.drawInteraction.off("resetMove", this.legendElement.moveInteraction.resetMove, this.legendElement.moveInteraction);
        this.drawInteraction.off("onMove", this.legendElement.moveInteraction.onMove, this.legendElement.moveInteraction);
        this.drawInteraction.off("endMove", this.legendElement.moveInteraction.endMove, this.legendElement.moveInteraction);
        // -- Draw
        this.drawInteraction.off("startDraw", this.legendElement.startDraw, this.legendElement);
        this.drawInteraction.off("startDrawScale", this.legendElement.startDrawScale, this.legendElement);
        this.drawInteraction.off("onDraw", this.legendElement.onDraw, this.legendElement);
        this.drawInteraction.off("onScale", this.onScale, this);
        this.drawInteraction.off("endDraw", this.endDraw, this);
        this.drawInteraction.off("abortDraw", this.abortDraw, this);
        //</editor-fold>

        // Parent end
        super.end();

    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Legend element
        if (this.legendElement && this.legendElement.modeInteraction.lastMode.endsWith('_new')) {
            this.legendElement.modeInteraction.setMode('delete_draw')
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Scale">
    /**
     * On scale
     * @param event
     */
    public onScale(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Setup is image moving
        this._isImageMoving = true;
    }
    //</editor-fold>

    //<editor-fold desc="Draw">
    /**
     * End draw
     * @param event
     */
    public endDraw(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        if (this._isImageMoving) {
            this._isImageMoving = false;
            this.drawInteraction.startDraw(event);
            return;
        }

        // Set mode
        this.legendElement.modeInteraction.setModeItem("draw_select_new");

        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort draw
     * @param event
     */
    public abortDraw(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On zoom
     */
    public onZoom(): void {
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        this.legendElement.position.x = (pos.x - rect.x) / this.target.scale.x;
        this.legendElement.position.y = (pos.y - rect.y) / this.target.scale.y;

        // Reset draw
        const event = new PIXI.InteractionEvent();
        event.data = new PIXI.InteractionData();
        event.data.global = pos;
        this.drawInteraction.resetDraw(event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.LegendCollection> {
        return this._drawInteraction;
    }

    /**
     * Get legend element6
     */
    public get legendElement(): MedsurfDraw.LegendCollection {
        return this._legendElement;
    }
    //</editor-fold>
}