import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";

/**
 * Position point generator model
 */
export class PositionPointGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    layerGroup: Models.LayerGroup;
    image: Models.Image;
    stickMode: MedsurfDraw.StickMode;
}

/**
 * Position point generator
 */
export class PositionPointGenerator extends BaseGenerator<MedsurfDraw.Image, PositionPointGeneratorModel> {
    /**
     * Members
     */
    private _isImageMoving: boolean;

    /**
     * Interactions
     */
    protected _drawInteraction: MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint>;

    /**
     * Elements
     */
    protected _positionPointElement: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: PositionPointGeneratorModel) {
        // Parent constructor
        super(model);

        // Members
        this._isImageMoving = false;

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        // -- Position point element
        this._positionPointElement = MedsurfDraw.PositionPoint.getInstance(this.target);
        this.positionPointElement.hideItem();
        this.target.addChild(this.positionPointElement);
        this.positionPointElement.controlDrawItem(true);
        //</editor-fold>

        //<editor-fold desc="Interactions">
        // -- Draw interaction
        this._drawInteraction = new MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint>(this.positionPointElement as MedsurfDraw.PositionPoint, this.stickMode);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Image
        this.target.on('imageZoom', this.onZoom, this);
        this.target.parent.on("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
        this.target.parent.on("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
        this.target.parent.on("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
        this.target.parent.on("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
        // -- Move
        this.drawInteraction.on("startMove", this.positionPointElement.moveInteraction.startMove, this.positionPointElement.moveInteraction);
        this.drawInteraction.on("resetMove", this.positionPointElement.moveInteraction.resetMove, this.positionPointElement.moveInteraction);
        this.drawInteraction.on("onMove", this.positionPointElement.moveInteraction.onMove, this.positionPointElement.moveInteraction);
        this.drawInteraction.on("endMove", this.positionPointElement.moveInteraction.endMove, this.positionPointElement.moveInteraction);
        // -- Draw
        this.drawInteraction.on("startDraw", (this.positionPointElement as MedsurfDraw.PositionPoint).startDraw, this.positionPointElement);
        this.drawInteraction.on("startDrawScale", (this.positionPointElement as MedsurfDraw.PositionPoint).startDrawScale, this.positionPointElement);
        this.drawInteraction.on("onDraw", (this.positionPointElement as MedsurfDraw.PositionPoint).onDraw, this.positionPointElement);
        this.drawInteraction.on("onScale", this.onScale, this);
        this.drawInteraction.on("updatePoint", this.updatePoint, this);
        this.drawInteraction.on("updateFill", this.updateFill, this);
        this.drawInteraction.on("endDraw", this.endDraw, this);
        this.drawInteraction.on("abortDraw", this.abortDraw, this);
        //</editor-fold>

        // Check pointerover
        // -- Check if mouse is in image
        const globalPoint = this.target.renderer.plugins.interaction.mouse.global;
        const hitTest = this.target.renderer.plugins.interaction.hitTest(globalPoint, this.target.parent);
        if (!hitTest) {
            // -- Not hit test result -> mouse outside image
            this.target.parent.once("pointerover", this.start, this);
        } else {
            // -- Hit rest result -> mouse inside image
            const event = new PIXI.InteractionEvent();
            event.data = new PIXI.InteractionData();
            event.data.global = globalPoint;
            this.emit("debounceStart", event);
        }
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        // Update position
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        this.positionPointElement.position.x += (pos.x - rect.x) / this.target.scale.x;
        this.positionPointElement.position.y += (pos.y - rect.y) / this.target.scale.y;

        // Start draw
        this.drawInteraction.startDraw(event);

        //<editor-fold desc="Elements">
        // -- Position point element
        this.positionPointElement.showItem();
        //</editor-fold>

        // Parent
        super.start(event);

        // Add Model
        this.target.addImageObject(this.positionPointElement.model, this.data.layerGroup, this.data.image);
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.off('imageZoom', this.onZoom, this);
        this.target.parent.off("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
        this.target.parent.off("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
        this.target.parent.off("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
        this.target.parent.off("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
        // -- Draw
        this.drawInteraction.removeAllListeners();
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Position point element
        if (this.positionPointElement && this.positionPointElement.modeInteraction.lastMode.endsWith('new')) {
            this.positionPointElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Scale">
    /**
     * On scale
     * @param event
     */
    public onScale(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Setup is image moving
        this._isImageMoving = true;
    }
    //</editor-fold>

    //<editor-fold desc="Draw">
    /**
     * Update point
     * @param event
     * @param positionPoint
     */
    public updatePoint(event: PIXI.InteractionEvent, positionPoint: MedsurfDraw.PositionPoint): void {
        event.stopPropagation();

        // Destroy element
        this.destroy();

        // Setup position point
        this._positionPointElement = positionPoint;

        // Set mode
        if (!this.positionPointElement.modeInteraction.lastMode.startsWith("draw_select")) {
            this.positionPointElement.modeInteraction.setModeItem('draw_select');
        }

        // Emit update point
        this.emit("updatePoint");
    }

    /**
     * Update fill
     * @param event
     * @param fillCollection
     */
    public updateFill(event: PIXI.InteractionEvent, fillCollection: MedsurfDraw.FillCollection): void {
        event.stopPropagation();

        // Destroy element
        this.destroy();

        // Setup position point
        this._positionPointElement = fillCollection;

        // Set mode
        if (!this.positionPointElement.modeInteraction.lastMode.startsWith("draw_select")) {
            this.positionPointElement.modeInteraction.setModeItem('draw_select');
        }

        // Emit update fill
        this.emit('updateFill');
    }

    /**
     * End draw
     * @param event
     */
    public endDraw(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        if (this._isImageMoving) {
            this._isImageMoving = false;
            this.drawInteraction.startDraw(event);
            return;
        }

        // Set mode
        this.positionPointElement.modeInteraction.setModeItem('draw_select_new');

        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort draw
     * @param event
     */
    public abortDraw(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Zoom">
    /**
     * On zoom
     */
    public onZoom(): void {
        const pos = this.target.renderer.plugins.interaction.mouse.global;
        const rect = this.target.getRectangle();
        this.positionPointElement.position.x = (pos.x - rect.x) / this.target.scale.x;
        this.positionPointElement.position.y = (pos.y - rect.y) / this.target.scale.y;

        // Reset draw
        const event = new PIXI.InteractionEvent();
        event.data = new PIXI.InteractionData();
        event.data.global = pos;
        this.drawInteraction.resetDraw(event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.PositionPoint> {
        return this._drawInteraction;
    }

    /**
     * Get position point element6
     */
    public get positionPointElement(): MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection {
        return this._positionPointElement;
    }

    /**
     * Get stick mode
     */
    public get stickMode(): MedsurfDraw.StickMode {
        return this.data.stickMode;
    }

    /**
     * Set stick mode
     * @param value
     */
    public set stickMode(value: MedsurfDraw.StickMode) {
        this.data.stickMode = value;
    }
    //</editor-fold>
}