import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {BaseGenerator} from "../../bases/generators/BaseGenerator";
import {PositionPointGeneratorModel} from "../positionings/PositionPointGenerator";

/**
 * Rectangle primitive generator model
 */
export class RectanglePrimitiveGeneratorModel extends PositionPointGeneratorModel {
    startElement?: MedsurfDraw.PositionPoint;
    rotateWithLine: boolean;
    withScaling: boolean;
    triggerImmediateEndDraw: boolean;
}

/**
 * Rectangle primitive generator
 */
export class RectanglePrimitiveGenerator extends BaseGenerator<MedsurfDraw.Image, RectanglePrimitiveGeneratorModel> {
    /**
     * Generators
     */
    protected _positionPointGenerator: MedsurfDraw.PositionPointGenerator;

    /**
     * Interactions
     */
    protected _drawInteraction: MedsurfDraw.DrawInteraction<MedsurfDraw.RectanglePrimitive>;

    /**
     * Elements
     */
    protected _rectanglePrimitiveElement: MedsurfDraw.RectanglePrimitive;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RectanglePrimitiveGeneratorModel) {
        // Parent constructor
        super(model);

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem("drawing");
        //</editor-fold>

        //<editor-fold desc="Generators">
        if (!this.startElement) {
            // Setup generator
            this._positionPointGenerator = new MedsurfDraw.PositionPointGenerator({
                target: this.target,
                layerGroup: this.data.layerGroup,
                image: this.data.image,
                stickMode: this.stickMode
            });
            this._positionPointGenerator.once("startGenerator", this.start, this);
            this._positionPointGenerator.once("startGenerator", this.startPositionPointGenerator, this);
            this._positionPointGenerator.once("updatePoint", this.updatePositionPointGenerator, this);
            if (this.withScaling) {
                this._positionPointGenerator.once("endGenerator", this.endPositionPointGeneratorScale, this);
                this._positionPointGenerator.once("endGenerator", this.startScale, this);
            } else {
                this._positionPointGenerator.once("endGenerator", this.endPositionPointGenerator, this);
            }
            this._positionPointGenerator.once("abortGenerator", this.abortPositionPointGenerator, this);
        } else {
            // Setup rectangle primitive
            this.startPositionPointGenerator();

            // Start rectangle primitive scale
            if (this.withScaling) {
                this.endPositionPointGeneratorScale();

                // Check pointerover
                // -- Check if mouse is in image
                const globalPoint = this.target.renderer.plugins.interaction.mouse.global;
                const hitTest = this.target.renderer.plugins.interaction.hitTest(globalPoint, this.target);
                if (!hitTest) {
                    // -- Not hit test result -> mouse outside image
                    this.target.once("pointerover", this.startScale, this);
                } else {
                    // -- Hit rest result -> mouse inside image
                    this.startScale();
                }
            } else {
                this.endPositionPointGenerator();
            }
        }
        //</editor-fold>
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        // Emit start generator
        this.emit("startGenerator");
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        // -- Image
        this.target.parent.off("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
        this.target.parent.off("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
        this.target.parent.off("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
        this.target.parent.off("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
        // -- Draw
        this.drawInteraction.removeAllListeners();
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Position point generator
        if (this._positionPointGenerator) {
            this._positionPointGenerator.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();

        //<editor-fold desc="Elements">
        // -- Rectangle primitive element
        if (this.rectanglePrimitiveElement) {
            this.rectanglePrimitiveElement.modeInteraction.setMode('delete_draw');
        }
        //</editor-fold>

        //<editor-fold desc="Generators">
        // -- Position point generator
        if (this._positionPointGenerator) {
            this._positionPointGenerator.destroy();
        }
        //</editor-fold>
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('draw_') && imageObject.modeInteraction.lastMode.endsWith('_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Generator">
    /**
     * Start position point generator
     */
    public startPositionPointGenerator(): void {
        //<editor-fold desc="Elements">
        // -- Rectangle primitive element
        this._rectanglePrimitiveElement = MedsurfDraw.RectanglePrimitive.getInstance(this.target, this.rotateWithLine, this.target.dimensions.width);
        this.positionPointElement.addChild(this.rectanglePrimitiveElement);
        this.rectanglePrimitiveElement.controlDrawItem();
        this.rectanglePrimitiveElement.showItem();
        //</editor-fold>

        //<editor-fold desc="Interactions">
        // -- Draw interaction
        this._drawInteraction = new MedsurfDraw.DrawInteraction<MedsurfDraw.RectanglePrimitive>(this.rectanglePrimitiveElement, MedsurfDraw.StickMode.NOELEMENTS);
        //</editor-fold>
    }

    /**
     * Update position point generator
     */
    public updatePositionPointGenerator(): void {
        // Update primitive
        this.startPositionPointGenerator();

        // End generator
        if (this.withScaling) {
            this.endPositionPointGeneratorScale();

            // Start scale
            this.startScale();
        } else {
            this.endPositionPointGenerator();
        }
    }

    /**
     * End position point generator scale
     */
    public endPositionPointGeneratorScale(): void {
        // Model
        if (!this.positionPointElement.model.primitives) {
            this.positionPointElement.model.primitives = [];
        }
        this.positionPointElement.model.primitives.push(this.rectanglePrimitiveElement.model as any);

        if (!this.withScaling && this.triggerImmediateEndDraw) {
            this.endRectanglePrimitiveDraw();
        } else {
            //<editor-fold desc="Events">
            // -- Image
            this.target.parent.on("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
            this.target.parent.on("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
            this.target.parent.on("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
            this.target.parent.on("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
            // ---- Scale
            this.drawInteraction.on("startScale", this.rectanglePrimitiveElement.scaleInteraction.startScale, this.rectanglePrimitiveElement.scaleInteraction);
            this.drawInteraction.on("onScale", this.rectanglePrimitiveElement.scaleInteraction.onScale, this.rectanglePrimitiveElement.scaleInteraction);
            this.drawInteraction.on("endScale", this.rectanglePrimitiveElement.scaleInteraction.endScale, this.rectanglePrimitiveElement.scaleInteraction);
            // -- Draw
            this.drawInteraction.on("endDraw", this.endRectanglePrimitiveDraw, this);
            this.drawInteraction.on("abortDraw", this.abortRectanglePrimitiveDraw, this);
            //</editor-fold>
        }
    }

    /**
     * End position point generator
     */
    public endPositionPointGenerator(): void {
        // Model
        if (!this.positionPointElement.model.primitives) {
            this.positionPointElement.model.primitives = [];
        }
        this.positionPointElement.model.primitives.push(this.rectanglePrimitiveElement.model as any);

        if (!this.withScaling && this.triggerImmediateEndDraw) {
            this.endRectanglePrimitiveDraw();
        } else {
            //<editor-fold desc="Events">
            // -- Image
            this.target.parent.on("mousedown", this.drawInteraction.startDrawScale, this.drawInteraction);
            this.target.parent.on("rightup", this.drawInteraction.abortDraw, this.drawInteraction);
            this.target.parent.on("pointermove", this.drawInteraction.onDraw, this.drawInteraction);
            this.target.parent.on("mouseup", this.drawInteraction.endDraw, this.drawInteraction);
            // -- Draw
            this.drawInteraction.on("endDraw", this.endRectanglePrimitiveDraw, this);
            this.drawInteraction.on("abortDraw", this.abortRectanglePrimitiveDraw, this);
            //</editor-fold>
        }
    }

    /**
     * Start scale
     */
    public startScale(): void {
        // Start draw scale
        const event = new PIXI.InteractionEvent();
        event.data = new PIXI.InteractionData();
        event.data.global = this.target.renderer.plugins.interaction.mouse.global;
        this.drawInteraction.startDrawScale(event);
    }

    /**
     * Abort position point generator
     */
    public abortPositionPointGenerator(): void {
        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Draw">
    /**
     * End rectangle primitive draw
     */
    public endRectanglePrimitiveDraw(): void {
        // Emit end generator
        this.emit("endGenerator");

        // End
        this.end();
    }

    /**
     * Abort rectangle primitive draw
     */
    public abortRectanglePrimitiveDraw(): void {
        // Emit abort generator
        this.emit("abortGenerator");

        // Destroy element
        this.destroy();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get draw interaction
     */
    public get drawInteraction(): MedsurfDraw.DrawInteraction<MedsurfDraw.RectanglePrimitive> {
        return this._drawInteraction;
    }

    /**
     * Get position point element6
     */
    public get positionPointElement(): MedsurfDraw.PositionPoint {
        return this.startElement ? this.startElement : this._positionPointGenerator.positionPointElement as MedsurfDraw.PositionPoint
    }

    /**
     * Get rectangle primitive element
     */
    public get rectanglePrimitiveElement(): MedsurfDraw.RectanglePrimitive {
        return this._rectanglePrimitiveElement;
    }

    /**
     * Get stick mode
     */
    public get stickMode(): MedsurfDraw.StickMode {
        return this.data.stickMode;
    }

    /**
     * Set stick mode
     * @param value
     */
    public set stickMode(value: MedsurfDraw.StickMode) {
        this.data.stickMode = value;
    }

    /**
     * Get start element
     */
    public get startElement(): MedsurfDraw.PositionPoint | undefined {
        return this.data.startElement;
    }

    /**
     * Set start element
     * @param value
     */
    public set startElement(value: MedsurfDraw.PositionPoint | undefined) {
        this.data.startElement = value;
    }

    /**
     * Get rotate with line
     */
    public get rotateWithLine(): boolean {
        return this.data.rotateWithLine;
    }

    /**
     * Set rotate with line
     * @param value
     */
    public set rotateWithLine(value: boolean) {
        this.data.rotateWithLine = value;
    }

    /**
     * Get with scaling
     */
    public get withScaling(): boolean {
        return this.data.withScaling;
    }

    /**
     * Set with scaling
     * @param value
     */
    public set withScaling(value: boolean) {
        this.data.withScaling = value;
    }

    /**
     * Get trigger immediate end draw
     */
    public get triggerImmediateEndDraw(): boolean {
        return this.data.triggerImmediateEndDraw;
    }

    /**
     * Set trigger immediate end draw
     * @param value
     */
    public set triggerImmediateEndDraw(value: boolean) {
        this.data.triggerImmediateEndDraw = value;
    }
    //</editor-fold>
}