import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

/**
 * Choose types
 */
export enum ChooseTypes {
    PositionPoint = 0,
    Lines = 1,
    Primitives = 2,
    Collections = 3,
}

/**
 * Choose interaction
 */
export class ChooseInteraction<TElement extends MedsurfDraw.DrawableObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    private _chooseType: ChooseTypes;

    /**
     * Constructor
     * @param element
     * @param chooseType
     */
    public constructor(element: TElement, chooseType: ChooseTypes) {
        // Parent constructor
        super(element);

        // Members
        this.chooseType = chooseType;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        return;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * Start choose
     * @param event
     */
    public startChoose(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        this.emit("startChooseStart", event);

        if (this.element.moveInteraction.isMoving) {
            return;
        }

        this.emit("startMove", event);
        this.emit("startChoose", event);
    }

    /**
     * On choose
     * @param event
     */
    public onChoose(event: PIXI.InteractionEvent): void {
        this.emit("onChooseStart", event);

        if (this.element.moveInteraction.isMoving) {
            this.emit("onMove", event);
        }

        this.emit("onChoose", event);
    }

    /**
     * On choose
     * @param event
     */
    public endChoose(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        this.emit("endChooseStart", event);

        if (this.element.moveInteraction.isMoving) {
            this.emit("endMove", event);
        }

        // Check for elements
        const image = this.element.image;
        const renderer = image.renderer;
        if (renderer) {
            const globalPoint = renderer.plugins.interaction.mouse.global;
            const element = renderer.plugins.interaction.hitTest(globalPoint, image);

            // Check type
            if (element instanceof MedsurfDraw.PositionPoint) {
                this.emit("chooseElement", event, element);
            } else if (this._chooseType >= ChooseTypes.Lines && element instanceof MedsurfDraw.Line) {
                this.emit("chooseElement", event, element);
            } else if (this._chooseType >= ChooseTypes.Primitives && (element instanceof MedsurfDraw.ArrowPrimitive
                || element instanceof MedsurfDraw.EllipsePrimitive
                || element instanceof MedsurfDraw.RectanglePrimitive
                || element instanceof MedsurfDraw.TextPrimitive)) {
                this.emit("chooseElement", event, element.parent as MedsurfDraw.PositionPoint);
            } else if (this._chooseType >= ChooseTypes.Collections && (element instanceof MedsurfDraw.LegendRow
                || element instanceof MedsurfDraw.FillCollection)) {
                this.emit("chooseElement", event, element);
            }
        }
    }

    /**
     * Abort choose
     * @param event
     */
    public abortChoose(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        this.emit("abortChooseStart", event);

        /*
        // @ts-ignore
        if (this.targetElement.moveInteraction._isMoving) {
            // @ts-ignore
            this.targetElement.moveInteraction.endMove(event);
        }
        */

        this.emit("abortChoose", event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get choose type
     */
    public get chooseType(): ChooseTypes {
        return this._chooseType;
    }

    /**
     * Set choose type
     * @param value
     */
    public set chooseType(value: ChooseTypes) {
        this._chooseType = value;
    }
    //</editor-fold>
}