import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

/**
 * Stick mode
 */
export enum StickMode {
    NOELEMENTS,
    POSITIONPOINTS,
    POSITIONPOINTS_FILLCOLLECTIONS,
}

/**
 * Draw interaction
 */
export class DrawInteraction<TElement extends MedsurfDraw.DrawableObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    private _stickMode: StickMode;
    protected _isScaling: boolean;

    /**
     * Constructor
     * @param element
     * @param stickMode
     */
    public constructor(element: TElement, stickMode: StickMode) {
        // Parent constructor
        super(element);

        // Members
        this.stickMode = stickMode;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._isScaling = false;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * Start draw
     * @param event
     */
    public startDraw(event: PIXI.InteractionEvent) {
        event.stopPropagation();

        this.emit("startDrawStart", event);

        if (this.element.moveInteraction.isMoving) {
            return;
        }

        this.emit("startMove", event);

        this._isScaling = false;

        this.emit("startDraw", event);
    }

    /**
     * Reset draw
     * @param event
     */
    public resetDraw(event: PIXI.InteractionEvent) {
        event.stopPropagation();

        this.emit("resetDrawStart", event);

        if (!this.element.moveInteraction.isMoving) {
            return;
        }

        this.emit("resetMove", event);
        this.emit("resetDraw", event);
    }

    /**
     * Start draw scale
     * @param event
     */
    public startDrawScale(event: PIXI.InteractionEvent) {
        event.stopPropagation();

        this.emit("startDrawScaleStart", event);

        if (this._isScaling) {
            return;
        }

        this.emit("startScale", event);

        this._isScaling = true;

        this.emit("startDrawScale", event);
    }

    /**
     * On draw
     * @param event
     */
    public onDraw(event: PIXI.InteractionEvent) {
        this.emit("onDrawStart", event);

        if (this._isScaling) {
            this.emit("onScale", event);
        }
        else if (this.element.moveInteraction.isMoving) {
            this.emit("onMove", event);
        }

        this.emit("onDraw", event);
    }

    /**
     * End draw
     * @param event
     */
    public endDraw(event: PIXI.InteractionEvent) {
        event.stopPropagation();

        this.emit("endDrawStart", event);

        if (this.element.moveInteraction.isMoving) {
            this.emit("endMove", event);
        }
        if (this._isScaling) {
            this._isScaling = false;
            this.emit("endScale", event);
        }

        // Check for other elements
        const renderer = this.element.image.renderer;
        if (renderer) {
            const globalPoint = renderer.plugins.interaction.mouse.global;
            const element = renderer.plugins.interaction.hitTest(globalPoint, this.element.image);
            if ((this.stickMode === StickMode.POSITIONPOINTS || this.stickMode === StickMode.POSITIONPOINTS_FILLCOLLECTIONS) && element instanceof MedsurfDraw.PositionPoint) {
                if (element.modeInteraction.lastMode.startsWith('draw')) {
                    this.emit("updatePoint", event, element);
                } else {
                    this.emit("startMove", event);
                }
            } else if ((this.stickMode === StickMode.POSITIONPOINTS || this.stickMode === StickMode.POSITIONPOINTS_FILLCOLLECTIONS) && (element instanceof MedsurfDraw.ArrowPrimitive
                || element instanceof MedsurfDraw.EllipsePrimitive
                || element instanceof MedsurfDraw.RectanglePrimitive
                || element instanceof MedsurfDraw.TextPrimitive)
                && element.parent !== this.element) {
                if ((element.parent as MedsurfDraw.PositionPoint).modeInteraction.lastMode.startsWith('draw')) {
                    this.emit("updatePoint", event, element.parent);
                } else {
                    this.emit("startMove", event);
                }
            } else if (this.stickMode === StickMode.POSITIONPOINTS_FILLCOLLECTIONS && element instanceof MedsurfDraw.FillCollection) {
                if (element.modeInteraction.lastMode.startsWith('draw')) {
                    this.emit("updateFill", event, element);
                } else {
                    this.emit("startMove", event);
                }
            } else {
                this.emit("endDraw", event);
            }
        }
    }

    /**
     * Abort draw
     * @param event
     */
    public abortDraw(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this.emit("abortDrawStart", event);
        this.emit("abortDraw", event);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    public get stickMode(): StickMode {
        return this._stickMode;
    }

    public set stickMode(value: StickMode) {
        this._stickMode = value;
    }
    //</editor-fold>
}