import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";
import {debounce} from 'debounce';

export const MOVE_HITTEST_EXTENSION: number = 200;

/**
 * Move interaction
 */
export class MoveInteraction<TElement extends MedsurfDraw.ImageObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    protected _isMoving: boolean;
    protected _isMoveLock: boolean;
    protected _isImage: boolean;
    protected _moveX: number;
    protected _moveY: number;
    protected _moveDeltaScaleCounter: number;
    protected _moveDeltaScale: number;

    /**
     * Design
     */
    protected readonly _moveDelta: number = 0.1;
    protected readonly _moveDeltaScaleCounterLimit: number = 5;

    /**
     * Constructor
     * @param element
     * @param isImage
     */
    public constructor(element: TElement, isImage: boolean = false) {
        // Parent constructor
        super(element);

        // Members
        this._isMoving = false;
        this._isMoveLock = false;
        this._isImage = isImage;
        this._moveDeltaScaleCounter = 0;
        this._moveDeltaScale = 1;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._moveX = 0;
        this._moveY = 0;
        this._isMoving = false;
        this._isMoveLock = false;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * On move hover
     * @param event
     */
    public onMoveHover(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "move";
        this.emit("onMoveHover", event);
    }

    /**
     * Start move
     * @param event
     */
    public startMove(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this.element.cursor = "move";
        this.emit("startMoveStart", event);

        if (this.isMoving) {
            return;
        }

        const scale = (this._isImage) ? {x: 1, y: 1} : this.element.imageScale;

        this._moveX = event.data.global.x / scale.x;
        this._moveY = event.data.global.y / scale.y;

        this._isMoving = true;
        this._isMoveLock = false;

        this.emit("startMove", event);
    }

    /**
     * Reset move
     * @param event
     */
    public resetMove(event: PIXI.InteractionEvent) {
        event.stopPropagation();

        this.emit("resetMoveStart", event);

        if (!this.isMoving) {
            return;
        }

        const scale = (this._isImage) ? {x: 1, y: 1} : this.element.imageScale;

        this._moveX = event.data.global.x / scale.x;
        this._moveY = event.data.global.y / scale.y;

        this.emit("resetMove", event);
    }

    /**
     * On move
     * @param event
     */
    public onMove(event: PIXI.InteractionEvent) {
        if (!this.isMoving) {
            return;
        }

        this._isMoveLock = true;

        const scale = (this._isImage) ? {x: 1, y: 1} : this.element.imageScale;

        let dX = event.data.global.x / scale.x - this._moveX;
        let dY = event.data.global.y / scale.y - this._moveY;

        this._moveX = event.data.global.x / scale.x;
        this._moveY = event.data.global.y / scale.y;

        this.emit("onMove", event, dX, dY);
    }

    /**
     * End move
     * @param event
     */
    public endMove(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this.element.cursor = "default";
        this.emit("endMoveStart", event);

        if (!this.isMoving) {
            return;
        }

        this._isMoving = false;

        this.emit("endMove", event);

        if (!this._isMoveLock) {
            this.emit("endMoveLock", event);
        } else {
            this._isMoveLock = false;
        }
    }

    /**
     * Move up
     */
    public moveUp(event: PIXI.InteractionEvent): void {
        if (!this._isMoving) {
            this.emit("startMove", event);
            this._isMoving = true;
        }
        if (this._moveDeltaScaleCounter++ % this._moveDeltaScaleCounterLimit == 0) {
            this._moveDeltaScale++;
        }
        this.emit("onMove", event, 0, -1 * this._moveDelta * this._moveDeltaScale);
    }

    /**
     * Move right
     */
    public moveRight(event: PIXI.InteractionEvent): void {
        if (!this._isMoving) {
            this.emit("startMove", event);
            this._isMoving = true;
        }
        if (this._moveDeltaScaleCounter++ % this._moveDeltaScaleCounterLimit == 0) {
            this._moveDeltaScale++;
        }
        this.emit("onMove", event, this._moveDelta * this._moveDeltaScale, 0);
    }

    /**
     * Move down
     */
    public moveDown(event: PIXI.InteractionEvent): void {
        if (!this._isMoving) {
            this.emit("startMove", event);
            this._isMoving = true;
        }
        if (this._moveDeltaScaleCounter++ % this._moveDeltaScaleCounterLimit == 0) {
            this._moveDeltaScale++;
        }
        this.emit("onMove", event, 0, this._moveDelta * this._moveDeltaScale);
    }

    /**
     * Move left
     */
    public moveLeft(event: PIXI.InteractionEvent): void {
        if (!this._isMoving) {
            this.emit("startMove", event);
            this._isMoving = true;
        }
        if (this._moveDeltaScaleCounter++ % this._moveDeltaScaleCounterLimit == 0) {
            this._moveDeltaScale++;
        }
        this.emit("onMove", event, -1 * this._moveDelta * this._moveDeltaScale, 0);
    }

    /**
     * Move release
     */
    public moveRelease(): void {
        this._moveDeltaScaleCounter = 0;
        this._moveDeltaScale = 1;

        this._isMoving = false;
        debounce(this.endMove.bind(this, new PIXI.InteractionEvent()), 50).bind(this)();
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    public get isMoving(): boolean {
        return this._isMoving;
    }
    //</editor-fold>
}