import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

/**
 * Rotate interaction
 */
export class RotateInteraction<TElement extends MedsurfDraw.ImageObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    private _rotateWithLine: boolean;
    protected _rotateX: number;
    protected _isRotating: boolean;

    /**
     * Constructor
     * @param element
     * @param rotateWithLine
     */
    public constructor(element: TElement, rotateWithLine?: boolean) {
        super(element);

        // Setup
        this.rotateWithLine = rotateWithLine || false;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._rotateX = 0;
        // this._rotateY = 0;
        this._isRotating = false;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * Start rotate
     * @param event
     */
    public startRotate(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        this.emit("startRotateStart", event);

        if (this._isRotating) {
            return;
        }

        const scale = this.element.imageScale;

        this._rotateX = event.data.global.x / scale.x;
        // TODO remove this._rotateY = event.data.global.x / scale.y;

        this._isRotating = true;

        this.emit("startRotate", event);
    }

    /**
     * On rotate
     * @param event
     */
    public onRotate(event: PIXI.InteractionEvent): void {
        if (!this._isRotating) {
            return;
        }

        const scale = this.element.imageScale;

        let dR = (event.data.global.x / scale.x - this._rotateX);
        this._rotateX = event.data.global.x / scale.x;

        // TODO snap points
        const degree = Math.round(dR) % 360 * (2 * Math.PI / 360);
        /* TODO
        this.element.rotation = degree;
        (parent.parent.parent as Primitives).rotateAnnotation(degree);
        */

        // TODO parent._rotateGraphic.degree = parent.parent.rotation;

        this.emit("onRotate", degree);
    }

    /**
     * End rotate
     * @param event
     */
    public endRotate(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        this.emit("endRotateStart", event);

        if (!this._isRotating) {
            return;
        }

        this._isRotating = false;

        this.emit("endRotate", event);
    }

    /**
     * Line rotate
     * @param event
     */
    public lineRotate(event: PIXI.InteractionEvent): void {
        if (!this.rotateWithLine) {
            return;
        }

        const parent = this.element.parent as MedsurfDraw.PositionPoint;

        let lineStartX: number = 0;
        let lineStartY: number = 0;
        let lineEndX: number = 0;
        let lineEndY: number = 0;

        const lines = parent.getLines();
        if (lines.length === 0) {
            return;
        }
        for (let line of lines) {
            if (line.start === parent) {
                if (!line.start || !line.end) {
                    return;
                }
                lineStartX += line.start.x;
                lineStartY += line.start.y;
                lineEndX += line.end.x;
                lineEndY += line.end.y;
            } else {
                if (!line.start || !line.end) {
                    return;
                }
                lineStartX += line.end.x;
                lineStartY += line.end.y;
                lineEndX += line.start.x;
                lineEndY += line.start.y;
            }
        }

        lineStartX /= lines.length;
        lineStartY /= lines.length;
        lineEndX /= lines.length;
        lineEndY /= lines.length;

        const degree = Math.atan2(lineEndY - lineStartY, lineEndX - lineStartX) - (90 * Math.PI / 180) || 0;

        this.emit("lineRotate", degree);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    public get rotateWithLine(): boolean {
        return this._rotateWithLine;
    }

    public set rotateWithLine(value: boolean) {
        this._rotateWithLine = value;
    }
    //</editor-fold>
}