import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

/**
 * Enum scale side types
 */
export enum ScaleSideTypes {
    X,
    Y,
    WIDTH,
    HEIGHT,
    X_Y,
    X_HEIGHT,
    WIDTH_Y,
    WIDTH_HEIGHT
}

/**
 * Scale interaction
 */
export class ScaleInteraction<TElement extends MedsurfDraw.ImageObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    protected _pivotCenter: boolean;
    protected _scaleX: number;
    protected _scaleY: number;
    protected _isScaling: boolean;
    protected _scaleSide: ScaleSideTypes;

    /**
     * Constructor
     * @param element
     * @param pivotCenter
     */
    public constructor(element: TElement, pivotCenter: boolean = false) {
        // Parent constructor
        super(element);

        // Members
        this._pivotCenter = pivotCenter;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._scaleX = 0;
        this._scaleY = 0;
        this._isScaling = false;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * Start scale
     * @param event
     * @param scaleSide
     */
    public startScale(event: PIXI.InteractionEvent, scaleSide: ScaleSideTypes = ScaleSideTypes.WIDTH_HEIGHT) {
        event.stopPropagation();

        this.emit("startScaleStart", event);

        if (this._isScaling) {
            return;
        }

        this._scaleSide = scaleSide;

        const scale = this.element.imageScale;

        this._scaleX = event.data.global.x / scale.x;
        this._scaleY = event.data.global.y / scale.y;

        this._isScaling = true;

        this.emit("startScale", event);
    }

    /**
     * On scale
     * @param event
     */
    public onScale(event: PIXI.InteractionEvent) {
        if (!this._isScaling) {
            return;
        }

        const scale = this.element.imageScale;

        let dX: number = 0;
        let dY: number = 0;
        let dW: number = 0;
        let dH: number = 0;

        let pW = 1;
        let pH = 1;
        if (this._pivotCenter) {
            pW = 2;
            pH = 2;
        }

        /*
        const scaleDirection = Math.atan2(event.data.global.y / scale.y - this._scaleY, event.data.global.x / scale.x - this._scaleX);
        if (scaleDirection + 0.2 > Math.PI && scaleDirection - 0.2 < Math.PI) {
            return;
        }
        const scaleDirection = 1 - 1 / Math.atan2(event.data.global.y / scale.y - this._scaleY, event.data.global.x / scale.x - this._scaleX);
        if ((scaleDirection - 0.2) > this.element.rotation && (scaleDirection + 0.2) < this.element.rotation) {
            return;
        }
        */

        /* TODO only scale in correct way
        console.log(event.data.global.x / scale.x, this._scaleX);
        console.log(event.data.global.y / scale.y, this._scaleY);
        console.log("answer:", Math.atan2(event.data.global.y / scale.y - this._scaleY, event.data.global.x / scale.x - this._scaleX));
        */

        let cosR = Math.cos(this.element.rotation);
        if (cosR === 0)
            cosR = 1;

        let sinR = Math.sin(this.element.rotation);
        if (sinR === 0)
            sinR = 1;

        /*
        console.log(cosR);
        console.log(sinR);
        console.log("answer:", 1 - 1 / Math.atan2(event.data.global.y / scale.y - this._scaleY, event.data.global.x / scale.x - this._scaleX));
        */

        if (this._scaleSide == ScaleSideTypes.X || this._scaleSide == ScaleSideTypes.X_Y
            || this._scaleSide == ScaleSideTypes.X_HEIGHT) {
            dX = (event.data.global.x / scale.x - this._scaleX) / cosR;
            dW = -1 * dX * pW;
            this._scaleX = event.data.global.x / scale.x;
        }
        if (this._scaleSide == ScaleSideTypes.Y || this._scaleSide == ScaleSideTypes.X_Y
            || this._scaleSide == ScaleSideTypes.WIDTH_Y) {
            dY = (event.data.global.y / scale.y - this._scaleY) / sinR;
            dH = -1 * dY * pH;
            this._scaleY = event.data.global.y / scale.y;
        }
        if (this._scaleSide == ScaleSideTypes.WIDTH || this._scaleSide == ScaleSideTypes.WIDTH_Y
            || this._scaleSide == ScaleSideTypes.WIDTH_HEIGHT) {
            dW = (event.data.global.x / scale.x - this._scaleX) * pW / cosR;
            this._scaleX = event.data.global.x / scale.x;
        }
        if (this._scaleSide == ScaleSideTypes.HEIGHT || this._scaleSide == ScaleSideTypes.X_HEIGHT
            || this._scaleSide == ScaleSideTypes.WIDTH_HEIGHT) {
            dH = (event.data.global.y / scale.y - this._scaleY) * pH / sinR;
            this._scaleY = event.data.global.y / scale.y;
        }

        // Keep aspect ratio
        if (!(MedsurfDraw.Keyboard.isKeyDown("ShiftLeft") || MedsurfDraw.Keyboard.isKeyDown("ShiftRight"))) {
            const rectangle = this.element.getRectangle();

            if (this._scaleSide == ScaleSideTypes.X_Y || this._scaleSide == ScaleSideTypes.WIDTH_Y) {
                dH = rectangle.height / rectangle.width * dW;
            }
            if (this._scaleSide == ScaleSideTypes.X_HEIGHT || this._scaleSide == ScaleSideTypes.WIDTH_HEIGHT) {
                dW = rectangle.width / rectangle.height * dH;
            }
        }

        this.emit("onScale", this._scaleSide, dX, dY, dW, dH);
    }

    /**
     * End scale
     * @param event
     */
    public endScale(event: PIXI.InteractionEvent) {
        event.stopPropagation();

        this.emit("endScaleStart", event);

        if (!this._isScaling) {
            return;
        }

        this._isScaling = false;

        this.emit("endScale", event);
    }
    //</editor-fold>
}